//$Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/AeExpressionLanguageFactory.java,v 1.1 2005/06/08 12:50:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel;

import java.io.File;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.net.URLClassLoader;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.def.expr.IAeExpressionAnalyzer;
import org.activebpel.rt.bpel.def.expr.xpath.AeXPathExpressionAnalyzer;
import org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidator;
import org.activebpel.rt.bpel.def.validation.expr.xpath.AeXPathExpressionValidator;
import org.activebpel.rt.bpel.impl.expr.IAeExpressionRunner;
import org.activebpel.rt.bpel.impl.expr.xpath.AeXPathExpressionRunner;
import org.activebpel.rt.util.AeUtil;

/**
 * This implementation of the expression language factory uses the engine configuration file to
 * map expression languages to implementations of validators and runners.
 */
public class AeExpressionLanguageFactory implements IAeExpressionLanguageFactory
{
   /** The value of "expressionLanguage" for XPath (also the default for BPEL 1.0). */
   public static final String XPATH_EXPR_LANGUAGE_URI = "http://www.w3.org/TR/1999/REC-xpath-19991116"; //$NON-NLS-1$
   /** The value of "expressionLanguage" for XQuery (extension language by AEI). */
   public static final String XQUERY_EXPR_LANGUAGE_URI = "urn:active-endpoints:expression-language:xquery1.0"; //$NON-NLS-1$
   /** The value of "expressionLanguage" for JavaScript (extension language by AEI). */
   public static final String JAVASCRIPT_EXPR_LANGUAGE_URI = "urn:active-endpoints:expression-language:javascript1.5"; //$NON-NLS-1$

   /** The config map key to get the default language. */
   public static final String DEFAULT_LANGUAGE_KEY = "DefaultLanguage"; //$NON-NLS-1$
   /** The config map key to get the list of languages. */
   public static final String LANGUAGES_KEY = "Languages"; //$NON-NLS-1$
   /** The config map key to get the language URI. */
   public static final String LANGUAGE_URI_KEY = "Uri"; //$NON-NLS-1$
   /** The config map key to get the language validator classname. */
   public static final String LANGUAGE_VALIDATOR_KEY = "Validator"; //$NON-NLS-1$
   /** The config map key to get the language runner classname. */
   public static final String LANGUAGE_RUNNER_KEY = "Runner"; //$NON-NLS-1$
   /** The config map key to get the language classpath. */
   public static final String LANGUAGE_CLASSPATH_KEY = "Classpath"; //$NON-NLS-1$
   /** The config map key to get the language util classname. */
   public static final String LANGUAGE_ANALYZER_KEY = "Analyzer"; //$NON-NLS-1$

   /** The map/list of default languages when no config info is found in aeEngineConfig.xml. */
   private static Map sDefaultLanguages = new HashMap();

   /**
    * Initializes above languages map.
    */
   static 
   {
      try
      {
         Map map = new HashMap();
         map.put(AeExpressionLanguageFactory.LANGUAGE_URI_KEY, AeExpressionLanguageFactory.XPATH_EXPR_LANGUAGE_URI);
         map.put(AeExpressionLanguageFactory.LANGUAGE_VALIDATOR_KEY, AeXPathExpressionValidator.class.getName());
         map.put(AeExpressionLanguageFactory.LANGUAGE_RUNNER_KEY, AeXPathExpressionRunner.class.getName());
         map.put(AeExpressionLanguageFactory.LANGUAGE_ANALYZER_KEY, AeXPathExpressionAnalyzer.class.getName());
         sDefaultLanguages.put("XPath1.0", map); //$NON-NLS-1$
      }
      catch (Throwable t)
      {
         AeException.logError(t, t.getLocalizedMessage());
      }
   }
   
   /** The map of language URI -> language info object. */
   private Map mLanguageMap;
   /** The default language to use if none is specified. */
   private String mDefaultLanguage;

   /**
    * Constructs an expression language factory using the given engine configuration map.
    * 
    * @param aConfig
    */
   public AeExpressionLanguageFactory(Map aConfig)
   {
      setLanguageMap(new HashMap());

      // Set the default expression language.
      String defLanguage = (String) aConfig.get(DEFAULT_LANGUAGE_KEY);
      if (AeUtil.isNullOrEmpty(defLanguage))
      {
         defLanguage = XPATH_EXPR_LANGUAGE_URI;
      }
      setDefaultLanguage(defLanguage);

      // Iterate through all the languages.
      Map languages = (Map) aConfig.get(LANGUAGES_KEY);
      try
      {
         if (languages == null)
         {
            addDefaultLanguages();
         }
         else
         {
            addLanguages(languages);
         }
      }
      catch (AeException e)
      {
         e.logError();
      }
   }

   /**
    * Constructs a default expression language factory that supports only XPath 1.0.
    *
    */
   public AeExpressionLanguageFactory()
   {
      setLanguageMap(new HashMap());
      setDefaultLanguage(XPATH_EXPR_LANGUAGE_URI);

      // Default factory knows about XPath 1.0
      try
      {
         addDefaultLanguages();
      }
      catch (AeException e)
      {
         e.logError();
      }
   }

   /**
    * Gets the map of default languages supported by this expression language factory.
    */
   public static Map getDefaultLanguages()
   {
      return sDefaultLanguages;
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeExpressionLanguageFactory#supportsLanguage(java.lang.String)
    */
   public boolean supportsLanguage(String aLanguageUri)
   {
      String lang = resolveLanguageUri(aLanguageUri);
      return getLanguageMap().containsKey(lang);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeExpressionLanguageFactory#getSupportedLanguages()
    */
   public List getSupportedLanguages()
   {
      return new LinkedList(getLanguageMap().keySet());
   }

   /**
    * Adds all of the languages found in the supplied Map of languages to the language map.
    * 
    * @param aLanguages
    */
   protected void addLanguages(Map aLanguages) throws AeException
   {
      for (Iterator iter = aLanguages.keySet().iterator(); iter.hasNext(); )
      {
         String entryName = (String) iter.next();
         Map map = (Map) aLanguages.get(entryName);
         addLanguage(entryName, map);
      }
   }

   /**
    * Adds the default languages to the language map.
    * 
    * @throws AeException
    */
   protected void addDefaultLanguages() throws AeException
   {
      for (Iterator iter = getDefaultLanguages().keySet().iterator(); iter.hasNext(); )
      {
         String name = (String) iter.next();
         Map language = (Map) getDefaultLanguages().get(name);
         addLanguage(name, language);
      }
   }

   /**
    * Adds a language to the factory.  The map contains all of the pieces of information (found
    * in the engine config) needed to fully describe support for a single expression language.
    * This includes the URL of the language, its name, and implementation classes for validation
    * and execution.
    * 
    * @param aMap
    */
   protected void addLanguage(String aName, Map aMap) throws AeException
   {
      String uri = (String) aMap.get(LANGUAGE_URI_KEY);

      getLanguageMap().put(uri, createLanguageInfo(aName, aMap));
   }

   /**
    * Creates a language info object for the language name and map of properties.
    * 
    * @param aName
    * @param aMap
    * @throws AeException
    */
   protected AeLanguageInfo createLanguageInfo(String aName, Map aMap) throws AeException
   {
      String uri = (String) aMap.get(LANGUAGE_URI_KEY);
      Object validator = aMap.get(LANGUAGE_VALIDATOR_KEY);
      Object runner = aMap.get(LANGUAGE_RUNNER_KEY);
      Object analyzer = aMap.get(LANGUAGE_ANALYZER_KEY);
      String classpath = (String) aMap.get(LANGUAGE_CLASSPATH_KEY);

      return createLanguageInfo(aName, uri, validator, runner, analyzer, classpath);
   }

   /**
    * Creates the language info object given the list of required items.
    * 
    * @param aName
    * @param aUri
    * @param aValidator
    * @param aRunner
    * @param aAnalyzer
    * @param aClasspath
    * @throws AeException
    */
   protected AeLanguageInfo createLanguageInfo(String aName, String aUri, Object aValidator, Object aRunner,
         Object aAnalyzer, String aClasspath) throws AeException
   {
      return new AeLanguageInfo(aName, aUri, aValidator, aRunner, aAnalyzer, aClasspath);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeExpressionLanguageFactory#createExpressionValidator(java.lang.String)
    */
   public IAeExpressionValidator createExpressionValidator(String aLanguage) throws AeException
   {
      String language = resolveLanguageUri(aLanguage);

      AeLanguageInfo info = (AeLanguageInfo) getLanguageMap().get(language);
      if (info == null)
      {
         throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.UNKNOWN_EXPR_LANGUAGE_ERROR") + language); //$NON-NLS-1$
      }
      IAeExpressionValidator validator = info.getValidator();
      if (validator == null)
      {
         throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.NO_VALIDATOR_ERROR") + language); //$NON-NLS-1$
      }
      return validator;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeExpressionLanguageFactory#createExpressionAnalyzer(java.lang.String)
    */
   public IAeExpressionAnalyzer createExpressionAnalyzer(String aLanguage) throws AeException
   {
      String language = resolveLanguageUri(aLanguage);

      AeLanguageInfo info = (AeLanguageInfo) getLanguageMap().get(language);
      if (info == null)
      {
         throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.UNKNOWN_EXPR_LANGUAGE_ERROR") + language); //$NON-NLS-1$
      }
      IAeExpressionAnalyzer analyzer = info.getAnalyzer();
      if (analyzer == null)
      {
         throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.COULD_NOT_FIND_ANALYZER_ERROR") + language); //$NON-NLS-1$
      }
      return analyzer;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeExpressionLanguageFactory#createExpressionRunner(java.lang.String)
    */
   public IAeExpressionRunner createExpressionRunner(String aLanguage) throws AeException
   {
      String language = resolveLanguageUri(aLanguage);

      AeLanguageInfo info = (AeLanguageInfo) getLanguageMap().get(language);
      if (info == null)
      {
         throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.UNKNOWN_EXPR_LANGUAGE_ERROR") + language); //$NON-NLS-1$
      }
      IAeExpressionRunner runner = info.getRunner();
      if (runner == null)
      {
         throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.NO_RUNNER_ERROR") + language); //$NON-NLS-1$
      }
      return runner;
   }
   
   /**
    * This method resolves a language into a language URI.  The value that is passed in
    * to this method is whatever is in the 'expressionLanguage' attribute of the BPEL
    * 'process' element.  In most cases, the value of this attribute will be the URI of
    * the language itself, so nothing is done.  However, an implementation of this 
    * factory could map short names to URIs, for example.  In addition, this method is
    * used to determine the default expression language if no value is specified in the
    * 'expressionLanguage' attribute.  By default (and unless overridden in the engine
    * config), the default language will be XPath 1.0 (as written in the BPEL spec).
    * 
    * @param aLanguage
    */
   protected String resolveLanguageUri(String aLanguage)
   {
      if (AeUtil.isNullOrEmpty(aLanguage))
      {
         return getDefaultLanguage();
      }
      else
      {
         return aLanguage;
      }
   }
   
   /**
    * @return Returns the languageMap.
    */
   public Map getLanguageMap()
   {
      return mLanguageMap;
   }
   
   /**
    * @param aLanguageMap The languageMap to set.
    */
   public void setLanguageMap(Map aLanguageMap)
   {
      mLanguageMap = aLanguageMap;
   }

   /**
    * @return Returns the defaultLanguage.
    */
   public String getDefaultLanguage()
   {
      return mDefaultLanguage;
   }
   
   /**
    * @param aDefaultLanguage The defaultLanguage to set.
    */
   public void setDefaultLanguage(String aDefaultLanguage)
   {
      mDefaultLanguage = aDefaultLanguage;
   }

   /**
    * This class contains information about a class from the engine config.  The class info can be
    * either a String or a Map.  If it is a String, then it represents the fully qualified class 
    * name of the class to instantiate.  If it is a Map, then the map must contain a string entry
    * called "Class" with the fully qualified classname.  In addition, the map may contain a map
    * entry called "Params".  This map will be passed to the class being instantiated during
    * construction, assuming the class has a qualified constructor.
    * 
    * From above, the first form of the XML in the engine config would be:
    * 
    * <pre>
    *   <entry name="Uri" value="urn:active-endpoints:expression-language:my-language" />
    * </pre>
    * 
    * The second form of the XML (with the Params map) would be:
    * 
    * <pre>
    *   <entry name="Uri">
    *      <entry name="Class" value="urn:active-endpoints:expression-language:my-language" />
    *      <entry name="Params">
    *         <entry name="Param1" value="some-value" />
    *         <entry name="Param2" value="some-value" />
    *         <entry name="Param3" value="some-value" />
    *      </entry>
    *   </entry>
    * </pre>
    */
   protected static class AeClassInfo
   {
      /** The classpath to use to find the class. */
      private String mClasspath;
      /** The constructor for this class. */
      private Constructor mConstructor;
      /** Params to send during construction. */
      private Map mParams;

      /**
       * Constructs the class info given either a String or Map.
       * 
       * @param aInfo
       * @param aClasspath
       * @throws AeException
       */
      public AeClassInfo(Object aInfo, String aClasspath) throws AeException
      {
         setClasspath(aClasspath);
         if (aInfo instanceof Map)
         {
            initFromMap((Map) aInfo);
         }
         else if (aInfo instanceof String)
         {
            initFromString((String) aInfo);
         }
         else
         {
            throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.INVALID_CLASSINFO_ERROR")); //$NON-NLS-1$
         }
      }

      /**
       * Constructs a class info from a fully qualified classname.
       * 
       * @param aClassname
       * @throws AeException
       */
      public void initFromString(String aClassname) throws AeException
      {
         try
         {
            Class c = null;
            if (AeUtil.isNullOrEmpty(getClasspath()))
            {
               c = getClassForName(aClassname);
            }
            else
            {
               ClassLoader loader = createClassLoader();
               c = loader.loadClass(aClassname);
            }
            setConstructor(c.getConstructor(new Class[] {}));
            setParams(null);
         }
         catch (Throwable t)
         {
            throw new AeException(t);
         }
      }
      
      /**
       * Constructs a class info from a Map.
       * 
       * @param aInfo
       * @throws AeException
       */
      public void initFromMap(Map aInfo) throws AeException
      {
         try
         {
            String classname = (String) aInfo.get("Class"); //$NON-NLS-1$
            if (AeUtil.isNullOrEmpty(classname))
            {
               throw new AeException(AeMessages.getString("AeExpressionLanguageFactory.CLASSNAME_NOT_SPECIFIED_ERROR")); //$NON-NLS-1$
            }
            Map params = (Map) aInfo.get("Params"); //$NON-NLS-1$

            Class c = null;
            if (AeUtil.isNullOrEmpty(getClasspath()))
            {
               c = getClassForName(classname);
            }
            else
            {
               ClassLoader loader = createClassLoader();
               c = loader.loadClass(classname);
            }
            if (AeUtil.isNullOrEmpty(params))
            {
               setConstructor(c.getConstructor(new Class[] {}));
               setParams(null);
            }
            else
            {
               setConstructor(c.getConstructor(new Class[] { Map.class }));
               setParams(params);
            }
         }
         catch (Throwable t)
         {
            throw new AeException(t);
         }
      }

      /**
       * Gets a class object given the full classname.
       * 
       * @param aClassname
       * @throws Exception
       */
      protected Class getClassForName(String aClassname) throws ClassNotFoundException
      {
         return Class.forName(aClassname);
      }
      
      /**
       * Creates a class loader from the configured classpath.
       */
      private ClassLoader createClassLoader() throws AeException
      {
         URL [] urls = null;
         try
         {
            StringTokenizer tokenizer = new StringTokenizer(getClasspath(), ";"); //$NON-NLS-1$
            List urlList = new LinkedList();
            while (tokenizer.hasMoreTokens())
            {
               String pathItem = tokenizer.nextToken();
               if (pathItem.startsWith("http")) //$NON-NLS-1$
               {
                  URL url = new URL(pathItem);
                  urlList.add(url);
               }
               else
               {
                  File file = new File(pathItem);
                  if (file.exists())
                  {
                     URL url = file.toURL();
                     urlList.add(url);
                  }
               }
            }
            urls = (URL []) urlList.toArray(new URL[urlList.size()]);
         }
         catch (Throwable t)
         {
            throw new AeException(t);
         }
         if (urls == null)
         {
            String msg = MessageFormat.format(AeMessages.getString("AeExpressionLanguageFactory.CLASSPATH_NOT_FOUND_ERROR"), //$NON-NLS-1$
                  new Object [] { getClasspath() });
            throw new AeException(msg);
         }
         else
         {
            return new URLClassLoader(urls, getClass().getClassLoader());
         }
      }
      
      /**
       * Creates a new instance of the class.
       */
      public Object newInstance() throws AeException
      {
         try
         {
            if (getParams() == null)
            {
               return getConstructor().newInstance(new Object[] {});
            }
            else
            {
               return getConstructor().newInstance(new Object[] { getParams() });
            }
         }
         catch (Throwable t)
         {
            throw new AeException(t);
         }
      }

      /**
       * @return Returns the constructor.
       */
      protected Constructor getConstructor()
      {
         return mConstructor;
      }
      
      /**
       * @param aConstructor The constructor to set.
       */
      protected void setConstructor(Constructor aConstructor)
      {
         mConstructor = aConstructor;
      }
      
      /**
       * @return Returns the params.
       */
      protected Map getParams()
      {
         return mParams;
      }
      
      /**
       * @param aParams The params to set.
       */
      protected void setParams(Map aParams)
      {
         mParams = aParams;
      }

      /**
       * @return Returns the classpath.
       */
      protected String getClasspath()
      {
         return mClasspath;
      }

      /**
       * @param aClasspath The classpath to set.
       */
      protected void setClasspath(String aClasspath)
      {
         mClasspath = aClasspath;
      }
   }
   
   /**
    * This internal class represents the information about a specific expression language as 
    * defined in the engine configuration.  Each expression language info will include the
    * URI of the language, the language name, and impl classes for validation and execution.
    */
   protected static class AeLanguageInfo
   {
      /** The name. */
      private String mName;
      /** The expression URI as found in the expressionLanguage attribute of the BPEL process element. */
      private String mUri;
      /** The classname of a class that implements IAeExpressionValidator. */
      private AeClassInfo mValidatorInfo;
      /** The classname of a class that implements IAeExpressionRunner. */
      private AeClassInfo mRunnerInfo;
      /** The classname of a class that implements IAeExpressionAnalyzer. */
      private AeClassInfo mAnalyzerInfo;

      /**
       * Constructs a language info with all the necessary info.  The validator and runner can be 
       * either a String or a Map.  If a String, it should be the fully qualified classname of a
       * class implementing the IAeExpressionValidator or IAeExpressionRunner interface (respectively).
       * If a Map, the map should contain an entry called "Class" which contains the aforementioned
       * class.  In addition, the Map can contain additional parameters (if the Map contains a map
       * entry called "Params") that will be passed to the constructor of the class.
       * 
       * @param aName
       * @param aUri
       * @param aValidator
       * @param aRunner
       * @param aAnalyzer
       * @param aClasspath
       * @throws AeException
       */
      public AeLanguageInfo(String aName, String aUri, Object aValidator, Object aRunner, Object aAnalyzer,
            String aClasspath) throws AeException
      {
         setName(aName);
         setUri(aUri);
         setValidatorInfo(createClassInfo(aValidator, aClasspath));
         setRunnerInfo(createClassInfo(aRunner, aClasspath));
         setAnalyzerInfo(createClassInfo(aAnalyzer, aClasspath));
      }

      /**
       * Creates a class info object from the given meta data and classpath.
       * 
       * @param aClassObj
       * @param aClasspath
       */
      protected AeClassInfo createClassInfo(Object aClassObj, String aClasspath) throws AeException
      {
         return new AeClassInfo(aClassObj, aClasspath);
      }
      
      /**
       * @return Returns the name.
       */
      public String getName()
      {
         return mName;
      }

      /**
       * @param aName The name to set.
       */
      protected void setName(String aName)
      {
         mName = aName;
      }

      /**
       * @return Returns the uri.
       */
      public String getUri()
      {
         return mUri;
      }

      /**
       * @param aUri The uri to set.
       */
      protected void setUri(String aUri)
      {
         mUri = aUri;
      }

      /**
       * Gets an expression runner for this expression language.
       */
      public IAeExpressionRunner getRunner()
      {
         try
         {
            return (IAeExpressionRunner) getRunnerInfo().newInstance();
         }
         catch (AeException ae)
         {
            ae.logError();
            return null;
         }
      }

      /**
       * Gets an expression validator for this expression language.
       */
      public IAeExpressionValidator getValidator()
      {
         try
         {
            return (IAeExpressionValidator) getValidatorInfo().newInstance();
         }
         catch (AeException ae)
         {
            ae.logError();
            return null;
         }
      }

      /**
       * Gets an expression analyzer for this expression language.
       */
      public IAeExpressionAnalyzer getAnalyzer()
      {
         try
         {
            return (IAeExpressionAnalyzer) getAnalyzerInfo().newInstance();
         }
         catch (AeException ae)
         {
            ae.logError();
            return null;
         }
      }
      
      /**
       * @return Returns the runnerInfo.
       */
      protected AeClassInfo getRunnerInfo()
      {
         return mRunnerInfo;
      }
      
      /**
       * @param aRunnerInfo The runnerInfo to set.
       */
      protected void setRunnerInfo(AeClassInfo aRunnerInfo)
      {
         mRunnerInfo = aRunnerInfo;
      }
      
      /**
       * @return Returns the validatorInfo.
       */
      protected AeClassInfo getValidatorInfo()
      {
         return mValidatorInfo;
      }
      
      /**
       * @param aValidatorInfo The validatorInfo to set.
       */
      protected void setValidatorInfo(AeClassInfo aValidatorInfo)
      {
         mValidatorInfo = aValidatorInfo;
      }
      
      /**
       * @return Returns the analyzerInfo.
       */
      protected AeClassInfo getAnalyzerInfo()
      {
         return mAnalyzerInfo;
      }
      
      /**
       * @param aAnalyzerInfo The analyzerInfo to set.
       */
      protected void setAnalyzerInfo(AeClassInfo aAnalyzerInfo)
      {
         mAnalyzerInfo = aAnalyzerInfo;
      }
   }

}
