// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/IAeBusinessProcess.java,v 1.64 2006/02/14 22:39:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel;

import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.impl.AeBpelState;
import org.activebpel.rt.bpel.impl.AeSuspendReason;
import org.activebpel.rt.bpel.impl.IAeAlarmReceiver;
import org.activebpel.rt.bpel.impl.IAeBpelObject;
import org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal;
import org.activebpel.rt.bpel.impl.IAeLink;
import org.activebpel.rt.bpel.impl.IAeProcessPlan;
import org.activebpel.rt.bpel.impl.IAeStateChangeDetail;
import org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity;
import org.activebpel.rt.bpel.impl.activity.support.IAeCompensationCallback;
import org.activebpel.rt.bpel.impl.queue.AeInboundReceive;
import org.activebpel.rt.bpel.impl.storage.IAeProcessSnapshot;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.wsdl.def.AeBPELExtendedWSDLDef;
import org.activebpel.rt.wsdl.def.IAePropertyAlias;
import org.w3c.dom.Document;

/** Describes the interface used for interacting with business processes */
public interface IAeBusinessProcess extends IAeActivity
{
   public static final int PROCESS_LOADED = 0;
   public static final int PROCESS_RUNNING = 1;
   public static final int PROCESS_SUSPENDED = 2;
   public static final int PROCESS_COMPLETE = 3;
   public static final int PROCESS_FAULTED = 4;
   public static final int PROCESS_COMPENSATABLE = 5;

   /** Reason code for processes without a specific state reason */
   public static final int PROCESS_REASON_NONE = -1;
   
   /**
    * Return the partnerRole endpoint reference or null if none is found.
    * @param aPartnerLinkPath
    * @throws AeBusinessProcessException
    */
   public IAeEndpointReference getPartnerRoleEndpointReference(String aPartnerLinkPath) throws AeBusinessProcessException;
   
   /**
    * Gets the def used to create this process.
    */
   public AeProcessDef getProcessDefinition();
   
   /**
    * Get the state of the process
    */
   public int getProcessState();

   /**
    * Get the reason code of process state
    */
   public int getProcessStateReason();
   
   /**
    * Get the process id for this process.
    */
   public long getProcessId();

   /**
    * The engine this process is running inside of.
    */
   public IAeBusinessProcessEngineInternal getEngine();

   /**
    * Suspend the business process.
    */
   public void suspend(AeSuspendReason aReasonCode) throws AeBusinessProcessException;

   /**
    * Returns correlation data on the process instance.
    * @param aLocationPath Identifies the correlation set within the process.
    * @return CorrelationData map
    */
   public Map getCorrelationData(String aLocationPath) throws AeBusinessProcessException;   
   
   /**
    * Set correlation data on the process instance.
    * @param aLocationPath Identifies the correlation set within the process.
    * @param aCorrelationData
    */
   public void setCorrelationData( String aLocationPath, Map aCorrelationData )
   throws AeBusinessProcessException;
   
   /**
    * Set the partnerRole partner link endpoint reference data.
    * @param aIsPartnerRole
    * @param aLocationPath
    * @param aPartnerEndpointRef
    * @throws AeBusinessProcessException
    */
   public void setPartnerLinkData( boolean aIsPartnerRole, String aLocationPath, Document aPartnerEndpointRef )
   throws AeBusinessProcessException;

   /**
    * Resumes the business process.
    *
    * @param aExecute <code>true</code> to execute the process immediately or
    * <code>false</code> to put the process in a running state without
    * executing immediately.
    */
   public void resume(boolean aExecute) throws AeBusinessProcessException;

   /**
    * Resume the business process for the suspended object associated
    * with the passed location path
    * @param aLocation the location path of the object to resume.
    */
   public void resume(String aLocation) throws AeBusinessProcessException;
   
   /**
    * Retry the activity associated with the passed location path or its 
    * enclosing scope.
    * @param aLocation
    * @param aAtScope
    * @throws AeBusinessProcessException
    */
   public void retryActivity( String aLocation, boolean aAtScope ) throws AeBusinessProcessException;
   
   /**
    * Step resume the process and mark the activity associated with passed location
    * path as complete.  The idea here is that the user has manually corrected
    * any issues with the suspended activity or its enclosing scope and has put 
    * the process in a state where it can continue (without actually executing 
    * suspended the activity).
    * @param aLocation
    * @throws AeBusinessProcessException
    */
   public void completeActivity( String aLocation ) throws AeBusinessProcessException;

   /**
    * Terminates the business process.
    */
   public void terminate() throws AeBusinessProcessException;
   
   /**
    * Cancels the process by throwing a forcedTerminatation fault.
    */
   public void cancelProcess() throws AeBusinessProcessException;

   /**
    * This method is a callback for objects to notify the process that it is done.
    * An object's completion is a two step process. The first step is reaching the
    * completion state which is done by calling this method. The second step is
    * the evaluation of an object's links. The parent should be notified of the child's
    * completion after the links have been evaluated in order to catch any faults during link
    * evaluation.
    * @param aObject The object which has completed.
    */
   public void objectCompleted(IAeBpelObject aObject) throws AeBusinessProcessException;
   
   /**
    * This method is a callback for objects to notify the process that it is done,
    * but that an error has occured as indicated by the passed fault object.
    * The process will then notify the parent and fire appropriate events.
    *
    * @param aObject The object which has completed.
    * @param aFaultObject The fault which caused the object to complete.
    */
   public void objectCompletedWithFault(IAeBpelObject aObject, IAeFault aFaultObject) throws AeBusinessProcessException;

   /**
    * This method allows a parent to queue a child activity to execute.
    * @param aObject The object that a parent says is ready to execute, links
    *                 may still prevent this object from immediately executing.
    */
   public void queueObjectToExecute(IAeBpelObject aObject) throws AeBusinessProcessException;


   /**
    * This method allows a runnable object to be queued for execution.
    * @param aObject 
    */
   public void queueObjectToExecute(Runnable aRunnable) throws AeBusinessProcessException;   
   
   /**
    * This method allows a callback for links, which allows for either deadpath
    * elimination or allowing a queued activity to execute.
    * @param aLink The link that has changed its status.
    */
   public void linkStatusChanged(IAeLink aLink);

   /**
    * Sets a message receiver for the given correlated business operation
    * from the specified partner link.
    * @param aMessageReceiver The receiver to call back when the message is received.
    * @param aGroupId
    * @param aMessageExchange Value used to match replies to receives
    * @throws AeBusinessProcessException Thrown if error occurs setting the receiver.
    */
   public void queueMessageReceiver(IAeMessageReceiverActivity aMessageReceiver, int aGroupId, String aMessageExchange)
         throws AeBusinessProcessException;

   /**
    * Returns true if there is data available for this receive. This is used in
    * hanlding multi-start processes.
    * @param aPartnerLinkName
    * @param aPortType
    * @param aOperation
    */
   public boolean isReceiveDataQueued(String aPartnerLinkName,
                                      QName aPortType,
                                      String aOperation);

   /**
    * Returns true if the start message is available for the process.
    */
   public boolean isStartMessageAvailable();

   /**
    * Queues the reply.
    * @param aInputMessage can be null
    * @param aFaultName can be null
    * @param aPartnerLinkName
    * @param aPortType
    * @param aOperation
    * @param aMessageExchange can be null
    * @param aCorrelationMap
    * @throws AeBusinessProcessException
    */
   public void queueReply(IAeMessageData aInputMessage,
                          QName aFaultName,
                          String aPartnerLinkName,
                          QName aPortType,
                          String aOperation,
                          String aMessageExchange,                          
                          Map aCorrelationMap)
         throws AeBusinessProcessException;


   /**
    * Removes a message receiver from the list of message receivers which are active.
    * @param aMessageReceiver The receiver to remove.
    * @throws AeBusinessProcessException Thrown if error occurs removing the receiver.
    */
   public void dequeueMessageReceiver(IAeMessageReceiverActivity aMessageReceiver)
         throws AeBusinessProcessException;

   /**
    * Queue an alarm to callback when the passed dealine has been reached.
    * @param aAlarm The alarm to callback when the dealine is reached.
    * @param aDeadline The deadline for the alarm.
    */
   public void queueAlarm(IAeAlarmReceiver aAlarm, Date aDeadline)
      throws AeBusinessProcessException;

   /**
    * Dequeues an alarm that has been queued.
    * @param aAlarm The alarm to dequeue.
    */
   public void dequeueAlarm(IAeAlarmReceiver aAlarm)
      throws AeBusinessProcessException;

   /**
    * Will queue an invoke to take place.  A framework in the engine
    * will typically perform the actual invoke and then call back
    * the invoke response receiver.
    * @param aInvoke The object to receive the callback when the invoke response is ready.
    * @param aInputMessage The input message for the invoke, null if none.
    * @param aPartnerLinkName The name of the partner link the invoke is to take plave via.
    * @param aPortType The port type of the operation to be invoked.
    * @param aOperation The operation to be invoked.
    * @throws AeBusinessProcessException Thrown if error occurs setting the receiver.
    */
   public void queueInvoke(IAeInvokeResponseReceiver aInvoke,
                            IAeMessageData aInputMessage,
                            String aPartnerLinkName,
                            QName aPortType,
                            String aOperation)
         throws AeBusinessProcessException;

   /**
    * This method is a callback for objects to notify the process that they're
    * state has changed.
    * @param aObject
    * @param aOldState
    * @param aDetailsObject  - object used to report any additional information along with the state change,
    *                          most notably the fault that caused the object to
    *                          become faulted.
    */
   public void objectStateChanged(IAeBpelObject aObject, AeBpelState aOldState, IAeStateChangeDetail aDetailsObject) throws AeBusinessProcessException;

   /**
    * Returns the data for the variable being referenced in the activity location path.
    * An optional part may be specified when dealing with a message type variable.
    * @param aActivityLoc the location path for the activity which references this variable
    * @param aVarName the name of the variable being referenced
    * @param aPart the message part we want data for or null if not message type variable
    * @return the data for the given variable.
    */
   public Object getVariableData(String aActivityLoc, String aVarName, String aPart);

   /**
    * Sets the message that created the process. The process will keep this until
    * its start activity executes at which time it'll be consumed by that
    * activity.
    * @param aReceiveQueueObject
    */
   public void setCreateMessage(AeInboundReceive aReceiveQueueObject);

   /**
    * Returns the property alias for the given message property. This method may
    * return null if a message property provider has not been defined.
    * @param aMessage the message we are processing
    * @param aProperty the property within the message we are looking for.
    */
   public IAePropertyAlias getMessageProperty(QName aMessage, QName aProperty) throws AeBusinessProcessException;

   /**
    * Sets the context WSDL provider in the process engine.
    * @param aProvider the provider of the context definitions
    */
   public void setContextWSDLProvider( IAeContextWSDLProvider aProvider );

   /**
    * Returns the WSDL definition given the message name or null if not found.
    * @param aMsgName the message name we are interested in.
    */
   public AeBPELExtendedWSDLDef getWSDLDefinition(QName aMsgName) throws AeBusinessProcessException;

   /**
    * Called by serializable scopes prior to execution to ensure that they have
    * all of the necessary locks on the variables being used by their enclosed
    * activities. If the locks are immediately obtained then the activity proceeds
    * with its normal execution path. If not, then execution is halted for the
    * activity until the lock can be acquired. At that time, the activity will
    * resume normal execution and release any of its locks after it is complete.
    * @param aSetOfVariablePaths
    * @param aOwnerXPath
    */
   public boolean addExclusiveLock(Set aSetOfVariablePaths, String aOwnerXPath);

   /**
    * Called by activities prior to execution to ensure that they have
    * all of the necessary locks on the variables being used by their enclosed
    * activities. If the locks are immediately obtained then the activity proceeds
    * with its normal execution path. If not, then execution is halted for the
    * activity until the lock can be acquired. At that time, the activity will
    * resume normal execution and release any of its locks after it is complete.
    * @param aSetOfVariablePaths
    * @param aOwnerXPath
    */
   public boolean addSharedLock(Set aSetOfVariablePaths, String aOwnerXPath);

   /**
    * Get the date/time that the process instance was started.
    */
   public Date getStartDate();

   /**
    * Gets the date/time that the process ended, either normally, through a fault,
    * or via termination.
    */
   public Date getEndDate();

   /**
    * Gets the qualified name of the process.
    */
   public QName getName();

   /**
    * Serialize and return the process state.
    * @param aForPersistence <code>true</code> to serialize for persistence.
    * @return Document
    * @throws AeBusinessProcessException
    */
   public Document serializeState(boolean aForPersistence) throws AeBusinessProcessException;

   /**
    * Serialize and return a process variable.
    * @param aLocationPath The location XPath of the variable's enclosing scope.
    * @return Document
    * @throws AeBusinessProcessException
    */
   public Document serializeVariable( String aLocationPath ) throws AeBusinessProcessException;
   
   /**
    * Set this process's state from a previously serialized process state document.
    *
    * @param aDocument
    * @throws AeBusinessProcessException
    */
   public void setProcessData(Document aDocument) throws AeBusinessProcessException;

   /**
    * Set a variable from a previously serialized variable document.
    *
    * @param aLocationPath
    * @param aDocument
    * @param aValidate
    * @throws AeBusinessProcessException
    */
   public void setVariableData(String aLocationPath, Document aDocument, boolean aValidate) throws AeBusinessProcessException;

   /**
    * Dispatches an alarm by location path.
    *
    * @param aLocationPath
    */
   public void dispatchAlarm(String aLocationPath);

   /**
    * Dispatches an alarm by location path id.
    *
    * @param aLocationPathId The location path id.
    */
   public void dispatchAlarm(int aLocationPathId);

   /**
    * Dispatches message data to a queued invoke.
    *
    * @param aLocationPath The path to the location awaiting the response
    * @param aData The data we have received from invoke.
    */
   public void dispatchInvokeData(String aLocationPath, IAeMessageData aData);

   /**
    * Dispatches a fault to a queued invoke.
    *
    * @param aLocationPath The path to the location awaiting the response
    * @param aFault The fault we received from invoke.
    */
   public void dispatchInvokeFault(String aLocationPath, IAeFault aFault);

   /**
    * Dispatches message data to a message receiver.
    *
    * @param aLocationPath The path to the location awaiting the message
    * @param aData The message data.
    */
   public void dispatchReceiveData(String aLocationPath, IAeMessageData aData );

   /**
    * Dispatches message data to a message receiver (by location path id).
    *
    * @param aLocationPathId The location path id to the location awaiting the message
    * @param aData The message data.
    */
   public void dispatchReceiveData(int aLocationPathId, IAeMessageData aData );

   /**
    * Returns the message that created the process or <code>null</code> if the
    * start activity has already consumed the message.
    *
    * @return AeInboundReceive
    */
   public AeInboundReceive getCreateMessage();

   /**
    * Returns location id corresponding to a location path or <code>-1</code> if not found.
    */
   public int getLocationId(String aLocationPath);

   /**
    * Returns location path corresponding to a location id or <code>null</code> if not found.
    */
   public String getLocationPath(int aLocationId);

   /**
    * Returns a process snapshot that contains all variables and correlation
    * sets that are live for normal or compensation processing.
    *
    * @throws AeBusinessProcessException
    */
   public IAeProcessSnapshot getProcessSnapshot() throws AeBusinessProcessException;

   /**
    * Gets the plan that this process was created with.
    */
   public IAeProcessPlan getProcessPlan();

   /**
    * Gets the expression language factory.
    */
   public IAeExpressionLanguageFactory getExpressionLanguageFactory() throws AeException;

   /**
    * Remove the business process property from the process.
    * @param aName
    * @return The business process property that was removed or null if no object was removed.
    */
   public String removeBusinessProcessProperty( String aName );

   /**
    * Add the business process property to the process.
    * @param aName
    * @param aProperty
    */
   public void setBusinessProcessProperty( String aName, String aProperty );

   /**
    * Retrieve the business process property object or null if none is found.
    * @param aName
    */
   public String getBusinessProcessProperty( String aName );

   /**
    * Returns an <code>Iterator</code> over the names of all of the business
    * process properties contained in the process.
    */
   public Iterator getBusinessProcessPropertyNames();

   /**
    * Returns the Bpel implementation object specified by the location id.
    * @param aLocationId the location id of the Bpel object
    * @return the Bpel impl object or null if not found
    */
   public IAeBpelObject findBpelObject(int aLocationId);

   /**
    * Returns the Bpel implementation object specified by the location path string.
    * @param aLocationPath the XPath location of the Bpel object
    * @return the Bpel impl object or null if not found
    */
   public IAeBpelObject findBpelObject(String aLocationPath);   
   
   /**
    * Returns the variable specified by the location id.
    * @param aLocationPath the location path of the Bpel object
    * @return the variable impl or null if not found
    */
   public IAeVariable findProcessVariable(String aLocationPath);
   
   /**
    * Sets the engine this process is running inside of.
    */
   public void setEngine(IAeBusinessProcessEngineInternal aEngine);
   
   /**
    * Returns true if the process has been suspended.
    */
   public boolean isSuspended();
   
   /**
    * Return the <code>List</code> of location paths for any activities that
    * are in the faulting state.  This <code>List</code> will be empty if there
    * are none.
    */
   public List getFaultingActivityLocationPaths();
   
   /**
    * Removes the receiver from the process's collection of receive activities used
    * to track bpws:conflictingReceives. This gets called by the receive or onMessage
    * when it receives its data. In the case of an onMessage within a scope, the
    * onMessage will not remove itself until it has been dequeued since events
    * remain queued in BPEL 2.0.
    * @param aMessageReceiver
    */
   public void removeReceiverKeyForConflictingReceives(IAeMessageReceiverActivity aMessageReceiver);
   
   /**
    * The process maintains unique values for each version of a variable. 
    * Each time a variable's data changes it gets a new id for the storage layer.
    */
   public int getNextVariableVersionNumber();
   
   /**
    * Sets the next id to use when getNextVariableVersionNumber() is called. 
    * @param aId
    */
   public void setNextVariableVersionNumber(int aId);
   
   /**
    * Starts compensation for the process
    * @param aCallback
    * @throws AeBusinessProcessException
    */
   public void compensate(IAeCompensationCallback aCallback) throws AeBusinessProcessException;

   /**
    * Releases any resources tied to process level compensation. Called when a compensatable process is
    * no longer compensatable (likely owing to its coordination ending without having called compensate). 
    */
   public void releaseCompensationResources();
}
