// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/config/AeDefaultEngineConfiguration.java,v 1.20 2006/01/06 23:55:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.config;

import java.io.File;
import java.io.InputStream;
import java.io.Writer;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeExpressionLanguageFactory;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeExpressionLanguageFactory;
import org.activebpel.rt.bpel.function.AeFunctionContextContainer;
import org.activebpel.rt.bpel.function.AeFunctionContextLocator;
import org.activebpel.rt.bpel.function.AeUnresolvableException;
import org.activebpel.rt.bpel.function.IAeFunction;
import org.activebpel.rt.bpel.function.IAeFunctionContext;
import org.activebpel.rt.bpel.function.IAeFunctionContextLocator;
import org.activebpel.rt.bpel.impl.function.AeBpelFunctionContext;
import org.activebpel.rt.bpel.impl.function.AeExtensionFunctionContext;
import org.activebpel.rt.util.AeUtil;
import org.activebpel.rt.xml.AeXMLParserBase;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Default engine configuration implementation.
 */
public class AeDefaultEngineConfiguration implements IAeEngineConfiguration, IAeUpdatableEngineConfig, Cloneable
{
   private static final String ERROR_LOADING_FUNCTION_CONTEXT_LOCATOR = "AeDefaultEngineConfiguration.ERROR_17"; //$NON-NLS-1$
   private static final String ERROR_LOADING_FUNCTION_CONTEXT_CLASS = "AeDefaultEngineConfiguration.ERROR_18"; //$NON-NLS-1$
   
   /** Default name for static engine config file. */
   public static final String DEFAULT_CONFIG_FILE = "aeEngineConfig.xml"; //$NON-NLS-1$
   
   ///////////////////////////////////////////////////////
   // constants for config file parsing
   ///////////////////////////////////////////////////////
   public static final String CONFIG_TAG = "config"; //$NON-NLS-1$
   public static final String ENTRY_TAG  = "entry"; //$NON-NLS-1$
   public static final String NAME_ATTR  = "name"; //$NON-NLS-1$
   public static final String VALUE_ATTR = "value"; //$NON-NLS-1$

   //////////////////////////////////////////////////////
   // default values for some of the main entries.
   //////////////////////////////////////////////////////
   /** Default description of the configuration. */
   private static final String DEFAULT_CONFIG_DESCRIPTION = ""; //$NON-NLS-1$
   /** default timeout value for unmatched correlated receives */
   public static final int UNMATCHED_RECEIVE_TIMEOUT_DEFAULT = 30;
   /** default min number of threads for the work manager's thread pool */
   public static final int WORKMANAGER_THREAD_MIN_DEFAULT = 1;
   /** default max number of threads for the work manager's thread pool */
   public static final int WORKMANAGER_THREAD_MAX_DEFAULT = 10;
   /** default maximum number of work requests to schedule per-process */
   public static final int PROCESS_WORK_COUNT_DEFAULT = 10;
   /** default wsdl cache value */
   private static final int DEFAULT_WSDL_CACHE_MAX = 100; 
   /** default partner defn address factory */
   private static final String PDEF_FACTORY_DEFAULT = "org.activebpel.rt.bpel.server.addressing.pdef.AeDefaultPartnerAddressingFactory"; //$NON-NLS-1$
   /** default process manager */
   private static final String PROCESS_MANAGER_DEFAULT = "org.activebpel.rt.bpel.impl.AeInMemoryProcessManager"; //$NON-NLS-1$
   /** Default logging dir - defaults to {user.home}/AeBpelEngine*/
   private static final String LOGGING_DIR_DEFAULT = new File(System.getProperty("user.home"), "AeBpelEngine").getPath(); //$NON-NLS-1$ //$NON-NLS-2$
   /** Map of entries in the configuration. */
   protected HashMap mEntries = new HashMap();
   /** Configuration change listeners. */
   protected ArrayList mListeners = new ArrayList();
   /** Container for custom function contexts. */
   protected AeFunctionContextContainer mContextContainer;
   /** A cached expression language factory. */
   protected IAeExpressionLanguageFactory mExpressionLanguageFactory;
   /** Storage listeners  */
   protected ArrayList mStorageListeners = new ArrayList();

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getExpressionLanguageFactory()
    */
   public IAeExpressionLanguageFactory getExpressionLanguageFactory() throws AeException
   {
      return mExpressionLanguageFactory;
   }

   /**
    * Creates the expression language factory using the current configuration map.  Defaults to the
    * standard AeExpressionLanguageFactory which includes support for XPath 1.0 as required by the 
    * BPEL spec.  This method is called on config load and the expression language factory is 
    * cached at that point.
    * 
    * @throws AeException
    */
   protected void createExpressionLanguageFactory() throws AeException
   {
      try
      {
         IAeExpressionLanguageFactory factory = (IAeExpressionLanguageFactory) createConfigSpecificClass(EXPRESSION_FACTORY);
         if (factory == null)
         {
            mExpressionLanguageFactory = createDefaultExpressionLanguageFactory();
         }
         else
         {
            mExpressionLanguageFactory = factory;
         }
      }
      catch (Throwable t)
      {
         throw new AeException(AeMessages.getString("AeDefaultEngineConfiguration.FAILED_TO_CREATE_EXPR_LANG_FACTORY_ERROR"), t); //$NON-NLS-1$
      }
   }

   /**
    * Creates the default expression language factory when none is found in the configuration
    * file.
    */
   protected IAeExpressionLanguageFactory createDefaultExpressionLanguageFactory()
   {
      return new AeExpressionLanguageFactory();
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getFunction(java.lang.String, java.lang.String)
    */
   public IAeFunction getFunction(String aFunctionName, String aNamespaceUri) throws AeUnresolvableException
   {
      IAeFunctionContext context = getFunctionContextContainer().getFunctionContext(aNamespaceUri);
      if (context == null)
      {
         return null;
      }
      
      try
      {
         return context.getFunction(aFunctionName);
      }
      catch( AeUnresolvableException ure )
      {
         AeException.logError(ure, ure.getLocalizedMessage());
         throw new AeUnresolvableException(ure.getLocalizedMessage());
      }
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getFunctionContextNamespaceList()
    */
   public List getFunctionContextNamespaceList()
   {
      return new LinkedList(getFunctionContextContainer().getFunctionContextNamespaces());
   }
   
   /**
    * Adds a new function context.
    * 
    * @param aName Name of the context group.
    * @param aNamespace namepsace for context, null if no default namespace for this context.
    * @param aContext The function context for the passed prefix or namespace.
    */
   protected void addFunctionContext(String aName, String aNamespace, IAeFunctionContext aContext )
   {
      getFunctionContextContainer().addFunctionContext(aName, aNamespace, aContext);
   }

   /**
    * Load entries from the passed element, note this routione is recursive.  It
    * builds maps for any nested config entries.
    * @param aElement The element to fill in the passed entry map for.
    */
   protected static void loadEntries(Element aElement, HashMap aEntries)
   {
      for(Node node=aElement.getFirstChild(); node != null; node = node.getNextSibling())
      {
         if(node.getNodeType() == Node.ELEMENT_NODE)
         {
            Element elem = (Element)node;
            if(ENTRY_TAG.equals(elem.getLocalName()))
            {
               String name = elem.getAttribute(NAME_ATTR);
               if (elem.hasAttribute(VALUE_ATTR))
               {
                  String value = elem.getAttribute(VALUE_ATTR);
                  aEntries.put(name, value);
               }
               else
               {
                  HashMap map = new HashMap();
                  aEntries.put(name, map);
                  loadEntries(elem, map);
               }
            }
         }
      }
   }
   
   /**
    * Process all the function context entries in the passed map.  Note that this method is
    * public for unit testing purposes (the engine unit test framework calls it - see 
    * AeEngineTestCase).
    * 
    * @param aContextEntries The map of xpath function context entries.
    */
   public void processFunctionContexts(Map aContextEntries)
   {
      for(Iterator iter = aContextEntries.keySet().iterator(); iter.hasNext(); )
      {
         Object key = iter.next();
         Object contextObj = aContextEntries.get(key);
         if (contextObj instanceof Map)
         {
            Map context = (Map)contextObj; 
            String namespace = (String)context.get(FUNCTION_CONTEXT_NAMESPACE_ENTRY); 
            String contextClassName = (String)context.get(CLASS_ENTRY);
            String location = (String)context.get(FUNCTION_CONTEXT_CLASSPATH_ENTRY);
            
            try
            {
               IAeFunctionContext fc = getFunctionContextContainer().loadFunctionContext( namespace, location, contextClassName );
               getFunctionContextContainer().addFunctionContext( (String)key, namespace, fc );
            }
            catch( AeException ae )
            {
               ae.logError();
            }
         }
         else
         {
            AeException.logError( new AeException(AeMessages.getString("AeDefaultEngineConfiguration.11")), AeMessages.getString("AeDefaultEngineConfiguration.ERROR_14") + key); //$NON-NLS-1$ //$NON-NLS-2$
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getPartnerAddressingFactoryClassName()
    */
   public String getPartnerAddressingFactoryClassName()
   {
      return getEntry(PDEF_FACTORY_ENTRY, PDEF_FACTORY_DEFAULT);
   }

   /**
    * Setter for the partner addressing factory class name.
    * @param aClassName
    */
   public void setPartnerAddressingFactoryClassName( String aClassName )
   {
      setEntry(PDEF_FACTORY_ENTRY, aClassName);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getProcessManagerClassName()
    */
   public String getProcessManagerClassName()
   {
      return getEntry(PROCESS_MANAGER_ENTRY, PROCESS_MANAGER_DEFAULT);
   }

   /**
    * Setter for the process manager class name.
    *
    * @param aClassName
    */
   public void setProcessManagerClassName(String aClassName)
   {
      setEntry(PROCESS_MANAGER_ENTRY, aClassName);
   }

   ///////////////////////////////////////////////////////
   // Main config accessor methods, with variant support
   ///////////////////////////////////////////////////////

   /**
    * Get a string entry in the config with the passed name.
    * Returning null if none available.
    */
   public String getEntry(String aName)
   {
      return getEntry(aName, null);
   }

   /**
    * Get a string entry in the config with the passed name.
    * Returning the passed default if entry is unavailable.
    */
   public String getEntry(String aName, String aDefault)
   {
      return (String)getEntryInternal( getEntries(), aName,  String.class, aDefault );
   }
   
   /**
    * Get a boolean entry in the config with the passed name.
    * Returning the passed default if entry is unavailable.
    */
   public boolean getBooleanEntry(String aName, boolean aDefault)
   {
      String trueFalseFlag = getEntry(aName, Boolean.toString(aDefault) );
      return AeUtil.toBoolean( trueFalseFlag );
   }
   
   /**
    * Get a boolean entry from the map with the passed name.
    * Returning the passed default if entry is unavailable.
    * @param aMap
    * @param aName
    * @param aDefault
    */
   protected boolean getBooleanEntryInternal( Map aMap, String aName, boolean aDefault )
   {
      String defaultFlag = Boolean.toString(aDefault);
      String trueFalseFlag = (String)getEntryInternal( aMap, aName, String.class, defaultFlag );
      return AeUtil.toBoolean( trueFalseFlag );
   }

   /**
    * Get an integer entry in the config with the passed name.
    * Returning the passed default if entry is unavailable.
    */
   public int getIntegerEntry(String aName, int aDefault)
   {
      return getIntegerEntryInternal( getEntries(), aName, aDefault );
   }

   /**
    * Internal utility method for accessing config entries with expected int values.
    * @param aMap
    * @param aKey
    * @param aDefaultValue
    */
   protected int getIntegerEntryInternal( Map aMap, String aKey, int aDefaultValue )
   {
      int retVal = aDefaultValue;
      try
      {
         String intValue = (String)getEntryInternal( aMap, aKey, String.class, String.valueOf(aDefaultValue) );
         retVal = Integer.parseInt( intValue );
      }
      catch( NumberFormatException nfe )
      {
         AeException.logError( nfe, nfe.getMessage() );
      }
      return retVal;
   }

   /**
    * Get a mapped entry in the config with the passed name.
    * Returns null if no map entry found or entry is not a map.
    */
   public Map getMapEntry(String aName)
   {
      return getMapEntry( aName, false );
   }
   
   /**
    * Retrieve the map entry from the config.
    * @param aName The config entry name.
    * @param aCreateFlag If true and the named map is null, it will be created
    * and added to the config map.
    */
   protected Map getMapEntry( String aName, boolean aCreateFlag )
   {
      Map map = (Map)getEntryInternal( getEntries(), aName, Map.class, null );
      if( map == null && aCreateFlag )
      {
         map = new HashMap();
         setEntry( aName, map );
      }
      return map;
   }

   /**
    * Internal utility method for accessing config entries.
    * @param aMap
    * @param aKey
    * @param aType
    * @param aDefault
    */
   protected Object getEntryInternal( Map aMap, String aKey, Class aType, Object aDefault )
   {
      Object retVal = aDefault;
      
      if( aMap != null )
      {
         Object entry = aMap.get( aKey );
         if( entry != null && aType.isAssignableFrom( entry.getClass() ) )
         {
            retVal = entry;
         }
      }
      return retVal;
   }

   ///////////////////////////////////////////////////////   
   // Main config mutator methods, with variant support
   ///////////////////////////////////////////////////////

   /** 
    * Removes an entry in the config using the given key object. 
    * @param aKey the key object
    */
   public void removeEntry(Object aKey)
   {
      getEntries().remove(aKey);
   }
   
   /** Set an entry in the config with a specific value. */
   public void setEntry(String aName, String aValue)
   {
      getEntries().put(aName, aValue);
   }

   /** Set an entry in the config with a specific value. */
   public void setBooleanEntry(String aName, boolean aBool)
   {
      getEntries().put(aName, Boolean.toString(aBool));
   }

   /** Set an entry in the config with a specific value. */
   public void setIntegerEntry(String aName, int aInt)
   {
      getEntries().put(aName, Integer.toString(aInt));
   }

   /** Set an entry in the config with a map of values. */
   public void setEntry(String aName, Map aMap)
   {
      getEntries().put(aName, aMap);
   }

   ////////////////////////////////////////////////////////
   // Methods for loading and writing a configuration.
   ////////////////////////////////////////////////////////

   /**
    * Writes the configuration file to the passed stream.
    * @param aWriter The stream to write the document to.
    * @throws AeException If an error occurs wsaving the config to the stream.
    */
   public void save(Writer aWriter) throws AeException
   {
      Document doc = (new AeXMLParserBase()).createDocument();
      Element root = doc.createElement(CONFIG_TAG);
      doc.appendChild(root);
      writeEntries(root, getEntries());
      AeXMLParserBase.saveDocument(doc, aWriter);
   }

   /**
    * Writes a map of entries as children of the passed element, recursively.
    * @param aParent The parent element to contain the child entries
    * @param aEntries the entries to write.
    */
   private static void writeEntries(Element aParent, Map aEntries)
   {
      for(Iterator iter = aEntries.entrySet().iterator(); iter.hasNext();)
      {
         Map.Entry entry = (Map.Entry)iter.next();
         Element entryElement = aParent.getOwnerDocument().createElement(ENTRY_TAG);
         aParent.appendChild(entryElement);
         entryElement.setAttribute(NAME_ATTR, entry.getKey().toString());
         if(entry.getValue() instanceof Map)
            writeEntries(entryElement, (Map)entry.getValue());
         else if(entry.getValue() != null)
            entryElement.setAttribute(VALUE_ATTR, entry.getValue().toString());
      }
   }

   /**
    * Load the engine configuration from the xml in the passed stream.
    * @param aConfigStream a stream containing the configuration xml.
    * @param aClassLoader the classloader for function contexts.
    * @return the config created from the passed stream.
    */
   public static AeDefaultEngineConfiguration loadConfig(InputStream aConfigStream, ClassLoader aClassLoader)
   {
      AeDefaultEngineConfiguration config = new AeDefaultEngineConfiguration();
      loadConfig(config, aConfigStream, aClassLoader);
      return config;
   }
   
   /**
    * Load the engine configuration from the xml in the passed stream.
    * @param aDefaultConfig the config to load
    * @param aConfigStream a stream containing the configuration xml.
    * @param aClassLoader the classloader for function contexts.
    */
   public static void loadConfig(AeDefaultEngineConfiguration aDefaultConfig, InputStream aConfigStream, ClassLoader aClassLoader)
   {
      if(aConfigStream == null)
         return;

      AeXMLParserBase parser = new AeXMLParserBase();
      parser.setValidating(false);
      parser.setNamespaceAware(true);

      try
      {
         Document doc = parser.loadDocument(aConfigStream, null);
         if(doc.getDocumentElement() != null)
         {
            HashMap entries = new HashMap();
            loadEntries(doc.getDocumentElement(), entries );
            aDefaultConfig.setEntries( entries );
            aDefaultConfig.createExpressionLanguageFactory();
            aDefaultConfig.initFunctionContexts(aClassLoader);
         }
      }
      catch(Throwable ex)
      {
         AeException.logError(ex, AeMessages.getString("AeDefaultEngineConfiguration.ERROR_16")); //$NON-NLS-1$
      }

      return;
   }
   
   /**
    * Initialize the function context container, the standard function contexts
    * and any custom function contexts.
    * @param aClassLoader
    */
   protected void initFunctionContexts( ClassLoader aClassLoader )
   {
      if( getFunctionContextContainer() == null )
      {
         initFunctionContextContainer( aClassLoader );
         initStandardFunctionContexts();
      }
      initCustomFunctionContexts();
   }
   
   /**
    * Initialize the function context container.  If a custom  <code>IAeFunctionContextLocator</code>
    * has been specified in the config, attempt to create it for the container.  If none has
    * been specified or if the creation of the custom impl fails, use the default
    * <code>AeFunctionContextLocator</code>.
    * @param aClassLoader
    */
   protected void initFunctionContextContainer( ClassLoader aClassLoader )
   {
      IAeFunctionContextLocator locator = null;
      String locatorClassName = getEntry( FUNCTION_CONTEXT_LOCATOR_ENTRY );
      try
      {
         if( AeUtil.notNullOrEmpty(locatorClassName) )
         {
            locator = (IAeFunctionContextLocator)Class.forName(locatorClassName).newInstance();
         }
      }
      catch( Throwable t )
      {
         AeException.logError( t, AeMessages.format(ERROR_LOADING_FUNCTION_CONTEXT_LOCATOR, locatorClassName) );
      }
      
      if( locator == null )
      {
         locator = new AeFunctionContextLocator();
      }
      
      if( locator instanceof AeFunctionContextLocator )
      {
         ((AeFunctionContextLocator)locator).setDefaultClassLoader( aClassLoader );
      }
      
      mContextContainer = new AeFunctionContextContainer( locator );
   }
   
   /**
    * Initialize the standard (default, BPEL and BPEL extension) <code>FunctionContext</code>.
    */
   protected void initStandardFunctionContexts()
   {
      initBpelFunctionContext();
      initBpelExtFunctionContext();
   }

   /**
    * Initialize the BPEL function context.  Attempt to load the BPEL 
    * context from the engine config if one was specified, otherwise, use
    * the default (which is <code>AeBpelFunctionContext</code>).
    */
   protected void initBpelFunctionContext()
   {
      Map functionContexts = getMapEntry( STANDARD_FUNCTION_CONTEXTS_ENTRY );
      IAeFunctionContext bpelFc = null;

      if( functionContexts != null )
      {
         Map bpelContextMap = (Map)functionContexts.get(IAeEngineConfiguration.BPEL_FUNCTION_CONTEXT);
         bpelFc = createStandardContext( bpelContextMap );
      }
      
      if( bpelFc == null )
      {
         bpelFc = new AeBpelFunctionContext();
      }
      getFunctionContextContainer().setBpelContext( bpelFc );
   }
   
   /**
    * Initialize the BPEL extension function context.  Attempt to load the 
    * BPEL extension context from the engine config if one was specified, 
    * otherwise, use the default (which is <code>AeExtensionFunctionContext</code>).
    */
   protected void initBpelExtFunctionContext()
   {
      Map functionContexts = getMapEntry( STANDARD_FUNCTION_CONTEXTS_ENTRY );
      IAeFunctionContext bpelExtFc = null;

      if( functionContexts != null )
      {
         Map bpelExtContextMap = (Map)functionContexts.get(IAeEngineConfiguration.BPEL_EXT_FUNCTION_CONTEXT);
         bpelExtFc = createStandardContext( bpelExtContextMap );
      }
   
      if( bpelExtFc == null )
      {
         bpelExtFc = new AeExtensionFunctionContext();
      }
      getFunctionContextContainer().setBpelExtContext( bpelExtFc );
   }

   /**
    * Utility method for creating the function context from the class and
    * location entries in the given map.  Null is returned if the map is null,
    * empty, does not contain a class entry, or if there is a problem creating 
    * the function context class.
    * @param defaultContextMap
    * @param aContainer
    */
   protected IAeFunctionContext createStandardContext( Map defaultContextMap )
   {
      IAeFunctionContext retVal = null;
      
      if( defaultContextMap != null )
      {
         String namespace = (String) defaultContextMap.get(IAeEngineConfiguration.FUNCTION_CONTEXT_NAMESPACE_ENTRY);
         String className = (String) defaultContextMap.get(IAeEngineConfiguration.CLASS_ENTRY);
         String location = (String) defaultContextMap.get(IAeEngineConfiguration.FUNCTION_CONTEXT_CLASSPATH_ENTRY);
         
         try
         {
            if( AeUtil.notNullOrEmpty(className) )
            {
               retVal = getFunctionContextContainer().loadFunctionContext( namespace, location, className );
            }
         }
         catch( AeException ae )
         {
            AeException.logError( ae.getCause(), AeMessages.format(ERROR_LOADING_FUNCTION_CONTEXT_CLASS, className) );
         }
      }
      return retVal;
   }

   /**
    * Initialize any custom <code>FunctionContext<code> impls specified in the config.
    */
   protected void initCustomFunctionContexts()
   {
      getFunctionContextContainer().clearCustomFunctions();
      Map customContexts = getMapEntry(FUNCTION_CONTEXTS_ENTRY);
      if(customContexts != null)
      {
         processFunctionContexts( customContexts );
      }
   }
   
   /**
    * Setter for the logging base dir.
    * @param aBaseDir The base directory for all system log files.
    */
   protected void setLoggingBaseDir( String aBaseDir )
   {
      setEntry( LOGGING_DIR_ENTRY, aBaseDir );
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getLoggingBaseDir()
    */
   public String getLoggingBaseDir()
   {
      return getEntry( LOGGING_DIR_ENTRY, LOGGING_DIR_DEFAULT );
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#isAutoStart()
    */
   public boolean isAutoStart()
   {
      return getBooleanEntry(AUTO_START_ENTRY, true);
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getWsdlCatalogFactoryClassName()
    */
   public String getWsdlCatalogFactoryClassName()
   {
      Map wsdlParams = getMapEntry( WSDL_CATALOG_ENTRY );
      return (String)getEntryInternal(wsdlParams, CLASS_ENTRY, String.class, null);
   }

   /**
    * Sets the description for this configuration.
    */
   public void setDescription(String aDescription)
   {
      setEntry( CONFIG_DESCRIPTION_ENTRY, aDescription );
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getDescription()
    */
   public String getDescription()
   {
      return getEntry( CONFIG_DESCRIPTION_ENTRY, DEFAULT_CONFIG_DESCRIPTION );
   }
   

   //----------[ MUTABLE CONFIGURATION SETTINGS ]-------------------------------
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getUnmatchedCorrelatedReceiveTimeout()
    */
   public int getUnmatchedCorrelatedReceiveTimeout()
   {
      return getIntegerEntry(UNMATCHED_RECEIVE_TIMEOUT_ENTRY, UNMATCHED_RECEIVE_TIMEOUT_DEFAULT);
   }

   /**
    * Sets the timeout value for unmatched correlated receives.
    * @param aTimeoutValue
    */
   public void setUnmatchedCorrelatedReceiveTimeout(int aTimeoutValue)
   {
      setIntegerEntry(UNMATCHED_RECEIVE_TIMEOUT_ENTRY, aTimeoutValue);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getWorkManagerThreadPoolMax()
    */
   public int getWorkManagerThreadPoolMax()
   {
      Map workMgrMap = getMapEntry(WORK_MANAGER_ENTRY);
      return getIntegerEntryInternal( workMgrMap, WORKMANAGER_THREAD_MAX_ENTRY, WORKMANAGER_THREAD_MAX_DEFAULT );
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getWorkManagerThreadPoolMin()
    */
   public int getWorkManagerThreadPoolMin()
   {
      Map workMgrMap = getMapEntry(WORK_MANAGER_ENTRY);
      return getIntegerEntryInternal( workMgrMap, WORKMANAGER_THREAD_MIN_ENTRY, WORKMANAGER_THREAD_MIN_DEFAULT );
   }
   
   /**
    * Setter for the work manager thread pool max thread count
    * @param aInt
    */
   public void setWorkManagerThreadPoolMax(int aInt)
   {
      Map workMgrMap = getMapEntry(WORK_MANAGER_ENTRY, true);
      workMgrMap.put(WORKMANAGER_THREAD_MAX_ENTRY, Integer.toString(aInt));
   }

   /**
    * Setter for the work manager thread pool min thread count
    * @param aInt
    */
   public void setWorkManagerThreadPoolMin(int aInt)
   {
      Map workMgrMap = getMapEntry(WORK_MANAGER_ENTRY, true);
      workMgrMap.put(WORKMANAGER_THREAD_MIN_ENTRY, Integer.toString(aInt));
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#allowEmptyQuerySelection()
    */
   public boolean allowEmptyQuerySelection()
   {
      return getBooleanEntry(ALLOW_EMPTY_QUERY_SELECTION_ENTRY, false);
   }

   /**
    * Sets the allow empty query selection results on or off.
    */
   public void setAllowEmptyQuerySelection(boolean aBoolean)
   {
      setBooleanEntry(ALLOW_EMPTY_QUERY_SELECTION_ENTRY, aBoolean);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#allowCreateXPath()
    */
   public boolean allowCreateXPath()
   {
      return getBooleanEntry(ALLOW_CREATE_XPATH_ENTRY, false);
   }

   /**
    * Sets the allow create xpath option on or off.
    */
   public void setAllowCreateXPath(boolean aBoolean)
   {
      setBooleanEntry(ALLOW_CREATE_XPATH_ENTRY, aBoolean);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#validateServiceMessages()
    */
   public boolean validateServiceMessages()
   {
      return getBooleanEntry(VALIDATE_SERVICE_MESSAGES_ENTRY, false);
   }

   /**
    * Sets whether inbound and outbound messages are validated.
    */
   public void setValidateServiceMessages(boolean aBoolean)
   {
      setBooleanEntry(VALIDATE_SERVICE_MESSAGES_ENTRY, aBoolean);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#isLoggingEnabled()
    */
   public boolean isLoggingEnabled()
   {
      return getBooleanEntry(LOGGING_ENTRY, false);
   }

   /**
    * Setter for logging property.
    * @param aEnabled
    */
   public void setLoggingEnabled(boolean aEnabled)
   {
      setBooleanEntry(LOGGING_ENTRY, aEnabled);
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#setWsdlCacheMax(int)
    */
   public void setWsdlCacheMax(int aMax)
   {
      Map wsdlParams = getMapEntry(WSDL_CATALOG_ENTRY);
      wsdlParams.put( WSDL_FACTORY_CACHE_SIZE_ENTRY, String.valueOf(aMax) );      

   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getWsdlCacheMax()
    */
   public int getWsdlCacheMax()
   {
      Map wsdlParams = getMapEntry( WSDL_CATALOG_ENTRY );
      return getIntegerEntryInternal( wsdlParams, WSDL_FACTORY_CACHE_SIZE_ENTRY, DEFAULT_WSDL_CACHE_MAX );
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#isAllowedRolesEnforced()
    */
   public boolean isAllowedRolesEnforced()
   {
      return getBooleanEntry(ALLOWED_ROLES_ENFORCED, true);
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#setAllowedRolesEnforced(boolean)
    */
   public void setAllowedRolesEnforced(boolean aFlag)
   {
      setBooleanEntry(ALLOWED_ROLES_ENFORCED, aFlag);
   }
   
   //----------[ CHANGE LISTENER METHODS ]--------------------------------------
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#update()
    */
   public void update()
   {
      notifyListeners();
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#addConfigChangeListener(org.activebpel.rt.bpel.config.IAeConfigChangeListener)
    */
   public synchronized void addConfigChangeListener(IAeConfigChangeListener aListener)
   {
      mListeners.add( aListener );
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#removeConfigChangeListener(org.activebpel.rt.bpel.config.IAeConfigChangeListener)
    */
   public synchronized void removeConfigChangeListener(IAeConfigChangeListener aListener)
   {
      mListeners.remove(aListener);
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#addStorageChangeListener(org.activebpel.rt.bpel.config.IAeStorageChangeListener)
    */
   public synchronized void addStorageChangeListener(IAeStorageChangeListener aListener)
   {
      mStorageListeners.add( aListener );
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#removeStorageChangeListener(org.activebpel.rt.bpel.config.IAeStorageChangeListener)
    */
   public synchronized void removeStorageChangeListener(IAeStorageChangeListener aListener)
   {
      mStorageListeners.remove(aListener);
   }

   /**
    * Notify any registered listeners of a config change.
    */
   protected void notifyListeners()
   {
      ArrayList listeners = null;
      synchronized( mListeners )
      {
         if( !mListeners.isEmpty() )
         {
            listeners = (ArrayList)mListeners.clone();
         }
      }
      
      if( listeners != null )
      {
         for( Iterator iter = listeners.iterator(); iter.hasNext(); )
         {
            IAeConfigChangeListener listener = (IAeConfigChangeListener) iter.next();
            listener.updateConfig( this );
         }
      }
   }
   
   /**
    * Walk all of the listeners and notify them of the change.
    * 
    * @param aMap
    */
   protected void notifyStorageListeners(Map aMap)
   {
      ArrayList listeners = null;
      synchronized( mStorageListeners )
      {
         if( !mStorageListeners.isEmpty() )
         {
            listeners = (ArrayList)mStorageListeners.clone();
         }
      }
      
      if( listeners != null )
      {
         for( Iterator iter = listeners.iterator(); iter.hasNext(); )
         {
            IAeStorageChangeListener listener = (IAeStorageChangeListener) iter.next();
            listener.storageConstantsChanged( aMap );
         }
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getUpdatableEngineConfig()
    */
   public IAeUpdatableEngineConfig getUpdatableEngineConfig()
   {
      return this;
   }
   
   /**
    * Accessor for <code>AeFunctionContextContainer</code>.
    */
   protected AeFunctionContextContainer getFunctionContextContainer()
   {
      return mContextContainer;
   }
   
   /**
    * Setter for top level entries map.
    * @param aEntries
    */
   protected void setEntries( HashMap aEntries )
   {
      mEntries = aEntries;
   }
   
   /**
    * Accessor for entries map.
    */
   protected Map getEntries()
   {
      return mEntries;
   }
   
   /**
    * This clone simply returns a config that contains the same map entries. It
    * does not include the listeners or the function context container.
    * 
    * @see java.lang.Object#clone()
    */
   public Object clone()
   {
      AeDefaultEngineConfiguration copy = new AeDefaultEngineConfiguration();
      
      HashMap map = new HashMap(getEntries());
      
      copy.setEntries(map);
      
      return copy;
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#isWsdlReplaceEnabled()
    */
   public boolean isWsdlReplaceEnabled()
   {
      Map wsdlParams = getMapEntry( WSDL_CATALOG_ENTRY );
      return getBooleanEntryInternal( wsdlParams, WSDL_REPLACE_EXISTING_ENTRY, false );
   }
   
   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#setWsdlReplaceEnabled(boolean)
    */
   public void setWsdlReplaceEnabled(boolean aFlag)
   {
      Map wsdlParams = getMapEntry(WSDL_CATALOG_ENTRY);
      wsdlParams.put( WSDL_REPLACE_EXISTING_ENTRY, Boolean.toString(aFlag) );      
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#isSuspendProcessOnUncaughtFault()
    */
   public boolean isSuspendProcessOnUncaughtFault()
   {
      return getBooleanEntry(SUSPEND_PROCESS_ENTRY, false);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#setSuspendProcessOnUncaughtFault(boolean)
    */
   public void setSuspendProcessOnUncaughtFault(boolean aFlag)
   {
      setBooleanEntry(SUSPEND_PROCESS_ENTRY, aFlag);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeEngineConfiguration#getProcessWorkCount()
    */
   public int getProcessWorkCount()
   {
      Map map = getMapEntry(PROCESS_WORK_MANAGER_ENTRY);
      return getIntegerEntryInternal(map, PROCESS_WORK_COUNT_ENTRY, PROCESS_WORK_COUNT_DEFAULT);
   }

   /**
    * @see org.activebpel.rt.bpel.config.IAeUpdatableEngineConfig#setProcessWorkCount(int)
    */
   public void setProcessWorkCount(int aProcessWorkCount)
   {
      Map map = getMapEntry(PROCESS_WORK_MANAGER_ENTRY);
      map.put(PROCESS_WORK_COUNT_ENTRY, String.valueOf(aProcessWorkCount));
   }

   /**
    * This method takes a configuration map for a config class and instantiates that
    * class.  This involves some simple java reflection to find the proper
    * constructer and then calling that constructor.  Any class can use this method, 
    * as long as the class in question has a constructor that takes the configuration
    * map.  Returns null if no entry is found.
    * 
    * @param aConfigName The name of a MapEntry in the config.
    * @throws AeException
    * @return A config class.
    */
   public Object createConfigSpecificClass(String aConfigName) throws AeException
   {
      return createConfigSpecificClass(getMapEntry(aConfigName));
   }

   /**
    * This method takes a configuration map for a config class and instantiates that
    * class.  This involves some simple java reflection to find the proper
    * constructer and then calling that constructor.  Any class can use this method, 
    * as long as the class in question has a constructor that takes the configuration
    * map.  Returns null if no config is specified.
    * 
    * @param aConfig The engine configuration map for the class.
    * @return A config class.
    */
   public static Object createConfigSpecificClass(Map aConfig) throws AeException
   {
      if (AeUtil.isNullOrEmpty(aConfig))
      {
         return null;
      }
      String className = (String) aConfig.get("Class"); //$NON-NLS-1$
      if (className == null)
      {
         throw new AeException(AeMessages.getString("AeDefaultEngineConfiguration.NO_CLASS_ERROR")); //$NON-NLS-1$
      }

      try
      {
         Class clazz = Class.forName(className);
         Constructor cons = clazz.getConstructor(new Class[] { Map.class });
         return cons.newInstance(new Object[] { aConfig });
      }
      catch (Exception e)
      {
         AeException.logError( e, AeMessages.getString("AeDefaultEngineConfiguration.INSTANTIATION_ERROR") + className ); //$NON-NLS-1$
         throw new AeException(AeMessages.getString("AeDefaultEngineConfiguration.INSTANTIATION_ERROR")+ className, e); //$NON-NLS-1$
      }
   }
}
