/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.config;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.activebpel.rt.AeException;

/**
 * This class implements an engine configuration that saves/loads a list of engine config differences rather
 * than an entire engine config document. It is loaded first with a base config file, then a set of
 * differences are applied to that base to get the final config. When saving the config, only those settings
 * that are different from the base are saved.
 * 
 * TODO (EPW) We should use this as a base for the AeStorageAwareEngineConfig.
 */
public abstract class AeDiffEngineConfiguration extends AeDefaultEngineConfiguration
{
   /** Copy of the original config entries. */
   protected HashMap mOriginalParams;

   /**
    * Constructor.
    */
   public AeDiffEngineConfiguration()
   {
      super();
   }

   /**
    * This method adds a map of "diff" config settings to the engine config.  The map will be
    * from engine config path (pseudo-xpath) to config value.
    * 
    * @param aDiffSettings
    */
   public void addConfigSettingsFromDiff(Map aDiffSettings)
   {
      if (!aDiffSettings.isEmpty())
      {
         syncInternalEntries(aDiffSettings);
         resetFunctionContexts();
         notifyListeners();
      }
   }

   /**
    * Add any customized entries to the map.
    * 
    * @param aConfigDiff
    * @param aEntries
    */
   protected void syncInternalEntries(Map aConfigDiff)
   {
      for (Iterator iter = aConfigDiff.entrySet().iterator(); iter.hasNext(); )
      {
         Map.Entry entry = (Map.Entry) iter.next();
         String pathToSetting = (String) entry.getKey();
         String customSetting = (String) entry.getValue();
         addCustomSetting(pathToSetting, customSetting, getEntries());
      }
   }

   /**
    * Descend into the config maps for the given path to set the custom value.
    * 
    * @param aPath The path to the map. The path format will be something like foo/bar/debug where foo and bar
    *           are key names for map and debug is the key that will be added to the bar map (with the setting
    *           as its value).
    * @param aSetting The value of the custom param.
    */
   protected void addCustomSetting(String aPath, String aSetting, Map aEntries)
   {
      AeCustomSetting setting = new AeCustomSetting(aPath, aSetting);
      Map parentMap = aEntries;
      Map currentMap = parentMap;

      for (Iterator iter = setting.getMapKeyIterator(); iter.hasNext();)
      {
         String mapKey = (String) iter.next();
         currentMap = (Map) parentMap.get(mapKey);
         if (currentMap == null)
         {
            currentMap = new HashMap();
            parentMap.put(mapKey, currentMap);
            break;
         }
         else
         {
            parentMap = currentMap;
         }
      }

      currentMap.put(setting.getKey(), setting.getValue());
   }

   /**
    * Clear current function contexts from the container and then reload them from the map.
    */
   protected void resetFunctionContexts()
   {
      getFunctionContextContainer().clearCustomFunctions();
      Map functionContexts = getMapEntry(FUNCTION_CONTEXTS_ENTRY);
      if (functionContexts != null)
      {
         processFunctionContexts(functionContexts);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.config.AeDefaultEngineConfiguration#update()
    */
   public void update()
   {
      try
      {
         Map customSettings = extractConfigDif();
         updateConfigData(customSettings);
      }
      catch (AeException se)
      {
         se.logError();
      }
      super.update();
   }

   /**
    * Extract the diff.
    */
   protected Map extractConfigDif()
   {
      Map source = getOriginalParams();
      Map target = getEntries();
      return AeConfigDifUtil.compare(source, target);
   }

   /**
    * @see org.activebpel.rt.bpel.config.AeDefaultEngineConfiguration#setEntries(java.util.HashMap)
    */
   protected void setEntries(HashMap aMap)
   {
      super.setEntries(aMap);

      // keep a clean copy of the original entries
      if (getOriginalParams() == null)
      {
         mOriginalParams = new HashMap();
         copy(getEntries(), mOriginalParams);
      }
   }

   /**
    * Utility copy method. Recurses into the source map if the value itself is a <code>Map</code>
    * 
    * @param aSource
    * @param aTarget
    */
   protected void copy(Map aSource, Map aTarget)
   {
      for (Iterator iter = aSource.entrySet().iterator(); iter.hasNext();)
      {
         Map.Entry entry = (Map.Entry) iter.next();
         Object key = entry.getKey();
         Object value = entry.getValue();

         if (value instanceof Map)
         {
            Map newCopy = new HashMap();
            copy((Map) value, newCopy);
            aTarget.put(key, newCopy);
         }
         else
         {
            aTarget.put(key, value);
         }
      }
   }

   /**
    * Accessor for original params. These remain "untouched" and are used as a reference to determine what has
    * changed during updates.
    */
   protected Map getOriginalParams()
   {
      return mOriginalParams;
   }

   /**
    * Subclasses must implement this method when
    * 
    * @param aCustomSettings
    * @throws AeException
    */
   protected abstract void updateConfigData(Map aCustomSettings) throws AeException;

   /**
    * Utility class for wrapping the config path and value.
    */
   protected class AeCustomSetting
   {
      /** The ordered list of map keys. */
      protected List mMapKeys;
      /** The setting name. */
      protected String mName;
      /** The setting value. */
      protected String mValue;
      
      /**
       * Constructor.
       * @param aPath
       * @param aValue
       */
      public AeCustomSetting( String aPath, String aValue )
      {
         initKeys( aPath );
         mValue = aValue;
      }
      
      /**
       * Parse the path.  A given path of foo/bar/debug
       * results in a list with the [foo,bar] entries
       * and the key property is set to debug.
       * @param aPath
       */
      protected void initKeys( String aPath )
      {
         StringTokenizer st = new StringTokenizer( aPath, "/" ); //$NON-NLS-1$
         int count = st.countTokens();
         int currentIndex = 0;
         ArrayList mapKeys = new ArrayList();
         while( st.hasMoreTokens() )
         {
            ++currentIndex;
            String pathElement = st.nextToken();
            if( currentIndex < count )
            {
               mapKeys.add( pathElement );
            }
            else
            {
               mName = pathElement;
            }
         }
         mMapKeys = mapKeys;
      }
      
      /**
       * Accessor for iterator over the map key list.
       */
      public Iterator getMapKeyIterator()
      {
         return mMapKeys.iterator();
      }
      
      /**
       * Key for the custom setting.
       */
      public String getKey()
      {
         return mName;
      }
      
      /**
       * Value of the custom setting.
       */
      public String getValue()
      {
         return mValue;
      }
   }
}
