// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/AeActivityDef.java,v 1.8 2005/06/08 12:50:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.activebpel.rt.bpel.def.activity.support.AeSourceDef;
import org.activebpel.rt.bpel.def.activity.support.AeTargetDef;


/**
 * Definition for bpel base activity definition.
 */
public abstract class AeActivityDef extends AeNamedDef
{
   // standard attributes of the process
   private String  mJoinCondition;
   private Boolean mSuppressFailure;

   // standard elements of the activity definition
   private ArrayList mSources;  // List of links where this is a source activity   
   private ArrayList mTargets;  // List of links where this is a target activity
   private ArrayList mTargetLinkNames;
   
   /** set of variables that this activity uses. Used in conjunction with serializable scopes and variable locking. */
   private Set mVariablesUsed;

   /**
    * Default constructor
    */
   public AeActivityDef()
   {
   }

   /**
    * Provides the ability to add a source link to an activity.
    * 
    * @param aSource source link to be added
    */
   public void addSource(AeSourceDef aSource)
   {
      if (mSources == null)
         mSources = new ArrayList();
      
      mSources.add(aSource);
   }

   /**
    * Provides the ability to delete a source link from an activity.
    * 
    * @param aSource the copy element to be removed
    * @return boolean status indicating if element was removed
    */
   public boolean removeSource(AeSourceDef aSource)
   {
      boolean found = false;
      for (Iterator iter = getSourceList(); iter.hasNext() && !found;)
      {
         if (aSource == iter.next())
         {
            iter.remove();
            found = true;
         }
      }
      
      return found;
   }

   /**
    * Provides the ability to remove all source references from this activity.
    */
   protected void clearSourceList()
   {
      mSources = null;
   }

   /**
    * Provide a list of the source links for this activity.
    * 
    * @return iterator over source link list
    */
   public Iterator getSourceList()
   {
		if (mSources == null || mSources.size() == 0)
		   return Collections.EMPTY_LIST.iterator();
		else
         return mSources.iterator();
   }

   /**
    * Provides the ability to add a target link to an activity.
    * 
    * @param aTarget target link to be added
    */
   public void addTarget(AeTargetDef aTarget)
   {
      if (mTargets == null)
         mTargets = new ArrayList();
      
      mTargets.add(aTarget);
   }

   /**
    * Provides the ability to delete a target link from an activity.
    * 
    * @param aTarget the target link to be removed
    * @return boolean status indicating if element was removed
    */
   public boolean removeTarget(AeTargetDef aTarget)
   {
      boolean found = false;
      for (Iterator iter = getSourceList(); iter.hasNext() && !found;)
      {
         if (aTarget == iter.next())
         {
            iter.remove();
            found = true;
         }
      }
      
      return found;
   }

   /**
    * Provides the ability to remove all target references from this activity.
    */
   protected void clearTargetList()
   {
      mTargets = null;
   }
   
   /**
    * Provide a list of the target links for this activity.
    * 
    * @return iterator over target link list
    */
   public Iterator getTargetList()
   {
      if (mTargets == null || mTargets.size() == 0)
         return Collections.EMPTY_LIST.iterator();
      else
         return mTargets.iterator();
   }

   /**
    * Gets a list of the targets.
    */
   public List getTargetLinkNames()
   {
      if (mTargetLinkNames == null)
      {
         ArrayList list = new ArrayList();
         for (Iterator titer = getTargetList(); titer.hasNext(); )
         {
            AeTargetDef target = (AeTargetDef) titer.next();
            list.add(target.getLinkName());
         }
         mTargetLinkNames = list;
      }
      return mTargetLinkNames;
   }

   /**
    * Accessor method to obtain joinCondition property of this object.
    * 
    * @return name of object
    */
   public String getJoinCondition()
   {
      return mJoinCondition;
   }

   /**
    * Mutator method to set the joinCondition of this object.
    * 
    * @param aJoinCondition of object
    */
   public void setJoinCondition(String aJoinCondition)
   {
      mJoinCondition = aJoinCondition;
   }

   /**
    * Accessor method to obtain suppressFailure property of this object.
    * 
    * @return suppressFailure flag (true="yes", false="no)
    */
   public Boolean getSuppressFailure()
   {
      return mSuppressFailure;
   }

   /**
    * Mutator method to set the suppressFailure property of this object.
    * 
    * @param aSuppressFailure (true="yes", false="no)
    */
   public void setSuppressFailure(Boolean aSuppressFailure)
   {
      mSuppressFailure = aSuppressFailure;
   }

   /**
    * Method which may be taken over by inheriting classes to wrap themselves
    * within a scope if deemed necessary. Default implementation returns this.
    * @return activity definition object
    */
   public AeActivityDef getScopedActivity()
   {
      return this;
   }
   
   /**
    * Getter for the variables used by this activity.
    */
   public Set getVariablesUsed()
   {
      return mVariablesUsed;
   }

   /**
    * Setter for the variables used. 
    * @param aSet
    */
   public void setVariablesUsed(Set aSet)
   {
      mVariablesUsed = aSet;
   }
}   
