// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/AeBaseDef.java,v 1.27 2005/06/08 12:50:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.activebpel.rt.bpel.def.io.ext.IAeExtensionElement;
import org.activebpel.rt.bpel.def.visitors.IAeDefVisitor;
import org.activebpel.rt.util.AeUtil;

/**
 * Base definition object for bpel definitions
 */
public abstract class AeBaseDef implements IAeBPELConstants
{
   /** A list of all referenced namespaces referenced at the current node level */
   protected HashMap mNamespaceMap;
   
   /** Path that uniquely identifies this element in the BPEL Process hierarchy. */
   private String mLocationPath ;

   /** Unique location identifier for this element in the BPEL Process hierarchy. */
   private int mLocationId ;

   /** Optional bpel entity comment */
   private String  mComment;
   
   /** Parent def, if any. */
   private AeBaseDef mParent;

   /** A list of extension elements */   
   private List mExtensions;

   /**
    * Default constructor
    */
   public AeBaseDef()
   {
   }
   
   /**
    * Set this instance's parent def.
    * 
    * @param aParent The parent to set.
    */
   public void setParent( AeBaseDef aParent )
   {
      mParent = aParent ;
   }
   
   /**
    * Get this instance's parent def.<p>
    * 
    * N.B.: the AeDefAssignParentVisitor must be used before assuming that
    * parent-child def relationships are valid.  These relationships are NOT
    * updated automatically (e.g., during modification by the process design user,
    * etc.).
    * 
    * @return AeBaseDef if parent exists, null otherwise.
    */
   public AeBaseDef getParent()
   {
      return mParent ;
   }
   
   /**
    * Provides the ability to add a namespace to the list.
    * 
    * @param aName  name of the namespace to be added 
    * @param aValue value of the namespace to be added
    */
   public void addNamespace(String aName, String aValue)
   {
      if (mNamespaceMap == null)
      	mNamespaceMap = new HashMap();
         
      mNamespaceMap.put(aName, aValue);
   }

   /**
    * Provides the ability to remove a namespace element from the activity.
    * 
    * @param aName the namespace to be removed
    * @return boolean status indicating if element was removed
    */
   public boolean removeNamespace(String aName)
   {
      boolean found = false;
      
      if (mNamespaceMap != null)
      	found = (mNamespaceMap.remove(aName) != null);
      
      return found;
   }

   /**
    * Provides the ability to find a namespace definition.
    * 
    * @param aName the namespace to search for.
    * @return the value of the namespace if found or null if not found
    */
   public String findNamespace(String aName)
   {
      String namespace = null;
      
      if (mNamespaceMap != null)
         namespace = (String)mNamespaceMap.get(aName);
      
      return namespace;
   }
   
   /**
    * Provide a list of the Namespaces for the user to iterate.
    * 
    * @return iterator of String objects 
    */
   public Iterator getNamespaceList()
   {
      if (mNamespaceMap == null)
         return Collections.EMPTY_LIST.iterator();
      else
         return mNamespaceMap.keySet().iterator();
   }


   /**
    * Returns a list of the namespace prefixes.
    */
   public Iterator getNamespacePrefixList()
   {
      if (mNamespaceMap == null)
      {
         return Collections.EMPTY_LIST.iterator();
      }
      List prefixes = new ArrayList();
      for (Iterator iter = mNamespaceMap.keySet().iterator(); iter.hasNext(); )
      {
         String prefix = (String) iter.next();
         if (AeUtil.notNullOrEmpty(prefix))
         {
            if (prefix.startsWith("xmlns:")) //$NON-NLS-1$
            {
               prefixes.add(prefix.substring(6));
            }
         }
      }
      return prefixes.iterator();
   }

   /**
    * Gets a set of prefixes mapped to the given namespace.
    * 
    * @param aNamespace
    */
   public Set getPrefixesForNamespace(String aNamespace)
   {
      HashSet set = new HashSet();
      if (mNamespaceMap != null && !AeUtil.isNullOrEmpty(aNamespace))
      {
         for (Iterator iter = mNamespaceMap.entrySet().iterator(); iter.hasNext(); )
         {
            Map.Entry entry = (Map.Entry) iter.next();
            String prefix = (String) entry.getKey();
            String ns = (String) entry.getValue();
            if (aNamespace.equals(ns) && AeUtil.notNullOrEmpty(prefix) && prefix.startsWith("xmlns:")) //$NON-NLS-1$
            {
               set.add(prefix.substring(6));
            }
         }
      }
      return set;
   }
   
   abstract public void accept(IAeDefVisitor aVisitor);
   
   /**
    * Get the unique location path for this element.
    * 
    * @return String
    */
   public String getLocationPath()
   {
      return mLocationPath;
   }

   /**
    * Set the unique location path for this element.
    * 
    * @param aString
    */
   public void setLocationPath(String aString)
   {
      mLocationPath = aString;
   }

   /**
    * Get the unique location id for this element.
    * 
    * @return int
    */
   public int getLocationId()
   {
      return mLocationId;
   }

   /**
    * Set the unique location id for this element.
    * 
    * @param aLocationId
    */
   public void setLocationId(int aLocationId)
   {
      mLocationId = aLocationId;
   }

   /**
    * Gets the BPEL comment.
    */
   public String getComment()
   {
      return mComment;
   }

   /**
    * Sets the BPEL comment.
    * @param aString
    */
   public void setComment(String aString)
   {
      mComment = aString;
   }

   /**
    * Mutator for adding an extension element.
    * @param aExtension the IAeExtensionElement impl to be added
    */
   public void addExtension( IAeExtensionElement aExtension )
   {
      if( mExtensions == null )
      {
         mExtensions = new ArrayList();      
      }
      mExtensions.add( aExtension );
   }
   
   /**
    * Returns an iterator for any extension elements
    * (or any empty iterator if none are present).
    * @return iterator over any extension element
    */
   public Iterator getExtensions()
   {
      if( mExtensions == null )
      {
         return Collections.EMPTY_LIST.iterator();
      }
      return mExtensions.iterator();
   }
}
