// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/AeProcessDef.java,v 1.43 2006/02/07 23:48:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeContextWSDLProvider;
import org.activebpel.rt.bpel.IAeExpressionLanguageFactory;
import org.activebpel.rt.bpel.def.visitors.AeDefAssignParentVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefCorrelatedReceiveVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefEntryPointInitialVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefEntryPointPropertiesVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefOneWayVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefPartnerLinkTypeVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefPathVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefVariableTypeVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefVariableUsageVisitor;
import org.activebpel.rt.bpel.def.visitors.AeInlinePropertyAliasVisitor;
import org.activebpel.rt.bpel.def.visitors.AeMessageExchangeVisitor;
import org.activebpel.rt.bpel.def.visitors.IAeDefVisitor;
import org.activebpel.rt.util.AeIntMap;
import org.activebpel.rt.wsdl.def.IAePropertyAlias;


/**
 * Definition for bpel process
 */
public class AeProcessDef extends AeScopeDef
{
   /** True if the definition has been preprocessed. */
   private boolean mPreProcessed;
   /** The target namespace of the process. */
   private String mTargetNamespace;
   /** The query language override. */
   private String mQueryLanguage;
   /** The expression language override. */
   private String mExpressionLanguage;
   /** The suppress join failure override. */
   private boolean mSuppressJoinFailure;
   /** True if this process was designed to enable instance compensation after completion. */
   private boolean mEnableInstanceCompensation;
   /** True if this is an abstract (non-executable) process. */
   private boolean mAbstractProcess;
   /** Comment associated with process. */
   private String mComment = ""; //$NON-NLS-1$
   /** Built-in comment from bpel editor. */
   private String mProcessPreambleComment = ""; //$NON-NLS-1$
   /** Partner links container. */
   private AePartnerLinksContainerDef mPartnerLinks = new AePartnerLinksContainerDef();
   /** Partners container. */
   private AePartnersContainerDef mPartners = new AePartnersContainerDef();
   /** Imports container. */
   private AeImportsContainerDef mImports = new AeImportsContainerDef();
   /** List of create instance activities. */
   private Collection mCreateInstances;
   /** Map of partner link properties. */
   private Map mPartnerLinkProperties;
   /** Process fully qualified name. */
   private QName mProcessName;
   /** indicates that this process has multiple start activities */
   private boolean mMultiStart;
   /** indicates that this process contains serializable scopes */
   private boolean mContainsSerializableScopes;
   /** cache of property names and message types to property aliases */
   private Map mPropertyAliasMap = new HashMap();
   /** cache of property names to property types. */
   private Map mPropertyTypeMap = new HashMap();
   /** Maps location paths to location ids */
   private final Map mLocationPathsToIds = new HashMap();
   /** Maps location ids to location paths */
   private final AeIntMap mLocationIdsToPaths = new AeIntMap();
   /** Collection of port types plus operations that are one-way receives. */
   private Set mOneWayReceives = new HashSet();
   /** Records the largest value used for the location id. */
   /** maps partner link name and operation to a list of property sets that can be in the request */
   private Map mCorrelationPropertiesMap = new HashMap();
   /** map of create instance activities and their message exchange values (if they're not-null) */
   private Map mMessageExchangeMap = new HashMap();
   
   /**
    * Default constructor
    */
   public AeProcessDef()
   {
   }
   
   /**
    * Adds the port type and operation to the collection of one way receives.
    * 
    * @param aPartnerLinkType
    * @param aOperation
    */
   public void addOneWayReceive(QName aPartnerLinkType, String aOperation)
   {
      mOneWayReceives.add(makeOneWayReceiveKey(aPartnerLinkType, aOperation));
   }
   
   /**
    * Returns true if the partner link's port type and operation is a one-way receive.
    * 
    * @param aPartnerLinkName
    * @param aOperation
    */
   public boolean isOneWayReceive(String aPartnerLinkName, String aOperation)
   {
      QName plinkDef = getPartnerLink(aPartnerLinkName).getPartnerLinkTypeName();
      return mOneWayReceives.contains(makeOneWayReceiveKey(plinkDef, aOperation));
   }
   
   /**
    * Makes a key from the concatenation of the port type and the operation.
    * 
    * @param aPartnerLinkType
    * @param aOperation
    */
   protected Object makeOneWayReceiveKey(QName aPartnerLinkType, String aOperation)
   {
      return aPartnerLinkType + aOperation;
   }
   
   /**
    * Caches a property alias with the process def for faster lookup at runtime.
    * @param aPropAlias
    */
   public void cachePropertyAlias(IAePropertyAlias aPropAlias)
   {
      mPropertyAliasMap.put(makeKey(aPropAlias), aPropAlias);
   }
   
   /**
    * Caches a property type.
    * 
    * @param aPropertyName The name of the property.
    * @param aPropertyType The type of the property.
    */
   public void cachePropertyType(QName aPropertyName, QName aPropertyType)
   {
      mPropertyTypeMap.put(aPropertyName, aPropertyType);
   }
   
   /**
    * Retrieves a previously cached property alias by its message type and its
    * property name.
    * @param aMessageType The message type for the property alias
    * @param aPropertyName The property name for the property alias
    */
   public IAePropertyAlias getPropertyAlias(QName aMessageType, QName aPropertyName)
   {
      return (IAePropertyAlias) mPropertyAliasMap.get(aMessageType.toString() + aPropertyName);
   }

   /**
    * Retrieves a previously cached property type by its property name.
    * @param aPropertyName The property name.
    * @return The property type.
    */
   public QName getPropertyType(QName aPropertyName)
   {
      return (QName) mPropertyTypeMap.get(aPropertyName);
   }
  
   /**
    * Makes a key for the property alias so it can be stored in our map.
    * @param aPropAlias
    */
   private Object makeKey(IAePropertyAlias aPropAlias)
   {
      return aPropAlias.getMessageType().toString() + aPropAlias.getPropertyName().toString(); 
   }

   /**
    * Setter for the multi-start field
    * @param aFlag
    */
   public void setMultiStart(boolean aFlag)
   {
      mMultiStart = aFlag;
   }
   
   /**
    * Getter for the multi start field
    */
   public boolean isMultiStart()
   {
      return mMultiStart;
   }

   /**
    * This is a post load step that assigns parents to all child defs and also gives
    * each child path a unique path.
    * @throws AeBusinessProcessException
    */
   public void processAssignParentAndLocationPathVisitors() throws AeBusinessProcessException
   {      
      // assign parents for definitions
      AeDefAssignParentVisitor parentVisitor = new AeDefAssignParentVisitor();
      parentVisitor.visit( this );

      // assign location paths for definitions
      AeDefPathVisitor pathVisitor = new AeDefPathVisitor();
      pathVisitor.visit(this);

      // populate bidirectional maps between location paths and location ids
      for (Iterator i = pathVisitor.getLocationPaths().iterator(); i.hasNext(); )
      {
         String locationPath = (String) i.next();
         int locationId = pathVisitor.getLocationId(locationPath);
         mLocationPathsToIds.put(locationPath, new Integer(locationId));
         mLocationIdsToPaths.put(locationId, locationPath);
      }      
   }
   
   /**
    * Runs all of the pre-processing that we need to do on the def objects. This
    * includes the creation of the xpaths on the def, the variable usage for 
    * serializable scopes and discovering type information for variables.
    * @param aProvider a WSDL provider used to located WSDL definitions
    * @param aExpressionLanguageFactory the expression language factory to use
    */
   public void preProcessDefinition( IAeContextWSDLProvider aContextProvider, IAeExpressionLanguageFactory aExpressionLanguageFactory ) 
      throws AeBusinessProcessException
   {

      // assign parent and build location paths
      processAssignParentAndLocationPathVisitors();
      
      // assign variable usages for definitions
      AeDefVariableUsageVisitor.createVariableUsageSets(this, aExpressionLanguageFactory);

      // assign variable type information for variables
      AeDefVariableTypeVisitor varVisitor = new AeDefVariableTypeVisitor(aContextProvider);
      varVisitor.findTypeInfo(this);

      // NOTE: parent assignments must be run before this
      // examine all receives and picks (for onMessages) 
      // whose createInstance property is set to true
      // list contents: partnerLinkName:portTypeQName:operationName
      AeDefEntryPointInitialVisitor entryPtVisitor = new AeDefEntryPointInitialVisitor();
      entryPtVisitor.visit( this );
      mCreateInstances = entryPtVisitor.getCreateInstanceCollection();
      setMultiStart(entryPtVisitor.getCreateInstanceActivityCount() > 1);

      AeDefOneWayVisitor oneWayReceiveVisitor = new AeDefOneWayVisitor(aContextProvider);
      oneWayReceiveVisitor.visit(this);

      AeDefCorrelatedReceiveVisitor duplicateReceiveVisitor = new AeDefCorrelatedReceiveVisitor();
      duplicateReceiveVisitor.visit(this);

      // NOTE: parent assignments must be run before this
      // build map for all properties
      // values are the set of properties
      // key is partnerLinkName:portTypeQName:operationName
      AeDefEntryPointPropertiesVisitor propsVisitor = new AeDefEntryPointPropertiesVisitor();
      propsVisitor.visit( this );
      mPartnerLinkProperties = propsVisitor.getPropertiesMap();
      
      // this visitor caches the property aliases so we don't have to resolve at runtime
      AeInlinePropertyAliasVisitor propAliasViz = new AeInlinePropertyAliasVisitor(aContextProvider,
            aExpressionLanguageFactory);
      accept(propAliasViz);
      
      // populate the partner link type objects on the partner link defs
      AeDefPartnerLinkTypeVisitor plTypeViz = new AeDefPartnerLinkTypeVisitor(aContextProvider);
      plTypeViz.visit(this);
      
      // populate message exchange map
      AeMessageExchangeVisitor msgExchgVisitor = new AeMessageExchangeVisitor();
      msgExchgVisitor.visit(this);
      
      // set that we are now preprocessed
      setPreProcessed(true);
   }
   
   /**
    * Gets the qname for this process. 
    */
   public QName getQName()
   {
      if( mProcessName == null )
      {
         mProcessName = new QName(getTargetNamespace(), getName()); 
      }
      return  mProcessName;
   }

   /**
    * Accessor method to obtain the target namespace for this process.
    * 
    * @return target namespace of the process
    */
   public final String getTargetNamespace()
   {
      return mTargetNamespace;
   }

   /**
    * Mutator method to set the target namespace for this process.
    * 
    * @param aTargetNamespace the target namespace for this process
    */
   public final void setTargetNamespace(String aTargetNamespace)
   {
      mTargetNamespace = aTargetNamespace;
   }

   /**
    * Accessor method to obtain the query language for this process.
    * 
    * @return query language for this process
    */
   public final String getQueryLanguage()
   {
      return mQueryLanguage;
   }

   /**
    * Mutator method to set the query language for this process.
    * 
    * @param aQueryLanguage the query language for this process
    */
   public final void setQueryLanguage(String aQueryLanguage)
   {
      mQueryLanguage = aQueryLanguage;
   }

   /**
    * Accessor method to obtain the expression language for this process.
    * 
    * @return expression language for this process
    */
   public final String getExpressionLanguage()
   {
      return mExpressionLanguage;
   }

   /**
    * Mutator method to set the expression language for this process.
    * 
    * @param aExpressionLanguage the expression language for this process
    */
   public final void setExpressionLanguage(String aExpressionLanguage)
   {
      mExpressionLanguage = aExpressionLanguage;
   }

   /**
    * Accessor method to determine if the suppress join failure flag is set 
    * for this process.
    * 
    * @return boolean flag indicating if the suppress join failure flag is set
    *          for this process
    */
   public final boolean getSuppressJoinFailure()
   {
      return mSuppressJoinFailure;
   }

   /**
    * Mutator method to set the suppress join failure flag for this process.
    * 
    * @param aSuppressJoinFailure flag indicating if join failure warning should be
    *         suppressed for this process
    */
   public final void setSuppressJoinFailure(boolean aSuppressJoinFailure)
   {
      mSuppressJoinFailure = aSuppressJoinFailure;
   }

   /**
    * Accessor method to determine if instance compensation is enabled for this process.
    * 
    * @return boolean flag indicating if instance compensation is enabled for this process
    */
   public final boolean getEnableInstanceCompensation()
   {
      return mEnableInstanceCompensation;
   }

   /**
    * Mutator method to set the enable instance compensation flag for this process.
    * 
    * @param aEnableInstanceCompensation the enable instance compensation flag for this process
    */
   public final void setEnableInstanceCompensation(boolean aEnableInstanceCompensation)
   {
      mEnableInstanceCompensation = aEnableInstanceCompensation;
   }

   /**
    * Accessor method to determine if this is an abstract process.
    * 
    * @return boolean flag indicating if this is an abstract process
    */
   public final boolean isAbstractProcess()
   {
      return mAbstractProcess;
   }

   /**
    * Mutator method to set the abstract process flag for this process.
    * 
    * @param aAbstractProcess the abstract process flag for this process
    */
   public final void setAbstractProcess(boolean aAbstractProcess)
   {
      mAbstractProcess = aAbstractProcess;
   }

   /**
    * Add a partner link to the current process.
    * @param aPartner the partner link to be added to the process
    */
   public void addPartnerLink(AePartnerLinkDef aPartnerLink)
   {
      getPartnerLinksContainer().add(aPartnerLink);
   }
   
   /**
    * Gets the partner link by name
    * @param aName
    */
   public AePartnerLinkDef getPartnerLink(String aName)
   {
      for (Iterator iter = getPartnerLinks(); iter.hasNext();)
      {
         AePartnerLinkDef def = (AePartnerLinkDef) iter.next();
         if (def.getName().equals(aName))
         {
            return def;
         }
      }
      return null;
   }

   /**
    * Get a list of partner links which are part of this process.
    * @return an iterator of the partner collection
    */
   public Iterator getPartnerLinks()
   {
      return getPartnerLinksContainer().getValues();
   }

   /**
    * Removes a partner link from the current process.
    * @param aPartnerLink the partner to be removed from the process
    */
   public void removePartnerLink(AePartnerLinkDef aPartnerLink)
   {
      getPartnerLinksContainer().remove(aPartnerLink);
   }

   /**
    * Add a partner to the current process.
    * @param aPartner the partner to be added to the process
    */
   public void addPartner(AePartnerDef aPartner)
   {
      getPartnersContainer().add(aPartner);
   }

   /**
    * Get a list of partners which are part of this process.
    * @return an iterator of the partner collection
    */
   public Iterator getPartners()
   {
      return getPartnersContainer().getValues();
   }

   /**
    * Removes a partner link from the current process.
    * @param aPartner the partner to be removed from the process
    */
   public void removePartner(AePartnerDef aPartner)
   {
      getPartnersContainer().remove(aPartner);
   }
   
   /**
    * Get a list of Imports which are part of this process.
    * @return an iterator of the import collection
    */
   public Iterator getImports()
   {
      return getImportsContainer().getValues();
   }
   
   /**
    * Gets the imports container definition.
    * @return AeImportsContainerDef
    */
   public AeImportsContainerDef getImportsContainer()
   {
      return mImports;
   }

   /**
    * @see org.activebpel.rt.bpel.def.AeBaseDef#accept(org.activebpel.rt.bpel.def.visitors.IAeDefVisitor)
    */
   public void accept(IAeDefVisitor aVisitor)
   {
      aVisitor.visit(this);
   }

   public AePartnerLinksContainerDef getPartnerLinksContainer()
   {
      return mPartnerLinks;
   }

   public AePartnersContainerDef getPartnersContainer()
   {
      return mPartners;
   }

   /**
    * Gets the BPEL comment.
    */
   public String getComment()
   {
      return mComment;
   }

   /**
    * Sets BPEL comment.
    * @param aString
    */
   public void setComment(String aString)
   {
      mComment = aString;
   }

   /**
    * Gets the preamble XML comment.
    * @return String the comment.
    */
   public String getProcessPreambleComment()
   {
      return mProcessPreambleComment;
   }

   /**
    * Sets the preable XML comment.
    * @param aString the comment.
    */
   public void setProcessPreambleComment(String aString)
   {
      mProcessPreambleComment = aString;
   }
   
   /**
    * Determines if the given partnerLink and operation target a create instance
    * activity.
    * @param aPartnerLinkName
    * @param aOperationName
    * @return true if this is an initial entry point
    */
   public boolean isCreateInstance( String aPartnerLinkName, String aOperationName )
   {
      return mCreateInstances.contains( makeKey(aPartnerLinkName, aOperationName) );
   }
   
   /**
    * Return the correlated property QNames for the
    * partnerLink, port type and operation
    * @param aKey partnerLink:portTypeQName:opName 
    * @return collection of property QNames
    */
   public Collection getCorrelatedPropertyNames( String aKey )
   {
      Collection props = (Collection)mPartnerLinkProperties.get(aKey);
      if( props == null )
      {
         props = Collections.EMPTY_SET;
      }
      return props;
   }
   
   /**
    * Returns location id corresponding to a location path or <code>-1</code> if not found.
    */
   public int getLocationId(String aLocationPath)
   {
      Integer id = (Integer) mLocationPathsToIds.get(aLocationPath);
      return (id == null) ? -1 : id.intValue();
   }

   /**
    * Returns location path corresponding to a location id or <code>null</code> if not found.
    */
   public String getLocationPath(int aLocationId)
   {
      return (String) mLocationIdsToPaths.get(aLocationId);
   }
   
   /**
    * @return True if the process contains serializable scopes.
    */
   public boolean containsSerializableScopes()
   {
      return mContainsSerializableScopes;
   }

   /**
    * @param aB flag indicating the process contains serializable scopes.
    */
   public void setContainsSerializableScopes(boolean aB)
   {
      mContainsSerializableScopes = aB;
   }
   
   /**
    * Getter for the max location id.
    */
   public int getMaxLocationId()
   {
      return mLocationIdsToPaths.size();
   }
   
   /**
    * Helper method for creating a key for maps that use the partnerLinkName and
    * operation as their keys.
    * @param aPartnerLinkName
    * @param aOperation
    */
   // TODO (MF) we should encapsulate the map key 
   public static String makeKey(String aPartnerLinkName, String aOperation)
   {
      return aPartnerLinkName + ":" + aOperation; //$NON-NLS-1$
   }
   
   /**
    * Gets a coll of correlation property sets that are applicable to the given
    * partner link and operation. 
    * @param aPartnerLinkName
    * @param aOperation
    */
   public Collection getCorrelationProperties(String aPartnerLinkName, String aOperation)
   {
      Collection coll = (Collection) mCorrelationPropertiesMap.get(makeKey(aPartnerLinkName,aOperation)); //$NON-NLS-1$
      if (coll == null)
      {
         coll = Collections.EMPTY_SET;
      }
      return coll;
   }
   
   /**
    * Adds a mapping for the given partner link name and operation to the set of
    * correlation property names that are part of the message.
    * @param aPartnerLinkName
    * @param aOperation
    * @param aListOfSets
    */
   public void setCorrelationProperties(String aPartnerLinkName, String aOperation, List aListOfSets)
   {
      mCorrelationPropertiesMap.put(makeKey(aPartnerLinkName,aOperation), aListOfSets); //$NON-NLS-1$
   }

   /**
    * Gets the messageExchange value for the createInstance receive with the given
    * partnerlink and operation. The messageExchange attribute is used to match
    * replies with receives. This method is only used during process creation. 
    * @param aPartnerLinkName
    * @param aOperation
    */
   public String getCreateInstanceMessageExchange(String aPartnerLinkName, String aOperation)
   {
      return (String) mMessageExchangeMap.get(makeKey(aPartnerLinkName,aOperation)); //$NON-NLS-1$
   }
   
   /**
    * Adds a mapping of partner link and operation to a messageExchange value. 
    * This is used by a visitor during the preprocessing stage of the process definition.
    * The visitor records the messageExchange value for createInstances since we'll
    * need that information later when creating a process.
    * @param aPartnerLinkName
    * @param aOperation
    * @param aMessageExchange
    */
   public void addCreateInstanceMessageExchange(String aPartnerLinkName, String aOperation, String aMessageExchange)
   {
      mMessageExchangeMap.put(makeKey(aPartnerLinkName,aOperation), aMessageExchange); //$NON-NLS-1$
   }

   /**
    * @param preProcessed The preProcessed state to set.
    */
   protected void setPreProcessed(boolean preProcessed)
   {
      mPreProcessed = preProcessed;
   }

   /**
    * True if the definition has been preprocessed.
    * @return Returns the preProcessed flag indicating whether the definition has been preprocessed.
    */
   public boolean isPreProcessed()
   {
      return mPreProcessed;
   }
}
