// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/AeScopeDef.java,v 1.14 2006/01/14 22:40:5
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def;

import java.util.Collections;
import java.util.Iterator;

import org.activebpel.rt.bpel.def.visitors.IAeDefVisitor;
import org.activebpel.rt.util.AeUtil;

/**
 * Definition for bpel scope.
 */
public class AeScopeDef
   extends AeNamedDef
   implements IAeSingleActivityContainerDef, IAeMessageExchangesParent
{
   /** contains events and alarms */
   private AeEventHandlersDef mEventHandlers;
   /** the single child activity for the scope */
   private AeActivityDef mActivity;
   /** container for the variables */
   private AeVariablesContainerDef mVariables;
   /** container for the correlation sets */
   private AeCorrelationSetContainerDef mCorrelationSets;
   /** container for the fault handlers */
   private AeFaultHandlersContainerDef mFaultHandlers;
   /** container for the compensation handler */
   private AeScopeCompensatorContainerDef mCompensationContainerDef;
   /** container for the message exchanges */
   private AeMessageExchangesDef mMessageExchangesDef;
   
   /**
    * Default constructor
    */
   public AeScopeDef()
   {
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.AeBaseDef#getLocationPath()
    */
   public String getLocationPath()
   {
      // The scope def is the base class for AeProcessDef and
      // a member of AeActivityScopeDef. If this def is the
      // process, then it'll have a location path. If it's
      // contained within a scope activity, then use the activity's
      // path for its path. 
      //
      // This came up with the addition of messageExchanges since we
      // need to calculate the fully qualified path of the message
      // exchange value when matching receives and replies.
      if (super.getLocationPath()==null && getParent() != null)
         return getParent().getLocationPath();
      else
         return super.getLocationPath();
   }

   /**
    * @return Returns the messageExchangesDef.
    */
   public AeMessageExchangesDef getMessageExchangesDef()
   {
      if (mMessageExchangesDef == null)
         mMessageExchangesDef = new AeMessageExchangesDef();
      return mMessageExchangesDef;
   }

   /**
    * @param aMessageExchangesDef The messageExchangesDef to set.
    */
   public void setMessageExchangesDef(AeMessageExchangesDef aMessageExchangesDef)
   {
      mMessageExchangesDef = aMessageExchangesDef;
   }
   
   /**
    * Returns true if the scope has the messageExchange explicitly declared.
    * @param aMessageExchangeValue
    */
   public boolean declaresMessageExchange(String aMessageExchangeValue)
   {
      return getMessageExchangesDef().declaresMessageExchange(aMessageExchangeValue);
   }
   
   /**
    * Returns true if the scope defines one or more messageExchange values.
    */
   public boolean hasMessageExchanges()
   {
      return mMessageExchangesDef != null && !AeUtil.isNullOrEmpty(getMessageExchangesDef().getMessageExchanges());
   }

   /**
    * Provides the ability to add a fault handler to the scope.
    *
    * @param aFault fault handler to be added
    */
   public void addFaultHandler(AeFaultHandlerDef aFault)
   {
      getFaultHandlerContainer().add(aFault);
   }

   /**
    * Provide a list of the fault handler objects for the user to iterate .
    *
    * @return iterator of AeFaultHandlerDef objects
    */
   public Iterator getFaultHandlerList()
   {
      if (mFaultHandlers == null)
      {
         return Collections.EMPTY_LIST.iterator();
      }
      return getFaultHandlerContainer().getFaultHandlers();
   }

   /**
    * Setter for the <code>catchAll</code> def.
    * @param aDef
    */
   public void setDefaultFaultHandler(AeDefaultFaultHandlerDef aDef)
   {
      getFaultHandlerContainer().setDefaultFaultHandler(aDef);
   }
   
   /**
    * Getter for the <code>catchAll</code> def.
    */
   public AeDefaultFaultHandlerDef getDefaultFaultHandler()
   {
      if (mFaultHandlers == null)
      {
         return null;
      }
      
      return getFaultHandlerContainer().getDefaultFaultHandler();
   }
   
   /**
    * Gets the fault handler container. Will create one if it's null.
    */
   public AeFaultHandlersContainerDef getFaultHandlerContainer()
   {
      if (mFaultHandlers == null)
      {
         mFaultHandlers = new AeFaultHandlersContainerDef();
      }
      return mFaultHandlers;
   }

   /**
    * Utility method to determine if a scoped activity has any fault handlers
    * @return true if has fault handlers false otherwise
    */
   public boolean hasFaultHandlers()
   {
      if (mFaultHandlers == null)
      {
         return false;
      }
      return getFaultHandlerContainer().hasFaultHandlers(); 
   }

   /**
    * Obtains the current activity associated with this object.
    *
    * @return an activity associated with this object
    * @see AeActivityDef
    */
   public final AeActivityDef getActivity()
   {
      return mActivity;
   }

   /**
    * Set the activity to execute if this scope is called
    *
    * @param aActivity activity to set
    * @see AeActivityDef
    */
   public final void setActivity(AeActivityDef aActivity)
   {
      mActivity = aActivity;
   }

   /**
    * Obtains the compensation handling activity associated with this object.
    *
    * @return an activity associated with this object
    * @see AeActivityDef
    */
   public final AeActivityDef getCompensationHandler()
   {
      if (mCompensationContainerDef == null)
      {
         return null;
      }
      return getCompensatorContainer().getActivity();
   }

   /**
    * Set the activity to execute for compensating this scope.
    *
    * @return an compensation handling activity associated with this object
    * @see AeActivityDef
    */
   public final void setCompensationHandler(AeActivityDef aCompensationHandler)
   {
      getCompensatorContainer().setActivity(aCompensationHandler);
   }

   /**
    * Returns the compensatorContainer.
    * @return AeScopeCompensatorDef
    */
   public AeScopeCompensatorContainerDef getCompensatorContainer()
   {
      if (mCompensationContainerDef == null)
      {
         mCompensationContainerDef = new AeScopeCompensatorContainerDef();
      }
      return mCompensationContainerDef;
   }
   
   /**
    * Returns true if the scope has a compensation handler defined. 
    */
   public boolean hasCompensator()
   {
      return getCompensationHandler() != null;
   }

   /**
    * Returns true if the evnt handlers object is not empty.
    */
   public boolean hasEventHandlers()
   {
      boolean hasEventHandler = false;
      if (mEventHandlers != null)
         hasEventHandler = mEventHandlers.hasEventHandler();
      return hasEventHandler;
   }

   /**
    * Returns the event handler container for this scope, can be null if none defined.
    */
   public AeEventHandlersDef getEventHandlers()
   {
      return mEventHandlers;
   }

   /**
    * Sets the event handler container for this scope.
    */
   public void setEventHandlers(AeEventHandlersDef aEventHandlers)
   {
      mEventHandlers = aEventHandlers;
   }

   /**
    * Get a specific variable from the variable collection.
    *
    * @param aVariableName the name of the variable we are searching for.
    *
    * @return a variable object reference or null if not found
    */
   public AeVariableDef getVariable(String aVariableName)
   {
      return getVariablesContainer().getVariable(aVariableName);
   }

   /**
    * Get a list of Variables which are part of this process.
    *
    * @return an iterator of the variable collection
    */
   public Iterator getVariables()
   {
      return getVariablesContainer().getValues();
   }

   /**
    * Get a list of correlation sets which are part of this process.
    *
    * @return an iterator of the correlation set collection
    */
   public Iterator getCorrelationSets()
   {
      return getCorrelationSetContainer().getValues();
   }

   /**
    * Add a variable to the current process.
    *
    * @param aVariable the variable to be added to the process
    */
   public void addVariable(AeVariableDef aVariable)
   {
      getVariablesContainer().add(aVariable);
   }

   /**
    * Add a correlation set to the current process.
    *
    * @param aCorrelationSet the correlation set to be added to the process
    */
   public void addCorrelationSet(AeCorrelationSetDef aCorrelationSet)
   {
      getCorrelationSetContainer().add(aCorrelationSet);
   }

   /**
    * Returns true if there are variables defined within the scope.
    */
   public boolean hasVariables()
   {
      boolean hasVars = false;
      if (mVariables != null)
         hasVars = (mVariables.getSize() != 0);
      return hasVars;
   }

   /**
    * Gets the <code>variables</code> container. Will create if it's null.
    */
   public AeVariablesContainerDef getVariablesContainer()
   {
      if (mVariables == null)
         mVariables = new AeVariablesContainerDef();

      return mVariables;
   }

   /**
    * Returns true if there are correlatiuon sets defined within the scope.
    */
   public boolean hasCorrelationSets()
   {
      boolean hasCorrSets = false;
      if (mCorrelationSets != null)
         hasCorrSets = (mCorrelationSets.getSize() != 0);
      return hasCorrSets;
   }

   /**
    * Gets the <code>correlationSets</code> container. Will create if it's null.
    */
   public AeCorrelationSetContainerDef getCorrelationSetContainer()
   {
      if (mCorrelationSets == null)
         mCorrelationSets = new AeCorrelationSetContainerDef();

      return mCorrelationSets;
   }

   /**
    * @see org.activebpel.rt.bpel.def.AeBaseDef#accept(org.activebpel.rt.bpel.def.visitors.IAeDefVisitor)
    */
   public void accept(IAeDefVisitor aVisitor)
   {
      aVisitor.visit(this);
   }
}
