//$Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/expr/AeAbstractExpressionParseResult.java,v 1.3 2006/01/10 22:18:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.expr;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.activebpel.rt.bpel.def.util.AeVariableData;
import org.activebpel.rt.bpel.def.util.AeVariableProperty;
import org.activebpel.rt.util.AeUtil;

/**
 * An abstract parse result object that can be extended by language specific impls.
 * 
 * Fundamentally, all of the implementations of methods in this base class will use the 
 * <code>getFunctions</code> method call, which any class extending this base class must
 * implement.  The <code>getFunctions</code> method will return a list of functions that
 * the other methods in this implementation will interrogate in various ways.  For example,
 * the <code>getLinkStatusFunctionList</code> implementation will get the full list of 
 * functions and filter out any that aren't "getLinkStatus".
 */
public abstract class AeAbstractExpressionParseResult implements IAeExpressionParseResult
{
   /** The expression. */
   private String mExpression;
   /** The parser context. */
   private IAeExpressionParserContext mParserContext;
   /** A list of errors found during parse. */
   private List mErrors;

   /**
    * Creates the parse result.
    * 
    * @param aExpression
    * @param aParserContext
    */
   public AeAbstractExpressionParseResult(String aExpression, IAeExpressionParserContext aParserContext)
   {
      setExpression(aExpression);
      setParserContext(aParserContext);
      setErrors(new ArrayList());
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getExpression()
    */
   public String getExpression()
   {
      return mExpression;
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getParseErrors()
    */
   public List getParseErrors()
   {
      return getErrors();
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#hasErrors()
    */
   public boolean hasErrors()
   {
      return getParseErrors().size() > 0;
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getLinkStatusFunctionList()
    */
   public List getLinkStatusFunctionList()
   {
      List list = new LinkedList();
      for (Iterator iter = getFunctions().iterator(); iter.hasNext();)
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         if (AeExpressionLanguageUtil.isLinkStatusFunction(function))
         {
            list.add(function);
         }
      }
      return list;
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getVarDataFunctionList()
    */
   public List getVarDataFunctionList()
   {
      List list = new LinkedList();
      for (Iterator iter = getFunctions().iterator(); iter.hasNext();)
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         if (AeExpressionLanguageUtil.isVarDataFunction(function))
         {
            list.add(function);
         }
      }
      return list;
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getVarPropertyFunctionList()
    */
   public List getVarPropertyFunctionList()
   {
      List list = new LinkedList();
      for (Iterator iter = getFunctions().iterator(); iter.hasNext();)
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         if (AeExpressionLanguageUtil.isVarPropertyFunction(function))
         {
            list.add(function);
         }
      }
      return list;
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getMyRolePropertyFunctionList()
    */
   public List getMyRolePropertyFunctionList()
   {
      List list = new LinkedList();
      for (Iterator iter = getFunctions().iterator(); iter.hasNext();)
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         if (AeExpressionLanguageUtil.isMyRolePropertyFunction(function))
         {
            list.add(function);
         }
      }
      return list;
   }
   
   
   
   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getVarDataList()
    */
   public List getVarDataList()
   {
      List list = new LinkedList();
      Collection functionList = getVarDataFunctionList();
      for (Iterator iter = functionList.iterator(); iter.hasNext();)
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();

         String var = function.getStringArgument(0);
         if (AeUtil.notNullOrEmpty(var))
         {
            String part = function.getStringArgument(1);
            String query = function.getStringArgument(2);
            list.add(new AeVariableData(var, part, query));
         }
      }
      return list;
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getVarPropertyList()
    */
   public List getVarPropertyList()
   {
      List list = new LinkedList();
      Collection functionList = getFunctions();
      for (Iterator iter = functionList.iterator(); iter.hasNext();)
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();

         if (AeExpressionLanguageUtil.isVarPropertyFunction(function))
         {
            String var = function.getStringArgument(0);
            if (AeUtil.notNullOrEmpty(var))
            {
               String property = function.getStringArgument(1);
               list.add(new AeVariableProperty(var, property));
            }
         }
      }
      return list;
   }

   /**
    * @see org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult#getVarNames()
    */
   public Set getVarNames()
   {
      Set rval = new LinkedHashSet();

      for (Iterator iter = getVarDataList().iterator(); iter.hasNext(); )
      {
         AeVariableData vd = (AeVariableData) iter.next();
         rval.add(vd.getVarName());
      }

      for (Iterator iter = getVarPropertyList().iterator(); iter.hasNext(); )
      {
         AeVariableProperty vp = (AeVariableProperty) iter.next();
         rval.add(vp.getVarName());
      }
      
      return rval;
   }
   
   /**
    * @return Returns the parserContext.
    */
   protected IAeExpressionParserContext getParserContext()
   {
      return mParserContext;
   }

   /**
    * @param aParserContext The parserContext to set.
    */
   protected void setParserContext(IAeExpressionParserContext aParserContext)
   {
      mParserContext = aParserContext;
   }

   /**
    * @param aExpression The expression to set.
    */
   protected void setExpression(String aExpression)
   {
      mExpression = aExpression;
   }

   /**
    * @return Returns the list of errors.
    */
   public List getErrors()
   {
      return mErrors;
   }

   /**
    * @param aErrors The errors to set.
    */
   protected void setErrors(List aErrors)
   {
      mErrors = aErrors;
   }
}
