//$Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/expr/xpath/AeXPathParseHandler.java,v 1.2 2005/06/14 17:06:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.expr.xpath;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeNamespaceContext;
import org.activebpel.rt.bpel.def.expr.AeScriptFuncDef;
import org.activebpel.rt.util.AeUtil;
import org.jaxen.saxpath.SAXPathException;
import org.jaxen.saxpath.XPathHandler;

/**
 * This class listens to a Jaxen parse of an XPath and constructs a list of functions that are used in the
 * expression.
 */
public class AeXPathParseHandler implements XPathHandler
{
   /** The namespace resolver to use while parsing. */
   private IAeNamespaceContext mNamespaceContext;
   /** A stack of functions - needed because function calls can be nested inside other function calls. */
   private Stack mFunctionStack = new Stack();
   /** The node stack. */
   private Stack mNodeStack = new Stack();
   /** A list of functions found during the parse. */
   private List mFunctions = new ArrayList();
   /** The full tree - set when endXPath() is called. */
   private AeXPathNode mTree = null;
   /** A list of errors found while handling the jaxen parse. */
   private List mErrors = new ArrayList();

   /**
    * Constructor.
    * 
    * @param aDef
    */
   public AeXPathParseHandler(IAeNamespaceContext aNamespaceContext)
   {
      setNamespaceContext(aNamespaceContext);
   }

   /**
    * Gets the list of functions found during parsing.
    */
   public List getFunctions()
   {
      return mFunctions;
   }

   /**
    * Returns the full parse tree.  This method is only valid once the parse is complete.
    */
   public AeXPathNode getTree()
   {
      return mTree;
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startAbsoluteLocationPath()
    */
   public void startAbsoluteLocationPath() throws SAXPathException
   {
      pushNode("AbsoluteLocationPath"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startAdditiveExpr()
    */
   public void startAdditiveExpr() throws SAXPathException
   {
      pushNode("AdditiveExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startAllNodeStep(int)
    */
   public void startAllNodeStep(int axis) throws SAXPathException
   {
      pushNode("AllNodeStep"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startAndExpr()
    */
   public void startAndExpr() throws SAXPathException
   {
      pushNode("AndExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startCommentNodeStep(int)
    */
   public void startCommentNodeStep(int axis) throws SAXPathException
   {
      pushNode("CommentNodeStep"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startEqualityExpr()
    */
   public void startEqualityExpr() throws SAXPathException
   {
      pushNode("EqualityExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startFilterExpr()
    */
   public void startFilterExpr() throws SAXPathException
   {
      pushNode("FilterExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startFunction(java.lang.String, java.lang.String)
    */
   public void startFunction(String aPrefix, String aFunctionName) throws SAXPathException
   {
      String ns = getNamespaceContext().resolvePrefixToNamespace(aPrefix);
      if (ns == null && AeUtil.notNullOrEmpty(aPrefix))
      {
         String msg = MessageFormat.format(AeMessages.getString("AeXPathParseHandler.PREFIX_NOT_DECLARED_ERROR"), //$NON-NLS-1$
               new Object [] { aPrefix, aPrefix + ":" + aFunctionName }); //$NON-NLS-1$
         getErrors().add(msg);
      }
      mFunctionStack.push(new AeScriptFuncDef(ns, aFunctionName));
      AeXPathFunctionNode node = new AeXPathFunctionNode();
      node.setName("Function"); //$NON-NLS-1$
      node.setValue(aPrefix + ":" + aFunctionName); //$NON-NLS-1$
      pushNode(node);
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startMultiplicativeExpr()
    */
   public void startMultiplicativeExpr() throws SAXPathException
   {
      pushNode("MultiplicativeExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startNameStep(int, java.lang.String, java.lang.String)
    */
   public void startNameStep(int axis, String aPrefix, String aLocalName) throws SAXPathException
   {
      pushNode("NameStep"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startOrExpr()
    */
   public void startOrExpr() throws SAXPathException
   {
      pushNode("OrExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startPathExpr()
    */
   public void startPathExpr() throws SAXPathException
   {
      pushNode("PathExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startPredicate()
    */
   public void startPredicate() throws SAXPathException
   {
      pushNode("Predicate"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startProcessingInstructionNodeStep(int, java.lang.String)
    */
   public void startProcessingInstructionNodeStep(int axis, String aName) throws SAXPathException
   {
      pushNode("ProcessingInstructionNodeStep"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startRelationalExpr()
    */
   public void startRelationalExpr() throws SAXPathException
   {
      pushNode("RelationalExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startRelativeLocationPath()
    */
   public void startRelativeLocationPath() throws SAXPathException
   {
      pushNode("RelativeLocationPath"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startTextNodeStep(int)
    */
   public void startTextNodeStep(int axis) throws SAXPathException
   {
      pushNode("TextNodeStep"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startUnaryExpr()
    */
   public void startUnaryExpr() throws SAXPathException
   {
      pushNode("UnaryExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startUnionExpr()
    */
   public void startUnionExpr() throws SAXPathException
   {
      pushNode("UnionExpr"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#startXPath()
    */
   public void startXPath() throws SAXPathException
   {
      pushNode("XPath"); //$NON-NLS-1$
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endAbsoluteLocationPath()
    */
   public void endAbsoluteLocationPath() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endAdditiveExpr(int)
    */
   public void endAdditiveExpr(int additiveOperator) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endAllNodeStep()
    */
   public void endAllNodeStep() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endAndExpr(boolean)
    */
   public void endAndExpr(boolean aCreate) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endCommentNodeStep()
    */
   public void endCommentNodeStep() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endEqualityExpr(int)
    */
   public void endEqualityExpr(int aEqualityOperator) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endFilterExpr()
    */
   public void endFilterExpr() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endFunction()
    */
   public void endFunction() throws SAXPathException
   {
      AeXPathNode node = popNode();
      AeScriptFuncDef func = (AeScriptFuncDef) mFunctionStack.pop();
      for (Iterator iter = node.mChildren.iterator(); iter.hasNext(); )
      {
         node = (AeXPathNode) iter.next();
         node = node.simplify();
         if (node instanceof AeXPathLiteralNode)
         {
            func.getArgs().add(node.mValue);
         }
         else
         {
            func.getArgs().add(AeScriptFuncDef.__EXPRESSION__);
         }
      }
      if (mFunctionStack.size() > 0)
      {
         func.setParent((AeScriptFuncDef) mFunctionStack.peek());
      }
      mFunctions.add(func);
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endMultiplicativeExpr(int)
    */
   public void endMultiplicativeExpr(int aMultiplicativeOperator) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endNameStep()
    */
   public void endNameStep() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endOrExpr(boolean)
    */
   public void endOrExpr(boolean aCreate) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endPathExpr()
    */
   public void endPathExpr() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endPredicate()
    */
   public void endPredicate() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endProcessingInstructionNodeStep()
    */
   public void endProcessingInstructionNodeStep() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endRelationalExpr(int)
    */
   public void endRelationalExpr(int aRelationalOperator) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endRelativeLocationPath()
    */
   public void endRelativeLocationPath() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endTextNodeStep()
    */
   public void endTextNodeStep() throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endUnaryExpr(int)
    */
   public void endUnaryExpr(int aUnaryOperator) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endUnionExpr(boolean)
    */
   public void endUnionExpr(boolean aCreate) throws SAXPathException
   {
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#endXPath()
    */
   public void endXPath() throws SAXPathException
   {
      mTree = popNode().simplify();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#literal(java.lang.String)
    */
   public void literal(String aLiteral) throws SAXPathException
   {
      AeXPathLiteralNode node = new AeXPathLiteralNode();
      node.setName("Literal"); //$NON-NLS-1$
      node.setValue(aLiteral);
      pushNode(node);
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#number(double)
    */
   public void number(double aNumber) throws SAXPathException
   {
      AeXPathLiteralNode node = new AeXPathLiteralNode();
      node.setName("Literal"); //$NON-NLS-1$
      node.setValue(new Double(aNumber));
      pushNode(node);
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#number(int)
    */
   public void number(int aNumber) throws SAXPathException
   {
      AeXPathLiteralNode node = new AeXPathLiteralNode();
      node.setName("Literal"); //$NON-NLS-1$
      node.setValue(new Integer(aNumber));
      pushNode(node);
      popNode();
   }

   /**
    * @see org.jaxen.saxpath.XPathHandler#variableReference(java.lang.String, java.lang.String)
    */
   public void variableReference(String aPrefix, String aVariableName) throws SAXPathException
   {
      pushNode("VariableReference", aPrefix + ":" + aVariableName); //$NON-NLS-1$ //$NON-NLS-2$
      popNode();
   }

   /**
    * Pushes a node onto the stack.
    * 
    * @param aName
    */
   protected void pushNode(String aName)
   {
      pushNode(aName, null);
   }

   /**
    * Pushes a node onto the stack.  This also adds the node as a child of the node currently at the
    * top of the stack.
    * 
    * @param aName
    * @param aValue
    */
   protected void pushNode(String aName, Object aValue)
   {
      AeXPathNode node = new AeXPathNode();
      node.mName = aName;
      node.mValue = aValue;
      pushNode(node);
   }
   
   /**
    * Pushes a node onto the stack.  This also adds the node as a child of the node currently at the
    * top of the stack.
    * 
    * @param aNode
    */
   protected void pushNode(AeXPathNode aNode)
   {
      if (mNodeStack.size() > 0)
      {
         ((AeXPathNode) mNodeStack.peek()).mChildren.add(aNode);
      }
      mNodeStack.push(aNode);
   }

   /**
    * Pops a node from the node stack.
    */
   protected AeXPathNode popNode()
   {
      return (AeXPathNode) mNodeStack.pop();
   }

   /**
    * A simple 'node' inner class.
    */
   public class AeXPathNode
   {
      /** The node name. */
      private String mName;
      /** The node value. */
      private Object mValue;
      /** The node's children. */
      private List mChildren = new ArrayList();

      /**
       * Simplifies this node.  If the node has only one child, then the simplified view of this
       * node is the child node.  If it has 0 children or multiple children, then this node is as
       * simplified as it can be.
       */
      public AeXPathNode simplify()
      {
         if (mChildren.size() == 1)
         {
            return ((AeXPathNode) mChildren.get(0)).simplify();
         }
         else
         {
            simplifyChildren();
            return this;
         }
      }
      
      /**
       * Simplifies this node's children.
       */
      protected void simplifyChildren()
      {
         List list = new ArrayList();
         for (Iterator iter = getChildren().iterator(); iter.hasNext(); )
         {
            AeXPathNode node = (AeXPathNode) iter.next();
            list.add(node.simplify());
         }
         setChildren(list);
      }
      
      /**
       * @see java.lang.Object#toString()
       */
      public String toString()
      {
         return MessageFormat.format("Node: {0}  Value: {1}\nNumChildren: {2}", new Object[] { mName, mValue, new Integer(mChildren.size()) }); //$NON-NLS-1$
      }
      
      /**
       * @return Returns the children.
       */
      public List getChildren()
      {
         return mChildren;
      }

      /**
       * @param aChildren The children to set.
       */
      public void setChildren(List aChildren)
      {
         mChildren = aChildren;
      }

      /**
       * @return Returns the name.
       */
      public String getName()
      {
         return mName;
      }

      /**
       * @param aName The name to set.
       */
      public void setName(String aName)
      {
         mName = aName;
      }

      /**
       * @return Returns the value.
       */
      public Object getValue()
      {
         return mValue;
      }

      /**
       * @param aValue The value to set.
       */
      public void setValue(Object aValue)
      {
         mValue = aValue;
      }
   }

   /**
    * A simple node to represent a function call.
    */
   public class AeXPathFunctionNode extends AeXPathNode
   {
      /**
       * @see org.activebpel.rt.bpel.def.expr.xpath.AeXPathParseHandler.AeXPathNode#simplify()
       */
      public AeXPathNode simplify()
      {
         return this;
      }
   }

   /**
    * A simple node to represent a literal.
    */
   public class AeXPathLiteralNode extends AeXPathNode
   {
      /**
       * @see org.activebpel.rt.bpel.def.expr.xpath.AeXPathParseHandler.AeXPathNode#simplify()
       */
      public AeXPathNode simplify()
      {
         return this;
      }
   }

   /**
    * @return Returns the namespaceContext.
    */
   protected IAeNamespaceContext getNamespaceContext()
   {
      return mNamespaceContext;
   }

   /**
    * @param aNamespaceContext The namespaceContext to set.
    */
   protected void setNamespaceContext(IAeNamespaceContext aNamespaceContext)
   {
      mNamespaceContext = aNamespaceContext;
   }
   
   /**
    * @return Returns the errors.
    */
   public List getErrors()
   {
      return mErrors;
   }
   
   /**
    * @param aErrors The errors to set.
    */
   protected void setErrors(List aErrors)
   {
      mErrors = aErrors;
   }
}
