// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/io/AeBpelFactory.java,v 1.9 2006/01/20 15:49:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.io;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.def.AeMessageExchangesDef;
import org.activebpel.rt.bpel.def.IAeBPELConstants;
import org.activebpel.rt.bpel.def.io.ext.AeDefaultExtensionElementIO;
import org.activebpel.rt.bpel.def.io.ext.IAeExtensionReader;
import org.activebpel.rt.bpel.def.io.registry.AeDefaultBpelRegistry;
import org.activebpel.rt.bpel.def.io.registry.AeDefaultExtensionRegistry;
import org.activebpel.rt.bpel.def.io.registry.IAeBpelRegistry;
import org.activebpel.rt.bpel.def.io.registry.IAeExtensionRegistry;

/**
 * Abstract factory whose impl classes provide access to
 * BPEL serializers and deserializers.
 */
public abstract class AeBpelFactory
{
   /** bpel reader/writer registry */
   private IAeBpelRegistry mBpelRegistry;

   /** extension element reader/writer registry */
   private IAeExtensionRegistry mExtensionRegistry;
   
   /**
    * Returns the default factory.
    * @return default <code>AeBpelFactoryImpl</code>
    * @throws AeBusinessProcessException
    */
   public static AeBpelFactory newInstance() throws AeBusinessProcessException
   {
      return new AeBpelFactoryImpl();
   }
   
   /**
    * Returns specified custom factory.
    * @param aFactoryImpl class name of the factory return
    * @return custom factory impl
    * @throws AeBusinessProcessException
    */
   public static AeBpelFactory newInstance( String aFactoryImpl )
   throws AeBusinessProcessException
   {
      if( aFactoryImpl != null )
      {
         return createFactory( aFactoryImpl );               
      }
      else
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeBpelFactory.ERROR_0")); //$NON-NLS-1$
      }
   }
   
   /**
    * Loads the factory class and returns a new instance.
    * @param aFactoryImpl
    * @return AeBpelFactory impl class
    * @throws AeBusinessProcessException wraps any exceptions thrown while creating the factory
    */
   private static AeBpelFactory createFactory( String aFactoryImpl )
   throws AeBusinessProcessException
   {
      try
      {
         Class factoryClass = Class.forName( aFactoryImpl );
         return (AeBpelFactory)factoryClass.newInstance();
      }
      catch (ClassNotFoundException e)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeBpelFactory.ERROR_1") + aFactoryImpl ); //$NON-NLS-1$
      }
      catch (InstantiationException e)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeBpelFactory.ERROR_2") + aFactoryImpl ); //$NON-NLS-1$
      }
      catch (IllegalAccessException e)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeBpelFactory.ERROR_3") + aFactoryImpl, e ); //$NON-NLS-1$
      }
   }
   
   
   /**
    * Accessor for deserializer.
    * @return the reader responsible for deserialing BPEL xml. 
    */
   public IAeBpelReader getBpelReader()
   {
      IAeBpelReader reader = createBpelReader();
      reader.setExtensionRegistry( getExtensionRegistry() );
      reader.setBpelRegistry( getBpelRegistry() );
      return reader;
   }
   
   
   /**
    * Subclasses are responsible for providing
    * IAeBpelReader impl classes.
    * @return the IAeBpelReader impl class
    */
   abstract protected IAeBpelReader createBpelReader();
   
   
   /**
    * Accessor for serializer.
    * @return the writer responsible for serialing AeDef objects. 
    */
   public IAeBpelWriter getBpelWriter()
   {
      IAeBpelWriter writer = createBpelWriter();
      writer.setBpelRegistry( getBpelRegistry() );
      writer.setExtensionRegistry( getExtensionRegistry() );
      return writer;
   }
   
   /**
    * Subclasses are responsible for providing
    * IAeBpelWriter impl classes.
    * @return the IAeBpelWriter impl class
    */
   abstract protected IAeBpelWriter createBpelWriter();
   
   
   /**
    * Allows caller to install a custom IAeBpelRegistry
    * if one is required.
    * @param aRegistry
    */
   public void installBpelRegistry( IAeBpelRegistry aRegistry )
   {
      mBpelRegistry = aRegistry;
   }
   
   /**
    * Accessor for the IAeBpelRegistry impl.
    * If none has been set, a new <code>AeDefaultBpelRegistry</code>
    * is created and set as the member registry.
    * <br />
    * This registry will be installed in all readers
    * and writers.
    * @return IAeBpelRegistry impl
    */
   protected IAeBpelRegistry getBpelRegistry()
   {
      if( mBpelRegistry == null )
      {
         mBpelRegistry = new AeDefaultBpelRegistry();
      }
      return mBpelRegistry;
   }
   
   /**
    * Allows caller to install a custom IAeExtensionRegistry
    * for handling BPEL extension elements.
    * @param aRegistry 
    */
   public void installExtensionRegistry( IAeExtensionRegistry aRegistry )
   {
      mExtensionRegistry = aRegistry; 
   }
   
   
   /**
    * Accessor for the IAeExtensionRegistry impl.
    * If none has been set, a new <code>AeDefaultExtensionRegistry</code>
    * is created and set as the member extension registry.  The
    * default extension registry is configured with the 
    * <code>AeDefaultExtensionElementIO</code> extension element
    * serializer/deserializer to handle all extension elements
    * generically.
    * <br />
    * This registry will be installed in all readers and writers.
    * @return IAeExtensionsRegistry impl
    */
   public IAeExtensionRegistry getExtensionRegistry()
   {
      if( mExtensionRegistry == null )
      {
         mExtensionRegistry = new AeDefaultExtensionRegistry();
         AeDefaultExtensionElementIO extIO = new AeDefaultExtensionElementIO();
         mExtensionRegistry.setDefaultExtensionReader( extIO );
         mExtensionRegistry.setDefaultExtensionWriter( extIO );   
         
         // add our extension points here
         mExtensionRegistry.registerExtensionReader(AeMessageExchangesDef.class, new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, IAeBPELConstants.TAG_SCOPE_MESSAGEEXCHANGE), IAeExtensionReader.NULL_READER);
         // legacy registration for the 2.0 beta that went out w/o a version qualifier in the namespace
         mExtensionRegistry.registerExtensionReader(AeMessageExchangesDef.class, new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, IAeBPELConstants.TAG_SCOPE_MESSAGEEXCHANGE), IAeExtensionReader.NULL_READER);         
      }
      return mExtensionRegistry;
   }

}
