// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/io/registry/AeDefReaderRegistry.java,v 1.9 2006/01/20 15:49:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.io.registry;

import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.def.AeActivityDef;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeMessageExchangesDef;
import org.activebpel.rt.bpel.def.IAeActivityContainerDef;
import org.activebpel.rt.bpel.def.IAeBPELConstants;
import org.activebpel.rt.bpel.def.activity.support.AeForEachBranchesDef;
import org.activebpel.rt.bpel.def.activity.support.AeForEachCompletionConditionDef;
import org.activebpel.rt.bpel.def.io.IAeBpelClassConstants;
import org.activebpel.rt.bpel.def.io.readers.def.AeContainerReaderFactory;
import org.activebpel.rt.bpel.def.io.readers.def.AeDispatchReader;
import org.activebpel.rt.bpel.def.io.readers.def.IAeBpelDefReader;

/**
 * Default registry for mapping BPEL elements to the
 * appropriate deserializer.
 * <br />
 * All readers, except activities and the standard
 * activity elements (target/source) are mapped to 
 * reader instances by parent def object and
 * QName.  This allows potential for registering
 * different readers for the same QName based on
 * parenting.
 * <br />
 * In this impl activities and and standard elements
 * are mapped directly via QName, however this behavior
 * can be overridden by installing a custom reader 
 * (mapped via the parent class to QName to reader)
 * in the main registry (specifically by overriding
 * the <code>initMainRegistry</code> method).
 */
public class AeDefReaderRegistry 
   implements IAeBPELConstants, IAeBpelClassConstants
{
   /** top level map for well know BPEL elements */
   private Map mRegistryMap;

   /** holds the standard element readers for activities (target/source) */
   private Map mStandardReadersMap;

   /** map of activity QNames to readers */
   private Map mActivityReadersMap;
   
   /**
    * Default constructor.
    */
   public AeDefReaderRegistry()
   {
      init();      
   }
   
   /*
    * Initializes the registry maps.
    */
   protected void init()
   {
      initMainRegistry();
      initBpelActivities();
      initBpelActivityStandardElements();
   }

   /**
    * Maps the parent AeDef class to its child readers based
    * on the child element QNames.
    */
   protected void initMainRegistry()
   {
      // maps readers for children of process element
      // exclusive of activity (which is handled sep)
      registerReader( PROCESS_CLASS,  TAG_VARIABLES,           AeContainerReaderFactory.getVariablesReader());
      registerReader( PROCESS_CLASS,  TAG_FAULT_HANDLERS,      AeContainerReaderFactory.getFaultHandlersReader());
      registerReader( PROCESS_CLASS,  TAG_EVENT_HANDLERS,      new AeDispatchReader(EVENT_HANDLERS_CLASS ));
      registerReader( PROCESS_CLASS,  TAG_PARTNER_LINKS,       AeContainerReaderFactory.getPartnerLinksReader());
      registerReader( PROCESS_CLASS,  TAG_PARTNERS,            AeContainerReaderFactory.getPartnersReader());
      registerReader( PROCESS_CLASS,  TAG_COMPENSATION_HANDLER,AeContainerReaderFactory.getCompensatorReader());      
      registerReader( PROCESS_CLASS,  TAG_CORRELATION_SETS,    AeContainerReaderFactory.getCorrelationSetsReader());      


      // map readers for children of copy element
      registerReader( ACTIVITY_ASSIGN_CLASS, TAG_COPY, new AeDispatchReader(ASSIGN_COPY_CLASS ) );
      registerReader( ASSIGN_COPY_CLASS, TAG_TO,   new AeDispatchReader( ASSIGN_TO_CLASS ) );
      registerReader( ASSIGN_COPY_CLASS, TAG_FROM, new AeDispatchReader( ASSIGN_FROM_CLASS ) );

      // maps readers for children of switch activity
      // exclusive of standard readers (source/target)
      registerReader( ACTIVITY_SWITCH_CLASS, TAG_CASE,      new AeDispatchReader(CASE_CLASS ));      
      registerReader( ACTIVITY_SWITCH_CLASS, TAG_OTHERWISE, new AeDispatchReader(OTHERWISE_CLASS ));      
      
      // maps readers for children of pick activity
      // exclusive of standard readers (source/target)
      registerReader( ACTIVITY_PICK_CLASS, TAG_ON_MESSAGE, new AeDispatchReader(ON_MESSAGE_CLASS ));      
      registerReader( ACTIVITY_PICK_CLASS, TAG_ON_ALARM,   new AeDispatchReader(ON_ALARM_CLASS ));      

      // maps readers for children of scope activity
      // exclusive of standard readers (source/target)
      registerReader( ACTIVITY_SCOPE_CLASS, TAG_COMPENSATION_HANDLER, AeContainerReaderFactory.getCompensatorReader());      
      registerReader( ACTIVITY_SCOPE_CLASS, TAG_FAULT_HANDLERS,       AeContainerReaderFactory.getFaultHandlersReader());      
      registerReader( ACTIVITY_SCOPE_CLASS, TAG_VARIABLES,            AeContainerReaderFactory.getVariablesReader());      
      registerReader( ACTIVITY_SCOPE_CLASS, TAG_CORRELATION_SETS,     AeContainerReaderFactory.getCorrelationSetsReader());      
      registerReader( ACTIVITY_SCOPE_CLASS, TAG_EVENT_HANDLERS,       new AeDispatchReader(EVENT_HANDLERS_CLASS ));

      // maps readers for children of invoke activity
      // exclusive of standard readers (source/target)
      registerReader( ACTIVITY_INVOKE_CLASS, TAG_CATCH,                new AeDispatchReader(FAULT_HANDLER_CLASS ));
      registerReader( ACTIVITY_INVOKE_CLASS, TAG_CATCH_ALL,            new AeDispatchReader(DEFAULT_FAULT_HANDLER_CLASS ));
      registerReader( ACTIVITY_INVOKE_CLASS, TAG_COMPENSATION_HANDLER, AeContainerReaderFactory.getCompensatorReader());
      registerReader( ACTIVITY_INVOKE_CLASS, TAG_CORRELATIONS,         AeContainerReaderFactory.getCorrelationsReader());
      
      // maps readers for children of eventHandlers element
      registerReader( EVENT_HANDLERS_CLASS, TAG_ON_MESSAGE, new AeDispatchReader(ON_MESSAGE_CLASS ));      
      registerReader( EVENT_HANDLERS_CLASS, TAG_ON_ALARM,   new AeDispatchReader(ON_ALARM_CLASS ));      

      // maps readers for children of faultHandlers element
      registerReader( FAULT_HANDLER_CONTAINER_CLASS, TAG_CATCH,     new AeDispatchReader(FAULT_HANDLER_CLASS ));
      registerReader( FAULT_HANDLER_CONTAINER_CLASS, TAG_CATCH_ALL, new AeDispatchReader(DEFAULT_FAULT_HANDLER_CLASS ) );


      // all of the various class to child element mappings
      registerReader( ACTIVITY_FLOW_CLASS,           TAG_LINKS,               AeContainerReaderFactory.getLinksContainerReader() );
      registerReader( ACTIVITY_RECEIVE_CLASS,        TAG_CORRELATIONS,        AeContainerReaderFactory.getCorrelationsReader() );
      registerReader( ACTIVITY_REPLY_CLASS,          TAG_CORRELATIONS,        AeContainerReaderFactory.getCorrelationsReader() );

      registerReader( PARTNER_CONTAINER_CLASS,       TAG_PARTNER,             new AeDispatchReader(PARTNER_CLASS ) );
      registerReader( PARNTER_LINKS_CONTAINER_CLASS, TAG_PARTNER_LINK,        new AeDispatchReader(PARTNER_LINK_CLASS ) );
      registerReader( PARTNER_CLASS,                 TAG_PARTNER_LINK,        new AeDispatchReader(PARTNER_LINK_CLASS ) );
      registerReader( VARIABLE_CONTAINER_CLASS,      TAG_VARIABLE,            new AeDispatchReader(VARIABLE_CLASS ) );

      registerReader( LINKS_CONTAINER_CLASS,         TAG_LINK,                new AeDispatchReader(LINK_CLASS ) );
      registerReader( ON_MESSAGE_CLASS,              TAG_CORRELATIONS,        AeContainerReaderFactory.getCorrelationsReader() );
      registerReader( ON_MESSAGE_CLASS,              TAG_CORRELATIONS,        AeContainerReaderFactory.getCorrelationsReader() );

      registerReader( CORRELATION_CONTAINER_CLASS,   TAG_CORRELATION,         new AeDispatchReader(CORRELATION_CLASS ) );      
      registerReader( CORRELATION_SETS_CLASS,        TAG_CORRELATION_SET,     new AeDispatchReader(CORRELATION_SET_CLASS ));
      
      // fourEach extensions
      registerReader( ACTIVITY_FOREACH_CLASS,        new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_FOREACH_STARTCOUNTER), new AeDispatchReader(ACTIVITY_FOREACH_START));
      registerReader( ACTIVITY_FOREACH_CLASS,        new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_FOREACH_FINALCOUNTER), new AeDispatchReader(ACTIVITY_FOREACH_FINAL));
      registerReader( ACTIVITY_FOREACH_CLASS,        new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_FOREACH_COMPLETION_CONDITION), new AeDispatchReader(AeForEachCompletionConditionDef.class));
      registerReader( AeForEachCompletionConditionDef.class, new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_FOREACH_BRANCHES), new AeDispatchReader(AeForEachBranchesDef.class));

      // messageExchange extension
      registerReader( PROCESS_CLASS, new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_SCOPE_MESSAGEEXCHANGES), new AeDispatchReader(AeMessageExchangesDef.class));
      registerReader( ACTIVITY_SCOPE_CLASS, new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_SCOPE_MESSAGEEXCHANGES), new AeDispatchReader(AeMessageExchangesDef.class));

      // legacy registration for the 2.0 beta that went out w/o a version qualifier in the namespace
      registerReader( PROCESS_CLASS, new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_SCOPE_MESSAGEEXCHANGES), new AeDispatchReader(AeMessageExchangesDef.class));
      registerReader( ACTIVITY_SCOPE_CLASS, new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_SCOPE_MESSAGEEXCHANGES), new AeDispatchReader(AeMessageExchangesDef.class));
      registerReader( ACTIVITY_FOREACH_CLASS,        new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_FOREACH_STARTCOUNTER), new AeDispatchReader(ACTIVITY_FOREACH_START));
      registerReader( ACTIVITY_FOREACH_CLASS,        new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_FOREACH_FINALCOUNTER), new AeDispatchReader(ACTIVITY_FOREACH_FINAL));
   }

   /**
    * Convenience method for adding readers to the <code>mRegistryMap</code>.
    * <br />
    * <code>mRegistryMap</code> is a map of maps where the inner map contains
    * the QName to IAeBpelDefReader mappings.  The top level map contains the
    * AeDef class to inner map mappings.
    * 
    * @param aClass the parent AeDef class
    * @param aElementName child BPEL element for this parent
    * @param aReader IAeBpelDefReader impl
    */
   protected void registerReader( Class aClass, String aElementName, IAeBpelDefReader aReader )
   {
      registerReader(aClass, makeDefaultQName(aElementName), aReader);
   }
   
   /**
    * Adds the readers to the map with the specified qname, allowing for our extensions
    * not assuming that they're in the bpel namespace.
    * @param aClass
    * @param aElementQName
    * @param aReader
    */
   protected void registerReader( Class aClass, QName aElementQName, IAeBpelDefReader aReader )
   {
      Map innerMap = (Map)getRegistry().get( aClass );
      if( innerMap == null )
      {
         innerMap = new HashMap();
         getRegistry().put( aClass, innerMap );
      }
      innerMap.put( aElementQName, aReader );
   }
   
   /**
    * Populates the <code>mActivityReadersMap</code> with all of the
    * activity readers. 
    */
   protected void initBpelActivities()
   {
      addActivityReader( TAG_ASSIGN,     ACTIVITY_ASSIGN_CLASS      );
      addActivityReader( TAG_COMPENSATE, ACTIVITY_COMPENSATE_CLASS  );
      addActivityReader( TAG_EMPTY,      ACTIVITY_EMPTY_CLASS       );
      addActivityReader( TAG_FLOW,       ACTIVITY_FLOW_CLASS        );
      addActivityReader( TAG_INVOKE,     ACTIVITY_INVOKE_CLASS      );
      addActivityReader( TAG_PICK,       ACTIVITY_PICK_CLASS        );
      addActivityReader( TAG_RECEIVE,    ACTIVITY_RECEIVE_CLASS     );
      addActivityReader( TAG_REPLY,      ACTIVITY_REPLY_CLASS       );
      addActivityReader( TAG_SCOPE,      ACTIVITY_SCOPE_CLASS       );
      addActivityReader( TAG_SEQUENCE,   ACTIVITY_SEQUENCE_CLASS    );
      addActivityReader( TAG_SWITCH,     ACTIVITY_SWITCH_CLASS      );
      addActivityReader( TAG_TERMINATE,  ACTIVITY_TERMINATE_CLASS   );
      addActivityReader( TAG_THROW,      ACTIVITY_THROW_CLASS       );
      addActivityReader( TAG_WAIT,       ACTIVITY_WAIT_CLASS        );
      addActivityReader( TAG_WHILE,      ACTIVITY_WHILE_CLASS       );

      // Our custom BPEL extensions
      addActivityReader(new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_CONTINUE), ACTIVITY_CONTINUE_CLASS);
      addActivityReader(new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_BREAK), ACTIVITY_BREAK_CLASS);
      addActivityReader(new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_FOREACH), ACTIVITY_FOREACH_CLASS);
      addActivityReader(new QName(IAeBPELConstants.ABX_2_0_NAMESPACE_URI, TAG_SUSPEND), ACTIVITY_SUSPEND_CLASS);
      
      // legacy registration for the 2.0 beta that went out w/o a version qualifier in the namespace
      addActivityReader(new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_CONTINUE), ACTIVITY_CONTINUE_CLASS);
      addActivityReader(new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_BREAK), ACTIVITY_BREAK_CLASS);
      addActivityReader(new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_FOREACH), ACTIVITY_FOREACH_CLASS);
      addActivityReader(new QName(IAeBPELConstants.ABX_FUNCTIONS_NAMESPACE_URI, TAG_SUSPEND), ACTIVITY_SUSPEND_CLASS);
   }
   
   /**
    * Convenience method for registering activity readers.
    * @param aElementName
    * @param aClass
    */
   protected void addActivityReader( String aElementName, Class aClass  )
   {
      QName qname = makeDefaultQName(aElementName);
      addActivityReader(qname, aClass);
   }
   
   /**
    * Overloaded to accept QName for the element and the class
    * @param aQName
    * @param aClass
    */
   protected void addActivityReader(QName aQName, Class aClass)
   {
      getActivityReaders().put( aQName, new AeDispatchReader( aClass ) );
   }

   /**
    * Populates the <code>mStandardReadersMap</code> with the
    * target and source element readers.
    */
   protected void initBpelActivityStandardElements()
   {
      getStandardReaders().put( makeDefaultQName(TAG_TARGET), new AeDispatchReader( TARGET_CLASS ) );
      getStandardReaders().put( makeDefaultQName(TAG_SOURCE), new AeDispatchReader( SOURCE_CLASS ) );
   }

   /**
    * Return the top level process reader
    * @return IABpelDef reader for deserializing process element
    */
   public IAeBpelDefReader getProcessDefReader()
   {
     return new AeDispatchReader(PROCESS_CLASS);
   }

   /**
    * Return the appropriate IAeBpelDefReader impl for this
    * parent def and QName mapping.
    * @param aParentDef parent AeDef in the object model
    * @param aElementQName the child element QName
    * @return IAeBpelReader impl for deserializing this element or null if not found
    */
   public IAeBpelDefReader getReader(AeBaseDef aParentDef, QName aElementQName)
   throws UnsupportedOperationException
   {
      // look for any specific reader mappings
      IAeBpelDefReader reader = null;
      Map childReadersMap = (Map)getRegistry().get(aParentDef.getClass());
      
      if( childReadersMap != null )
      {
         reader = (IAeBpelDefReader)childReadersMap.get(aElementQName);
      }
      
      // if a reader was installed, return it
      // otherwise continue looking
      if( reader != null )
      {
         return reader;
      }


      // if my parent def class is an activity,
      // check for standard elements first (target/source)
      if( aParentDef instanceof AeActivityDef )
      {
         IAeBpelDefReader standardReader = 
                 (IAeBpelDefReader)getStandardReaders().get( aElementQName );

         if( standardReader != null )
         {
            return standardReader;
         }         
      }

      // if my parent def class is an activity container
      // look for my standard activities
      if(aParentDef instanceof IAeActivityContainerDef) 
      {
         IAeBpelDefReader activityReader = 
                 (IAeBpelDefReader)getActivityReaders().get( aElementQName );

         if( activityReader != null )
         {
            return activityReader;
         }         
      }
      
      return null;
   }
   
   /**
    * Convenience method for QName creation based on default 
    * BPEL namespace.
    * @see org.activebpel.rt.bpel.def.IAeBPELConstants#BPWS_NAMESPACE_URI
    * @param aElementName
    * @return QName with BPEL 1.1 NS uri
    */
   protected QName makeDefaultQName( String aElementName )
   {
      return new QName( IAeBPELConstants.BPWS_NAMESPACE_URI, aElementName );
   }

   /**
    * Accessor for bpel registry map.
    * @return registry map
    */
   protected Map getRegistry()
   {
      if( mRegistryMap == null )
      {
         mRegistryMap = new HashMap();
      }
      return mRegistryMap;
   }
   
   /**
    * Accessor for target/source registry map.
    * @return map for target and source readers
    */
   protected Map getStandardReaders()
   {
      if( mStandardReadersMap == null )
      {
         mStandardReadersMap = new HashMap();
      }
      return mStandardReadersMap;
   }
   
   /**
    * Accessor for activity readers map.
    * @return map of QNames to activities
    */
   protected Map getActivityReaders()
   {
      if( mActivityReadersMap == null )
      {
         mActivityReadersMap = new HashMap();
      }
      return mActivityReadersMap;
   }
}
