// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/io/writers/AeImplicitScopeUtil.java,v 1.3 2005/02/01 19:53:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.io.writers;

import org.activebpel.rt.bpel.def.AeActivityDef;
import org.activebpel.rt.bpel.def.AeScopeDef;
import org.activebpel.rt.bpel.def.IAeBPELConstants;
import org.activebpel.rt.bpel.def.activity.AeActivityInvokeDef;
import org.activebpel.rt.bpel.def.activity.AeActivityScopeDef;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Utility class for serializing/deserializing 
 * scope activities that can be inlined and invoke
 * activities with inlined scopes.
 */
public class AeImplicitScopeUtil implements IAeBPELConstants
{
   
   /**
    * Indicates if the invoke activity contains an
    * implicit scope.
    * @param aInvoke
    * @return boolean indicating if the ivoke's parent
    * is an implicit scope
    */
   public static boolean isParentInlinedScope(AeActivityInvokeDef aInvoke)
   {
      return aInvoke.getImplicitScope() != null; 
   }

   
   /**
    * @return boolean indicating if the scope
    * activity is actually an implicit scope
    * (can be written out directly under invoke element)
    */
   public static boolean isInlinedInvokeScope( AeActivityScopeDef aScope )
   {
      // determine if its even worth looking
      // any deeper...
      if( !isScopeOkToInline(aScope) )
      {
         return false;
      }
      
      // check to see if scope has an invoke child..      
      if( hasInvokeChild(aScope) )
      {
         AeActivityInvokeDef invoke = 
                     (AeActivityInvokeDef)aScope.getScope().getActivity();

         // check state of invoke...
         if( !isInvokeOkToInline(invoke) )
         {
            return false;         
         }
         
         // everything ok so far - check names
         return isNameMatch( aScope, invoke );
      }
      return false;
   }
   
   /**
    * @return true if the scope activity has an invoke child activity
    */
   private static boolean hasInvokeChild( AeActivityScopeDef aDef )
   {
      AeActivityDef activity = aDef.getScope().getActivity();
      return activity != null && activity instanceof AeActivityInvokeDef;
   }
   
   
   /**
    * @return true if the scope contains no eventHandlers,
    * variables or correlationSets
    */
   private static boolean isScopeOkToInline( AeActivityScopeDef aDef )
   {
      AeScopeDef scope = aDef.getScope();
      return
          ! scope.hasEventHandlers()
      &&  ! scope.hasVariables()
      &&  ! scope.hasCorrelationSets() 
      &&  ! aDef.getVariableAccessSerializable();
   }
   
   /**
    * @return true if the invoke has no source or target
    * children
    */
   private static boolean isInvokeOkToInline( AeActivityInvokeDef aInvoke )
   {
      return
          ! aInvoke.getSourceList().hasNext()
      &&  ! aInvoke.getTargetList().hasNext();
   }
   
   /**
    * @return true if the invoke and the scope activity have
    * the same name or if the invoke name is an empty string
    */
   private static boolean isNameMatch( AeActivityScopeDef aScope, AeActivityInvokeDef aInvoke)
   {
      String invokeName = aInvoke.getName().trim();
      String scopeName = aScope.getName().trim();

      return scopeName.equals(invokeName) || 
            invokeName.length() == 0;
   }
   
   /**
    * Sets the name of the invoke to match the
    * name of the scope activity and sets the scope
    * activity as the implicit scope just prior to
    * a write.
    * <br />
    * Should only be called if the scope activity
    * can be inlined.
    */
   public static void syncObjectStates( AeActivityScopeDef aScope )
   {
      getActivityInvoke(aScope).setImplicitScope( aScope );
      getActivityInvoke(aScope).setName( aScope.getName() );
   }
   
   /**
    * Sets the invoke name back to an empty string.
    */
   public static void restoreObjectStates( AeActivityScopeDef aDef )
   {
      getActivityInvoke(aDef).setName(""); //$NON-NLS-1$
   }
   
   /**
    * Accessor for invoke
    * @return invoke child activity of scope activity
    */
   public static AeActivityInvokeDef getActivityInvoke( AeActivityScopeDef aDef )
   {
      return (AeActivityInvokeDef)aDef.getActivity();
   }
   
   
   /**
    * Convenience method for examining the child elements 
    * of the invoke element to see if it contains an 
    * implicit scope.
    * @param aInvokeElement
    * @return true if the invoke contains an implicit scope
    */
   public static boolean isImplicitScope(Element aInvokeElement)
   {
      NodeList children = aInvokeElement.getChildNodes();
      for( int i = 0; i < children.getLength(); i++ )
      {
         Node node = children.item(i);
         if( Node.ELEMENT_NODE != node.getNodeType() )
         {
            continue;
         }
         
         Element element = (Element)node;
         if( isImplicitScopeElement(element) )
         {
            return true;
         }
      }

      return false;      
   }
   
   /**
    * Look for specific tag names.
    * @param aElement
    * @return true if element names matches catch, catchAll or compensationHandler
    */
   private static boolean isImplicitScopeElement(Element aElement)
   {
      String tagName = aElement.getTagName();
      return 
            TAG_CATCH.equals(tagName) || 
            TAG_CATCH_ALL.equals(tagName) ||
            TAG_COMPENSATION_HANDLER.equals(tagName);
   }


}
