// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/util/AeDefUtil.java,v 1.8 2005/10/17 19:44:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.util;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeCorrelationSetDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.AeScopeDef;
import org.activebpel.rt.bpel.def.AeVariableDef;

/**
 * Contains some helper methods for the def visitors and validation
 */
public class AeDefUtil
{
   /**
    * Locate a correlation set among the process' definitions, by name.
    * 
    * @param aCorrSet The name to search.
    * @param aDef The def object from which to search (up).
    * 
    * @return AeCorrelationSetDef if found, otherwise null.
    */   
   public static AeCorrelationSetDef getCorrSetByName( String aCorrSet, AeBaseDef aDef )
   {
      // Search the enclosing scopes
      for(AeScopeDef parent = getParentScope( aDef ); parent != null; parent = getParentScope(parent))
      {
         for ( Iterator iter = parent.getCorrelationSets() ; iter.hasNext() ; )
         {
            AeCorrelationSetDef corrSet = (AeCorrelationSetDef)iter.next();
            if ( corrSet.getName().equals( aCorrSet))
            {
               return corrSet ;
            }
         }
      }

      return null ;
   }
   
   /**
    * Uses the stack to find the element's enclosing scope, or null if there is no enclosing scope.
    * 
    * @return AeScopeDef
    */
   public static AeScopeDef getParentScope( AeBaseDef aChild )
   {
      AeBaseDef parent = aChild.getParent();
      if ( parent != null )
      {
         if ( parent instanceof AeScopeDef )
            return (AeScopeDef)parent;
         else
            return ( getParentScope( parent ));
      }
      
      return null ;
   }

   /**
    * Retrieve a variable from the process scope heirarchy, by name.
    *  
    * @param aVariable The name of the variable to search.
    * @param aDef The def object from which to search (up).
    * 
    * @return AeVariableDef or null if not found.
    */
   public static AeVariableDef getVariableByName( String aVariable, AeBaseDef aDef )
   {
      // Search all enclosing scopes.
      //
      AeScopeDef parent = getParentScope( aDef );
      while (parent != null)
      {
         AeVariableDef var = parent.getVariable(aVariable);
         if (var != null)
         {
            return var;   
         }
         else
         {
            parent = getParentScope(parent);
         }
      }
      
      return null ;
   }
   
   /**
    * Gets the path for the scope (or process) that declares the message exchange value
    * and appends the message exchange value to create a fully qualified path for
    * the message exchange.
    * @param aMessageExchange
    * @param aDef - the def here is the receive, onMessage, or reply def
    */
   public static String getFullPathForMessageExchange(String aMessageExchange, AeBaseDef aDef)
   {
      // Search all enclosing scopes.
      //
      String found = null;
      AeScopeDef parent = getParentScope( aDef );
      while (parent != null && found == null)
      {
         found = parent.declaresMessageExchange(aMessageExchange) ? parent.getLocationPath()  + "/" + aMessageExchange: null; //$NON-NLS-1$
         if (found == null)
         {
            parent = getParentScope(parent);
         }
      }
      
      return found;
   }

   /**
    * Given a string which represents a qualified name and a context, this method will parse
    * it and return a QName object. Note this method returns null if the string
    * is not properly formed.
    * @param aDef the def from which to resolve the prefix to a namespace 
    * @param aQstr the string to be parsed 
    */
   public static QName parseQName(AeBaseDef aDef, String aQstr)
   {
      QName qname;
      
      int colon = aQstr.indexOf(":"); //$NON-NLS-1$
      if (colon < 0 )
         qname = new QName(null, aQstr);
      else
      {
         String nsURI = translateNamespacePrefixToUri(aDef, aQstr.substring(0, colon));
         qname = new QName(nsURI, aQstr.substring(colon+1));
      }
      
      return qname;
   }
   
   /**
    * Returns the namespace associated with the prefix from the associated model.
    */
   private static String translateNamespacePrefixToUriInternal(AeBaseDef aDef, String aPrefix)
   {
      // check def for the prefix we are looking for
      String nsURI = aDef.findNamespace(aPrefix);
      
      // if we haven't found the prefix ask our parent if it can find it
      if(nsURI == null)
      {
         AeBaseDef parent = aDef.getParent(); 
         if(parent != null)
            nsURI = translateNamespacePrefixToUriInternal(parent, aPrefix);
      }
      return nsURI;
   }
   
   /**
    * Returns the namespace associated with the prefix from the associated model.
    * TODO we should cache prefixes in process for non-conflicting prefixes and only do brute force search for conflicts.
    */
   public static String translateNamespacePrefixToUri(AeBaseDef aDef, String aPrefix)
   {
      String nsURI = null;
      
      // if we have a definiton test if it has the prefix we are looking for
      if(aDef != null)
      {
         String fullPrefix = "xmlns:" + aPrefix; //$NON-NLS-1$
         nsURI = translateNamespacePrefixToUriInternal(aDef, fullPrefix);
      }
      return nsURI;
   }

   /**
    * Returns a set of prefixes mapped to the given namespace.
    * 
    * @param aDef
    * @param aNamespace
    */
   public static Set getPrefixesForNamespaceInternal(AeBaseDef aDef, String aNamespace)
   {
      HashSet set = new HashSet();

      // First add the parent prefixes if there is a parent.
      if (aDef.getParent() != null)
      {
         set.addAll(getPrefixesForNamespaceInternal(aDef.getParent(), aNamespace));
      }
      set.addAll(aDef.getPrefixesForNamespace(aNamespace));

      return set;
   }

   /**
    * Returns a set of prefixes that are mapped to the given namespace.
    * 
    * @param aDef
    * @param aNamespace
    */
   public static Set getPrefixesForNamespace(AeBaseDef aDef, String aNamespace)
   {
      if (aDef != null)
      {
         return getPrefixesForNamespaceInternal(aDef, aNamespace);
      }
      else
      {
         return Collections.EMPTY_SET;
      }

   }

   /**
    * Gets the root process def parent of a given def object.
    * 
    * @param aDef
    */
   public static AeProcessDef getProcessDef(AeBaseDef aDef)
   {
      if (aDef instanceof AeProcessDef)
      {
         return (AeProcessDef) aDef;
      }

      AeBaseDef def = null;
      for (def = aDef; def != null && !(def instanceof AeProcessDef); def = def.getParent() );
      return (AeProcessDef) def;
   }
   
}
