// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/validation/AeLinkValidator.java,v 1.12 2005/10/17 19:44:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.validation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.def.AeActivityDef;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeEventHandlersDef;
import org.activebpel.rt.bpel.def.AeFaultHandlerDef;
import org.activebpel.rt.bpel.def.AeFaultHandlersContainerDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.AeScopeCompensatorContainerDef;
import org.activebpel.rt.bpel.def.IAeMultipleActivityContainerDef;
import org.activebpel.rt.bpel.def.IAeSingleActivityContainerDef;
import org.activebpel.rt.bpel.def.activity.AeActivityFlowDef;
import org.activebpel.rt.bpel.def.activity.AeActivityPickDef;
import org.activebpel.rt.bpel.def.activity.AeActivityScopeDef;
import org.activebpel.rt.bpel.def.activity.AeActivitySwitchDef;
import org.activebpel.rt.bpel.def.activity.AeActivityWhileDef;
import org.activebpel.rt.bpel.def.activity.support.AeLinkDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnAlarmDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef;
import org.activebpel.rt.bpel.def.activity.support.AeOtherwiseDef;
import org.activebpel.rt.bpel.def.activity.support.AeSourceDef;
import org.activebpel.rt.bpel.def.activity.support.AeSwitchCaseDef;
import org.activebpel.rt.bpel.def.activity.support.AeTargetDef;

/**
 * Validation helper for link collection and testing.
 */
public class AeLinkValidator implements IAeValidationDefs
{
   /** Error reporting instance. */
   private IAeBaseErrorReporter mReporter ;

   /** Map holding link composites to test for validity. */
   private HashMap mLinkMap ;
   
   /** List of source composites to test for cycles. */
   private ArrayList mLinkSources ;

   /** Process Def. */
   private AeProcessDef mProcessDef;

   /** Root scope locations for boundary cross checks. */   
   private AeBaseDef mActyScope ;
   private AeBaseDef mFaultScope ;
   private AeBaseDef mEventScope ;
   private AeBaseDef mCompScope ;

   /**
    * Ctor to set the error reporter.
    * 
    * @param aReporter The reporter to set.
    */
   public AeLinkValidator( AeProcessDef aProcessDef, IAeBaseErrorReporter aReporter )
   {
      mProcessDef = aProcessDef;
      setReporter( aReporter );
   }
   
   /**
    * Get the error reporter for this instance.
    * 
    * @return IAeErrorReporter
    */
   public IAeBaseErrorReporter getReporter()
   {
      return mReporter;
   }

   /**
    * Set the error reporter for this instance.
    * 
    * @param aReporter The reporter to set.
    */
   public void setReporter(IAeBaseErrorReporter aReporter)
   {
      mReporter = aReporter;
   }

   /**
    * Find the element's enclosing flow, or null if there is no enclosing flow.
    * 
    * @return AeActivityFlowDef
    */
   private AeActivityFlowDef getParentFlow( AeBaseDef aChild )
   {
      AeBaseDef parent = aChild.getParent();
      if ( parent != null )
      {
         if ( parent instanceof AeActivityFlowDef )
            return (AeActivityFlowDef)parent ;
         else
            return getParentFlow( parent );
      }

      return null ;
   }

   /**
    * Check for cycles and illegal boundary crosses in the graph's links.
    * 
    * Note: this should only be called after the validation visitor has traversed
    * the entire def graph, to ensure that all links and their sources/targets have
    * been resolved.  Wherever this hasn't occurred, the composite will be incomplete,
    * which signifies an erroneously defined link.
    */
   public void checkLinks()
   {
      for ( Iterator iter = getLinkMap().keySet().iterator() ; iter.hasNext() ; )
      {
         String key = (String)iter.next();
         AeLinkComposite comp = (AeLinkComposite)getLinkMap().get( key );
         if ( comp.isComplete())
         {
            checkBoundaryCrossings( comp );
            checkBadTarget( comp );
         }
         else
         {
            getReporter().addError( ERROR_BAD_LINK, 
                                    new String[] { comp.getLink().getLocationPath() },
                                    comp.getLink());
         }
      }

      checkLinkCycles();
   }
   
   /**
    * Recursively look for a cycle in the graph.
    * 
    * @param aStartNode The node from which to start checking.
    * @param aNextNode The next node in the chain we're testing.
    * @param aCycle The array of links that form the cycle, otherwise null.
    * @param aNotCycle The array of links that aren't part of a cycle.
    * @param aChecked The array of nodes checked for this target.
    * 
    * @return boolean True if the links eventually point back to the target.
    * 
    * TODO This should be revisited and implemented with a visitor pattern
    * 
    * TODO 'Automatic' (non-model) links can be part of a link cycle. Right now we detect most container 
    * type cycles, but we do miss the Sequence activity when we link to an activity within the sequence. 
    */
   private boolean findCycleInGraph( AeActivityDef aStartNode, AeActivityDef aNextNode, ArrayList aCycle, ArrayList aNotCycle, ArrayList aChecked )
   {
      // Follow all links from the next node to see if any of them eventually point to the target node.
      boolean foundCycle = false ;
      if ( aNextNode != null && getLinkSource( aNextNode ) != null )
      {
         for(Iterator iter = getLinkSource( aNextNode ).getSourceConnections();
             iter.hasNext(); )
         {
            AeLinkComposite link = (AeLinkComposite)iter.next();
            if ( link.isComplete())
            {
               if ( link.getTarget().equals( aStartNode ))
               {
                  // This link points back to the original target node.
                  aCycle.add( link );
                  foundCycle = true ;
               }
               else if ( !link.getTarget().equals( aNextNode ) && 
                           !aChecked.contains( link.getTarget() ))
               {
                  // Temporarily add the current target as 'checked', since we are checking it for this path.
                  // This is in case there's a link that points straight back to it, which will cause problems.
                  aChecked.add( link.getTarget());
                  if ( findCycleInGraph( aStartNode, link.getTarget(), aCycle, aNotCycle, aChecked ))
                  {
                     // This link is part of a cycle that points back to the original target node.
                     aCycle.add( link );
                     foundCycle = true ;
                  }
                  
                  aChecked.remove( link.getTarget());
               }
               else
               {
                  // Add this link to the 'checked' list so we don't follow
                  // it again as part of a different cycle.
                  aNotCycle.add( link );
               }
            }
         }
      }
      
      // if we haven't found a cycle for direct links check if children cycle
      if(! foundCycle)
          foundCycle = findCycleInContainer(aStartNode, aNextNode, aCycle, aNotCycle, aChecked);
         
      return foundCycle ;
   }
   
   /**
    * Recursively look for a cycle in container activities of the graph.
    * 
    * @param aStartNode The node from which to start checking.
    * @param aNextNode The next node in the chain we're testing.
    * @param aCycle The array of links that form the cycle, otherwise null.
    * @param aNotCycle The array of links that aren't part of a cycle.
    * @param aChecked The array of nodes checked for this target.
    * 
    * @return boolean True if the links eventually point back to the target.
    */
   private boolean findCycleInContainer(AeActivityDef aStartNode, AeActivityDef aNextNode, ArrayList aCycle, ArrayList aNotCycle, ArrayList aChecked)
   {
      boolean foundCycle = false;
      
      // Handles WHILE and SCOPE activities
      if(aNextNode instanceof IAeSingleActivityContainerDef)
      {
         AeActivityDef def = ((IAeSingleActivityContainerDef)aNextNode).getActivity();
         if (def != null)
            foundCycle = findCycleInGraph(aStartNode, def, aCycle, aNotCycle, aChecked);
         
         // Handle SCOPE activities
         if(! foundCycle && aNextNode instanceof AeActivityScopeDef)
         {
            AeActivityScopeDef scope = (AeActivityScopeDef)aNextNode;
            for (Iterator iter = scope.getFaultHandlerContainer().getFaultHandlers(); iter.hasNext() && ! foundCycle;)
            {
               AeFaultHandlerDef faultHandler = (AeFaultHandlerDef)iter.next();
               def = (AeActivityDef)faultHandler.getActivity();
               if (def != null)
                  foundCycle = findCycleInGraph(aStartNode, def, aCycle, aNotCycle, aChecked);
            }
            
            if (! foundCycle && scope.getFaultHandlerContainer().getDefaultFaultHandler() != null)
            {
               def = scope.getFaultHandlerContainer().getDefaultFaultHandler().getActivity();
               if (def != null)
                  foundCycle = findCycleInGraph(aStartNode, def, aCycle, aNotCycle, aChecked);
            }
            
            AeEventHandlersDef evtDef = scope.getScope().getEventHandlers();
            if (! foundCycle && evtDef != null)
            {
               for(Iterator iter = evtDef.getAlarmList(); iter.hasNext() && ! foundCycle; )
               {
                  AeOnAlarmDef alarm = (AeOnAlarmDef)iter.next();
                  if (alarm.getActivity() != null)
                     foundCycle = findCycleInGraph(aStartNode, alarm.getActivity(), aCycle, aNotCycle, aChecked);
               }
               for(Iterator iter = evtDef.getMessageList(); iter.hasNext() && ! foundCycle;)
               {
                  AeOnMessageDef msg = (AeOnMessageDef)iter.next();
                  if (msg.getActivity() != null)
                     foundCycle = findCycleInGraph(aStartNode, msg.getActivity(), aCycle, aNotCycle, aChecked);
               }
            }
         }
      }
      // Handle FLOW and SEQUENCE activities
      else if(aNextNode instanceof IAeMultipleActivityContainerDef)
      {
         Iterator iter = ((IAeMultipleActivityContainerDef)aNextNode).getActivityList();  
         while (iter.hasNext() && ! foundCycle)
            foundCycle = findCycleInGraph(aStartNode, (AeActivityDef)iter.next(), aCycle, aNotCycle, aChecked);
      }
      // Handle PICK activities
      else if(aNextNode instanceof AeActivityPickDef)
      {
         AeActivityPickDef pick = (AeActivityPickDef)aNextNode;
         for(Iterator iter = pick.getAlarmList(); iter.hasNext() && ! foundCycle; )
         {
            AeOnAlarmDef alarm = (AeOnAlarmDef)iter.next();
            if (alarm.getActivity() != null)
               foundCycle = findCycleInGraph(aStartNode, alarm.getActivity(), aCycle, aNotCycle, aChecked);
         }
         for(Iterator iter = pick.getMessageList(); iter.hasNext() && ! foundCycle;)
         {
            AeOnMessageDef msg = (AeOnMessageDef)iter.next();
            if (msg.getActivity() != null)
               foundCycle = findCycleInGraph(aStartNode, msg.getActivity(), aCycle, aNotCycle, aChecked);
         }
      }
      // Handle SWITCH activities
      else if(aNextNode instanceof AeActivitySwitchDef)
      {
         AeActivitySwitchDef swtch = (AeActivitySwitchDef)aNextNode;
         for (Iterator iter = swtch.getCaseList(); iter.hasNext() && ! foundCycle;)
         {
            AeSwitchCaseDef switchCase = (AeSwitchCaseDef)iter.next();
            if (switchCase.getActivity() != null)
               foundCycle = findCycleInGraph(aStartNode, switchCase.getActivity(), aCycle, aNotCycle, aChecked);
         }
         
         if (swtch.getOtherwise() != null)
         {
            AeOtherwiseDef otherwise = (AeOtherwiseDef)swtch.getOtherwise();
            if (otherwise.getActivity() != null)
               foundCycle = findCycleInGraph(aStartNode, otherwise.getActivity(), aCycle, aNotCycle, aChecked);
         }
      }
      
      return foundCycle;
   }
   
   /**
    * Check for cycles in the graph's links.
    */
   private void checkLinkCycles()
   {
      ArrayList flagged = new ArrayList();
      ArrayList checked = new ArrayList();
      if ( getLinkSources() != null && getLinkSources().size() > 0 )
      {
         Iterator iter = getLinkSources().iterator();
         while ( iter.hasNext())
         {
            // Check the link path from each source until it either
            //  dead-ends or comes back to this source.
            //
            ArrayList cycle = new ArrayList();
            ArrayList notCycle = new ArrayList();
            AeActivityDef targetNode = ((AeLinkSource)iter.next()).getSourceDef();
            if ( findCycleInGraph( targetNode, targetNode, cycle, notCycle, checked ))
            {
               // Flag all the unflagged links in the cycle.
               //
               Iterator iter2 = cycle.iterator();
               while ( iter2.hasNext())
               {
                  AeLinkDef link = ((AeLinkComposite)iter2.next()).getLink();
                  if ( !flagged.contains( link ))
                  {
                     getReporter().addError( ERROR_LINK_CYCLE, new String[] { link.getLocationPath() }, link );
                     flagged.add( link );
                  }
               }
            }

            checked.add( targetNode );
         }
      }
   }
   
   /**
    * Returns a string array containing the appropriate error text.
    * 
    * @param aParent Node type involved in the error.
    * 
    * return String
    */
   private String getCrossingError( AeBaseDef aParent )
   {
      if ( aParent instanceof AeActivityWhileDef )
      {
         return AeMessages.getString("AeLinkValidator.0"); //$NON-NLS-1$
      }
      else if ( aParent instanceof AeActivityScopeDef && ((AeActivityScopeDef)aParent).getVariableAccessSerializable())
      {
         return AeMessages.getString("AeLinkValidator.1"); //$NON-NLS-1$
      }
      else if ( aParent instanceof AeFaultHandlersContainerDef )
      {
         return AeMessages.getString("AeLinkValidator.2"); //$NON-NLS-1$
      }
      else if ( aParent instanceof AeEventHandlersDef )
      {
         return AeMessages.getString("AeLinkValidator.3"); //$NON-NLS-1$
      }
      else if ( aParent instanceof AeScopeCompensatorContainerDef )
      {
         return AeMessages.getString("AeLinkValidator.4"); //$NON-NLS-1$
      }

      return AeMessages.getString("AeLinkValidator.5") + aParent.getClass().toString(); //$NON-NLS-1$
   }
   
   /**
    * Determine whether or not this node resides in an uncrossable container.
    * 
    * @param aNode The node to check.
    * 
    * @return AeBaseDef A While, etc., activity that has uncrossable boundary.
    */
   private AeBaseDef getInnermostUncrossableContainer( AeBaseDef aNode )
   {
      if ( mActyScope == null )
      {
         // Init scopes.
         //
         mActyScope = getProcessDefNode();
         mFaultScope = getProcessDefNode().getFaultHandlerContainer();
         mEventScope = getProcessDefNode().getEventHandlers();
         mCompScope = getProcessDefNode().getCompensatorContainer();
      }

      AeBaseDef parent = aNode.getParent();
      if ( parent == null )
      {
         // No parent - none found.
         //
         return null ;
      }
      else if ( parent.equals( mActyScope) ||
                  parent.equals( mFaultScope ) || 
                  parent.equals( mEventScope ) || 
                  parent.equals( mCompScope )) 
      {
         // Root scope - none found.
         //
         return null ;
      }
      else if ( parent instanceof AeActivityWhileDef ||
                  parent instanceof AeFaultHandlersContainerDef ||
                  parent instanceof AeScopeCompensatorContainerDef  ||
                  parent instanceof AeEventHandlersDef ||
                  ( parent instanceof AeActivityScopeDef && 
                     ((AeActivityScopeDef)parent).getVariableAccessSerializable()) )
      {
         // Uncrossable type - return it.
         //
         return parent ;
      }
      else
      {
         // None of the above - recurse.
         //
         return getInnermostUncrossableContainer( parent ); 
      }
   }

   /**
    * Verify that a link doesn't cross an uncrossable boundary.  Flag an error
    * if so.
    * 
    * @param aLink The link to check.
    */   
   private void checkBoundaryCrossings( AeLinkComposite aLink )
   {
      AeActivityDef src = aLink.getSource();
      AeActivityDef dst = aLink.getTarget();
      AeBaseDef srcParent = getInnermostUncrossableContainer( src );
      AeBaseDef dstParent = getInnermostUncrossableContainer( dst );
      if ( srcParent != null )
      {
         // Check for fault handler / direction.
         //
         if ( srcParent instanceof AeFaultHandlersContainerDef )
            return ;
         
         if ( dstParent == null || ( dstParent == null && !dstParent.equals( srcParent )))
         {
            getReporter().addError( ERROR_LINK_CROSSING, 
                                    new String[] { aLink.getLink().getLocationPath(), getCrossingError( srcParent )},
                                    aLink.getLink());
         }
      }
      else if ( dstParent != null )
      {
         if (( dstParent instanceof AeFaultHandlersContainerDef ) ||
             ( srcParent == null || (srcParent == null && !srcParent.equals( dstParent ))))
            getReporter().addError( ERROR_LINK_CROSSING, 
                                    new String[] { aLink.getLink().getLocationPath(), getCrossingError( dstParent )},
                                    aLink.getLink());
      }
   }

   /**
    * Verify that link target is not nested in src.
    * @param aLink The link to check.
    */   
   private void checkBadTarget( AeLinkComposite aLink )
   {
      AeActivityDef src = aLink.getSource();
      AeActivityDef dst = aLink.getTarget();
      
      // If the source of this link is a parent of the destination it is an error
      for(AeBaseDef dstModel = dst; dstModel != null; dstModel = dstModel.getParent() )
      {
         if ( dstModel == src )
         {
            getReporter().addError( ERROR_SCOPE_LINK, new String[] {aLink.getLink().getLocationPath()},
                                    aLink.getLink() );
            break;
         }
      }
   }

   /**
    * Find an AeLinkDef by name within an enclosing Flow.
    * 
    * @param aLinkName The name of the link to search for.
    * @param aDef The def object 'above' which the search starts.
    * 
    * @return AeLinkDef
    */
   public AeLinkDef findLink( String aLinkName, AeBaseDef aDef )
   {
      AeActivityFlowDef parent = getParentFlow( aDef );
      while ( parent != null )
      {
         for ( Iterator iter = parent.getLinkList() ; iter.hasNext() ; )
         {
            AeLinkDef link = (AeLinkDef)iter.next();
            if ( link.getName().equals( aLinkName ))
               return( link ); 
         }
         
         parent = getParentFlow( parent );
      }

      return null ;
   }

   /**
    * Get the map used to manage the AeLinkComposite instances.
    * 
    * @return HashMap
    */   
   public HashMap getLinkMap()
   {
      if ( mLinkMap == null )
         mLinkMap = new HashMap();
         
      return mLinkMap ;
   }

   /**
    * Gets the set of link names from the current link map.
    * 
    * @return Set
    */
   public Set getLinkNames()
   {
      HashSet set = new HashSet();
      HashMap map = getLinkMap();
      for (Iterator iter = map.values().iterator(); iter.hasNext(); )
      {
         AeLinkComposite composite = (AeLinkComposite) iter.next();
         set.add(composite.getLink().getName());
      }
      return set;
   }
   
   /**
    * Get the list used to manage the AeLinkSource instances.
    * 
    * @return ArrayList
    */   
   public ArrayList getLinkSources()
   {
      if ( mLinkSources == null )
         mLinkSources = new ArrayList();
         
      return mLinkSources ;
   }

   /**
    * Create a link map key from the parent flow's location path and link name.
    *  
    * @param aDef The object whose parent flow is used.
    * @param aLinkName The link name used.
    * 
    * @return String
    */   
   protected String makeLinkMapKey( AeBaseDef aDef, String aLinkName )
   {
      AeLinkDef link = findLink( aLinkName, aDef );
      if ( link != null )
         return link.getLocationPath();
      else
         return null ;
   }
   
   /**
    * Get the link composite from the map, if there, otherwise create it.
    * 
    * @param aKey Key to search the map.
    * 
    * @return AeLinkComposite
    */
   protected AeLinkComposite getLinkComposite( String aKey )
   {
      AeLinkComposite comp = (AeLinkComposite)getLinkMap().get( aKey );
      if ( comp == null )
         comp = new AeLinkComposite( aKey );

      return comp ;
   }
      
   /**
    * Add an AeLinkDef to the map.
    * 
    * @param aLink The link to add.
    */
   public void addLink( AeLinkDef aLink )
   {
      AeLinkComposite comp = getLinkComposite( aLink.getLocationPath());
      comp.mLink = aLink ;
   }
      
   /**
    * Add a source activity to the map.
    * 
    * @param aSource The source to add.
    * @param aDef The activity def that owns the source.
    */
   public void addSource( AeSourceDef aSource, AeActivityDef aDef )
   {
      String key = makeLinkMapKey( aDef, aSource.getLinkName());
      if ( key != null )
      {
         AeLinkComposite comp = getLinkComposite( key );
         
         // Make certain that this source doesn't already exist.
         if ( comp.mSource != null)
         {
            // Source already exists, therefore this link erroreously has multiple sources!
            getReporter().addError( ERROR_MULTI_SRC_LINK, 
                                    new String[] { comp.getLink().getLocationPath() },
                                    comp.getLink() );
            return;
         }
         
         comp.mSource = aDef ;
   
         // Record this link source def for later cycle check.
         //      
         AeLinkSource src = getLinkSource( aDef );
         if ( src == null )
         {
            src = new AeLinkSource( aDef );
            addLinkSource( src );
         }
   
         src.addConnection( comp );
      }
      else
      {
         // the key was null so we didn't find our link definition
         // this is a problem.
         getReporter().addError( ERROR_BAD_LINK, 
               new String[] { aDef.getLocationPath() }, aDef );
      }
         
   }

   /**
    * Add a target activity to the map.
    * 
    * @param aTarget The target to add.
    * @param aDef The activity def that owns the target.
    */
   public void addTarget( AeTargetDef aTarget, AeActivityDef aDef )
   {
      String key = makeLinkMapKey( aDef, aTarget.getLinkName());
      if ( key != null )
      {
         AeLinkComposite comp = getLinkComposite( key );
         
         // Make certain that this target doesn't already exist.
         if ( comp.mTarget != null)
         {
            // Target already exists, therefore this link erroreously has multiple targets!
            getReporter().addError( ERROR_MULTI_TARGET_LINK, 
                                    new String[] { comp.getLink().getLocationPath() },
                                    comp.getLink() );
            return;
         }
         comp.mTarget = aDef ;
      }
      else
      {
         // the key was null so we didn't find our link definition
         // this is a problem.
         getReporter().addError( ERROR_BAD_LINK, 
               new String[] { aDef.getLocationPath() }, aDef );
      }
   }
   
   /**
    * Search for the link source that corresponds to the def provided.
    * 
    * @param aDef The source def to search for.
    * 
    * @return AeLinkSource
    */
   private AeLinkSource getLinkSource( AeActivityDef aDef )
   {
      for ( Iterator iter = getLinkSources().iterator() ; iter.hasNext() ; )
      {
         AeLinkSource src = (AeLinkSource)iter.next();
         if ( aDef.equals( src.getSourceDef() ))
            return src ;
      }
      
      return null ;
   }
   
   /**
    * Add a link source instance to the list.
    * 
    * @param aSrc The link source to add.
    */
   private void addLinkSource( AeLinkSource aSrc )
   {
      if ( !getLinkSources().contains( aSrc ))
         getLinkSources().add( aSrc );
   }
   
   /**
    * Accessor for the root process def.
    */
   private AeProcessDef getProcessDefNode()
   {
      return mProcessDef;
   }
   
   /**
    * Inner class to facilitate cycle checking.
    *
    */
   public class AeLinkSource
   {
      /** List of link composites for which this instance is a source. */
      private ArrayList mSources = new ArrayList();
      
      /** Activity associated with this source. */
      private AeActivityDef mSource ;
      
      /**
       * Ctor to set the source def.
       * 
       * @param aSource The source activity to set.
       */
      public AeLinkSource( AeActivityDef aSource )
      {
         mSource = aSource ;
      }
      
      /**
       * Add a link / connection to this source's list of links.
       * 
       * @param aConnection The link to add.
       */
      public void addConnection( AeLinkComposite aConnection )
      {
         mSources.add( aConnection );
      }
      
      /**
       * Get the iterator over current list of link sources.
       * 
       * @return Iterator
       */
      public Iterator getSourceConnections()
      {
         return mSources.iterator();
      }
      
      /**
       * Get the activity associated with this list of source links.
       * 
       * @return AeActivityDef
       */
      public AeActivityDef getSourceDef()
      {
         return mSource ;
      }
   }

   /**
    * Composite link inner class used to manage validation of link definitions.
    */
   public class AeLinkComposite
   {
      private AeActivityDef mSource ;
      private AeActivityDef mTarget ;
      private AeLinkDef mLink ;

      /**
       * Create a composite with link name.
       * 
       * @param aName The name of the link for this composite.
       */
      public AeLinkComposite( String aName )
      {
         getLinkMap().put( aName, this );
      }
      
      /**
       * Indicate whether the composite has all its parts.
       * 
       * @return boolean <code>true</code> if the source, target and link elements are defined.
       */
      public boolean isComplete()
      {
         if ( mSource == null || mTarget == null || mLink == null )
            return false ;
            
         return true ;
      }
      
      /**
       * Get the link for this composite.
       * 
       * @return AeLinkDef
       */
      public AeLinkDef getLink()
      {
         return mLink;
      }

      /**
       * Get the source node for this composite.
       * 
       * @return AeActivityDef
       */
      public AeActivityDef getSource()
      {
         return mSource;
      }

      /**
       * Get the target node for this composite.
       * 
       * @return AeActivityDef
       */
      public AeActivityDef getTarget()
      {
         return mTarget;
      }
   }
   
}
