//$Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/validation/AeMessageExchangeValidationVisitor.java,v 1.1 2006/01/11 21:01:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.validation; 

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;

import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReplyDef;
import org.activebpel.rt.bpel.def.activity.IAeReceiveActivityDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef;
import org.activebpel.rt.bpel.def.util.AeDefUtil;
import org.activebpel.rt.bpel.def.visitors.AeAbstractDefVisitor;
import org.activebpel.rt.bpel.def.visitors.AeDefTraverser;
import org.activebpel.rt.bpel.def.visitors.AeTraversalVisitor;
import org.activebpel.rt.util.AeUtil;

public class AeMessageExchangeValidationVisitor extends AeAbstractDefVisitor
{
   /** error message for unmatched message exchange value */
   protected static final String UNMATCHED_MESSAGE_EXCHANGE = AeMessages.getString("AeMessageExchangeValidationVisitor.UnmatchedMessageExchange"); //$NON-NLS-1$
   
   protected static final String MISSING_REPLY = AeMessages.getString("AeMessageExchangeValidationVisitor.MissingReply"); //$NON-NLS-1$

   /** error message for a receive/onMessage/reply that references a message exchange value that is not declared w/in the enclosing scope(s) */
   protected static final String UNDECLARED_MESSAGE_EXCHANGE = AeMessages.getString("AeMessageExchangeValidationVisitor.UndeclaredMessageExchange"); //$NON-NLS-1$
   
   /** Error reporter used during validation process. */
   private IAeBaseErrorReporter mReporter ;

   /** set of plink.operation.messageExchange, all replies are matched against
    *  this set to assert that we have matched receives/replies */
   private Set mMessageExchangeReceives = new HashSet();

   /** Collection of replies that have been visited - used to assert a match w/ receives/replies for messageExchange */
   private Collection mReplies = new LinkedList();
   
   /** collection of receives and onMessages */
   private Collection mReceiveDefs = new LinkedList();

   public AeMessageExchangeValidationVisitor(IAeBaseErrorReporter aErrorReporter)
   {
      setTraversalVisitor( new AeTraversalVisitor( new AeDefTraverser(), this ) );
      setReporter(aErrorReporter);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.AeAbstractDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef)
    */
   public void visit(AeActivityReceiveDef aDef)
   {
      visit((IAeReceiveActivityDef)aDef);
      super.visit(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.AeAbstractDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef)
    */
   public void visit(AeOnMessageDef aDef)
   {
      visit((IAeReceiveActivityDef)aDef);
      super.visit(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.AeAbstractDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReplyDef)
    */
   public void visit(AeActivityReplyDef aDef)
   {
      // Note: the message exchange values for replies are validated after the 
      //       whole process has been visited.
      mReplies.add(aDef);
      super.visit(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.AeAbstractDefVisitor#visit(org.activebpel.rt.bpel.def.AeProcessDef)
    */
   public void visit(AeProcessDef aDef)
   {
      super.visit(aDef);
      validateMessageExchangeForReplies();
   }
   
   /**
    * Validates synchronous receives
    * @param aDef
    */
   protected void visit(IAeReceiveActivityDef aDef)
   {
      if (!aDef.isOneWay())
      {
         if (validateReceivingWithMessageExchange(aDef))
            mReceiveDefs.add(aDef);
      }
   }
   
   /**
    * Validates that the receive or onMessage def has a messageExchange value
    * that is currently within scope and then adds the fully qualified path of
    * the messageExchange to a collection that we'll validate the replies against
    * later.
    * @param aDef
    * @return boolean if messageExchange value is valid
    */
   protected boolean validateReceivingWithMessageExchange(IAeReceiveActivityDef aDef)
   {
      String fullPath = validateMessageExchangeDeclared((AeBaseDef)aDef, aDef.getMessageExchange());
      
      if (fullPath != null)
         mMessageExchangeReceives.add(makeMessageExchangeKey(aDef.getPartnerLink(), aDef.getOperation(), fullPath));
      
      return fullPath != null;
   }
   

   /**
    * Gets the fully qualified path for the message exchange and returns that path.
    * An error message will be reported if it's not found.
    * @param aDef
    * @param aMessageExchange value from the messageExchange attribute on the receive/onMessage/reply (must not be null)
    * @return fully qualified path to messageExchange declaration or null if not found
    */
   protected String validateMessageExchangeDeclared(AeBaseDef aDef, String aMessageExchange)
   {
      // determine if the message exchange is declared w/in our scope
      String fullPath = AeDefUtil.getFullPathForMessageExchange(aMessageExchange, aDef);
      
      if (AeUtil.isNullOrEmpty(fullPath))
      {
         // The receive/onMessage/reply activity at ''{0}'' contains a messageExchange value of ''{1}'' that is not in scope.
         Object[] args = { aDef.getLocationPath(), aMessageExchange };
         getReporter().addWarning( UNDECLARED_MESSAGE_EXCHANGE, args, aDef );//$NON-NLS-1$
         return null;
      }
      else
         return fullPath;
   }

   /**
    * Makes a key for the message exchange set.
    * @param aPartnerLink
    * @param aOperation
    * @param aMessageExchange
    */
   protected String makeMessageExchangeKey(String aPartnerLink, String aOperation, String aMessageExchange)
   {
      StringBuffer buffer = new StringBuffer(aPartnerLink);
      buffer.append('.');
      buffer.append(aOperation);
      buffer.append('.');
      buffer.append(aMessageExchange);
      return buffer.toString();
   }

   /**
    * walks all of the replies and asserts that they're properly matched to receives.
    */
   protected void validateMessageExchangeForReplies()
   {
      // validates that the replies match to receives. This will find any replies that have invalid messageExchange values
      Set replyPaths = new HashSet();
      for (Iterator iter = mReplies.iterator(); iter.hasNext();)
      {
         AeActivityReplyDef def = (AeActivityReplyDef) iter.next();
         
         String fullMessageExchangePath = AeDefUtil.getFullPathForMessageExchange(def.getMessageExchange(), def);
         
         String key = makeMessageExchangeKey(def.getPartnerLink(), def.getOperation(), fullMessageExchangePath);
         if (!mMessageExchangeReceives.contains(key))
         {
            // The reply activity {0} has no corresponding receive with partner link ''{1}'', operation ''{2}'', and message exchange ''{3}''
            Object[] args = { def.getLocationPath(), def.getPartnerLink(), def.getOperation(), def.getMessageExchange() };
            getReporter().addWarning( UNMATCHED_MESSAGE_EXCHANGE, args, def );//$NON-NLS-1$
         }
         else
         {
            replyPaths.add(key);
         }
      }
      
      // walk the receives that had valid messageExchange values and make sure that
      // they match to replies
      for (Iterator iter = mReceiveDefs.iterator(); iter.hasNext();)
      {
         IAeReceiveActivityDef receiveDef = (IAeReceiveActivityDef) iter.next();
         String fullMessageExchangePath = AeDefUtil.getFullPathForMessageExchange(receiveDef.getMessageExchange(), (AeBaseDef) receiveDef);
         String key = makeMessageExchangeKey(receiveDef.getPartnerLink(), receiveDef.getOperation(), fullMessageExchangePath);
         if (!replyPaths.contains(key))
         {
            // The receive activity {0} has no corresponding reply with partner link ''{1}'', operation ''{2}'', and message exchange ''{3}''
            Object[] args = { receiveDef.getLocationPath(), receiveDef.getPartnerLink(), receiveDef.getOperation(), receiveDef.getMessageExchange() };
            getReporter().addWarning( MISSING_REPLY, args, receiveDef );//$NON-NLS-1$
         }
      }
   }

   /**
    * @return Returns the reporter.
    */
   protected IAeBaseErrorReporter getReporter()
   {
      return mReporter;
   }

   /**
    * @param aReporter The reporter to set.
    */
   protected void setReporter(IAeBaseErrorReporter aReporter)
   {
      mReporter = aReporter;
   }
}
