//$Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/validation/expr/AeAbstractExpressionValidator.java,v 1.4 2006/01/31 20:14:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.validation.expr;

import java.text.MessageFormat;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeNamespaceContext;
import org.activebpel.rt.bpel.def.AeActivityDef;
import org.activebpel.rt.bpel.def.AeBaseDefNamespaceContext;
import org.activebpel.rt.bpel.def.expr.AeExpressionLanguageUtil;
import org.activebpel.rt.bpel.def.expr.AeExpressionParserContext;
import org.activebpel.rt.bpel.def.expr.AeScriptFuncDef;
import org.activebpel.rt.bpel.def.expr.IAeExpressionParseResult;
import org.activebpel.rt.bpel.def.expr.IAeExpressionParser;
import org.activebpel.rt.bpel.def.expr.IAeExpressionParserContext;
import org.activebpel.rt.bpel.def.util.AeVariableData;
import org.activebpel.rt.bpel.def.util.AeVariableProperty;
import org.activebpel.rt.bpel.def.validation.IAeValidationDefs;
import org.activebpel.rt.util.AeUtil;

/**
 * Base class that expression validators may use to perform common functionality.
 */
public abstract class AeAbstractExpressionValidator implements IAeExpressionValidator, IAeValidationDefs
{
   /** ActiveBPEL function  namespace. */
   public final static String ACTIVE_BPEL_FUNCTION_NAMESPACE = "http://www.activebpel.org/bpel/extension"; //$NON-NLS-1$
   
   /** The set of functions allowed in join conditions. */
   public static Set sJoinConditionAllowedFunctions = new HashSet();
   /** The namespace that all built in functions are a part of. */
   public final static String BUILTIN_FUNCTION_NAMESPACE = null;

   /**
    * Populate the sJoinConditionAllowedFunctions map.
    */
   static
   {
      sJoinConditionAllowedFunctions.add(AeExpressionLanguageUtil.LINK_STATUS_FUNC);
      sJoinConditionAllowedFunctions.add(new QName(BUILTIN_FUNCTION_NAMESPACE, "boolean")); //$NON-NLS-1$
      sJoinConditionAllowedFunctions.add(new QName(BUILTIN_FUNCTION_NAMESPACE, "not")); //$NON-NLS-1$
      sJoinConditionAllowedFunctions.add(new QName(BUILTIN_FUNCTION_NAMESPACE, "true")); //$NON-NLS-1$
      sJoinConditionAllowedFunctions.add(new QName(BUILTIN_FUNCTION_NAMESPACE, "false")); //$NON-NLS-1$
   }

   /**
    * Constructs an expression validator.
    */
   protected AeAbstractExpressionValidator()
   {
   }

   /**
    * @see org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidator#validateExpression(org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidationContext, java.lang.String)
    */
   public IAeExpressionValidationResult validateExpression(IAeExpressionValidationContext aContext, String aExpression) throws AeException
   {
      AeExpressionValidationResult result = new AeExpressionValidationResult();
      if (AeUtil.isNullOrEmpty(aExpression))
         return result;

      IAeExpressionParser parser = createExpressionParser(createParserContext(aContext));
      IAeExpressionParseResult parseResult = parser.parse(aExpression);
      if (parseResult.hasErrors())
      {
         result.addErrors(parseResult.getParseErrors());
      }

      doCommonExpressionValidation(parseResult, result, aContext);
      doNonJoinConditionValidation(parseResult, result, aContext);

      return result;
   }

   /**
    * @see org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidator#validateBooleanExpression(org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidationContext, java.lang.String)
    */
   public IAeExpressionValidationResult validateBooleanExpression(IAeExpressionValidationContext aContext, String aExpression)
         throws AeException
   {
      return validateExpression(aContext, aExpression);
   }

   /**
    * @see org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidator#validateDeadlineExpression(org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidationContext, java.lang.String)
    */
   public IAeExpressionValidationResult validateDeadlineExpression(IAeExpressionValidationContext aContext, String aExpression)
         throws AeException
   {
      return validateExpression(aContext, aExpression);
   }

   /**
    * @see org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidator#validateDurationExpression(org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidationContext, java.lang.String)
    */
   public IAeExpressionValidationResult validateDurationExpression(IAeExpressionValidationContext aContext, String aExpression)
         throws AeException
   {
      return validateExpression(aContext, aExpression);
   }

   /**
    * @see org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidator#validateJoinConditionExpression(org.activebpel.rt.bpel.def.validation.expr.IAeExpressionValidationContext, java.lang.String)
    */
   public IAeExpressionValidationResult validateJoinConditionExpression(IAeExpressionValidationContext aContext, String aExpression)
         throws AeException
   {
      AeExpressionValidationResult result = new AeExpressionValidationResult();
      if (AeUtil.isNullOrEmpty(aExpression))
         return result;

      IAeExpressionParser parser = createExpressionParser(createParserContext(aContext));
      IAeExpressionParseResult parseResult = parser.parse(aExpression);
      if (parseResult.hasErrors())
      {
         result.addErrors(parseResult.getParseErrors());
      }

      doCommonExpressionValidation(parseResult, result, aContext);
      doJoinConditionValidation(parseResult, result, aContext);

      return result;
   }

   /**
    * Does some common validation steps.  This includes checking the variable data syntax and 
    * using a variable expression parser to check for correct variable and property usage.
    * 
    * @param aParseResult
    * @param aValidationResult
    * @param aContext
    */
   protected void doCommonExpressionValidation(IAeExpressionParseResult aParseResult,
         AeExpressionValidationResult aValidationResult, IAeExpressionValidationContext aContext)
   {
      // Check the bpel function syntax
      checkBpelFunctionSyntax(aParseResult, aValidationResult);
      // Check usage of active-bpel specific funtions
      checkActiveBpelExtensionFunctions(aParseResult, aValidationResult);
   }

   /**
    * Do validation for 'normal' expressions.  This basically means all expressions that aren't 
    * joinConditions.
    * 
    * @param aParseResult
    * @param aValidationResult
    * @param aContext
    */
   protected void doNonJoinConditionValidation(IAeExpressionParseResult aParseResult,
         AeExpressionValidationResult aValidationResult, IAeExpressionValidationContext aContext)
   {
      // Set the var data and var prop info on the result container.
      for (Iterator iter = aParseResult.getVarDataList().iterator(); iter.hasNext(); )
         aValidationResult.addVarData((AeVariableData) iter.next());
      for (Iterator iter = aParseResult.getVarPropertyList().iterator(); iter.hasNext(); )
         aValidationResult.addVarProperty((AeVariableProperty) iter.next());

      // getLinkStatus can only be used in joinCondition expressions.
      if (aParseResult.getLinkStatusFunctionList().size() > 0)
      {
         addError(aValidationResult, 
               AeMessages.getString("AeAbstractExpressionValidator.INVALID_USE_OF_LINKSTATUS"), //$NON-NLS-1$
               new Object[] { aParseResult, aParseResult.getExpression() });
      }
   }

   /**
    * There are additional restrictions imposed on join condition expressions.  This method will check
    * for those restrictions.
    * 
    * @param aParseResult
    * @param aValidationResult
    * @param aContext
    */
   protected void doJoinConditionValidation(IAeExpressionParseResult aParseResult,
         AeExpressionValidationResult aValidationResult, IAeExpressionValidationContext aContext)
   {
      Collection functionList = aParseResult.getFunctions();
      if (functionList.size() == 0)
      {
         addError(aValidationResult,
               AeMessages.getString("AeAbstractExpressionValidator.INVALID_JOIN_CONDITION1_ERROR"),  //$NON-NLS-1$
               new Object[] { aParseResult.getExpression() });
      }
      for (Iterator iter = functionList.iterator(); iter.hasNext(); )
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         int numArgs = function.getArgs().size();
         if (!getJoinConditionAllowedFunctions().contains(function.getQName()))
         {
            addError(aValidationResult, 
                  AeMessages.getString("AeAbstractExpressionValidator.ILLEGAL_CALL_IN_JOIN_CONDITION_ERROR"), //$NON-NLS-1$
                  new Object[] { function.getQName(), aParseResult.getExpression() });
         }
         else if (isLinkStatusFunction(function))
         {
            if (numArgs != 1)
            {
               addError(aValidationResult,
                     AeMessages.getString("AeAbstractExpressionValidator.WRONG_NUM_ARGS_IN_LINKSTATUS_ERROR"),  //$NON-NLS-1$
                     new Object [] { new Integer(numArgs), aParseResult.getExpression() });
            }
            else if ( !(function.getArgument(0) instanceof String) )
            {
               addError(aValidationResult,
                     AeMessages.getString("AeAbstractExpressionValidator.INVALID_LINKSTATUS_ARG_ERROR"), //$NON-NLS-1$
                     new Object [] { aParseResult.getExpression() });
            }
            else
            {
               AeActivityDef def = (AeActivityDef) aContext.getBaseDef();
               String linkName = function.getStringArgument(0);
               if (!def.getTargetLinkNames().contains(linkName))
               {
                  addError(aValidationResult,
                        AeMessages.getString("AeAbstractExpressionValidator.LINK_DOES_NOT_EXIST_ERROR"), //$NON-NLS-1$
                        new Object[] { linkName, aParseResult.getExpression() });
               }
            }
         }
      }
   }

   /**
    * Ensure that the first arg of each getVariableData() function call is correct.
    * 
    * @param aParseResult
    */
   protected void checkBpelFunctionSyntax(IAeExpressionParseResult aParseResult, AeExpressionValidationResult aResult)
   {
      // Check the bpws:getVariableData function syntax.
      for (Iterator iter = aParseResult.getVarDataFunctionList().iterator(); iter.hasNext(); )
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         int numArgs = function.getArgs().size();

         Object param1 = function.getArgument(0);
         if (param1 == AeScriptFuncDef.__EXPRESSION__ || !(param1 instanceof String))
         {
            addWarning(aResult,
                  AeMessages.getString("AeAbstractExpressionValidator.NON_CONSTANT_REF_IN_GETVARDATA_WARNING"), //$NON-NLS-1$
                  new String[] {});
         }

         if (numArgs < 1 || numArgs > 3)
         {
            addError(aResult, AeMessages.getString("AeAbstractExpressionValidator.INVALID_GETVARIABLEDATA_PARAMS"),  //$NON-NLS-1$
                  new Object[] { new Integer(numArgs), aParseResult.getExpression() });
         }
      }

      // Check the bpws:getVariableProperty function syntax.
      for (Iterator iter = aParseResult.getVarPropertyFunctionList().iterator(); iter.hasNext(); )
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         int numArgs = function.getArgs().size();
         if (numArgs != 2)
         {
            addError(aResult,
                  AeMessages.getString("AeAbstractExpressionValidator.INVALID_GETVARDATA_ARGS"), //$NON-NLS-1$
                  new Object[] { new Integer(numArgs), aParseResult.getExpression() });
         }
      }

      // Check the abx:getMyRoleProperty function syntax.
      for (Iterator iter = aParseResult.getMyRolePropertyFunctionList().iterator(); iter.hasNext(); )
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         int numArgs = function.getArgs().size();
         if (numArgs < 2 || numArgs > 3)
         {
            addError(aResult,
                  AeMessages.getString("AeAbstractBpelObject.ERROR_45"), //$NON-NLS-1$ 
                  new Object[] { new Integer(numArgs), aParseResult.getExpression() });
         }
      }
   }
   
   /**
    * Checks to see if any of the ActiveBPEL specific functions are used..
    * 
    * @param aParseResult
    */
   protected void checkActiveBpelExtensionFunctions(IAeExpressionParseResult aParseResult, AeExpressionValidationResult aResult)
   {

      // Check the usage of active-bpel specific function usage. (and display these as INFO messages).
      for (Iterator iter = aParseResult.getFunctions().iterator(); iter.hasNext(); )
      {
         AeScriptFuncDef function = (AeScriptFuncDef) iter.next();
         if (ACTIVE_BPEL_FUNCTION_NAMESPACE.equals( function.getNamespace() ) )
         {
            addInfo(aResult,
                  AeMessages.getString("AeAbstractExpressionValidator.ACTIVE_BPEL_EXT_FUNCTION"), //$NON-NLS-1$ 
                  new Object[] { function.getName() });
            
         }
      }
           
   }   
   
   /**
    * Returns true if the function is a bpws:getVariableData function.
    * 
    * @param aFunction
    */
   protected boolean isVarDataFunction(AeScriptFuncDef aFunction)
   {
      return getVarDataFunctionQName().equals(aFunction.getQName());
   }

   /**
    * Gets the QName of the bpws:getVariableData function.
    */
   protected QName getVarDataFunctionQName()
   {
      return AeExpressionLanguageUtil.VAR_DATA_FUNC;
   }
   
   /**
    * Returns true if the function is a bpws:getVariableProperty function.
    * 
    * @param aFunction
    */
   protected boolean isVarPropertyFunction(AeScriptFuncDef aFunction)
   {
      return getVarPropertyFunctionQName().equals(aFunction.getQName());
   }

   /**
    * Gets the QName of the bpws:getVariableProperty function.
    */
   protected QName getVarPropertyFunctionQName()
   {
      return AeExpressionLanguageUtil.VAR_PROPERTY_FUNC;
   }

   /**
    * Returns true if the function is a bpws:getLinkStatus function.
    * 
    * @param aFunction
    */
   protected boolean isLinkStatusFunction(AeScriptFuncDef aFunction)
   {
      return getLinkStatusFunctionQName().equals(aFunction.getQName());
   }

   /**
    * Gets the QName of the bpws:getLinkStatus function.
    */
   protected QName getLinkStatusFunctionQName()
   {
      return AeExpressionLanguageUtil.LINK_STATUS_FUNC;
   }

   /**
    * Gets the set of functions allowed in join condition expressions.
    */
   protected Set getJoinConditionAllowedFunctions()
   {
      return sJoinConditionAllowedFunctions;
   }

   /**
    * Adds an info to the validation result.
    * 
    * @param aMessage
    * @param aArgs
    */
   protected void addInfo(AeExpressionValidationResult aResult, String aMessage, Object [] aArgs)
   {
      String msg = MessageFormat.format(aMessage, aArgs);
      aResult.addInfo(msg);
   }   
   
   /**
    * Adds an error to the validation result.
    * 
    * @param aMessage
    * @param aArgs
    */
   protected void addError(AeExpressionValidationResult aResult, String aMessage, Object [] aArgs)
   {
      String msg = MessageFormat.format(aMessage, aArgs);
      aResult.addError(msg);
   }

   /**
    * Adds a warning to the validation result.
    * 
    * @param aMessage
    * @param aArgs
    */
   protected void addWarning(AeExpressionValidationResult aResult, String aMessage, Object [] aArgs)
   {
      String msg = MessageFormat.format(aMessage, aArgs);
      aResult.addWarning(msg);
   }
   
   /**
    * Creates the expression parser context from the validation context.
    * 
    * @param aValidationContext
    */
   protected IAeExpressionParserContext createParserContext(IAeExpressionValidationContext aValidationContext)
   {
      IAeNamespaceContext nsCtx = new AeBaseDefNamespaceContext(aValidationContext.getBaseDef());
      return new AeExpressionParserContext(nsCtx);
   }
   
   /**
    * Creates the parser to use to parse the expression.
    * 
    * @param aContext
    */
   abstract protected IAeExpressionParser createExpressionParser(IAeExpressionParserContext aContext);
}
