// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/validation/query/AeXPathQueryValidator.java,v 1.3 2006/02/13 23:09:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.validation.query;

import java.util.Stack;

import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.util.AeUtil;
import org.exolab.castor.xml.schema.ComplexType;
import org.exolab.castor.xml.schema.ContentModelGroup;
import org.exolab.castor.xml.schema.ElementDecl;
import org.exolab.castor.xml.schema.Particle;
import org.exolab.castor.xml.schema.XMLType;
import org.jaxen.saxpath.Axis;
import org.jaxen.saxpath.SAXPathException;
import org.jaxen.saxpath.XPathReader;
import org.jaxen.saxpath.helpers.XPathReaderFactory;

/**
 * This class validates xpath queries against the types declared for their contexts.
 */
public class AeXPathQueryValidator extends AeXPathSyntaxValidator
{
   /** The type associated with the context. */
   protected XMLType mRootContext;

   /** The root name of the query context. */ 
   protected QName mRootName;
   
   /** The current parent within the context (null at start of validation). */   
   protected XMLType mCurrentParent = null;
   
   /** Stack of contexts for queries which use dot-dot (..) type constructs. */
   protected Stack mContextStack = new Stack();
   
   /**
    * Constructor for the validator.
    */
   public AeXPathQueryValidator()
   {
   }
   
   /** 
    * Call to perform validation.  This is the main entry point for validation
    * and can be called multiple times by a process to validate its queries.
    * However this entry point is not reentrant and should only be called again 
    * after it returns. 
    */
   public void validate(String aExpr, XMLType aContext, QName aRoot) throws AeException
   {
      try
      {
         // set new context
         mRootContext = aContext;
         mRootName = aRoot;
         
         // reset instance data 
         mStopValidate = false;
         mCurrentParent = null;
         mContextStack.clear();

         // parse the passed xpath and validate through our handler implementation         
         XPathReader reader = XPathReaderFactory.createReader();
         reader.setXPathHandler( this );
         reader.parse( aExpr );
      }
      catch (Exception e)
      {
         throw new AeException(e.getMessage(), e);
      }
   }

   /**
    * Checks the name passed for applicability at the current level.
    * @see org.jaxen.saxpath.XPathHandler#startNameStep(int, java.lang.String, java.lang.String)
    */
   public void startNameStep(int axis, String prefix, String localName) throws SAXPathException
   {
      // if wqe have stopped trying to validate then do a short return
      if(mStopValidate)
         return;
         
      switch(axis)
      {
         case Axis.CHILD:
            if(mCurrentParent == null)
            {
               if(! AeUtil.compareObjects(localName, mRootName.getLocalPart()))
                  throw new SAXPathException(AeMessages.getString("AeXPathQueryValidator.ERROR_0") + localName); //$NON-NLS-1$
               // TODO validate namespace
               mCurrentParent = mRootContext; 
            }
            else
            {
               mContextStack.push(mCurrentParent);
               if(mCurrentParent.isSimpleType())
               {
                  throw new SAXPathException(AeMessages.getString("AeXPathQueryValidator.2") + localName + AeMessages.getString("AeXPathQueryValidator.1")); //$NON-NLS-1$ //$NON-NLS-2$
               }
               else if(mCurrentParent.isComplexType())
               {
                  if(! findChild((ComplexType)mCurrentParent, prefix, localName))
                     throw new SAXPathException(AeMessages.getString("AeXPathQueryValidator.2") + localName + AeMessages.getString("AeXPathQueryValidator.3")); //$NON-NLS-1$ //$NON-NLS-2$
               }
            }
            
         break;
         
         case Axis.ATTRIBUTE:
         // TODO validate attributes
         //         if(aType.isComplexType())
         //         {
         //            ComplexType complexType = (ComplexType)aType;
         //            for(Enumeration en=complexType.getAttributeDecls(); en.hasMoreElements();)
         //               aParent.addChild(new AevSchemaTypeModel(aParent, aPart, (AttributeDecl)en.nextElement()));
            mStopValidate = true;
         break;
      
         default:
            // axis passed which we are not handing so stop validating   
            mStopValidate = true;
         break;
      }
   }

   /**
    * Checks the schema model to see if the passed child element name is applicable
    * for the current type.  Note this is recursive as one schema group can be 
    * defined as extending another group and a type can have a base type.
    */
   protected boolean findChild(ComplexType type, String prefix, String localName) throws SAXPathException
   {
      boolean found = findChild((ContentModelGroup)type, prefix, localName);
      
      if(! found && type.getBaseType() != null)
          found = findChild((ComplexType)type.getBaseType(), prefix, localName);
      
      return found;
   }

   /**
    * Checks the schema model to see if the passed child element name is applicable
    * for the current level.  Note this is recursive as one schema group can be 
    * defined as extending another group. 
    */
   protected boolean findChild(ContentModelGroup group, String prefix, String localName) throws SAXPathException
   {
      boolean found = false;
      for(int i=0; !found && i < group.getParticleCount(); ++i)
      {
         Particle particle = group.getParticle(i);
         if(particle.getStructureType() == Particle.ELEMENT)
         {
            // TODO validate namespace
            if(localName.equals(((ElementDecl)particle).getName()))
            {
               mCurrentParent = ((ElementDecl)particle).getType();
               found = true; 
            }
         }
         else if(particle.getStructureType() == Particle.GROUP)
         {
            found = findChild((ContentModelGroup)particle, prefix, localName);
         }
         else 
         {
            // Wildcard (extensibility elements) assume we are good, but stop validating
            mStopValidate = true;
            found = true;
         }
      }
      return found;
   }
   
   /**
    * @see org.jaxen.saxpath.XPathHandler#startAbsoluteLocationPath()
    */
   public void startAbsoluteLocationPath() throws SAXPathException
   {
      mCurrentParent = null;
      mContextStack.clear();
   }
}
