//$Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/visitors/AeCheckStartActivityVisitor.java,v 1.3 2005/09/14 17:50:4
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.visitors;

import java.util.Iterator;
import java.util.List;

import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.def.AeActivityDef;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeCorrelationContainerDef;
import org.activebpel.rt.bpel.def.AeCorrelationSetContainerDef;
import org.activebpel.rt.bpel.def.AeEventHandlersDef;
import org.activebpel.rt.bpel.def.AeFaultHandlersContainerDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.AeScopeDef;
import org.activebpel.rt.bpel.def.activity.AeActivityAssignDef;
import org.activebpel.rt.bpel.def.activity.AeActivityCompensateDef;
import org.activebpel.rt.bpel.def.activity.AeActivityEmptyDef;
import org.activebpel.rt.bpel.def.activity.AeActivityFlowDef;
import org.activebpel.rt.bpel.def.activity.AeActivityInvokeDef;
import org.activebpel.rt.bpel.def.activity.AeActivityPickDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReplyDef;
import org.activebpel.rt.bpel.def.activity.AeActivityScopeDef;
import org.activebpel.rt.bpel.def.activity.AeActivitySequenceDef;
import org.activebpel.rt.bpel.def.activity.AeActivitySwitchDef;
import org.activebpel.rt.bpel.def.activity.AeActivityTerminateDef;
import org.activebpel.rt.bpel.def.activity.AeActivityThrowDef;
import org.activebpel.rt.bpel.def.activity.AeActivityWaitDef;
import org.activebpel.rt.bpel.def.activity.AeActivityWhileDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnAlarmDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef;
import org.activebpel.rt.bpel.def.activity.support.AeOtherwiseDef;
import org.activebpel.rt.bpel.def.activity.support.AeSwitchCaseDef;
import org.activebpel.rt.bpel.def.validation.IAeBaseErrorReporter;

/**
 * Class which implements a visitor pattern used to validate that the current process
 * conforms to the BPEL constraint regarding Create Instance usage. This contraint stipulates
 * that if a process activity (Pick or Receive) is annotated with the Create Instance property, 
 * that no other basic activities precede them. Furthermore, all or none of the activities which 
 * start in parallel for the process must be specified as Create Instance. 
 * 
 * This visitor differs from our normal visitor pattern in that the traversal direction 
 * navigates from a given node towards the top of the model. The intention is to assure that 
 * we have no basic activities preceding us, and only containers are allowed as parents.
 * The basic validition is initiated with the list of nodes which are know to be instance
 * creators, and each node is traversed to assure it is the first basic activity. All invalid
 * points which are hit have been equipped with an error reporter to log with the process
 * validation log.    
 */
public class AeCheckStartActivityVisitor extends AeAbstractDefVisitor implements IAeDefVisitor
{
   /** The error reporter specified during creation. */
   private IAeBaseErrorReporter mErrorReporter ;
   /** The current activity being processed. */
   private AeActivityDef mCurrentActivity;
   /** Flag indicating current direction of traversal. */
   private boolean mAscending;
   
   /**
    * Constructor which requires an error reporter to be used during traversal.
    * @param aErrorReporter The required error reporter
    */
   public AeCheckStartActivityVisitor(IAeBaseErrorReporter aErrorReporter)
   {
      mErrorReporter = aErrorReporter;
      
      setTraversalVisitor(new AeTraversalVisitor(new AeDefTraverser(), this));
   }
   
   /**
    * Kicks of the validation process given the list of create instance activities.
    * @param aCreateInstanceActivities the list of activities which have been annotated with Create Instance
    */
   public void doValidation(List aCreateInstanceActivities)
   {
      for (Iterator iter=aCreateInstanceActivities.iterator(); iter.hasNext();)
      {
         mCurrentActivity = (AeActivityDef)iter.next();
         mAscending = true;
         mCurrentActivity.getParent().accept(this);      
      }
   }
      
   /**
    * Override to handle special processing of Scope container.
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeScopeDef)
    */
   public void visit(AeScopeDef aDef)
   {
      if (mAscending)
         aDef.getParent().accept(this);
      else
         aDef.getActivity().accept(this);
   }
   
   /**
    * Override to handle special processing of Flow container.
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityFlowDef)
    */
   public void visit(AeActivityFlowDef aDef)
   {
      // Loop through all activities contained in this flow
      for (Iterator iter=aDef.getActivityList(); iter.hasNext();)
      {
         // If the activity is not a target for any other activity, then it is an
         // initial activity in the flow and must be evaluated.
         AeActivityDef activity = (AeActivityDef)iter.next();
         if (! activity.getTargetList().hasNext())
         {
            // If the current activity is not the same as the activity we are processing
            // we will need to switch direction and descend into the activity to find it.
            if (activity != mCurrentActivity)
            {
               boolean currMode = mAscending;
               mAscending = false;
               activity.accept(this);
               mAscending = currMode;
            }
         }
      }
      
      // Only process if ascending
      if (mAscending)
      {
         mCurrentActivity = aDef;
         aDef.getParent().accept(this);
      }
   }

   /**
    * Override to handle special processing of Sequence container.
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivitySequenceDef)
    */
   public void visit(AeActivitySequenceDef aDef)
   {
      // If no activities in sequence, return since this error is flagged elsewhere
      if (! aDef.getActivityList().hasNext())
         return;
      
      // We only need to process the first activity of the sequence 
      AeActivityDef firstActivity = (AeActivityDef)aDef.getActivityList().next();
      
      if (mAscending)
      {
         // If the current activity we are looking for is the first activity or we are dealing with
         // a nested sequence, this activity becomes current activity and we traverse to the parent.
         // Otherwise, this is an error.
         if(mCurrentActivity == firstActivity || firstActivity instanceof AeActivitySequenceDef)
         {
            mCurrentActivity = firstActivity;
            aDef.getParent().accept(this);
         }
         else
         {
            reportError(firstActivity);
         }
      }
      else
      {
         // Descend into the first activity
         firstActivity.accept(this);
      }
   }
   
   /**
    * Override to handle special processing of While container.
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityWhileDef)
    */
   public void visit(AeActivityWhileDef aDef)
   {
      // If we are not ascending this is an error
      if (mAscending)
         aDef.getParent().accept(this);
      else
         reportError(aDef);
   }
   
   /**
    * Overrides method to prevent traversal from performing the normal top down navigation. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeProcessDef)
    */
   public void visit(AeProcessDef aDef)
   {
   }
   
   /**
    * Overrides method to prevent traversal from performing the normal top down navigation. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityScopeDef)
    */
   public void visit(AeActivityScopeDef aDef)
   {
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityAssignDef)
    */
   public void visit(AeActivityAssignDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityCompensateDef)
    */
   public void visit(AeActivityCompensateDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityEmptyDef)
    */
   public void visit(AeActivityEmptyDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityInvokeDef)
    */
   public void visit(AeActivityInvokeDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityPickDef)
    */
   public void visit(AeActivityPickDef aDef)
   {
      if (! aDef.isCreateInstance())
         reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef)
    */
   public void visit(AeActivityReceiveDef aDef)
   {
      if (! aDef.isCreateInstance())
         reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReplyDef)
    */
   public void visit(AeActivityReplyDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivitySwitchDef)
    */
   public void visit(AeActivitySwitchDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityTerminateDef)
    */
   public void visit(AeActivityTerminateDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityThrowDef)
    */
   public void visit(AeActivityThrowDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityWaitDef)
    */
   public void visit(AeActivityWaitDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeCorrelationContainerDef)
    */
   public void visit(AeCorrelationContainerDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeCorrelationSetContainerDef)
    */
   public void visit(AeCorrelationSetContainerDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeEventHandlersDef)
    */
   public void visit(AeEventHandlersDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeFaultHandlersContainerDef)
    */
   public void visit(AeFaultHandlersContainerDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOnAlarmDef)
    */
   public void visit(AeOnAlarmDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef)
    */
   public void visit(AeOnMessageDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOtherwiseDef)
    */
   public void visit(AeOtherwiseDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Overrides method to report error during traversal. 
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeSwitchCaseDef)
    */
   public void visit(AeSwitchCaseDef aDef)
   {
      reportError(aDef);
   }
   
   /**
    * Report error encountered during traversal. 
    * @param aDef the activity which the error was detected while processing
    */
   private void reportError(AeBaseDef aDef)
   {
      mErrorReporter.addError(AeMessages.getString("AeCheckStartActivityVisitor.ERROR_CREATE_INSTANCE_VALIDATION"), new String [] {}, aDef); //$NON-NLS-1$
   }
}
