// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/visitors/AeDefCorrelatedReceiveVisitor.java,v 1.5 2005/10/17 19:44:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.visitors;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeCorrelationSetDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef;
import org.activebpel.rt.bpel.def.activity.AeActivityScopeDef;
import org.activebpel.rt.bpel.def.activity.support.AeCorrelationDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef;

/**
 * This visitor walks the process def collecting information about correlated receives.
 * It is possible to have multiple activities receiving data for the same partner link,
 * operation but with different correlation sets. 
 * 
 * In order to ensure that the engine routes the inbound receive to the correct 
 * activity, we need to know about all of the possible correlation sets in play 
 * for the given partner link and operation. The matching algorithm will pass
 * in the partner link and operation being invoked and get back a collection
 * of properties to match against. This collection will be in descending order
 * by size since we want to attempt to match against the correlation sets with
 * the most specific information - this being determined by the number of properties
 * within the set. 
 *  
 */
public class AeDefCorrelatedReceiveVisitor extends AeAbstractDefVisitor
{
   /** The process def we're visiting */
   private AeProcessDef mProcessDef;
   /** stack of scopes we've visited, used to resolve corelationSets */
   private Stack mScopeStack = new Stack();
   /** an internal map that uses the concat of the partner link and operation as its
    *  and maps to a Set of correlation sets. */
   private Map mPartnerLinkOperationToCorrSets = new HashMap(); 

   /**
    * Constructs the visitor with the given WSDL provider.
    */
   public AeDefCorrelatedReceiveVisitor()
   {
      setTraversalVisitor(new AeTraversalVisitor(new AeDefTraverser(), this));
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeProcessDef)
    */
   public void visit(AeProcessDef aDef)
   {
      mProcessDef = aDef;
      super.visit(aDef);
      
      // The comparator is used to sort each collection of correlation sets by size
      // to ensure that we attempt to match against the most properties in the collection first
      Comparator comp = new Comparator()
      {
         public int compare(Object aOne, Object aTwo)
         {
            Set one = (Set) aOne;
            Set two = (Set) aTwo;
            // i want descending order, so flip the comparison
            return two.size() - one.size();
         }
      };
      
      // walk the map and set all of the plink/operation sets on the def
      for (Iterator iter = mPartnerLinkOperationToCorrSets.entrySet().iterator(); iter.hasNext();)
      {
         Map.Entry entry = (Map.Entry) iter.next();
         int offset = entry.getKey().toString().indexOf('.');
         String partnerLink = entry.getKey().toString().substring(0, offset);
         String operation = entry.getKey().toString().substring(offset+1);
         LinkedList list = new LinkedList((Collection) entry.getValue());
         Collections.sort(list, comp);
         mProcessDef.setCorrelationProperties(partnerLink, operation, list);
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityScopeDef)
    */
   public void visit(AeActivityScopeDef aDef)
   {
      mScopeStack.push(aDef);
      super.visit(aDef);
      mScopeStack.pop();
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef)
    */
   public void visit(AeActivityReceiveDef aDef)
   {
      addCorrProps(aDef, aDef.getPartnerLink(), aDef.getOperation(), aDef.getCorrelationList());
      super.visit(aDef);
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef)
    */
   public void visit(AeOnMessageDef aDef)
   {
      addCorrProps(aDef, aDef.getPartnerLink(), aDef.getOperation(), aDef.getCorrelationList());
      super.visit(aDef);
   }
   
   /**
    * Adds all of the correlation properties used for the given def object
    * @param aDef
    * @param plinkName
    * @param operation
    * @param correlationIterator
    */
   protected void addCorrProps(AeBaseDef aDef, String plinkName, String operation, Iterator correlationIterator)
   {
      Set corrSetProps = null;
      for (Iterator it=correlationIterator; it.hasNext();)
      {
         AeCorrelationDef def = (AeCorrelationDef) it.next();
         AeCorrelationSetDef corrSetDef = getCorrelationSetDef(def.getCorrelationSetName());
         
         if (corrSetProps == null)
            corrSetProps = new HashSet();
         
         corrSetProps.addAll(corrSetDef.getProperties());
      }
      
      if (corrSetProps != null)
      {
         // this map is used internally to track the activities so no worries about the key we're using.
         Set set = (Set) mPartnerLinkOperationToCorrSets.get(plinkName + "." + operation); //$NON-NLS-1$
         if (set == null)
         {
            set = new HashSet();
            mPartnerLinkOperationToCorrSets.put(plinkName + "." + operation, set); //$NON-NLS-1$
         }
         set.add(corrSetProps);
      }
   }

   /**
    * Walks up the parent hierarchy looking for the correlationSet definition
    * @param aName
    * @param aDef
    */
   protected AeCorrelationSetDef getCorrelationSetDef(String aName)
   {
      AeCorrelationSetDef corrSetDef = null;
      
      for (int i = mScopeStack.size()-1; corrSetDef == null && i >= 0; i--)
      {
         AeActivityScopeDef scope = (AeActivityScopeDef) mScopeStack.get(i);
         if (scope.getScope().hasCorrelationSets())
         {
            corrSetDef = scope.getScope().getCorrelationSetContainer().getCorrelationSet(aName);
         }
      }
      
      if (corrSetDef == null && mProcessDef.hasCorrelationSets())
      {
         // check with the process
         corrSetDef = mProcessDef.getCorrelationSetContainer().getCorrelationSet(aName);
      }
      
      // should never be null
      return corrSetDef;
   }
}
