 // $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/def/visitors/AeDefVariableUsageVisitor.java,v 1.12 2005/10/24 16:08:0
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.def.visitors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Stack;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeExpressionLanguageFactory;
import org.activebpel.rt.bpel.def.AeActivityDef;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.AeBaseDefNamespaceContext;
import org.activebpel.rt.bpel.def.AeFaultHandlerDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.AeScopeCompensatorContainerDef;
import org.activebpel.rt.bpel.def.AeScopeDef;
import org.activebpel.rt.bpel.def.AeVariableDef;
import org.activebpel.rt.bpel.def.activity.AeActivityAssignDef;
import org.activebpel.rt.bpel.def.activity.AeActivityInvokeDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef;
import org.activebpel.rt.bpel.def.activity.AeActivityReplyDef;
import org.activebpel.rt.bpel.def.activity.AeActivityScopeDef;
import org.activebpel.rt.bpel.def.activity.AeActivityThrowDef;
import org.activebpel.rt.bpel.def.activity.AeActivityWaitDef;
import org.activebpel.rt.bpel.def.activity.AeActivityWhileDef;
import org.activebpel.rt.bpel.def.activity.support.AeAssignFromDef;
import org.activebpel.rt.bpel.def.activity.support.AeAssignToDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnAlarmDef;
import org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef;
import org.activebpel.rt.bpel.def.activity.support.AeSourceDef;
import org.activebpel.rt.bpel.def.activity.support.AeSwitchCaseDef;
import org.activebpel.rt.bpel.def.expr.AeExpressionAnalyzerContext;
import org.activebpel.rt.bpel.def.expr.IAeExpressionAnalyzer;
import org.activebpel.rt.bpel.def.expr.IAeExpressionAnalyzerContext;
import org.activebpel.rt.bpel.def.util.AeDefUtil;
import org.activebpel.rt.bpel.def.util.AeVariableData;
import org.activebpel.rt.bpel.def.util.AeVariableProperty;
import org.activebpel.rt.util.AeUtil;

/**
 * Visitor that compiles a set of variables used for each activity. This set
 * will be stored on each activity's def object and is used as part of the 
 * variable locking process and the implementation of serializable scopes.
 */
public class AeDefVariableUsageVisitor extends AeAbstractDefVisitor implements IAeDefVisitor
{
   /** Def objects are pushed and popped from the stack as we traverse the tree */
   private Stack mStack = new Stack();
   /** Serves as a flag for the current serializable scope that we're traversing */
   private AeActivityScopeDef mSerializableScope;
   /** Collection of visited activities, useful in case we don't find a serializable scope and want to reset all of the used variable sets*/
   private Collection mVisitedActivities = new ArrayList();
   /** Set to true if come across a serializable scope which requires us to do variable locking */
   private boolean mVariableLockingRequired = false;
   /** Gets set to true while we're in a compensation handler. Avoids setting write lock in these cases */
   private boolean mInCompensationHandler;
   /** The expression language factory. */
   private IAeExpressionLanguageFactory mExpressionLanguageFactory;

   /**
    * Creates the visitor with the default traversal implementation. 
    */
   protected AeDefVariableUsageVisitor(IAeExpressionLanguageFactory aExpressionLanguageFactory)
   {
      setExpressionLanguageFactory(aExpressionLanguageFactory);
      setTraversalVisitor(new AeTraversalVisitor(new AeDefTraverser(), this));
   }
   
   /**
    * Visits the def and records the variables that need to be locked for 
    * each activity in the process. If there are no serializable scopes within
    * the process def then the visitor will install Collections.EMPTY_SET for
    * each of the activity defs since no variable locking is required.
    * 
    * @param aDef
    * @param aExpressionLanguageFactory
    */
   public static void createVariableUsageSets(AeProcessDef aDef, IAeExpressionLanguageFactory aExpressionLanguageFactory)
   {
      AeDefVariableUsageVisitor viz = new AeDefVariableUsageVisitor(aExpressionLanguageFactory);
      aDef.accept(viz);
      aDef.setContainsSerializableScopes(viz.isVariableLockingRequired());
      
      if (!viz.isVariableLockingRequired())
      {
         for (Iterator iter = viz.getVisitedActivities().iterator(); iter.hasNext();)
         {
            AeActivityDef def = (AeActivityDef) iter.next();
            def.setVariablesUsed(Collections.EMPTY_SET);
         }
      }
   }
   
   /**
    * Gets the visited activities.
    */
   protected Collection getVisitedActivities()
   {
      return mVisitedActivities;
   }

   /** 
    * Returns true variable locking is required, meaning we've discovered a 
    * serializable scope within the process 
    */
   protected boolean isVariableLockingRequired()
   {
      return mVariableLockingRequired;
   }
   
   /**
    * Gets called when the visitor discovers a serializable scope within the 
    * process which necessitates variable locking.
    * @param aBool
    */
   protected void setVariableLockingRequired(boolean aBool)
   {
      mVariableLockingRequired = aBool;
   }

   /**
    * An enumerated constant for read locks and write locks, not used outside of 
    * this class. 
    * 
    * Read locks are required when an activity (like an Invoke) reads data from 
    * a variable but does not update it. If these operations exist OUTSIDE the 
    * bounds of a serializable scope then we do not need to account for any locks. 
    * If these operations occur within the bounds of a serializable scope, then 
    * we have to account for the lock since all of the variable usage within a 
    * serializable scope is accounted for during the execution of a scope. 
    * 
    * Write locks are required when an activity (like an Assign) writes data
    * to a variable. Even activities that exist outside the serializable scope 
    * are required to use locks since the variable's data cannot change by an
    * activity external to the scope once the serializable scope begins execution.
    * This inlcudes other serializable scopes as well.
    */
   private static class AeLockType
   {
      /** constant for a read lock */
      static final AeLockType Read = new AeLockType();
      /** constant for a write lock */
      static final AeLockType Write = new AeLockType();
      
      /** private ctor to prevent instantiation */
      private AeLockType() { }
   }

   /**
    * Called by all of the visit methods in order to traverse the tree. Serves
    * as a good place to record the def if it's an activity because we'll be 
    * revisiting all of the def objects to install empty sets if there are no
    * serialized scopes within this process def.
    * @param aDef
    */
   protected void traverse(AeBaseDef aDef)
   {
      // keep track of the activities since we'll revisit them in the event
      // that we don't discover a serializable scope so we can set their 
      // used variable set to Collections.EMPTY_SET
      if (aDef instanceof AeActivityDef)
         getVisitedActivities().add(aDef);

      // normal traverse code      
      getStack().push(aDef);
      aDef.accept(getTraversalVisitor());
      getStack().pop();
   }

   /**
    * Uses the expression parser to look for any signs of variable usage within
    * the string passed in. This includes the built in bpel functions getVariableProperty
    * and getVariableData. If these are found then a shared lock is added since
    * we're only reading the variable and not updating it.
    * @param aDef
    * @param aExpression
    */
   private void parseForVariables(AeActivityDef aDef, String aExpression)
   {
      // we don't parse for variables when the expression is empty or if
      // the enclosing activity is null. In the latter case, it indicates
      // that the parent was the process and we don't acquire locks on variables
      // at this level. With BPEL 2.0, the variables for eventHandlers and faultHandlers
      // will be implicitly defined.
      if (AeUtil.isNullOrEmpty(aExpression) || aDef == null)
         return;

      try
      {
         String language = AeDefUtil.getProcessDef(aDef).getExpressionLanguage();
         IAeExpressionAnalyzer analyzer = getExpressionLanguageFactory().createExpressionAnalyzer(language);
         IAeExpressionAnalyzerContext ctx = new AeExpressionAnalyzerContext(new AeBaseDefNamespaceContext(aDef));

         for (Iterator it = analyzer.getVarPropertyList(ctx, aExpression).iterator(); it.hasNext();)
         {
            AeVariableProperty varProp = (AeVariableProperty) it.next();
            addLock(aDef, varProp.getVarName(), AeLockType.Read);
         }

         for (Iterator it = analyzer.getVarDataList(ctx, aExpression).iterator(); it.hasNext();)
         {
            AeVariableData varData = (AeVariableData) it.next();
            addLock(aDef, varData.getVarName(), AeLockType.Read);
         }

      }
      catch (AeException e)
      {
         AeException.logError(e, e.getLocalizedMessage());
      }
   }
   
   /**
    * We'll record the usage of the variable based on the lock type and
    * whether or not the activity is nested within a serializable scope. In all
    * cases, we record a lock if it's a write lock. This is because the rules for
    * a serializable scope say that a variable should not change once the scope
    * is entered. If an activity outside of the serializable scope wants to write to
    * a variable then it must obtain a write lock and as such, we're recording that
    * requirement here.
    * If we're in a serializable scope however, we acquire a lock irrespective of 
    * the lock flag since the serializable scope acquires locks on all of the 
    * variables its using to prevent another serializable scope from accessing 
    * these same variables until the locks are released. This prevents the issue
    * of dirty reads and non-repeatable reads.  
    * @param aDef - the def that uses the variable
    * @param aVarName - the name of the variable
    * @param aLockType - true if we're adding a write lock
    */
   private void addLock(AeActivityDef aDef, String aVarName, AeLockType aLockType)
   {
      // if def is null we must be a child of the main process, which
      // doesn't lock its event handlers or fault handler variables.
      // @todo may want to revisit based on oasis wsbpel event handler logic
      if (aDef != null && (! AeUtil.isNullOrEmpty(aVarName)) )
      {
         if (aLockType == AeLockType.Write || isWithinSerializableScope())
         {
            String variableLocationPath = getPathForVariable(aVarName);
            AeActivityDef targetForLock = getTargetForLock(aDef);
            // if the variable in question is defined locally within the activity
            // then there is no need to acquire a lock since the variable is not
            // accessible to other scopes.
            if (!variableLocationPath.startsWith(targetForLock.getLocationPath()))
            {
               getVariablesUsedSet(targetForLock).add(variableLocationPath);
            }
         }
      }
   }

   /**
    * Gets or creates the the set that stores the variables used for this def.
    * @param targetForLock
    */
   private Set getVariablesUsedSet(AeActivityDef aActivityDef)
   {
      Set variablesUsed = aActivityDef.getVariablesUsed();
      if (variablesUsed == null)
      {
         variablesUsed = new HashSet();
         aActivityDef.setVariablesUsed(variablesUsed);
      }
      return variablesUsed;
   }
   
   /**
    * Walks up the stack looking for an AeScopeDef that defines the variable passed in.
    * We should ALWAYS find one or we have a bad bpel file.
    * @param aVarName
    */
   private String getPathForVariable(String aVarName)
   {
      for(int i=getStack().size()-1; i>=0; i--)
      {
         if (getStack().get(i) instanceof AeScopeDef)
         {
            AeScopeDef scopeDef = (AeScopeDef) getStack().get(i);
            AeVariableDef varDef = scopeDef.getVariable(aVarName);
            if (varDef != null)
               return varDef.getLocationPath();
         }
      }
      throw new RuntimeException( AeMessages.format( "AeDefVariableUsageVisitor.ERROR_0", new Object[] {aVarName, getCurrentLocationPath()} ) ); //$NON-NLS-1$
   }

   /**
    * Gets the location path of the item on the stack
    */
   private String getCurrentLocationPath()
   {
      AeBaseDef def = (AeBaseDef) getStack().peek();
      return def.getLocationPath(); 
   }

   /**
    * Getter for the serializable scope, only set for the duration of visiting
    * a serializable scope.
    */
   private AeActivityDef getSerializableScope()
   {
      return mSerializableScope;
   }
   
   /**
    * Setter for serializable scope, set during the visit of a serializable scope
    * and then unset after the traverse is done.
    * @param aDef
    */
   private void setSerializableScope(AeActivityScopeDef aDef)
   {
      mSerializableScope = aDef;
   }

   /**
    * Returns true if we're within a serializable scope. This also takes into
    * consideration the possibility of being within the serializable scope's 
    * compensation handler which is technically not part of the serializable
    * requirement. 
    */
   private boolean isWithinSerializableScope()
   {
      return mSerializableScope != null && !isInCompensationHandler();      
   }
   
   /**
    * The target for the lock is either going to be the def passed in or it'll
    * be the serializable scope if we're within a serializable scope. The latter
    * is because the serializable scope will acquire the locks for all of its nested
    * activities prior to execution. This ensures that we won't have any deadlocks
    * and also frees the individual child activities from having to store any 
    * lock information.
    * @param aDef
    */
   private AeActivityDef getTargetForLock(AeActivityDef aDef)
   {
      if (isWithinSerializableScope())
      {
         return getSerializableScope();
      }
      return aDef;
   }
   

   /**
    * Walks the stack from the current position up til it finds an instanceof
    * AeActivityDef to return. 
    */
   private AeActivityDef findParentActivityDef()
   {
      for(int i=getStack().size()-1; i!=0; i--)
      {
         if (getStack().get(i) instanceof AeActivityDef)
            return (AeActivityDef) getStack().get(i);
      }
      return null;
   }
   
   /**
    * walks the stack looking for the assignment activity def. This is only called
    * from the copy/from and copy/to so the assign should only be two steps up the
    * stack.
    */
   private AeActivityAssignDef getAssignFromStack()
   {
      return (AeActivityAssignDef) getStack().get(getStack().size() - 2);
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeScopeCompensatorContainerDef)
    */
   public void visit(AeScopeCompensatorContainerDef aDef)
   {
      setInCompensationHandler(true);
      super.visit(aDef);
      setInCompensationHandler(false);
   }

   /**
    * Sets the flag that we're in a compensation handler. As stated in the spec,
    * serializable scopes do not need to acquire locks for activities within their compensation 
    * handler. This is because the execution of a compensation handler occurs 
    * after the normal execution of the scope and is therefore outside of the
    * bounds of the scope's serializable requirements.
    * @param aB
    */
   private void setInCompensationHandler(boolean aFlag)
   {
      mInCompensationHandler = aFlag;
   }
   
   private boolean isInCompensationHandler()
   {
      return mInCompensationHandler;
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityInvokeDef)
    */
   public void visit(AeActivityInvokeDef aDef)
   {
      addLock(aDef, aDef.getInputVariable(), AeLockType.Read);
      addLock(aDef, aDef.getOutputVariable(), AeLockType.Write);
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef)
    */
   public void visit(AeActivityReceiveDef aDef)
   {
      addLock(aDef, aDef.getVariable(), AeLockType.Write);
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityScopeDef)
    */
   public void visit(AeActivityScopeDef aDef)
   {
      if (aDef.getVariableAccessSerializable())
      {
         setSerializableScope(aDef);
         setVariableLockingRequired(true);
      }
      traverse(aDef);
      setSerializableScope(null);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.AeFaultHandlerDef)
    */
   public void visit(AeFaultHandlerDef aDef)
   {
      addLock(findParentActivityDef(), aDef.getFaultVariable(), AeLockType.Read);
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOnMessageDef)
    */
   public void visit(AeOnMessageDef aDef)
   {
      addLock(findParentActivityDef(), aDef.getVariable(), AeLockType.Write);
      traverse(aDef);
   }
   
   

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeOnAlarmDef)
    */
   public void visit(AeOnAlarmDef aDef)
   {
      parseForVariables(findParentActivityDef(), aDef.getFor());
      parseForVariables(findParentActivityDef(), aDef.getUntil());
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityThrowDef)
    */
   public void visit(AeActivityThrowDef aDef)
   {
      addLock(aDef, aDef.getFaultVariable(), AeLockType.Read);
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityWaitDef)
    */
   public void visit(AeActivityWaitDef aDef)
   {
      parseForVariables(aDef, aDef.getFor());
      parseForVariables(aDef, aDef.getUntil());
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityWhileDef)
    */
   public void visit(AeActivityWhileDef aDef)
   {
      parseForVariables(aDef, aDef.getCondition());
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeSwitchCaseDef)
    */
   public void visit(AeSwitchCaseDef aDef)
   {
      parseForVariables(findParentActivityDef(), aDef.getCondition());
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeSourceDef)
    */
   public void visit(AeSourceDef aDef)
   {
      parseForVariables(findParentActivityDef(), aDef.getTransitionCondition());
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeAssignFromDef)
    */
   public void visit(AeAssignFromDef aDef)
   {
      addLock(getAssignFromStack(), aDef.getVariable(), AeLockType.Read);
      parseForVariables(getAssignFromStack(), aDef.getExpression());
      traverse(aDef);
   }

   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.support.AeAssignToDef)
    */
   public void visit(AeAssignToDef aDef)
   {
      addLock(getAssignFromStack(), aDef.getVariable(), AeLockType.Write);
      traverse(aDef);
   }
   
   /**
    * @see org.activebpel.rt.bpel.def.visitors.IAeDefVisitor#visit(org.activebpel.rt.bpel.def.activity.AeActivityReplyDef)
    */
   public void visit(AeActivityReplyDef aDef)
   {
      addLock(aDef, aDef.getVariable(), AeLockType.Read);
      traverse(aDef);
   }

   /**
    * Getter for the stack
    */
   protected Stack getStack()
   {
      return mStack;
   }

   /**
    * @return Returns the expressionLanguageFactory.
    */
   protected IAeExpressionLanguageFactory getExpressionLanguageFactory()
   {
      return mExpressionLanguageFactory;
   }
   
   /**
    * @param aExpressionLanguageFactory The expressionLanguageFactory to set.
    */
   protected void setExpressionLanguageFactory(IAeExpressionLanguageFactory aExpressionLanguageFactory)
   {
      mExpressionLanguageFactory = aExpressionLanguageFactory;
   }
}

