// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeAbstractBpelObject.java,v 1.61 2006/01/31 16:40:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.text.ParseException;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.AeStaticAnalysisException;
import org.activebpel.rt.bpel.IAeBusinessProcess;
import org.activebpel.rt.bpel.IAeExpressionLanguageFactory;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.IAeNamespaceContext;
import org.activebpel.rt.bpel.IAeProcessInfoEvent;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.activity.IAeTimedDef;
import org.activebpel.rt.bpel.def.util.AeDefUtil;
import org.activebpel.rt.bpel.function.AeFunctionExecutionContext;
import org.activebpel.rt.bpel.function.AeUnresolvableException;
import org.activebpel.rt.bpel.function.IAeFunction;
import org.activebpel.rt.bpel.function.IAeFunctionExecutionContext;
import org.activebpel.rt.bpel.function.IAeFunctionFactory;
import org.activebpel.rt.bpel.impl.activity.AeActivityScopeImpl;
import org.activebpel.rt.bpel.impl.activity.AeActivityWaitImpl;
import org.activebpel.rt.bpel.impl.activity.support.AeCorrelationSet;
import org.activebpel.rt.bpel.impl.activity.support.AeLink;
import org.activebpel.rt.bpel.impl.expr.AeExpressionRunnerContext;
import org.activebpel.rt.bpel.impl.expr.IAeExpressionRunner;
import org.activebpel.rt.bpel.impl.expr.IAeExpressionRunnerContext;
import org.activebpel.rt.bpel.impl.expr.xpath.AeXPathFunctionContext;
import org.activebpel.rt.bpel.impl.expr.xpath.AeXPathNamespaceContext;
import org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor;
import org.activebpel.rt.bpel.impl.visitors.IAeVisitable;
import org.activebpel.rt.bpel.xpath.AeXPathHelper;
import org.activebpel.rt.util.AeUtil;
import org.activebpel.rt.wsdl.def.IAePropertyAlias;
import org.w3c.dom.Node;

/**
 * Base class for BPEL objects. Provides access to a parent object and 
 * base definition object. Subclasses will include Activities as well as supporting
 * objects like those that model <code>case</code>s, <code>catch</code>...etc.
 */
abstract public class AeAbstractBpelObject implements IAeExecutableBpelObject, IAeVisitable,
      IAeNamespaceContext, IAeFunctionFactory
{
   /** non-null when we're in the middle of handling a fault and need to terminate our child activities asynchronously */
   private IAeFault mFault;

   /** The parent object. */
   private IAeBpelObject mParent;

   /** The process object. */
   private IAeBusinessProcess mProcess;

   /** Base definition object. */
   private AeBaseDef mDef;

   /** The state of the bpel object. */
   private AeBpelState mState = AeBpelState.INACTIVE;

   /**
    * flag that indicates we're in the process of terminating. Gets set to true
    * in terminate method and then set back to false once the object's state changes. 
    */
   private boolean mTerminating;

   /** stores the custom location path */
   private String mLocationPath;

   /** stores the custom location id */
   private int mLocationId;

   /**
    * Requires the base definition object and
    * @param aDef
    * @param aParent
    */
   public AeAbstractBpelObject(AeBaseDef aDef, IAeBpelObject aParent)
   {
      mParent = (IAeBpelObject) aParent;
      mDef = aDef;
   }

   /**
    * Accept a visit from an IAeImplVisitor, e.g, a validator, etc.
    * 
    * @param aVisitor The visitor to accept.
    */
   public abstract void accept( IAeImplVisitor aVisitor ) throws AeBusinessProcessException;

   /**
    * Returns true if we're suppressing the join condition failure, false
    * otherwise.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#isSuppressJoinConditionFailure()
    */
   public boolean isSuppressJoinConditionFailure()
   {
      return getParent().isSuppressJoinConditionFailure();
   }

   /**
    * Getter for the definition
    */
   public AeBaseDef getDefinition()
   {
      return mDef;
   }

   /**
    * Returns the location path associated with the definition object.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getLocationPath()
    */
   public String getLocationPath()
   {
      if (hasCustomLocationPath())
      {
         return mLocationPath;
      }
      else
      {
         return getDefinition().getLocationPath();
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#setLocationPath(java.lang.String)
    */
   public void setLocationPath(String aPath)
   {
      mLocationPath = aPath;
   }

   /**
    * Returns the location id assoicated with the definition object.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getLocationId()
    */
   public int getLocationId()
   {
      if (hasCustomLocationPath())
      {
         return mLocationId;
      }
      else
      {
         return getDefinition().getLocationId();
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#setLocationId(int)
    */
   public void setLocationId(int aId)
   {
      mLocationId = aId;
   }

   /**
    * Returns true if the implementation object's location path differs from the
    * definition object. This will be true in cases where we create multiple
    * implementation objects from a single definition object (parallel forEach)
    */
   public boolean hasCustomLocationPath()
   {
      return mLocationPath != null;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#hasLocationId()
    */
   public boolean hasLocationId()
   {
      return true;
   }

   /**
    * Getter for the parent object.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getParent()
    */
   public IAeBpelObject getParent()
   {
      return mParent;
   }

   /**
    * Base implementation which returns null. Activities should extend this to
    * return the appropriate link.
    * @param aLinkName the name of the link we are interested in.
    */
   public AeLink findTargetLink(String aLinkName)
   {
      return null;
   }

   /**
    * Delegates the call to the parent.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getVariable(java.lang.String)
    */
   public IAeVariable getVariable(String aName)
   {
      return getParent().getVariable(aName);
   }

   /**
    * Delegates the call to the parent.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getPartnerLink(java.lang.String)
    */
   public AePartnerLink getPartnerLink(String aName)
   {
      return getParent().getPartnerLink(aName);
   }

   /**
    * Delegates the call to the parent and then caches.
    * TODO we may want to set this automatically when constructed as it is called a lot.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getProcess()
    */
   public IAeBusinessProcess getProcess()
   {
      if (mProcess == null)
         mProcess = getParent().getProcess();
      return mProcess;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getCorrelationSet(java.lang.String)
    */
   public AeCorrelationSet getCorrelationSet(String aName) throws AeBusinessProcessException
   {
      return getParent().getCorrelationSet(aName);
   }

   /**
    * Does nothing, if the derived class has no children it need not override.
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#childComplete(org.activebpel.rt.bpel.impl.IAeBpelObject)
    */
   public void childComplete(IAeBpelObject aChild) throws AeBusinessProcessException
   {
      // do nothing
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#childTerminated(org.activebpel.rt.bpel.impl.IAeBpelObject)
    */
   public void childTerminated(IAeBpelObject aChild)
         throws AeBusinessProcessException
   {
      terminateNextChild();
   }

   /**
    * Passes child completion fault to parent for handling.
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#childCompleteWithFault(org.activebpel.rt.bpel.impl.IAeBpelObject, org.activebpel.rt.bpel.IAeFault)
    */
   public void childCompleteWithFault(IAeBpelObject aChild, IAeFault aFaultObject) throws AeBusinessProcessException
   {
      objectCompletedWithFault(aFaultObject);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#childIsFaulting(org.activebpel.rt.bpel.impl.IAeBpelObject, org.activebpel.rt.bpel.IAeFault)
    */
   public void childIsFaulting(IAeBpelObject aChild, IAeFault aFaultObject)
   {
      setFault(aFaultObject);
   }

   /**
    * Indicates that the object has completed. Sets the objects state to
    * completed and then propagates the call through the process.
    * @throws AeBusinessProcessException
    */
   protected void objectCompleted() throws AeBusinessProcessException
   {
      getProcess().objectCompleted(this);
      notifyParentOfCompletion();
   }

   /**
    * Notifies the parent of our completion
    */
   protected void notifyParentOfCompletion() throws AeBusinessProcessException
   {
      if (getParent() != null && !getParent().getState().isFinal())
      {
         ((IAeExecutableBpelObject) getParent()).childComplete(this);
      }
   }

   /**
    * Determines is process should be suspended due to an uncaught fault or
    * should the process procees with normal fault completion.
    * @param aFault
    * @throws AeBusinessProcessException
    */
   public void objectCompletedWithFault(IAeFault aFault) throws AeBusinessProcessException
   {
      if( isSuspendable(aFault) )
      {
         setFault( aFault );
         setState( AeBpelState.FAULTING );
         getProcess().suspend(new AeSuspendReason(AeSuspendReason.SUSPEND_CODE_AUTOMATIC, getLocationPath(), null));
      }
      else
      {
         objectCompletedWithFaultInternal(aFault);
      }
   }

   /**
    * Indicates that the object has completed with a fault. Sets the objects state 
    * to faulted and propagates the fault through the process.
    * @param aFault
    * @throws AeBusinessProcessException
    */
   protected void objectCompletedWithFaultInternal( IAeFault aFault ) throws AeBusinessProcessException
   {
      setFault(aFault);
      startTermination();
   }

   /**
    * Return true if the fault is uncaught and the associated process instance
    * should be suspended if it encounters an uncaught fault.
    */
   protected boolean isSuspendable(IAeFault aFault)
   {
      return 
         aFault.isSuspendable() &&
         isFaultUncaught( aFault, false, findEnclosingScope() ) && 
         getProcess().getProcessPlan().isSuspendProcessOnUncaughtFaultEnabled();
   }

   /**
    * Return true if the enclosing scope does not handle the
    * given fault.
    * @param aFault
    */
   protected boolean isFaultUncaught(IAeFault aFault, boolean aIsHandled, AeActivityScopeImpl aScope )
   {
      if (aScope != null)
      {
         aIsHandled = aScope.isFaultHandledByScope(aFault);

         if (!aIsHandled)
         {
            aIsHandled = !isFaultUncaught( aFault, aIsHandled, aScope.findEnclosingScope() );
         }
      }
      return !aIsHandled;
   }

   /**
    * Walks up the object hierarchy until it finds the enclosing scope for this
    * object.
    */
   public AeActivityScopeImpl findEnclosingScope()
   {
      for (IAeBpelObject parent = getParent(); parent != null; parent = parent.getParent())
      {
         if (parent instanceof AeActivityScopeImpl)
         {
            return (AeActivityScopeImpl) parent;
         }
      }

      // should never get here since we'll hit the process at some point
      // unless of course we ARE the process at...
      return null;
   }

   /**
    * Propagates the terminate call to all of its child objects.
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#terminate()
    */
   public void terminate() throws AeBusinessProcessException
   {
      // The activity may be in a state where it has a fault due to processing
      // a fault raised by one of its children. If we have a fault present and
      // are asked to terminate, then we should clear the fault and proceed with
      // our standard termination routine. This will result in all of the child
      // nodes being terminated and the container then reporting its terminated
      // state to its parent. One case where this arises is a fault during DPE
      // where an uncaught bpws:joinFailure bubbles up to a top level fault handler
      // or the implicit one.
      setFault(null);
      startTermination();
   }

   /**
    * Called when an activity is faulting and the user opts to retry the activity.
    * The activity will terminate and at the end of its termination, the process
    * will transition the activity back into an executing state.
    * @throws AeBusinessProcessException
    */
   public void terminateForRetry() throws AeBusinessProcessException
   {
      terminate();
   }

   /**
    * @throws AeBusinessProcessException
    */
   protected void startTermination() throws AeBusinessProcessException
   {
      setTerminating(true);
      terminateNextChild();
   }

   /**
    * Terminates the next available child. If there are no active children left
    * then we're done terminating. At this point, we either set our state to terminated
    * or we propagate the fault which triggered the termination.
    * @throws AeBusinessProcessException
    */
   protected void terminateNextChild() throws AeBusinessProcessException
   {
      IAeBpelObject child = getChildForTermination();
      if (child == null)
      {
         // see note below for why this extra check is necessary
         if (allChildrenAreTerminated())
         {
            if (isTerminating() || getFault() != null)
            {
               noMoreChildrenToTerminate();
            }
            else
            {
               // if the terminating flag wasn't set then we must have been terminating
               // due to an early completion signal (continue or break). As such, we'll
               // complete normally and we're considered compensatable.
               objectCompleted();
            }
         }
      }
      else
      {
         ((IAeExecutableBpelObject) child).terminate();
      }
   }

   /**
    * Returns true if all of the children are terminated.
    * 
    * This extra check is necessary to handle the situation where a scope is asked
    * to terminate and it's in the middle of executing a fault handler. The scope's
    * parent will have asked the scope to terminate but the scope will ignore the request
    * since it's currently in the middle of handling a fault. As such, the parent
    * will move on thinking that the scope has been terminated. This check was
    * added as a way of preventing the parent from reporting that all of its 
    * children were terminated until the catch had a chance to finish executing.
    * 
    * When the catch finishes executing, it'll notify the scope (which was asked
    * to terminate) and the scope will in turn notify its parent that it's been
    * terminated. The parent will then execute terminateNextChild() again, but this
    * time make it through this extra check and then call noMoreChildrenToTerminate()
    */
   private boolean allChildrenAreTerminated()
   {
      for (Iterator iter = getChildrenForStateChange(); iter.hasNext();)
      {
         IAeBpelObject child = (IAeBpelObject) iter.next();
         if (child.isTerminatable())
         {
            return false;
         }
      }
      return true;
   }

   /**
    * Gets called from the termination routine when there are no more children
    * to terminate. If our terminate call originated from a fault, then we're
    * holding onto a fault that needs to be propagated to our parent.
    * If this is false, then the terminate originated from our parent and we
    * should change our state to terminated since all of our children are terminated
    */
   protected void noMoreChildrenToTerminate() throws AeBusinessProcessException
   {
      if (getFault() != null)
      {
         // a fault was the cause of our termination, propagate fault to parent
         IAeFault fault = getFault();
         setFault(null);
         getProcess().objectCompletedWithFault(this, fault);
      }
      else
      {
         // It's possible that we're already in a final state by way of a callback
         // from our terminate work and subsequent firing of link changes.
         if (!getState().isFinal())
         {
            if (getState().canTransitionToDeadPath())
            {
               setState(AeBpelState.DEAD_PATH);
            }
            else
            {
               // set our state to terminated which will trigger a callback to our parent
               // notifiying them that we've terminated
               setState(AeBpelState.TERMINATED);
            }
         }
      }
   }

   /**
    * Returns the next child eligible for termination or null if there are no
    * child objects that need to be terminated.
    */
   protected IAeBpelObject getChildForTermination() throws AeBusinessProcessException
   {
      for (Iterator it = getChildrenForStateChange(); it.hasNext();)
      {
         IAeBpelObject child = (IAeBpelObject) it.next();
         // If the child can transition to dead path then we'll do that here.
         // This includes children that are currently queued by their parent,
         // ready to execute, or inactive. Setting them to dead path here will
         // ensure that they enter a final state and have their children go dead path
         // and any links go to false.
         if (child.getState().canTransitionToDeadPath())
         {
            child.setState(AeBpelState.DEAD_PATH);
         }
         else if (child.isTerminatable())
         {
            return child;
         }
      }
      return null;
   }

   /**
    * Returns true if the bpel object has been queued by its parent. There is an
    * additional check here to prevent bpel activities that are in the process
    * of terminating from executing.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#isReadyToExecute()
    */
   public boolean isReadyToExecute()
   {
      return (!isTerminating()) && getState() == AeBpelState.QUEUED_BY_PARENT;
   }

   /**
    * Base returns true, activities should override.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#isNotDeadPath()
    */
   public boolean isNotDeadPath() throws AeBusinessProcessException
   {
      return true;
   }

   /**
    * Creates an expression runner context when running an expression.
    * 
    * @param aContext
    */
   protected IAeExpressionRunnerContext createExpressionRunnerContext(Object aContext)
   {
      return new AeExpressionRunnerContext(this, aContext, this, this);
   }

   /**
    * Creates a function execution context.
    * 
    * @param aContext
    */
   protected IAeFunctionExecutionContext createFunctionExecutionContext(Object aContext)
   {
      return new AeFunctionExecutionContext(this, aContext, this, this);
   }

   /**
    * Gets the expression language configured for this bpel activity.  In BPEL 1.0 this means looking
    * at the <code>expressionLanguage</code> attribute of the root <code>process</code> element.
    */
   protected String getExpressionLanguage()
   {
      return getProcess().getProcessDefinition().getExpressionLanguage();
   }

   /**
    * Executes the given query.
    * 
    * @param aQuery
    * @throws AeBusinessProcessException
    */
   protected Object executeQuery(String aQuery) throws AeBusinessProcessException
   {
      return executeQuery(aQuery, null);
   }

   /**
    * Executes the given query within the given context.
    * 
    * @param aQuery
    * @param aContext
    * @throws AeBusinessProcessException
    */
   protected Object executeQuery(String aQuery, Object aContext) throws AeBusinessProcessException
   {
      return executeQuery(aQuery, aContext, this);
   }

   /**
    * Executes the given query within the given context and using the given namespace context.
    * 
    * todo This would be where we hook in generic query support (for multiple query languages, without the XPath assumption)
    * 
    * @param aQuery
    * @param aContext
    * @param aNamespaceContext
    * @throws AeBusinessProcessException
    */
   protected Object executeQuery(String aQuery, Object aContext, IAeNamespaceContext aNamespaceContext)
         throws AeBusinessProcessException
   {
      return AeXPathHelper.executeXPathExpression(aQuery, aContext, 
            new AeXPathFunctionContext(createFunctionExecutionContext(aContext)),
            new AeXPathNamespaceContext(aNamespaceContext, (Node) aContext));
   }

   /**
    * Executes the given expression in the expression language specified by the 'expressionLanguage'
    * attribute of the process element.
    * 
    * @param aExpression
    * @throws AeBusinessProcessException
    */
   public Object executeExpression(String aExpression) throws AeBusinessProcessException
   {
      return executeExpression(aExpression, null);
   }

   /**
    * Executes an expression and returns the result as an Object.  This method checks
    * the expression language defined in the 'expressionLanguage' attribute of the
    * process element to figure out what expression language to use during expression
    * execution.
    * 
    * @param aExpression
    */
   public Object executeExpression(String aExpression, Object aContext) throws AeBusinessProcessException
   {
      return executeExpression(aExpression, aContext, null);
   }

   /**
    * Executes an expression and returns the result as an Object.  This method uses
    * the expression language passed in to load an <code>IAeExpressionLanguageFactory</code>
    * 
    * @param aExpression
    * @param aContext
    * @param aExpressionLanguage
    * @throws AeBusinessProcessException
    */
   public Object executeExpression(String aExpression, Object aContext, String aExpressionLanguage) throws AeBusinessProcessException
   {
      try
      {
         String expressionLanguage = AeUtil.isNullOrEmpty(aExpressionLanguage)? getExpressionLanguage() : aExpressionLanguage;
         IAeExpressionLanguageFactory factory = getProcess().getExpressionLanguageFactory();
         IAeExpressionRunner runner = factory.createExpressionRunner(expressionLanguage);
         return runner.executeExpression(createExpressionRunnerContext(aContext), aExpression);
      }
      catch (Exception e)
      {
         throw new AeBusinessProcessException(e.getLocalizedMessage(), e);
      }
   }

   /**
    * Executes an XPath expression that we expect to return a Boolean. Any other
    * return value from the evaluation will result in an exception being thrown
    * since there is something terribly wrong.
    * 
    * @param aExpression
    * @return boolean or bust
    * @throws AeBusinessProcessException if not boolean return or other error
    */
   public boolean executeBooleanExpression(String aExpression) throws AeBusinessProcessException
   {
      String expressionLanguage = getExpressionLanguage();
      try
      {
         IAeExpressionLanguageFactory factory = getProcess().getExpressionLanguageFactory();
         IAeExpressionRunner runner = factory.createExpressionRunner(expressionLanguage);
         return runner.executeBooleanExpression(createExpressionRunnerContext(null), aExpression).booleanValue();
      }
      catch (AeBusinessProcessException e)
      {
         throw e;
      }
      catch (Exception e)
      {
         throw new AeBusinessProcessException(e.getLocalizedMessage(), e);
      }
   }

   /**
    * Helper method to extract a deadline from a time based activity definition
    * like an alarm or wait.
    * @throws AeBusinessProcessException
    */
   protected Date extractDeadline(IAeTimedDef aTimedDefinition) throws AeBusinessProcessException
   {
      Date deadline;

      StringBuffer alarmInfo = new StringBuffer();
      alarmInfo.append('\n');
      alarmInfo.append(aTimedDefinition.getClass().getName());
      alarmInfo.append(AeMessages.getString("AeAbstractBpelObject.0")); //$NON-NLS-1$
      alarmInfo.append(AeMessages.getString("AeAbstractBpelObject.1")).append((new Date()).toString()).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$

      if (!AeUtil.isNullOrEmpty(aTimedDefinition.getFor()))
      {
         alarmInfo.append(AeMessages.getString("AeAbstractBpelObject.4")).append(aTimedDefinition.getFor()).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
         deadline = executeDurationExpression(aTimedDefinition.getFor());
      }
      else
      {
         alarmInfo.append(AeMessages.getString("AeAbstractBpelObject.6")).append(aTimedDefinition.getUntil()).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
         deadline = executeDeadlineExpression(aTimedDefinition.getUntil());
      }

      alarmInfo.append(AeMessages.getString("AeAbstractBpelObject.8")).append(deadline.toString()).append("\n"); //$NON-NLS-1$ //$NON-NLS-2$
      getProcess().getEngine().fireEvaluationEvent(
            getProcess().getProcessId(), aTimedDefinition.getFor(), this, 
            ( this instanceof AeActivityWaitImpl ) ? IAeProcessInfoEvent.INFO_WAIT : IAeProcessInfoEvent.INFO_ON_ALARM,
            getLocationPath(), deadline.toString());

      AeException.info(alarmInfo.toString());
      return deadline;
   }

   /**
    * Executes an XPath expression that we expect to return a duration. Any other
    * return value from the evaluation will result in an exception being thrown
    * since there is something terribly wrong.
    * Duration formats are PnYnMnDTnHnMnS ISO 8601.
    * 
    * @param aExpression
    * @return Date representing the time now that the duration represents.
    * @throws AeBusinessProcessException if not a duration return or other error
    */
   public Date executeDurationExpression(String aExpression) throws AeBusinessProcessException
   {
      String expressionLanguage = getExpressionLanguage();
      try
      {
         IAeExpressionLanguageFactory factory = getProcess().getExpressionLanguageFactory();
         IAeExpressionRunner runner = factory.createExpressionRunner(expressionLanguage);
         return runner.executeDurationExpression(createExpressionRunnerContext(null), aExpression);
      }
      catch (Exception e)
      {
         if (e.getCause() instanceof ParseException) 
            throw new AeBpelException(e.getLocalizedMessage(), AeFaultFactory.XPATH_DURATION_FORMAT_ERROR, e);
         else
            throw new AeBusinessProcessException(e.getLocalizedMessage(), e);
      }
   }

   /**
    * Executes an XPath expression that we expect to return a deadline. Any other
    * return value from the evaluation will result in an exception being thrown
    * since there is something terribly wrong.
    * Deadline format is yyyy-MM-ddTHH:mm:ssZ ISO 8601 subset (XML schema dateTime).
    * 
    * @param aExpression
    * @return Date representing the time now that the duration represents.
    * @throws AeBusinessProcessException if not a duration return or other error
    */
   public Date executeDeadlineExpression(String aExpression) throws AeBusinessProcessException
   {
      String expressionLanguage = getExpressionLanguage();
      try
      {
         IAeExpressionLanguageFactory factory = getProcess().getExpressionLanguageFactory();
         IAeExpressionRunner runner = factory.createExpressionRunner(expressionLanguage);
         return runner.executeDeadlineExpression(createExpressionRunnerContext(null), aExpression);
      }
      catch (Exception e)
      {
         if (e.getCause() instanceof ParseException) 
            throw new AeBpelException(e.getLocalizedMessage(), AeFaultFactory.XPATH_DATE_PARSE_ERROR, e);
         else
            throw new AeBusinessProcessException(e.getLocalizedMessage(), e);
      }
   }

   /**
    * Given a string which represents a qualified name, this method will parse
    * it and return a QName object. Note this method returns null if the string
    * is not properly formed.
    * @param aQstr the string to be parsed 
    */
   public QName parseQName(String aQstr)
   {
      return AeDefUtil.parseQName(getDefinition(), aQstr);
   }

   /**
    * Returns the namespace associated with the prefix from the associated model.
    * 
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#translateNamespacePrefixToUri(java.lang.String)
    */
   public String translateNamespacePrefixToUri(String aPrefix)
   {
      return AeDefUtil.translateNamespacePrefixToUri(getDefinition(), aPrefix);
   }

   /**
    * Return true if <code>IAeMutableConfig</code> settings for engine allow empty query selection.
    */
   protected boolean isEmptyQuerySelectionAllowed()
   {
      return getProcess().getEngine().getEngineConfiguration().getUpdatableEngineConfig().allowEmptyQuerySelection();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getState()
    */
   public AeBpelState getState()
   {
      return mState;
   }

   /**
    * Sets the state of the object <i>without</i> reporting the change to the process.
    * This is used when restoring the state of a process from its serialized
    * representation.
    * 
    * @param aNewState The new state to set
    */
   public void setRawState(AeBpelState aNewState)
   {
      mState = aNewState;
   }

   /**
    * Sets the state on the object and reports the change to the process.
    * In order to properly log the details of the fault object, you need to call
    * setFaultedState(IAeFault).
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#setState(org.activebpel.rt.bpel.impl.AeBpelState)
    */
   public void setState(AeBpelState aNewState) throws AeBusinessProcessException
   {
      if (aNewState == AeBpelState.FAULTED)
      {
         throw new IllegalArgumentException(AeMessages.getString("AeAbstractBpelObject.ERROR_41")); //$NON-NLS-1$
      }

      IAeStateChangeDetail details;
      // If Faulting, provide the causal fault's name.
      if (aNewState == AeBpelState.FAULTING)
      {
         String faultName = getFault() == null ? "" : getFault().getFaultName().getLocalPart(); //$NON-NLS-1$
         String faultInfo = getFault() == null ? "" : getFault().getInfo(); //$NON-NLS-1$
         details = new AeStateChangeDetail(faultName, faultInfo);
      }
      else
         details = AeStateChangeDetail.NONE;

      setState(aNewState, details);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#setFaultedState(org.activebpel.rt.bpel.IAeFault)
    */
   public void setFaultedState(IAeFault aFault) throws AeBusinessProcessException
   {
      // @todo ? can the fault name ever be null or empty qname for a fault ?
      String faultName = aFault.getFaultName() != null ? aFault.getFaultName().getLocalPart() : "[unnamed]"; //$NON-NLS-1$
      String info = (aFault.getSource() != null && aFault.getSource() != this) ? "..." : aFault.getInfo(); //$NON-NLS-1$
      setState(AeBpelState.FAULTED, new AeStateChangeDetail(faultName, info));
   }

   /**
    * Private overloaded setter for the state which takes an additional param to
    * convey any extra information associated with this state change. For the
    * time being this is limited to the fault object that comes with entering the
    * faulted state but this could conceiveably be expanded to include other
    * values (perhaps the originating node in a dead path chain?).
    * @param aNewState
    * @param aAdditionalValue
    * @param aDetailsObject
    */
   private void setState(AeBpelState aNewState, IAeStateChangeDetail aDetailsObject) throws AeBusinessProcessException
   {
      AeBpelState oldState = getState();

      if (oldState != aNewState)
      {
         validateStateChange(aNewState);

         mState = aNewState;

         // any kind of state change means that we're no longer in the process of terminating         
         setTerminating(false);

         if (mState.isPropagable())
         {
            mState.propagate(this);
         }

         getProcess().objectStateChanged(this, oldState, aDetailsObject);
      }
   }

   /**
    * Validates the state change to the new state is allowable.  This class
    * calls the state to determine if the change is allowable derived classes
    * may override.
    * @param aNewState The state we have been asked to change to.
    * @throws AeBusinessProcessException
    */
   protected void validateStateChange(AeBpelState aNewState) throws AeBusinessProcessException
   {
      AeBpelState.validateStateChange(this, aNewState);
   }

   /**
    * Returns true if the state of this object is faulted.
    */
   protected boolean isFaulted()
   {
      return getState() == AeBpelState.FAULTED;
   }

   /**
    * Setter for terminating flag
    * @param aTerminatingFlag
    */
   public void setTerminating(boolean aTerminatingFlag)
   {
      mTerminating = aTerminatingFlag;
   }

   /**
    * Getter for terminating flag
    */
   public boolean isTerminating()
   {
      return mTerminating;
   }

   /**
    * Gets called when we're in the middle of terminating our child activity in
    * order to process the fault. Terminating the child's execution will result
    * in a callback to the parent process once the termination is complete. We're
    * storing the fault that triggered the termination so we can run the proper
    * fault handling code after the callback is received.
    * @param fault
    */
   public void setFault(IAeFault aFault)
   {
      if (aFault != null)
      {
         // We don't want to overwrite a fault if one is already set.
         if (getFault() == null)
            mFault = aFault;
      }
      else
      {
         mFault = aFault;
      }
   }

   /**
    * Getter for the fault that we're storing. This gets called from the childTerminated
    * callback since we know that the scope's only child has terminated its execution
    * and we're now ready to execute the fault handling code.
    */
   public IAeFault getFault()
   {
      return mFault;
   }

   /**
    * Call this method in order to report a failure that should have been caught
    * during static analysis of the bpel process.
    * @param aMessage
    * @throws AeStaticAnalysisException
    */
   protected void staticAnalysisFailure(String aMessage) throws AeStaticAnalysisException
   {
      throw new AeStaticAnalysisException(aMessage);
   }

   /**
    * Gets the message property from the wsdl
    * @param aName
    * @param propName
    */
   protected IAePropertyAlias getMessageProperty(QName aName, QName aPropName) throws AeBusinessProcessException
   {
      return getProcess().getMessageProperty(aName, aPropName);
   }

   /**
    * Returns true if the object isn't currently terminating and its state is
    * one that can be terminated.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#isTerminatable()
    */
   public boolean isTerminatable()
   {
      if (!getState().isTerminatable())
         return false;
      if (((AeBusinessProcess) getProcess()).isTerminating())
         return true;
      return !isTerminating();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#isFaultingOrRetrying()
    */
   public boolean isFaultingOrRetrying()
   {
      return getState() == AeBpelState.FAULTING;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeNamespaceContext#resolvePrefixToNamespace(java.lang.String)
    */
   public String resolvePrefixToNamespace(String aPrefix)
   {
      return AeDefUtil.translateNamespacePrefixToUri(getDefinition(), aPrefix);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeNamespaceContext#resolveNamespaceToPrefixes(java.lang.String)
    */
   public Set resolveNamespaceToPrefixes(String aNamespace)
   {
      return AeDefUtil.getPrefixesForNamespace(getDefinition(), aNamespace);
   }

   /**
    * @see org.activebpel.rt.bpel.function.IAeFunctionFactory#getFunction(java.lang.String, java.lang.String)
    */
   public IAeFunction getFunction(String aNamespaceURI, String aLocalName) throws AeUnresolvableException
   {
      return getProcess().getEngine().getEngineConfiguration().getFunction(aLocalName, aNamespaceURI);
   }

   /**
    * @see org.activebpel.rt.bpel.function.IAeFunctionFactory#getFunctionContextNamespaceList()
    */
   public List getFunctionContextNamespaceList()
   {
      return getProcess().getEngine().getEngineConfiguration().getFunctionContextNamespaceList();
   }
}
