// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeBaseQueueManager.java,v 1.16 2006/01/06 23:55:5
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import commonj.timers.Timer;
import commonj.timers.TimerListener;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeBusinessProcess;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity;
import org.activebpel.rt.bpel.impl.activity.support.AeFault;
import org.activebpel.rt.bpel.impl.list.AeAlarmFilter;
import org.activebpel.rt.bpel.impl.list.AeAlarmFilterManager;
import org.activebpel.rt.bpel.impl.list.AeAlarmListResult;
import org.activebpel.rt.bpel.impl.list.AeMessageReceiverFilter;
import org.activebpel.rt.bpel.impl.list.AeMessageReceiverFilterManager;
import org.activebpel.rt.bpel.impl.list.AeMessageReceiverListResult;
import org.activebpel.rt.bpel.impl.queue.AeAlarm;
import org.activebpel.rt.bpel.impl.queue.AeInboundReceive;
import org.activebpel.rt.bpel.impl.queue.AeInvoke;
import org.activebpel.rt.bpel.impl.queue.AeMessageReceiver;
import org.activebpel.rt.bpel.impl.queue.AeReply;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeUtil;
import org.activebpel.wsio.invoke.IAeInvoke;

/**
 * An in-memory implementation of the queue manager. All of the queues are implemented
 * with java.util.Collection implementations.
 */
public abstract class AeBaseQueueManager extends AeManagerAdapter implements IAeQueueManager
{
   public static final String CONFIG_UNMATCHED_RECEIVES_COUNT = "UnmatchedReceivesCount"; //$NON-NLS-1$
   public static final int DEFAULT_UNMATCHED_RECEIVES_COUNT = 50;

   /** resolves pid and path to a message receiver */
   private IAeMessageReceiverResolver mMessageReceiverResolver = new AeDefaultMessageReceiverResolver();
   /** inbound receives that have not been matched to activities */
   private List mUnmatchedReceives = Collections.synchronizedList(new LinkedList());
   /** activites that are waiting to receive data */
   private Collection mMessageReceivers = Collections.synchronizedSet(new TreeSet(sMessageReceiverComparator));

   /** invokes that are currently executing and waiting for a response */
   private List mInvokes = Collections.synchronizedList(new LinkedList());
   /** contains the reply portion of an inbound receive that has been matched and is waiting to receive its response from its call */
   private List mReplies = Collections.synchronizedList(new LinkedList());
   /** map of replies that have not been queued yet because there was no matching receive for the reply */
   private Map mUnmatchedReplies = new HashMap();
   /** next id for unmatched replies */
   private long mUnmatchedReplyId = 0;
   /** next id for unmatched receives */
   private long mQueueId = 0;

   /** comparator used to keep the message receiver list sorted by # of correlated properties */
   private static final Comparator sMessageReceiverComparator = new AeMessageReceiverComparator();

   /** the maximum number of unmatched receives */
   private int mMaxUnmatchedReceivesCount;
   /** flag that let's us know if we've been started */
   private boolean mStarted = false;
   /** Maps AeAlarm objects to the scheduled task's timer id. */
   private Map mLookup = new HashMap();
   /** Saved map for in-memory starts and stops. */
   private Map mSavedLookup;

   /**
    * Constructs an in-memory queue manager.
    *
    * @param aConfig The configuration map for this manager.
    */
   public AeBaseQueueManager(Map aConfig)
   {
      setConfig(aConfig);
   }

   /**
    * Gets the next journal id.
    */
   public long getNextJournalId()
   {
      // Note: the journal ID is meaningless in in-memory config.
      return IAeProcessManager.NULL_JOURNAL_ID + 1;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#faultOutstandingReplies(long, org.activebpel.rt.bpel.IAeFault, java.util.Map)
    */
   public void faultOutstandingReplies(long aProcessId, IAeFault fault, Map aBusinessProcessProperties )
   {
      List list = new ArrayList();

      // Go through the list of replies and find any with the given process id.
      synchronized(getReplies())
      {
         for (Iterator iter = getReplies().iterator(); iter.hasNext();)
         {
            AeReply reply = (AeReply) iter.next();
            if (reply.getProcessId() == aProcessId)
            {
               list.add(reply);
               iter.remove();
            }
         }
      }

      // Now iterate through the resulting list and call onFault on each reply object.
      for (Iterator iter = list.iterator(); iter.hasNext();)
      {
         AeReply reply = (AeReply) iter.next();
         try
         {
            reply.getReplyReceiver().onFault(fault, aBusinessProcessProperties);
         }
         catch (AeBusinessProcessException e)
         {
            AeException.logError(e, AeMessages.getString("AeBaseQueueManager.ERROR_0")); //$NON-NLS-1$
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#getUnmatchedReceivesIterator()
    */
   public Iterator getUnmatchedReceivesIterator()
   {
      synchronized(getUnmatchedReceives())
      {
         return Collections.unmodifiableList( getUnmatchedReceives() ).iterator();
      }
   }

   /**
    * Attempts to find a match for the message receiver among the unmatched
    * inbound receives. If the message receiver doesn't have any correlation data
    * then we're only matching on the partnerlink, port type, and operation. Otherwise,
    * we include the correlation information in the check. If there was no match
    * then we return null.
    * @param aMessageReceiver
    */
   protected AeInboundReceive findMatchForMessageReceiver(AeMessageReceiver aMessageReceiver)
   {
      // if the message receiver is correlated, then look for an exact match
      // among the inbound receives
      if (aMessageReceiver.isCorrelated())
      {
         for (Iterator iter = getUnmatchedReceives().iterator(); iter.hasNext();)
         {
            AeInboundReceive inboundReceive = (AeInboundReceive) iter.next();
            if (aMessageReceiver.correlatesTo(inboundReceive))
            {
               iter.remove();
               return inboundReceive;
            }
         }
      }
      else
      {
         // if the message receiver is not correlated, then we only need to match
         // on the partner link, port type, and operation.
         for (Iterator iter = getUnmatchedReceives().iterator(); iter.hasNext();)
         {
            AeInboundReceive inboundReceive = (AeInboundReceive) iter.next();
            if (aMessageReceiver.matches(inboundReceive))
            {
               iter.remove();
               return inboundReceive;
            }
         }
      }

      return null;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#addMessageReceiver(org.activebpel.rt.bpel.impl.queue.AeMessageReceiver)
    */
   public void addMessageReceiver(AeMessageReceiver aMessageReceiver) throws AeBusinessProcessException
   {
      // check for a match before adding
      synchronized(getUnmatchedReceives())
      {
         synchronized(getMessageReceivers())
         {
            // Always queue the message receiver.
            addMessageReceiverInternal(aMessageReceiver);

            // This will look to see if we already have an inbound message waiting that matches the activity.
            // If we already have a message waiting, we will dispatch that message to the process via
            // the engine's dispatchReceiveData() method - this allows us to route ALL 'incoming' messages
            // to the engine in the same way.
            
            // Note that the inbound receive (if found) is also removed from the inbound receive queue, so it
            // MUST be properly handled here.
            AeInboundReceive foundMatch = findMatchForMessageReceiver(aMessageReceiver);
            if (foundMatch != null)
            {
               // Note: no need for locking here because the engine ensures that only one
               // business process activity will be executed at a time.
               IAeReplyReceiver reply = matchInboundReceiveToProcess(foundMatch, aMessageReceiver);
               queueReceiveDataInEngine(foundMatch, reply);
            }
         }
      }
   }

   /**
    * Queues the receive data in the business process engine.  In addition, this method will send a fault to
    * the reply if an exception is thrown when queueing the receive data.
    * 
    * @param aFoundMatch
    * @param aReply
    */
   protected void queueReceiveDataInEngine(AeInboundReceive aFoundMatch, IAeReplyReceiver aReply)
   {
      try
      {
         getEngine().queueReceiveData(aFoundMatch, aReply);
      }
      catch (Throwable t)
      {
         if (aReply != null)
         {
            IAeFault fault = AeFaultFactory.getSystemErrorFault(t);
            try
            {
               aReply.onFault(fault, aFoundMatch.getContext().getBusinessProcessProperties());
            }
            catch (AeBusinessProcessException ex)
            {
               // Not much left to do but log it.
               AeException.logError(ex, ex.getLocalizedMessage());
            }
         }
      }
   }

   /**
    * This method is called to put a receive (message receiver) on the receive queue.
    *
    * @param aMessageReceiver The receive to put on the queue.
    */
   protected void addMessageReceiverInternal(AeMessageReceiver aMessageReceiver)
   {
      // the caller will have sync'd on the message receivers coll so it's safe to add/sort
      getMessageReceivers().add(aMessageReceiver);
   }

   /**
    * Matches an inbound receive to a process id. If the inbound receive has a reply
    * that is in the unmatched reply queue, remove that reply and add it to the
    * regular reply queue with the pid.
    * @param aInboundReceive The inbound message
    * @param aMessageReceiver The activity in a process that the inbound receive matched.
    * @throws AeConflictingRequestException If there is already an outstanding reply with the same pid, plink, porttype, and operation
    */
   protected IAeReplyReceiver matchInboundReceiveToProcess(
      AeInboundReceive aInboundReceive, AeMessageReceiver aMessageReceiver) throws AeConflictingRequestException
   {
      IAeReplyReceiver found = null;
      if (aInboundReceive.getResponseReceiverId() != null)
      {
         found = removeUnmatchedReply(aInboundReceive.getResponseReceiverId());
      }
      return found;
   }

   /**
    * Attempts to add an unmatched receive which will get matched at a later point.
    * This implementation will reject an inbound receive that is lacking correlation
    * data, opting to throw a bpws:correlationViolation fault instead.
    * @param aInboundReceive An inbound receive that didn't match up to a waiting message receiver
    * @param aReply A possible reply that is waiting to gett called back.
    * @throws AeCorrelationViolationException
    */
   protected void addUnmatchedReceive(AeInboundReceive aInboundReceive, IAeReplyReceiver aReply) throws AeCorrelationViolationException
   {
      if (aInboundReceive.isCorrelated())
      {
         getUnmatchedReceives().add(aInboundReceive);
         if (aInboundReceive.getQueueId() == null)
            aInboundReceive.setQueueId(getNextQueueId());
         if (aReply != null)
         {
            String key = addUnmatchedReply(aReply);
            aInboundReceive.setResponseReceiverId(key);
         }

         // Limit the size of the unmatched receives list.
         if (getUnmatchedReceives().size() > getMaxUnmatchedReceivesCount())
         {
            AeInboundReceive first = (AeInboundReceive) getUnmatchedReceives().get(0);
            removeUnmatchedReceive(first.getQueueId());

            // Make sure it's really gone from the list at this point;
            // otherwise, we could be trying to remove the first one over and
            // over again.
            getUnmatchedReceives().remove(first);
         }
      }
      else
      {
         throw new AeCorrelationViolationException(AeCorrelationViolationException.UNMATCHED_RECEIVE);
      }
   }

   /**
    * Gets the next id to use for queued objects.
    */
   private String getNextQueueId()
   {
      mQueueId++;
      return String.valueOf(mQueueId);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#removeInvoke(org.activebpel.rt.bpel.impl.queue.AeInvoke)
    */
   public AeInvoke removeInvoke(AeInvoke aInvokeQueueObject)
   {
      return (AeInvoke) removeFromQueue(getInvokes(), aInvokeQueueObject);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#matchInboundReceive(org.activebpel.rt.bpel.impl.queue.AeInboundReceive, org.activebpel.rt.bpel.impl.IAeReplyReceiver, boolean)
    */
   public AeMessageReceiver matchInboundReceive(AeInboundReceive aInboundReceive,
                  IAeReplyReceiver aReply, boolean aQueueIfNotFoundFlag)
                   throws AeCorrelationViolationException, AeConflictingRequestException, AeBusinessProcessException
   {
      synchronized(getUnmatchedReceives())
      {
         synchronized(getMessageReceivers())
         {
            AeMessageReceiver found = findMatchForInboundReceive(aInboundReceive, aReply != null);

            if (found == null && aQueueIfNotFoundFlag)
            {
               addUnmatchedReceive(aInboundReceive, aReply);
            }
            else if (found != null && aReply != null)
            {
               addReply(new AeReply(found.getProcessId(), aReply, aInboundReceive, found.getMessageExchange()), found);
            }
            return found;
         }
      }
   }

   /**
    * Finds a matching queued message receiver for the given inbound receive.
    * <br>
    * Note that the caller is responsible for any necessary synchronization.  This
    * method assumes that its work will always be done atomically (at least within
    * the same VM).
    *
    * @param aInboundReceive An inbound receive.
    * @param aReplyWaiting True if there is a reply waiting.
    * @return A matching queued receive, or null if not found.
    */
   protected AeMessageReceiver findMatchForInboundReceive(AeInboundReceive aInboundReceive,
         boolean aReplyWaiting) throws AeBusinessProcessException
   {
      AeMessageReceiver found = null;

      // Find an exact match first
      if (aInboundReceive.isCorrelated())
      {
         for (Iterator iter = getMessageReceivers().iterator(); found == null && iter.hasNext();)
         {
            AeMessageReceiver queuedReceive = (AeMessageReceiver) iter.next();
            if (queuedReceive.correlatesTo(aInboundReceive))
            {
               found = queuedReceive;
            }
         }
      }

      // Find partial match among uncorrelated message receivers
      if (found == null)
      {
         for (Iterator iter = getMessageReceivers().iterator(); found == null && iter.hasNext();)
         {
            AeMessageReceiver queuedReceive = (AeMessageReceiver) iter.next();

            if (!queuedReceive.isCorrelated() && queuedReceive.matches(aInboundReceive))
            {
               found = queuedReceive;
            }
         }
      }

      // If a match was found, make sure we 'consume' it by deleting it and any alarms in its group
      if (found != null)
      {
         removeMessageReceiver(found.getProcessId(), found.getMessageReceiverPathId());
         removeAlarmsInGroup(found.getProcessId(), found.getGroupId());
         found.setJournalId(getNextJournalId());
      }

      return found;
   }

   /**
    * Notifies the sender of the message and the process of the conflicting request fault.
    * @param aMessageReceiver
    * @param aReply
    * @param aFault
    */
   protected void handleConflictingRequest(AeMessageReceiver aMessageReceiver,
      IAeReplyReceiver aReply,
      IAeFault aFault)
   {
      try
      {
         // business process properties are null here because there was
         // no interaction with the business process
         aReply.onFault(aFault, null);
      }
      catch (AeBusinessProcessException e)
      {
         AeException.logError(e, AeMessages.getString("AeBaseQueueManager.ERROR_1")); //$NON-NLS-1$
      }

      try
      {
         getMessageReceiverResolver().onFault(aMessageReceiver, aFault);
      }
      catch (AeBusinessProcessException e)
      {
         AeException.logError(e, AeMessages.getString("AeBaseQueueManager.ERROR_1")); //$NON-NLS-1$
      }
   }

   /**
    * Removes a previously queued unmatched receive. If the unmatched receive is found
    * then we'll check to see if it also had a reply waiting. If so, this reply
    * is notified with a fault.
    * @param aQueueId Id that was assigned to the previously unmatched receive
    */
   protected void removeUnmatchedReceive(String aQueueId)
   {
      AeInboundReceive found = null;
      synchronized(getUnmatchedReceives())
      {
         for (Iterator iter = getUnmatchedReceives().iterator(); found == null && iter.hasNext();)
         {
            AeInboundReceive queuedReceive = (AeInboundReceive) iter.next();
            if (aQueueId.equals(queuedReceive.getQueueId()))
            {
               iter.remove();
               found = queuedReceive;
            }
         }
      }
      if (found != null)
      {
         // the receive was dequeued, now clean up the reply if there was one
         if (found.getResponseReceiverId() != null)
         {
            try
            {
               // business process properties are not set here because fault is
               // correlation violation - so no interaction with business process
               // took place
               IAeReplyReceiver responseReceiver = removeUnmatchedReply(found.getResponseReceiverId());
               responseReceiver.onFault(AeFaultFactory.getFault(AeFaultFactory.CORRELATION_VIOLATION), null);
            }
            catch (AeBusinessProcessException e)
            {
               AeException.logError(e, AeMessages.getString("AeBaseQueueManager.ERROR_3")); //$NON-NLS-1$
            }
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#removeMessageReceiver(long, int)
    */
   public boolean removeMessageReceiver(long aProcessId, int aMessageReceiverPathId) throws AeBusinessProcessException
   {
      synchronized(getMessageReceivers())
      {
         for(Iterator iter = getMessageReceivers().iterator(); iter.hasNext(); )
         {
            AeMessageReceiver entry = (AeMessageReceiver)iter.next();
            if (aProcessId == entry.getProcessId() && aMessageReceiverPathId == entry.getMessageReceiverPathId())
            {
               iter.remove();
               return true;
            }
         }
      }
      return false;
   }

   /**
    * Removes all message receivers in the group.
    * 
    * @param aProcessId
    * @param aGroupId
    * @param aLocationPathId
    * @throws AeBusinessProcessException
    */
   protected int removeMessageReceiversInGroup(long aProcessId, int aGroupId, int aLocationPathId)
         throws AeBusinessProcessException
   {
      int count = 0;
      synchronized(getMessageReceivers())
      {
         for(Iterator iter = getMessageReceivers().iterator(); iter.hasNext(); )
         {
            AeMessageReceiver entry = (AeMessageReceiver)iter.next();
            if (aProcessId == entry.getProcessId() && aGroupId == entry.getGroupId())
            {
               iter.remove();
               count++;
            }
         }
      }

      // For backwards compatibility, try to remove the specific message receiver if removing the
      // group didn't turn up any items.
      if (count == 0 && aLocationPathId != -1)
      {
         if (removeMessageReceiver(aProcessId, aLocationPathId))
         {
            count++;
         }
      }
      
      return count;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#sendReply(org.activebpel.rt.bpel.impl.queue.AeReply, org.activebpel.rt.message.IAeMessageData, javax.xml.namespace.QName, java.util.Map)
    */
   public boolean sendReply(AeReply aReplyObject, IAeMessageData aData, QName aFaultName, Map aProcessProperties ) throws AeBusinessProcessException
   {
      AeReply waitingReply = (AeReply) removeFromQueue(getReplies(), aReplyObject);

      if (waitingReply == null)
         throw new AeInvalidReplyException();
      if (aFaultName != null)
         waitingReply.getReplyReceiver().onFault(new AeFault(aFaultName, aData), aProcessProperties );
      else
         waitingReply.getReplyReceiver().onMessage(aData, aProcessProperties);

      return true;
   }

   /**
    * Walks the queue looking for a match, removes if found.
    * @param aQueue the queue to search
    * @param aQueueObject the object we're looking to remove
    * @return the removed object or null if not found
    */
   private Object removeFromQueue(Collection aQueue, Object aQueueObject)
   {
      synchronized(aQueue)
      {
         for(Iterator iter = aQueue.iterator(); iter.hasNext(); )
         {
            Object entry = iter.next();
            if(entry.equals(aQueueObject))
            {
               aQueue.remove(entry);
               return entry;
            }
         }
      }
      return null;
   }

   /**
    * Adds an unmatched reply so that we'll have a place to store it in the event that
    * we're able to match a message receiver to the originating inbound receive or
    * until that inbound receive times out.
    * @param aResponseReceiver
    */
   protected String addUnmatchedReply(IAeReplyReceiver aResponseReceiver)
   {
      String key = null;
      synchronized(getUnmatchedReplies())
      {
         key = String.valueOf(getNextUnmatchedReplyId());
         getUnmatchedReplies().put(key, aResponseReceiver);
      }
      return key;
   }

   /**
    * Removes an unmatched reply. Unmatched replies are removed from the queue when
    * the inbound receive that they came in with is matched or is timed out.
    * @param aReplyId
    */
   protected IAeReplyReceiver removeUnmatchedReply(String aReplyId)
   {
      synchronized(getUnmatchedReplies())
      {
         return (IAeReplyReceiver) getUnmatchedReplies().remove(aReplyId);
      }
   }

   /**
    * returns the next value to use as an unmatched reply id
    */
   protected long getNextUnmatchedReplyId()
   {
      return mUnmatchedReplyId++;
   }

   /**
    * Getter for the unmatched receives collection
    */
   protected List getUnmatchedReceives()
   {
      return mUnmatchedReceives;
   }

   /**
    * Getter for the message receivers collection
    */
   protected Collection getMessageReceivers()
   {
      return mMessageReceivers;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#getMessageReceivers(org.activebpel.rt.bpel.impl.list.AeMessageReceiverFilter)
    */
   public AeMessageReceiverListResult getMessageReceivers(AeMessageReceiverFilter aFilter)
         throws AeBusinessProcessException
   {
      AeMessageReceiverFilter filter = aFilter == null ? AeMessageReceiverFilter.NULL_FILTER : aFilter;
      synchronized( getMessageReceivers() )
      {
         return AeMessageReceiverFilterManager.filter( filter, getMessageReceivers() );
      }
   }

   /**
    * Getter for the message receiver resolver. This interface is used to resolve
    * a pid and message path to a message receiver activity.
    */
   public IAeMessageReceiverResolver getMessageReceiverResolver()
   {
      return mMessageReceiverResolver;
   }

   /**
    * Setter for the message receiver resolver.
    * @param aResolver
    */
   public void setMessageReceiverResolver(IAeMessageReceiverResolver aResolver)
   {
      mMessageReceiverResolver = aResolver;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#addReply(org.activebpel.rt.bpel.impl.queue.AeReply, org.activebpel.rt.bpel.impl.queue.AeMessageReceiver)
    */
   public void addReply(AeReply aReply, AeMessageReceiver aMessageReceiver) throws AeConflictingRequestException
   {
      synchronized(getReplies())
      {
         if (getReplies().contains(aReply))
         {
            AeConflictingRequestException exception = new AeConflictingRequestException();
            if (aMessageReceiver != null)
            {
               handleConflictingRequest(aMessageReceiver, aReply.getReplyReceiver(), exception.getFault());
            }
            throw exception;
         }
         else
         {
            getReplies().add(aReply);
         }
      }
   }

   /**
    * Getter for the unmatched replies map
    */
   protected Map getUnmatchedReplies()
   {
      return mUnmatchedReplies;
   }

   /**
    * Getter for the replies list
    */
   protected List getReplies()
   {
      return mReplies;
   }

   /**
    * Getter for the invokes list
    */
   protected List getInvokes()
   {
      return mInvokes;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#getQueuedInvokeLocationPaths(long)
    */
   public Set getQueuedInvokeLocationPaths(long aProcessId)
   {
      Set locationPaths = new HashSet();

      // TODO (KR) Index invokes by process id
      synchronized (getInvokes())
      {
         for (Iterator i = getInvokes().iterator(); i.hasNext(); )
         {
            IAeInvoke invoke = (IAeInvoke) i.next();

            if (invoke.getProcessId() == aProcessId)
            {
               String locationPath = invoke.getLocationPath();
               locationPaths.add(locationPath);
            }
         }
      }

      return locationPaths;
   }

   /**
    * Returns <code>int</code> value from configuration <code>Map</code>.
    */
   protected static int getConfigInt(Map aConfig, String aKey, int aDefaultValue)
   {
      String value = (String) aConfig.get(aKey);
      if (!AeUtil.isNullOrEmpty(value))
      {
         try
         {
            return Integer.parseInt(value);
         }
         catch (NumberFormatException e)
         {
            AeException.logError(e, AeMessages.format("AeBaseQueueManager.ERROR_BAD_CONFIG_KEY_INT", new Object[] { value, aKey })); //$NON-NLS-1$
         }
      }

      return aDefaultValue;
   }

   /**
    * Returns maximum number of unmatched receives.
    */
   protected int getMaxUnmatchedReceivesCount()
   {
      return mMaxUnmatchedReceivesCount;
   }

   /**
    * Sets configuration.
    */
   protected void setConfig(Map aConfig)
   {
      setMaxUnmatchedReceivesCount(getConfigInt(aConfig, CONFIG_UNMATCHED_RECEIVES_COUNT, DEFAULT_UNMATCHED_RECEIVES_COUNT));
   }

   /**
    * Sets maximum number of unmatched receives.
    */
   protected void setMaxUnmatchedReceivesCount(int aMaxUnmatchedReceivesCount)
   {
      if (aMaxUnmatchedReceivesCount > 0)
      {
         mMaxUnmatchedReceivesCount = aMaxUnmatchedReceivesCount;
      }
      else
      {
         // Specifying 0 or less means no limit.
         mMaxUnmatchedReceivesCount = Integer.MAX_VALUE;
      }
   }

   /**
    * Provides a way of resolving a queue entry to its bpel object and then invoking
    * the targeted method. This encapsulates the business about having to lock
    * and then release the process.
    */
   public interface IAeMessageReceiverResolver
   {
      /**
       * Resolves the message receiver's pid and path to a bpel activity and then
       * calls onMessage on that activity with the data.
       * @param aMessageReceiver
       * @param aMessage
       * @param aReplyWaitingFlag
       * @throws AeBusinessProcessException
       */
      public void onMessage(AeMessageReceiver aMessageReceiver, AeInboundReceive aMessage, boolean aReplyWaitingFlag) throws AeBusinessProcessException;

      /**
       * Resolves the message receiver's pid and path to a bpel activity and then
       * calls onFault on that activity with the provided fault.
       * @param aMessageReceiver
       * @param aFault
       * @throws AeBusinessProcessException
       */
      public void onFault(AeMessageReceiver aMessageReceiver, IAeFault aFault) throws AeBusinessProcessException;
   }

   /**
    * Default impl for the resolver interfaces assumes that we're dealing with
    * an AeBusinessProcessEngine and casts to get access to the engine's methods
    * for getting the process and message receiver.
    */
   protected class AeDefaultMessageReceiverResolver implements IAeMessageReceiverResolver
   {
      /**
       * @see org.activebpel.rt.bpel.impl.AeBaseQueueManager.IAeMessageReceiverResolver#onMessage(org.activebpel.rt.bpel.impl.queue.AeMessageReceiver, org.activebpel.rt.bpel.impl.queue.AeInboundReceive, boolean)
       */
      public void onMessage(AeMessageReceiver aMessageReceiver, AeInboundReceive aMessage, boolean aReplyWaitingFlag) throws AeBusinessProcessException
      {
         getEngine().handleMatchedReceive(
            aMessage,
            aMessageReceiver,
            aReplyWaitingFlag
         );
      }

      /**
       * @see org.activebpel.rt.bpel.impl.AeBaseQueueManager.IAeMessageReceiverResolver#onFault(org.activebpel.rt.bpel.impl.queue.AeMessageReceiver, org.activebpel.rt.bpel.IAeFault)
       */
      public void onFault(AeMessageReceiver aMessageReceiver, IAeFault aFault) throws AeBusinessProcessException
      {
         IAeBusinessProcess process = ((AeBusinessProcessEngine)getEngine()).getProcessById(aMessageReceiver.getProcessId());
         try
         {
            IAeMessageReceiverActivity receiver = (IAeMessageReceiverActivity) ((AeBusinessProcess)process).findBpelObject(aMessageReceiver.getMessageReceiverPathId());
            receiver.onFault(aFault);
         }
         finally
         {
            ((AeBusinessProcessEngine)getEngine()).releaseProcess(process);
         }
      }
   }

   /**
    * Returns true if the manager has been started. This is to avoid scheduling 
    * an alarm while the manager is stopped - something that is possible in some
    * conditions during a shutdown.
    */
   protected boolean isStarted()
   {
      return mStarted;
   }

   /**
    * Used to actually perform the scheduling of the alarm. Derived classes should  
    * implement this to schedule with their timer implementation.
    * @param aListener the listener to be scheduled
    * @param aDeadline the date this alarm should execute
    * @return Timer object created by the scheduler
    */
   protected abstract Timer schedule(TimerListener aListener, Date aDeadline);
   
   /**
    * Creates the alarm listener which encapsulates the alarm work to be performed.
    * Derived classes should implement this such that the work is performed in a 
    * separate thread.
    * @param aAlarm The alarm definition
    */
   protected abstract TimerListener createAlarmListener(AeAlarm aAlarm);

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#scheduleAlarm(long, int, int, java.util.Date)
    */
   public void scheduleAlarm(long aProcessId, int aLocationPathId, int aGroupId, Date aDeadline)
         throws AeBusinessProcessException
   {
      AeAlarm alarm = new AeAlarm(aProcessId, aLocationPathId, aGroupId, aDeadline);

      // By synchronizing on the Hashtable, the entry can be added before a fired alarm
      // goes to remove it.  This will avoid the case where the alarm manager immediately
      // fires the scheduled alarm prior to it being added to the table.
      synchronized (mLookup)
      {
         // Make sure we don't double-schedule the alarm.
         if (! mLookup.containsKey(alarm))
         {
            Timer timer = schedule(createAlarmListener(alarm), aDeadline);
            if (timer != null)
               mLookup.put(alarm, timer);
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#removeAlarmForDispatch(long, int, int)
    */
   public long removeAlarmForDispatch(long aProcessId, int aGroupId, int aLocationPathId)
         throws AeBusinessProcessException
   {
      // Synchronize on the message receiver queue - the 'find match for inbound receive' will also
      // synchronize on the message receiver queue.  This eliminates a race condition where
      // an alarm fires at the same instant as an inbound receive is handled.
      synchronized (getMessageReceivers())
      {
         // Remove and unschedule the alarm.
         if (internalRemoveAlarm(aProcessId, aLocationPathId))
         {
            // Remove any message receivers in this group.
            removeMessageReceiversInGroup(aProcessId, aGroupId, aLocationPathId);
            // Pretend to journal the alarm in the process journal.
            return getNextJournalId();
         }
         else
         {
            return IAeProcessManager.NULL_JOURNAL_ID;
         }
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#removeAlarm(long, int)
    */
   public boolean removeAlarm(long aProcessId, int aLocationPathId) throws AeBusinessProcessException
   {
      return internalRemoveAlarm(aProcessId, aLocationPathId);
   }

   /**
    * Removes all of the alarms in the group.
    * 
    * @param aProcessId
    * @param aGroupId
    * @throws AeBusinessProcessException
    */
   protected int removeAlarmsInGroup(long aProcessId, int aGroupId) throws AeBusinessProcessException
   {
      int count = 0;
      synchronized (mLookup)
      {
         for (Iterator iter = mLookup.entrySet().iterator(); iter.hasNext(); )
         {
            Map.Entry entry = (Map.Entry) iter.next();
            AeAlarm alarm = (AeAlarm) entry.getKey();
            if (alarm.getProcessId() == aProcessId && alarm.getGroupId() == aGroupId)
            {
               count++;
               Timer timer = (Timer) entry.getValue();
               timer.cancel();
               iter.remove();
            }
         }
      }
      return count;
   }

   /**
    * This method removes the alarm from its internal representation.  In the case of the
    * in-memory alarm manager, this means removing it from the cache.  Once removed, this
    * method returns the Timer object that will be used to cancel the alarm from the
    * timer service.
    * 
    * @param aProcessId The process id of the alarm.
    * @param aLocationPathId The location path of the alarm.
    * @return True if the alarm was successfully removed.
    * todo should this get refactored to just be part of removeAlarm 
    */
   protected boolean internalRemoveAlarm(long aProcessId, int aLocationPathId) throws AeBusinessProcessException
   {
      Timer timer = null;
      synchronized (mLookup)
      {
         timer = (Timer)mLookup.remove(new AeAlarm(aProcessId, aLocationPathId));
      }

      if (timer == null)
         return false;
      else
      {
         // Cancel the timer in case it is executing, then return true indicating it has been removed 
         timer.cancel();
         return true;
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeManager#start()
    */
   public void start() throws Exception
   {
      super.start();
      mStarted = true;
      doInitialAlarmLoad();
   }

   /**
    * Called when the manager starts to do the initial loadup of saved alarms.  These saved
    * alarms were scheduled prior to stopping the manager.  Now that it's started up again, 
    * the alarms must be loaded and scheduled.
    */
   protected void doInitialAlarmLoad() throws AeBusinessProcessException
   {
      if(mSavedLookup != null)
      {
         for(Iterator iter = mSavedLookup.values().iterator(); iter.hasNext(); )
         {
            AeAlarm alarm = (AeAlarm)iter.next();
            scheduleAlarm(alarm.getProcessId(), alarm.getPathId(), alarm.getGroupId(), alarm.getDeadline());
         }
         mSavedLookup = null;
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeManager#stop()
    */
   public void stop()
   {
      super.stop();
      
      mStarted = false;

      synchronized (mLookup)
      {
         for (Iterator iter = mLookup.values().iterator(); iter.hasNext();)
            ((Timer)iter.next()).cancel();

         saveLookup();
         mLookup.clear();
      }
   }

   /**
    * Saves lookup for in-memory start and stops. 
    */
   protected void saveLookup()
   {
      mSavedLookup = new HashMap(mLookup);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeManager#destroy()
    */
   public void destroy()
   {
      stop();

      super.destroy();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeQueueManager#getAlarms(org.activebpel.rt.bpel.impl.list.AeAlarmFilter)
    */
   public AeAlarmListResult getAlarms(AeAlarmFilter aFilter) throws AeBusinessProcessException
   {
      AeAlarmFilter filter = aFilter == null ? AeAlarmFilter.NULL_FILTER : aFilter;
      synchronized( mLookup )
      {
         // The keyset is a collection of AeAlarm objects
         ArrayList list = new ArrayList(mLookup.keySet());
         return AeAlarmFilterManager.filter(getEngine(), filter, list);
      }
   }

   /**
    * Comparator that brings the correlated message receivers with the most
    * properties to the top so we can match against those first. 
    */
   private static class AeMessageReceiverComparator implements Comparator
   {
      public int compare(Object aO1, Object aO2)
      {
         AeMessageReceiver one = (AeMessageReceiver) aO1;
         AeMessageReceiver two = (AeMessageReceiver) aO2;
         
         int twoSize = two.isCorrelated() ? two.getCorrelation().size() : 0;
         int oneSize = one.isCorrelated() ? one.getCorrelation().size() : 0;

         int result = twoSize - oneSize;
         
         if (result == 0)
         {
            // if they have the same number of correlations, then compare using
            // their hash values
            result = two.hashCode() - one.hashCode();
         }
         
         return result;
      }
   }
}
