// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeBusinessProcessEngine.java,v 1.129 2006/02/14 22:39:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;

import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.AeWSDLDefHelper;
import org.activebpel.rt.bpel.IAeBusinessProcess;
import org.activebpel.rt.bpel.IAeEndpointReference;
import org.activebpel.rt.bpel.IAeEngineAlert;
import org.activebpel.rt.bpel.IAeEngineEvent;
import org.activebpel.rt.bpel.IAeEngineListener;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.IAePlanManager;
import org.activebpel.rt.bpel.IAeProcessEvent;
import org.activebpel.rt.bpel.IAeProcessInfoEvent;
import org.activebpel.rt.bpel.IAeProcessListener;
import org.activebpel.rt.bpel.config.IAeEngineConfiguration;
import org.activebpel.rt.bpel.impl.list.AeProcessInstanceDetail;
import org.activebpel.rt.bpel.impl.queue.AeInboundReceive;
import org.activebpel.rt.bpel.impl.queue.AeInvoke;
import org.activebpel.rt.bpel.impl.queue.AeMessageReceiver;
import org.activebpel.rt.bpel.impl.queue.AeReply;
import org.activebpel.rt.bpel.impl.visitors.AeOutstandingRepliesVisitor;
import org.activebpel.rt.bpel.xpath.AeXPathHelper;
import org.activebpel.rt.message.AeMessageData;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeSafelyViewableCollection;
import org.activebpel.rt.util.AeXmlUtil;
import org.activebpel.rt.wsdl.def.AeBPELExtendedWSDLDef;
import org.activebpel.rt.wsdl.def.AeOperationSignatureMatcher;
import org.activebpel.rt.wsdl.def.IAeProperty;
import org.activebpel.rt.wsdl.def.IAePropertyAlias;
import org.activebpel.wsio.IAeWebServiceMessageData;
import org.activebpel.wsio.IAeWebServiceResponse;
import org.activebpel.wsio.invoke.AeInvokeResponse;
import org.activebpel.wsio.invoke.IAeInvoke;
import org.activebpel.wsio.receive.AeMessageContext;
import org.activebpel.wsio.receive.AeTimeoutException;
import org.activebpel.wsio.receive.IAeMessageContext;
import org.exolab.castor.xml.schema.XMLType;
import org.w3c.dom.Document;

/** The class implementing the business process execution engine. */
public class AeBusinessProcessEngine implements IAeBusinessProcessEngineInternal, IAeBusinessProcessEngineCallback
{
   /** Engine listeners */
   protected Collection mEngineListeners = new AeSafelyViewableCollection(new LinkedHashSet());

   /** Global process listeners */
   protected Collection mGlobalProcessListeners = new AeSafelyViewableCollection(new LinkedHashSet());

   /** Map of listeners, keyed by the process ID */
   protected HashMap mProcessListeners = new HashMap();

   /** The queue for incoming data that gets dispatched to processes */
   protected IAeQueueManager mQueueManager;

   /** Process manager */
   private IAeProcessManager mProcessManager;

   /** The lock manager for handling acquiring and releasing locks. */
   protected IAeLockManager mLockManager;

   /** Maps process QName to its process plan */
   protected IAePlanManager mPlanManager;
   
   /** The type mapping to use when converting from schema to java and back. */
   private AeTypeMapping mTypeMapping = new AeTypeMapping();

   /** The engine configuration supplied during engine construction. */
   protected IAeEngineConfiguration mEngineConfiguration;

   /** The date the engine started */
   protected Date mStarted;

   /** Strategy for managing partner links. */
   private IAeEnginePartnerLinkStrategy mPartnerLinkStrategy;
   
   /** Coordination manager */
   private IAeCoordinationManagerInternal mCoordinationManager;
   
   /**
    * Process coordination handler. 
    */
   protected IAeProcessCoordination mProcessCoordination;

   /**
    * Constructs a new engine with the passed configuration, queue manager,
    * process manager, and alarm manager.
    *
    * @param aEngineConfiguration The engine configuration to use for this engine.
    * @param aQueueManager The queue manager to be associated with this engine.
    * @param aProcessManager The process manager to be associated with this engine.
    * @param aLockManager The lock manager to be associated with this engine.
    */
   public AeBusinessProcessEngine(
      IAeEngineConfiguration aEngineConfiguration,
      IAeQueueManager aQueueManager,
      IAeProcessManager aProcessManager,
      IAeLockManager aLockManager)
   {
      mEngineConfiguration = aEngineConfiguration;
      mQueueManager = aQueueManager;
      mProcessManager = aProcessManager;
      mLockManager = aLockManager;
      if (mQueueManager != null)
         mQueueManager.setEngine(this);
      if (mProcessManager != null)
         mProcessManager.setEngine(this);
      if (mLockManager != null)
         mLockManager.setEngine(this);
      mStarted = new Date();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#getProcessManager()
    */
   public IAeProcessManager getProcessManager()
   {
      return mProcessManager;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#getPartnerRoleEndpointReference(long, java.lang.String)
    */
   public IAeEndpointReference getPartnerRoleEndpointReference(long aPid, String aPartnerLinkPath) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         return process.getPartnerRoleEndpointReference( aPartnerLinkPath );
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * Executes a business process
    * @param aPid - the pid for the process to execute
    * @param aExecuteImmediatelyFlag - a hint regarding whether a reply is waiting for the process to execute or not.
    */
   protected void executeProcess(long aPid, boolean aExecuteImmediatelyFlag) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         if (process == null)
            throw new AeBusinessProcessException(AeMessages.getString("AeBusinessProcessEngine.ERROR_0") + aPid); //$NON-NLS-1$

         // tell the process to queue itself to execute
         process.queueObjectToExecute(process);
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * Returns the business process with the specified process id, locking the
    * process into memory. <em>Each call to {@link #getProcessById(long)} must
    * be followed eventually by a matching call to {@link
    * #releaseProcess(IAeBusinessProcess)}</em>.
    *
    * @param aId
    */
   protected IAeBusinessProcess getProcessById(long aId) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessManager().getProcess(aId);
      if (process == null)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeBusinessProcessEngine.ERROR_0") + aId); //$NON-NLS-1$
      }

      return process;
   }

   /**
    * Releases a process locked by {@link #getProcessById(long)}.
    */
   protected void releaseProcess(IAeBusinessProcess aProcess)
   {
      getProcessManager().releaseProcess(aProcess);
   }

   /**
    * Convenience method that calls {@link
    * IAeProcessManager#journalEntryDone(long, long)}.
    */
   protected void journalEntryDone(IAeBusinessProcess aProcess, long aJournalId)
   {
      if (aProcess != null)
      {
         getProcessManager().journalEntryDone(aProcess.getProcessId(), aJournalId);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#setVariableData(long, java.lang.String, org.w3c.dom.Document)
    */
   public void setVariableData(long aPid, String aVariablePath, Document aVariableData)
      throws AeBusinessProcessException
   {
      IAeBusinessProcess process = null;
      try
      {
         process = getProcessById(aPid);
         process.setVariableData( aVariablePath, aVariableData, true );
      }
      finally
      {
         if(process != null)
            releaseProcess(process);
      }
   }
   
   /** 
    * Overrides method to 
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#getCorrelationData(long, java.lang.String)
    */
   public Map getCorrelationData( long aPid, String aCorrsetPath) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = null;
      try
      {
         process = getProcessById(aPid);
         return process.getCorrelationData(aCorrsetPath);
      }
      finally
      {
         if (process != null)
            releaseProcess(process);
      }      
   }
      
   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#setCorrelationData(long, java.lang.String, java.util.Map)
    */
   public void setCorrelationData(long aPid, String aCorrsetPath,
         Map aCorrelationData) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = null;
      try
      {
         process = getProcessById(aPid);
         process.setCorrelationData( aCorrsetPath, aCorrelationData );
      }
      finally
      {
         if (process != null)
            releaseProcess(process);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#setPartnerLinkData(long, boolean, java.lang.String, org.w3c.dom.Document)
    */
   public void setPartnerLinkData(long aPid, boolean aIsPartnerRole, String aPartnerLinkPath,
                                  Document aPartnerEndpointRef) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = null;
      try
      {
         process = getProcessById(aPid);
         process.setPartnerLinkData( aIsPartnerRole, aPartnerLinkPath, aPartnerEndpointRef );
      }
      finally
      {
         if (process != null)
            releaseProcess(process);
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#suspendProcess(long)
    */
   public void suspendProcess(long aPid) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         // Only suspend if the process is running
         if (process.getProcessState() == IAeBusinessProcess.PROCESS_RUNNING)
         {
            // Pass in null to indicate a manual type suspend
            process.suspend(null);
         }
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#resumeProcess(long)
    */
   public void resumeProcess(long aPid) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         // Only resume if the process is suspended
         if (process.isSuspended())
         {
            process.resume(true);
         }
      }
      finally
      {
         releaseProcess(process);
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#retryActivity(long, java.lang.String, boolean)
    */
   public void retryActivity( long aPid, String aLocation, boolean aAtScope )
   throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         // Only resume if the process is suspended
         if (process.isSuspended())
         {
            process.retryActivity( aLocation, aAtScope );
         }
      }
      finally
      {
         releaseProcess(process);
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#completeActivity(long, java.lang.String)
    */
   public void completeActivity( long aPid, String aLocation )
   throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         // Only resume if the process is suspended
         if (process.isSuspended())
         {
            process.completeActivity( aLocation );
         }
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#terminateProcess(long)
    */
   public void terminateProcess(long aPid) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         // Only terminate if the process is not completed
         if (process.getProcessState() != IAeBusinessProcess.PROCESS_COMPLETE && process.getProcessState() != IAeBusinessProcess.PROCESS_FAULTED && process.getProcessState() != IAeBusinessProcess.PROCESS_COMPENSATABLE)
         {
            process.terminate();
         }
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#resumeProcessObject(long, java.lang.String)
    */
   public void resumeProcessObject(long aPid, String aLocation) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aPid);
      try
      {
         // Only resume if the process is suspended
         if (process.isSuspended())
         {
            process.resume(aLocation);
         }
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueReceiveData(javax.xml.namespace.QName, org.activebpel.rt.message.IAeMessageData, java.lang.String, javax.xml.namespace.QName, java.lang.String, org.activebpel.rt.bpel.impl.IAeMessageReceiver, java.lang.String, org.activebpel.rt.bpel.IAeEndpointReference)
    * @deprecated
    */
   public void queueReceiveData(QName aProcessQName, IAeMessageData aMessageData, String aPartnerLinkName,
            QName aPortType, String aOperation, IAeMessageReceiver aReply, String aPrincipal, IAeEndpointReference aEmbeddedEndpointReference) throws AeBusinessProcessException
   {
      IAeProcessPlan plan = getPlanManager().findCurrentPlan(aProcessQName);
      queueReceiveData(plan, aMessageData, aPartnerLinkName, aPortType, aOperation, aReply, aPrincipal, aEmbeddedEndpointReference);
   }

   /**
    * Receives messages that get dispatched to a specific process instance.
    * @param aProcessPlan
    * @param aMessageData
    * @param aPartnerLinkName
    * @param aPortType
    * @param aOperation
    * @param aReply
    * @param aPrincipal
    * @param aEmbeddedEndpointRef
    * @throws AeBusinessProcessException
    * @deprecated
    */
   public void queueReceiveData(IAeProcessPlan aProcessPlan, IAeMessageData aMessageData, String aPartnerLinkName,
            QName aPortType, String aOperation, IAeMessageReceiver aReply, String aPrincipal, IAeEndpointReference aEmbeddedEndpointRef) throws AeBusinessProcessException
   {
      AeMessageContext context = new AeMessageContext();
      context.setEmbeddedEndpointReference(aEmbeddedEndpointRef);
      context.setOperation(aOperation);
      context.setPartnerLinkName(aPartnerLinkName);
      context.setPrincipal(aPrincipal);
      context.setProcessName(aProcessPlan.getProcessDef().getQName());
      queueReceiveData(aProcessPlan, aMessageData, aReply, context);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueReceiveData(org.activebpel.rt.bpel.impl.IAeProcessPlan, org.activebpel.rt.message.IAeMessageData, org.activebpel.rt.bpel.impl.IAeMessageReceiver, org.activebpel.wsio.receive.IAeMessageContext)
    * @deprecated
    */
   public void queueReceiveData(IAeProcessPlan aProcessPlan,
         IAeMessageData aMessageData, IAeMessageReceiver aReply,
         IAeMessageContext aContext) throws AeBusinessProcessException
   {
      AeReplyReceiverAdapter replyReceiver = new AeReplyReceiverAdapter(aReply);
      queueReceiveData( aProcessPlan, aMessageData, replyReceiver, aContext );
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueReceiveData(org.activebpel.wsio.receive.IAeMessageContext, org.activebpel.wsio.IAeWebServiceMessageData)
    */
   public IAeWebServiceResponse queueReceiveData(IAeMessageContext aContext, IAeWebServiceMessageData aData) 
      throws AeBusinessProcessException, AeTimeoutException
   {
      IAeMessageData data = AeDataConverter.convert(aData);
      AeMessageContext context = AeMessageContext.convert(aContext);
      IAeProcessPlan plan = getProcessPlan(context);
      return queueReceiveData(plan, context, data);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueReceiveData(org.activebpel.wsio.receive.IAeMessageContext, org.w3c.dom.Document[])
    */
   public IAeWebServiceResponse queueReceiveData(IAeMessageContext aContext, Document[] aDataArray)  
      throws AeBusinessProcessException, AeTimeoutException
   {
      AeMessageContext context = AeMessageContext.convert(aContext);
      IAeProcessPlan plan = getProcessPlan(context);
      
      QName portTypeQName = plan.getMyRolePortType(context.getPartnerLinkName());
      
      AeBPELExtendedWSDLDef portTypeDef = AeWSDLDefHelper.getWSDLDefinitionForPortType(plan, portTypeQName);
      
      AeOperationSignatureMatcher opFinder = new AeOperationSignatureMatcher(portTypeDef, portTypeQName, context.getOperation(), aDataArray);
      if (!opFinder.foundMatch())
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeBusinessProcessEngine.NO_OPERATION_FOUND")); //$NON-NLS-1$
      }

      context.setOperation(opFinder.getOperationName());
      Map partAsKeyMap = opFinder.getPartsMap();
      Map nameAsKeyMap = new HashMap();
      for (Iterator iter = partAsKeyMap.entrySet().iterator(); iter.hasNext();)
      {
         Map.Entry entry = (Map.Entry) iter.next();
         Part part = (Part) entry.getKey();
         
         Object value = entry.getValue();
         if (part.getTypeName() != null)
         {
            // store a string instead of the doc if it's a simple type
            try
            {
               XMLType type = portTypeDef.findType(part.getTypeName());
               if (type != null && type.isSimpleType())
               {
                  value = AeXmlUtil.getText(((Document)value).getDocumentElement());
               }
            }
            catch (AeException e)
            {
               e.logError();
            }
         }
         
         nameAsKeyMap.put(part.getName(), value);
      }
      
      AeMessageData data = new AeMessageData(opFinder.getMessageName(), nameAsKeyMap);
      return queueReceiveData(plan, context, data);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#queueReceiveData(org.activebpel.rt.bpel.impl.IAeProcessPlan, org.activebpel.wsio.receive.IAeMessageContext, org.activebpel.rt.message.IAeMessageData)
    */
   public IAeWebServiceResponse queueReceiveData(IAeProcessPlan aPlan, IAeMessageContext aContext, IAeMessageData aData) 
      throws AeBusinessProcessException, AeTimeoutException
   {
      if (aPlan == null)
      {
         aPlan = getProcessPlan(AeMessageContext.convert(aContext));
      }
      
      boolean isOneWay = aPlan.getProcessDef().isOneWayReceive(aContext.getPartnerLinkName(), aContext.getOperation());
      
      AeWaitableReplyReceiver replyReceiver = isOneWay ? null : new AeWaitableReplyReceiver(); 

      AeInvokeResponse response = new AeInvokeResponse();
      
      queueReceiveData(aPlan, aData, replyReceiver, aContext);
      
      if (replyReceiver != null)
      {
         synchronized(replyReceiver)
         {
            if (replyReceiver.isWaitable())
            {
               try
               {
                  replyReceiver.wait(600000);
               }
               catch (InterruptedException e)
               {
                  throw new AeBusinessProcessException(AeMessages.getString("AeBusinessProcessEngine.WAIT_INTERRUPT")); //$NON-NLS-1$
               }
            }
         }

         // set the process properties back into the response
         response.setBusinessProcessProperties( replyReceiver.getBusinessProcessProperties() );

         // map the output message or fault
         if (replyReceiver.getMessageData() != null)
            response.setMessageData(AeDataConverter.convert(replyReceiver.getMessageData()));
         else if (replyReceiver.getFault() != null)
         {
            response.setFaultData(replyReceiver.getFault().getFaultName(), AeDataConverter.convert(replyReceiver.getFault().getMessageData()));
            response.setErrorString(replyReceiver.getFault().getInfo());
            response.setErrorDetail(replyReceiver.getFault().getDetailedInfo());
         }
         else
         {
            String msg = AeMessages.format("AeBusinessProcessEngine.WARNING_TIMEOUT_WAITING_FOR_REPLY", new Object[] { aPlan.getProcessDef().getName() }); //$NON-NLS-1$
            throw new AeTimeoutException(msg);
         }
      }
      return response;
   }
   
   /**
    * Gets the plan used to route the incoming message based on the information in the context.
    * At this layer, the plan is based on the process QName only. This will change with
    * other engines that support more advanced deployment options.
    * 
    * @param aContext
    * @throws AeBusinessProcessException
    * @throws AeException
    */
   protected IAeProcessPlan getProcessPlan(AeMessageContext aContext) throws AeBusinessProcessException
   {
      return getPlanManager().findCurrentPlan(aContext.getProcessName());
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueReceiveData(org.activebpel.rt.bpel.impl.IAeProcessPlan, org.activebpel.rt.message.IAeMessageData, org.activebpel.rt.bpel.impl.IAeReplyReceiver, org.activebpel.wsio.receive.IAeMessageContext)
    */
   public void queueReceiveData(IAeProcessPlan aProcessPlan,
         IAeMessageData aMessageData, IAeReplyReceiver aReply,
         IAeMessageContext aContext) throws AeBusinessProcessException
   {
      if (aProcessPlan == null)
      {
         // Process plan is optional. For example, the custom invoker for process and subprocess invokes
         // using the service name instead of the process plan.
         aProcessPlan = getProcessPlan(AeMessageContext.convert(aContext));
      }
      QName portType = aProcessPlan.getMyRolePortType(aContext.getPartnerLinkName());

      Map correlationMap = createCorrelationMap(aProcessPlan, aMessageData, aContext, portType);
      AeInboundReceive inboundReceive = new AeInboundReceive(correlationMap, aProcessPlan, aMessageData, aContext);

      queueReceiveData(inboundReceive, aReply);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueReceiveData(org.activebpel.rt.bpel.impl.queue.AeInboundReceive, org.activebpel.rt.bpel.impl.IAeReplyReceiver)
    */
   public void queueReceiveData(AeInboundReceive aInboundReceive, IAeReplyReceiver aReply)
         throws AeCorrelationViolationException, AeConflictingRequestException, AeBusinessProcessException
   {
      IAeProcessPlan plan = aInboundReceive.getProcessPlan();
      boolean canCreateInstance = plan.isCreateInstance(aInboundReceive.getPartnerLinkName(), aInboundReceive.getOperation());
      boolean coordinated = isCoordinated(aInboundReceive.getContext());

      AeMessageReceiver found = null;
      // if this is not a coordinated process, check to see if an entry is waiting.
      if (!coordinated)
      {
         found = getQueueManager().matchInboundReceive(aInboundReceive, aReply, !canCreateInstance);
      }

      // If a match was found, dispatch the receive data.
      // Note that at this point the match has been consumed from the queue and journalled!
      if (found != null)
      {
         boolean replyWaiting = aReply != null;
         boolean runSynchronously = replyWaiting; // do it synchronously if there is a reply waiting.
         dispatchReceiveData(found, aInboundReceive, replyWaiting, runSynchronously, found.getJournalId());
      }
      // If no match was found AND the message is capable of creating a new instance, then create the instance
      else if (found == null && canCreateInstance)
      {
         createProcessWithMessage(aInboundReceive, aReply);
      }
      else if (coordinated)
      {
         // subprocess's must be a createInstance and hence should have created an instance above.
         throw new AeCorrelationViolationException(AeCorrelationViolationException.NOT_CREATEINSTANCE);
      }
      else
      {
         // this is the case for (found == null && !canCreateInstance) or rather "no message receiver found 
         // and can not create instance".  Since that case will either result in the inbound receive 
         // having been added to the IAeQueueManager or having resulted in a bpws:correlationViolation,
         // we don't need to do anything here.
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#handleMatchedReceive(org.activebpel.rt.bpel.impl.queue.AeInboundReceive, org.activebpel.rt.bpel.impl.queue.AeMessageReceiver, boolean)
    */
   public void handleMatchedReceive(
      AeInboundReceive aInboundReceive,
      AeMessageReceiver aFound,
      boolean aReplyWaitingFlag)
      throws AeBusinessProcessException
   {
      dispatchReceiveData(aFound, aInboundReceive, aReplyWaitingFlag, true, IAeProcessManager.NULL_JOURNAL_ID);
   }

   /**
    * Dispatches the message data to the process's message receiver.
    * 
    * @param aMessageReceiver
    * @param aInboundReceive
    * @param aReplyWaitingFlag
    * @param aRunSynchronously Serves as a hint to how the dispatch might be handled.
    * @param aJournalId journal id of the message if it has been journalled (it may have already been) or {@link IAeProcessManager#NULL_JOURNAL_ID}
    */
   protected void dispatchReceiveData(AeMessageReceiver aMessageReceiver, AeInboundReceive aInboundReceive, 
         boolean aReplyWaitingFlag, boolean aRunSynchronously, long aJournalId) throws AeBusinessProcessException
   {
      long processId = aMessageReceiver.getProcessId();
      int locationPathId = aMessageReceiver.getMessageReceiverPathId();
      dispatchReceiveData(processId, locationPathId, aInboundReceive, aReplyWaitingFlag, aJournalId);
   }

   /**
    * Dispatches the message data to the process's message receiver.
    * 
    * @param aProcessId
    * @param aLocationPathId
    * @param aInboundReceive
    * @param aReplyWaitingFlag
    * @param aJournalId journal id of the message if it has been journalled (it may have already been) or {@link IAeProcessManager#NULL_JOURNAL_ID}
    */
   protected void dispatchReceiveData(long aProcessId, int aLocationPathId, AeInboundReceive aInboundReceive, 
         boolean aReplyWaitingFlag, long aJournalId) throws AeBusinessProcessException
   {
      long journalId = aJournalId;
      if (journalId == IAeProcessManager.NULL_JOURNAL_ID)
         journalId = getProcessManager().journalInboundReceive(aProcessId, aLocationPathId, aInboundReceive, aReplyWaitingFlag);
      IAeBusinessProcess process = getProcessById(aProcessId);

      try
      {
         getPartnerLinkStrategy().updatePartnerLink(process, aInboundReceive);

         setProcessProperties( process, aInboundReceive.getContext().getBusinessProcessProperties() );
         process.dispatchReceiveData(aLocationPathId, aInboundReceive.getMessageData());
         journalEntryDone(process, journalId);
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * Creates the correlation map for the incoming receive.
    * @param aDesc
    * @param aData
    * @param aContext
    * @param aPortType
    */
   private Map createCorrelationMap(IAeProcessPlan aDesc, IAeMessageData aData,
                                    IAeMessageContext aContext, QName aPortType) throws AeBusinessProcessException
   {
      Map map = new HashMap();
      for (Iterator iter = aDesc.getCorrelatedPropertyNames(aContext.getPartnerLinkName(), aContext.getOperation()).iterator(); iter.hasNext();)
      {
         QName propName = (QName) iter.next();
         IAePropertyAlias propAlias = aDesc.getProcessDef().getPropertyAlias(aData.getMessageType(), propName);
         IAeProperty prop = AeWSDLDefHelper.getProperty(aDesc, propName);
         if (prop == null)
            throw new AeBusinessProcessException(AeMessages.format("AeBusinessProcessEngine.ERROR_MISSING_PROPERTY", propName)); //$NON-NLS-1$
         
         map.put(propName, AeXPathHelper.evaluatePropertyToSimpleType(propAlias, aData, getTypeMapping(), prop.getTypeName()));
      }
      return map;
   }

   /** 
    * @return Returns the coordination manager.
    */
   public IAeCoordinationManagerInternal getCoordinationManager()
   {
      return mCoordinationManager;
   }
   
   /**
    * Sets the coordination manager.
    * @param aCoordinationManager coordination manager.
    */
   public void setCoordinationManager(IAeCoordinationManagerInternal aCoordinationManager)
   {
      mCoordinationManager = aCoordinationManager;
      if (mCoordinationManager != null)
      {
         mCoordinationManager.setEngine(this);
      }
   }
   
   /**
    * Registers the given context for coordination with the coordination manager.
    * @param aContext message context
    * @param aProcess process
    * @throws AeBusinessProcessException
    */
   public void registerForCoordination(IAeMessageContext aContext, IAeBusinessProcess aProcess) throws AeBusinessProcessException
   {
      // ignore
   }
   
   /**
    * Returns true if message context indicates that request should be coordinated.
    * @param aContext
    * @return returns true if the message context's business properties has a coordination context
    */
   public boolean isCoordinated(IAeMessageContext aContext)
   {
      return false;
   }

   /**
    * Creates the process with the data received.
    *
    * @param aInboundReceive The inbound receive object.
    * @param aReply The message receiver reply object.
    */
   protected void createProcessWithMessage(AeInboundReceive aInboundReceive, IAeReplyReceiver aReply)
      throws AeBusinessProcessException
   {
      // create the business process
      IAeBusinessProcess process = getProcessManager().createBusinessProcess( aInboundReceive.getProcessPlan() );
      // TODO (PJ): access coord context once and pass it as a param on method calls.
      // coordination!
      if ( isCoordinated( aInboundReceive.getContext() ) )
      {
         try
         {
            // register coordination
            registerForCoordination(aInboundReceive.getContext() , process);
         } catch (Exception e)
         {
            AeException.logError(e, e.getMessage());
            process.terminate();
            AeBusinessProcessException bpe = new AeBusinessProcessException();
            bpe.setRootCause(e);
            throw bpe;
         }
      }// if
      
      long journalId = IAeProcessManager.NULL_JOURNAL_ID;
      try
      {
         long processId = process.getProcessId();
         
         // TODO (KR) Combine this with IAeProcessManager#createBusinessProcess.
         journalId = getProcessManager().journalInboundReceive(processId, 0, aInboundReceive, aReply != null);

         // Now it's okay to announce that the process has been created.
         fireEngineEvent(new AeEngineEvent(processId, IAeEngineEvent.PROCESS_CREATED, aInboundReceive.getProcessName()));

         if (aReply != null)
         {
            String messageExchange = aInboundReceive.getProcessPlan().getProcessDef().getCreateInstanceMessageExchange(aInboundReceive.getPartnerLinkName(), aInboundReceive.getOperation());
            getQueueManager().addReply(new AeReply(processId, aReply, aInboundReceive, messageExchange), null);
         }

         // set any process instance properties attached to the message context
         // on the process instance
         setProcessProperties( process, aInboundReceive.getContext().getBusinessProcessProperties() );

         // set the create message on the process, this will get picked up by
         // the createInstance activity once the process executes
         process.setCreateMessage(aInboundReceive);

         // initialize the partner links from the deployment object
         getPartnerLinkStrategy().initPartnerLinks(process, aInboundReceive.getProcessPlan());

         getPartnerLinkStrategy().updatePartnerLink(process, aInboundReceive);

         // process needs reference to the WSDLProvider and message property provider
         process.setContextWSDLProvider(aInboundReceive.getProcessPlan());

         executeProcess(processId, aReply != null);
         journalEntryDone(process, journalId);
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * Set any process properties on the <code>IAeBusinessProcess</code> instance.
    * @param aProcess
    * @param aProperties
    */
   protected void setProcessProperties( IAeBusinessProcess aProcess, Map aProperties )
   {
      if( aProperties != null )
      {
         for( Iterator iter = aProperties.entrySet().iterator(); iter.hasNext(); )
         {
            Map.Entry propertyEntry = (Map.Entry)iter.next();
            aProcess.setBusinessProcessProperty( propertyEntry.getKey().toString(), (String)propertyEntry.getValue() );
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueInvokeData(long, java.lang.String, org.activebpel.rt.message.IAeMessageData, java.util.Map)
    */
   public void queueInvokeData(long aProcessId, String aLocationPath,
         IAeMessageData aMessageData, Map aProperties ) throws AeBusinessProcessException
   {
      AeInvoke invokeQueueObject = new AeInvoke(aProcessId, aLocationPath, aMessageData);

      // Is entry already waiting
      AeInvoke found = getQueueManager().removeInvoke(invokeQueueObject);

      if(found != null && found.getMessageReceiverPath() != null)
      {
         IAeBusinessProcess process = getProcessById(aProcessId);
         long journalId = IAeProcessManager.NULL_JOURNAL_ID;

         try
         {
            int locationId = process.getLocationId(aLocationPath);
            journalId = getProcessManager().journalInvokeData(aProcessId, locationId, aMessageData, aProperties);

            setProcessProperties( process, aProperties );
            process.dispatchInvokeData(aLocationPath, aMessageData);
            journalEntryDone(process, journalId);
         }
         finally
         {
            releaseProcess(process);
         }
      }
      else
      {
         unmatchedInvoke(invokeQueueObject);
      }
   }

   /**
    * Called when we got some invoke data that we weren't expecting. This should
    * never happen.
    * @param aInvokeQueueObject
    * @throws AeBusinessProcessException
    */
   protected void unmatchedInvoke(AeInvoke aInvokeQueueObject)
      throws AeBusinessProcessException
   {
      throw new AeBusinessProcessException(AeMessages.getString("AeBusinessProcessEngine.ERROR_2")); //$NON-NLS-1$
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#queueInvokeFault(long, java.lang.String, org.activebpel.rt.bpel.IAeFault, java.util.Map)
    */
   public void queueInvokeFault(long aProcessId, String aLocationPath, IAeFault aFault, Map aProperties) throws AeBusinessProcessException
   {
      AeInvoke invokeQueueObject = new AeInvoke(aProcessId, aLocationPath, aFault);

      // Is entry already waiting
      AeInvoke found = getQueueManager().removeInvoke(invokeQueueObject);

      if(found != null && found.getMessageReceiverPath() != null)
      {
         IAeBusinessProcess process = getProcessById(aProcessId);
         long journalId = IAeProcessManager.NULL_JOURNAL_ID;

         try
         {
            int locationId = process.getLocationId(aLocationPath);
            journalId = getProcessManager().journalInvokeFault(aProcessId, locationId, aFault, aProperties);

            setProcessProperties( process, aProperties );
            process.dispatchInvokeFault(aLocationPath, aFault);
            journalEntryDone(process, journalId);
         }
         finally
         {
            releaseProcess(process);
         }
      }
      else
      {
         unmatchedInvoke(invokeQueueObject);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#getQueueManager()
    */
   public IAeQueueManager getQueueManager()
   {
      return mQueueManager;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#getLockManager()
    */
   public IAeLockManager getLockManager()
   {
      return mLockManager;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#setPlanManager(org.activebpel.rt.bpel.IAePlanManager)
    */
   public void setPlanManager(IAePlanManager aPlanManager)
   {
      mPlanManager = aPlanManager;
   }

   /**
    * Getter for the plan manager.
    */
   protected IAePlanManager getPlanManager()
   {
      return mPlanManager;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#getEngineConfiguration()
    */
   public IAeEngineConfiguration getEngineConfiguration()
   {
      return mEngineConfiguration;
   }

   /**
    * Sets the configruation to use for this engine, note this is usually set during construction.
    * @param aConfiguration The configuration to use for this engine.
    */
   public void setEngineConfiguration(IAeEngineConfiguration aConfiguration)
   {
      mEngineConfiguration = aConfiguration;
   }

   /** 
    * Overrides method to return self. 
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#getEngineCallback()
    */
   public IAeBusinessProcessEngineCallback getEngineCallback()
   {
      return this;
   }

   /**
    * Returns handler for responsible for process coordination operations.
    */
   public IAeProcessCoordination getProcessCoordination()
   {
      if (mProcessCoordination == null)
      {
         // Create an empty (stub/void) process coordination handler.
         // Note: AeBpelEngine overrides this method to return a different implementation).
         mProcessCoordination = new AeProcessCoordinationStub();
      }
      return mProcessCoordination;
   }
   
   /**
    * Callback which signals that the process has ended. The engine will remove the process from the ProcessManager.
    * @param aProcess process
    */
   public void processEnded(AeBusinessProcess aProcess) throws AeBusinessProcessException
   {
      try
      {
         removeProcess(aProcess);
      }
      catch(Throwable t)
      {
         AeBusinessProcessException bpe = new AeBusinessProcessException(t.getMessage(), t);
         throw bpe;
      }      
   }
   
   /**
    * Removes process from ProcessManager.
    * @param aProcess process
    */
   protected void removeProcess(AeBusinessProcess aProcess) throws Exception
   {
      if (aProcess.getState() == AeBpelState.FAULTED)
      {
         IAeFault fault = aProcess.getFault();
         // if the process has faulted, then notify all of the existing replies
         AeOutstandingRepliesVisitor outstandingRepliesVisitor = new AeOutstandingRepliesVisitor();
         aProcess.accept(outstandingRepliesVisitor);
         if (outstandingRepliesVisitor.containsOutstandingReplies())
         {
            Map processProperties = aProcess.cloneBusinessProcessProperties();
            getQueueManager().faultOutstandingReplies(aProcess.getProcessId(), fault, processProperties );
         }
      }
      getProcessManager().processEnded(aProcess.getProcessId());      
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#getStartDate()
    */
   public Date getStartDate()
   {
      return mStarted;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#getProcessState(long)
    */
   public Document getProcessState(long aProcessId) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aProcessId);
      // TODO (KR) Add a call to process manager to boost release lag for this process
      try
      {
         return process.serializeState(false);
      }
      finally
      {
         releaseProcess(process);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#getProcessVariable(long, java.lang.String)
    */
   public Document getProcessVariable(long aProcessId, String aLocationPath) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = getProcessById(aProcessId);
      try
      {
         return process.serializeVariable( aLocationPath );
      }
      finally
      {
         releaseProcess(process);
      }
   }
   
   /* 
    * Overrides method to the location path given the process id and the location id. 
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#getLocationPathById(long, int)
    */
   public String getLocationPathById(long aProcessId, int aLocationId) throws AeBusinessProcessException
   {
      IAeBusinessProcess process = null;
      try
      {
         process = getProcessById(aProcessId);
         return process.getLocationPath(aLocationId);
      }
      finally
      {
         if (process != null)
         {
            releaseProcess(process);
         }
      }      
   }   

   /**
    * Dispatches alarm to a process by the location path of the alarm receiver.
    *
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#dispatchAlarm(long, int, int)
    */
   public void dispatchAlarm(long aProcessId, int aLocationPathId, int aGroupId) throws AeBusinessProcessException
   {
      // Only dispatch the alarm if it could be successfully removed from the
      // queue manager. If the remove fails, it means the alarm was already
      // cancelled by some other activity.
      long journalId = getQueueManager().removeAlarmForDispatch(aProcessId, aGroupId, aLocationPathId);
      if (journalId != IAeProcessManager.NULL_JOURNAL_ID)
      {
         AeBusinessProcess process = (AeBusinessProcess) getProcessById(aProcessId);

         try
         {
            process.dispatchAlarm(aLocationPathId);
            journalEntryDone(process, journalId);
         }
         finally
         {
            releaseProcess(process);
         }
      }
   }

   /**
    * Returns the process instance details for the process with the specified process id.
    *
    * @param aProcessId
    * @return AeProcessInstanceDetail
    */
   public AeProcessInstanceDetail getProcessInstanceDetails(long aProcessId)
   {
      return getProcessManager().getProcessInstanceDetails(aProcessId);
   }

   /////////////////////////////////////////////////////////////////////////////////
   // Support for adding, removing and firing ENGINE listeners
   /////////////////////////////////////////////////////////////////////////////////

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#addEngineListener(org.activebpel.rt.bpel.IAeEngineListener)
    */
   public void addEngineListener(IAeEngineListener aListener)
   {
      mEngineListeners.add(aListener);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#removeEngineListener(org.activebpel.rt.bpel.IAeEngineListener)
    */
   public void removeEngineListener(IAeEngineListener aListener)
   {
      mEngineListeners.remove(aListener);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#fireEngineEvent(org.activebpel.rt.bpel.IAeEngineEvent)
    */
   public void fireEngineEvent(IAeEngineEvent aEvent)
   {
      boolean suspend = false;
      for (Iterator iter=mEngineListeners.iterator(); iter.hasNext();)
         suspend |= ((IAeEngineListener)iter.next()).handleEngineEvent(aEvent);

      // if an engine event listener suggests we should supend the process then attempt to suspend it
      if(suspend && aEvent.getPID() >= 0)
      {
         try
         {
            suspendProcess(aEvent.getPID());
         }
         catch (AeBusinessProcessException ex)
         {
            ex.logError();
         }
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#fireEngineAlert(org.activebpel.rt.bpel.IAeEngineAlert)
    */
   public void fireEngineAlert(IAeEngineAlert aAlert)
   {
      for (Iterator iter=mEngineListeners.iterator(); iter.hasNext();)
         ((IAeEngineListener)iter.next()).handleAlert(aAlert);
   }
   
   /////////////////////////////////////////////////////////////////////////////////
   // Support for adding, removing and firing PROCESS listeners
   /////////////////////////////////////////////////////////////////////////////////

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#addProcessListener(org.activebpel.rt.bpel.IAeProcessListener)
    */
   public void addProcessListener(IAeProcessListener aListener)
   {
      mGlobalProcessListeners.add(aListener);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#removeProcessListener(org.activebpel.rt.bpel.IAeProcessListener)
    */
   public void removeProcessListener(IAeProcessListener aListener)
   {
      mGlobalProcessListeners.remove(aListener);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#addProcessListener(org.activebpel.rt.bpel.IAeProcessListener, long)
    */
   public void addProcessListener(IAeProcessListener aListener, long aPid)
   {
      String key = Long.toString(aPid);

      synchronized (mProcessListeners)
      {
         Collection listeners = (Collection) mProcessListeners.get(key);
         if (listeners == null)
            mProcessListeners.put(key, (listeners = new AeSafelyViewableCollection(new LinkedHashSet())));

         listeners.add(aListener);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeBusinessProcessEngine#removeProcessListener(org.activebpel.rt.bpel.IAeProcessListener, long)
    */
   public void removeProcessListener(IAeProcessListener aListener, long aPid)
   {
      String key = Long.toString(aPid);

      synchronized (mProcessListeners)
      {
         Collection listeners = (Collection) mProcessListeners.get(key);
         if (listeners != null)
         {
            listeners.remove(aListener);
            if (listeners.size() == 0)
               mProcessListeners.put(key, null);
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#fireInfoEvent(org.activebpel.rt.bpel.IAeProcessInfoEvent)
    */
   public void fireInfoEvent(IAeProcessInfoEvent aInfoEvent)
   {
      Collection listeners;
      String key = Long.toString(aInfoEvent.getPID());

      synchronized(mProcessListeners)
      {
         // Get listeners for the specified process
         listeners = (Collection) mProcessListeners.get(key);
      }

      // If we have listeners for this process, fire the notifications
      if (listeners != null)
      {
         for (Iterator iter=listeners.iterator(); iter.hasNext();)
            ((IAeProcessListener)iter.next()).handleProcessInfoEvent(aInfoEvent);
      }

      for (Iterator iter=mGlobalProcessListeners.iterator(); iter.hasNext();)
         ((IAeProcessListener)iter.next()).handleProcessInfoEvent(aInfoEvent);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#fireEvaluationEvent(long, java.lang.String, org.activebpel.rt.bpel.impl.AeAbstractBpelObject, int, java.lang.String, java.lang.String)
    */
   public void fireEvaluationEvent(long aPID, String aExpression, AeAbstractBpelObject aBPELObject, int aEventID, String aNodePath, String aResult)
   {
      // TODO (MF) we're not doing anything with the expression but we should be.
      // The designer code doesn't need it since it'll use the location path to
      // pull the expression from the def/model layer. While I could do the same
      // thing on the server side for process logging, I don't want to since it
      // adds unnecessary overhead during process logging. One solution would be
      // to put the expression on the AeProcessInfoEvent - all of the callers of
      // this method are already passing the expression in anyway.
      fireInfoEvent(new AeProcessInfoEvent(aPID, aNodePath, aEventID, "", aResult)); //$NON-NLS-1$
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#fireEvent(org.activebpel.rt.bpel.IAeProcessEvent)
    */
   public void fireEvent(IAeProcessEvent aEvent)
   {
      Collection listeners;
      String key = Long.toString(aEvent.getPID());
      boolean suspend = false;

      synchronized(mProcessListeners)
      {
         // Get listeners for the specified process
         listeners = (Collection) mProcessListeners.get(key);
      }

      // If we have listeners for this process, fire the notifications
      if (listeners != null)
      {
         for (Iterator iter=listeners.iterator(); iter.hasNext();)
            suspend |= ((IAeProcessListener)iter.next()).handleProcessEvent(aEvent);
      }

      for (Iterator iter=mGlobalProcessListeners.iterator(); iter.hasNext();)
         suspend |= ((IAeProcessListener)iter.next()).handleProcessEvent(aEvent);

      // if event handlers asked for suspend then suspend the process
      if(suspend)
      {
         try
         {
            suspendProcess(aEvent.getPID());
         }
         catch (AeBusinessProcessException e)
         {
            // attempt to suspend fails log error
            e.logError();
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessFactory#createProcess(long, org.activebpel.rt.bpel.impl.IAeProcessPlan)
    */
   public IAeBusinessProcess createProcess(long aPid, IAeProcessPlan aPlan)
   {
      AeBusinessProcess process = new AeBusinessProcess(aPid, this, aPlan);
      return process;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#start()
    */
   public void start() throws AeBusinessProcessException
   {
      mStarted = new Date();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#stop()
    */
   public void stop() throws AeBusinessProcessException
   {
      mStarted = null;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#shutDown()
    */
   public void shutDown() throws AeBusinessProcessException
   {
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#getTypeMapping()
    */
   public AeTypeMapping getTypeMapping()
   {
      return mTypeMapping;
   }

   /**
    * Returns the strategy for managing partner links.
    */
   public IAeEnginePartnerLinkStrategy getPartnerLinkStrategy()
   {
      if (mPartnerLinkStrategy == null)
      {
         // If the strategy is not defined yet, then construct the default
         // strategy.
         mPartnerLinkStrategy = new AeDefaultPartnerLinkStrategy();
      }

      return mPartnerLinkStrategy;
   }

   /**
    * Sets the strategy for managing partner link.
    */
   protected void setPartnerLinkStrategy(IAeEnginePartnerLinkStrategy aPartnerLinkStrategy)
   {
      mPartnerLinkStrategy = aPartnerLinkStrategy;
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#scheduleAlarm(long, int, int, java.util.Date)
    */
   public void scheduleAlarm(long aProcessId, int aLocationPathId, int aGroupId, Date aDeadline) throws AeBusinessProcessException
   {
      getQueueManager().scheduleAlarm(aProcessId, aLocationPathId, aGroupId, aDeadline);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#removeAlarm(long, int)
    */
   public boolean removeAlarm(long aProcessId, int aLocationPathId) throws AeBusinessProcessException
   {
      return getQueueManager().removeAlarm(aProcessId, aLocationPathId);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#addInvoke(org.activebpel.rt.bpel.impl.IAeProcessPlan, org.activebpel.wsio.invoke.IAeInvoke)
    */
   public void addInvoke(IAeProcessPlan aPlan, IAeInvoke aInvokeQueueObject) throws AeBusinessProcessException
   {
      getQueueManager().addInvoke(aPlan, aInvokeQueueObject);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#addMessageReceiver(org.activebpel.rt.bpel.impl.queue.AeMessageReceiver)
    */
   public void addMessageReceiver(AeMessageReceiver aMessageReceiver) throws AeBusinessProcessException
   {
      getQueueManager().addMessageReceiver(aMessageReceiver);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#removeMessageReceiver(long, int)
    */
   public boolean removeMessageReceiver(long aProcessId, int aMessageReceiverPathId) throws AeBusinessProcessException
   {
      return getQueueManager().removeMessageReceiver(aProcessId, aMessageReceiverPathId);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBusinessProcessEngineInternal#sendReply(long, org.activebpel.rt.bpel.impl.queue.AeReply, org.activebpel.rt.message.IAeMessageData, javax.xml.namespace.QName, java.util.Map)
    */
   public boolean sendReply(long aProcessId, AeReply aReplyObject, IAeMessageData aData, QName aFaultName, Map aProcessProperties) throws AeBusinessProcessException
   {
      boolean sentReply = getQueueManager().sendReply(aReplyObject, aData, aFaultName, aProcessProperties);

      if (sentReply)
      {
         getProcessManager().journalSentReply(aProcessId, aReplyObject, aProcessProperties);
      }

      return sentReply;
   }

   /**
    * Implements a default "do nothing" strategy for managing partner links.
    */
   protected static class AeDefaultPartnerLinkStrategy implements IAeEnginePartnerLinkStrategy
   {
      /**
       * @see org.activebpel.rt.bpel.impl.IAeEnginePartnerLinkStrategy#initPartnerLinks(org.activebpel.rt.bpel.IAeBusinessProcess, org.activebpel.rt.bpel.impl.IAeProcessPlan)
       */
      public void initPartnerLinks(IAeBusinessProcess aProcess, IAeProcessPlan aDesc) throws AeBusinessProcessException
      {
      }

      /**
       * @see org.activebpel.rt.bpel.impl.IAeEnginePartnerLinkStrategy#updatePartnerLink(org.activebpel.rt.bpel.IAeBusinessProcess, org.activebpel.rt.bpel.impl.queue.AeInboundReceive)
       */
      public void updatePartnerLink(IAeBusinessProcess aProcess, AeInboundReceive aInboundReceive) throws AeBusinessProcessException
      {
      }
   }
}
