// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeFaultFactory.java,v 1.14 2006/02/10 17:02:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.def.IAeBPELConstants;
import org.activebpel.rt.bpel.impl.activity.support.AeFault;
import org.activebpel.rt.util.AeUtil;

/**
 * Factory used to provide the common BPEL faults.
 */
public class AeFaultFactory
{
   //
   // Standard BPEL4WS Faults
   //

   /** Identifier for bpws:selectionFailure fault */
   public static final String SELECTION_FAILURE   = "selectionFailure"; //$NON-NLS-1$
   /** Identifier for bpws:conflictingReceive fault */
   public static final String CONFLICTING_RECEIVE = "conflictingReceive"; //$NON-NLS-1$
   /** Identifier for bpws:conflictingRequest fault */
   public static final String CONFLICTING_REQUEST = "conflictingRequest"; //$NON-NLS-1$
   /** Identifier for bpws:mismatchedAssignmentFailure fault */
   public static final String MISMATCHED_ASSIGNMENT_FAILURE = "mismatchedAssignmentFailure"; //$NON-NLS-1$
   /** Identifier for bpws:joinFailure fault */
   public static final String JOIN_FAILURE = "joinFailure"; //$NON-NLS-1$
   /** Identifier for bpws:forcedTermination fault */
   public static final String FORCED_TERMINATION = "forcedTermination"; //$NON-NLS-1$
   /** Identifier for bpws:correlationViolation fault */
   public static final String CORRELATION_VIOLATION = "correlationViolation"; //$NON-NLS-1$
   /** Identifier for bpws:uninitializedVariable fault */
   public static final String UNINITIALIZED_VARIABLE = "uninitializedVariable"; //$NON-NLS-1$
   /** Identifier for bpws:repeatedCompensation fault */
   public static final String REPEATED_COMPENSATION = "repeatedCompensation"; //$NON-NLS-1$
   /** Identifier for bpws:invalidReply fault */
   public static final String INVALID_REPLY = "invalidReply"; //$NON-NLS-1$
   /** Identifier for bpws:missingReply fault as per Issue 50 */
   public static final String MISSING_REPLY = "missingReply"; //$NON-NLS-1$
   /** Identifier for bpws:forEachCounterError */
   public static final String FOREACH_COUNTER_ERROR = "forEachCounterError"; //$NON-NLS-1$
   /** Identifier for bpws:invalidBranchCondition */
   public static final String FOREACH_INVALID_BRANCH_CONDITION = "invalidBranchCondition"; //$NON-NLS-1$
   /** Identifier for bpws:completionConditionFailure */
   public static final String FOREACH_COMPLETION_CONDITION_FAILURE = "completionConditionFailure"; //$NON-NLS-1$
   
   /** array of std bpel faults */
   public static final String[] STANDARD_BPEL_FAULTS =
   {
      SELECTION_FAILURE,
      CONFLICTING_RECEIVE,
      CONFLICTING_REQUEST,
      MISMATCHED_ASSIGNMENT_FAILURE,
      JOIN_FAILURE,
      FORCED_TERMINATION,
      CORRELATION_VIOLATION,
      UNINITIALIZED_VARIABLE,
      REPEATED_COMPENSATION,
      INVALID_REPLY,
      MISSING_REPLY,
      FOREACH_COUNTER_ERROR,
      FOREACH_INVALID_BRANCH_CONDITION,
      FOREACH_COMPLETION_CONDITION_FAILURE
   };
   
   //
   // Internal Faults
   //

   /** Identifier for internal ae:systemError fault: used for unrecoverable system errors */
   public static final String SYSTEM_ERROR = "systemError"; //$NON-NLS-1$
   /** Identifier for internal ae:bad process fault: used when we encounter invalid BPEL (static analysis should catch all of these issues */
   public static final String BAD_PROCESS = "badProcess"; //$NON-NLS-1$
   /** Identifier for internal ae:validationError: used to report validation errors in variable data */
   public static final String VALIDATION_ERROR = "validationError"; //$NON-NLS-1$
   /** Identifier for internal ae:xpathFunctionError: used to report error in executing xpath function */
   public static final String XPATH_FUNCTION_ERROR = "xpathFunctionError"; //$NON-NLS-1$
   /** Identifier for internal ae:invalidTransitionCondition: used to report a non boolean return from an xpath evaluation of a transition condition*/
   public static final String INVALID_TRANS_CONDITION = "invalidTransitionCondition"; //$NON-NLS-1$
   /** Identifier for internal ae:xpathDateParseError: used to report error in parsing an xsd:date or xsd:datetime */
   public static final String XPATH_DATE_PARSE_ERROR = "xpathDateParseError"; //$NON-NLS-1$
   /** Identifier for internal ae:xpathDurationFormatError: used to report error in parsing an xsd:duration */
   public static final String XPATH_DURATION_FORMAT_ERROR = "xpathDurationFormatError"; //$NON-NLS-1$
   /** Identifier for internal ae:retryFault: used to trigger a scope's compensation handlers if it's being retried */
   public static final String RETRY_FAULT = "retryFault"; //$NON-NLS-1$
   
   
   /** provides for lookup of faults by name */
   private static Map mFaultMap = new HashMap();
   
   /**
    * Private constructor to prevent instantiation. 
    */
   private AeFaultFactory()
   {
   }
   
   /**
    * Helper method to make a new fault in the bpel namespace
    * @param aName
    */
   private static IAeFault makeBpelFault(String aName)
   {
      return new AeFault(new QName(IAeBPELConstants.BPWS_NAMESPACE_URI, aName), null); 
   }

   /**
    * Makes a fault in the system namespace
    * @param aName
    */
   private static IAeFault makeSystemFault(String aName)
   {
      return new AeFault(new QName(IAeBPELConstants.AE_NAMESPACE_URI, aName), null); 
   }

   static 
   { 
      mFaultMap.put(SELECTION_FAILURE, makeBpelFault(SELECTION_FAILURE));
      mFaultMap.put(CONFLICTING_RECEIVE, makeBpelFault(CONFLICTING_RECEIVE));
      mFaultMap.put(CONFLICTING_REQUEST, makeBpelFault(CONFLICTING_REQUEST));
      mFaultMap.put(MISMATCHED_ASSIGNMENT_FAILURE, makeBpelFault(MISMATCHED_ASSIGNMENT_FAILURE));
      mFaultMap.put(JOIN_FAILURE, makeBpelFault(JOIN_FAILURE));
      mFaultMap.put(FORCED_TERMINATION, makeBpelFault(FORCED_TERMINATION));
      mFaultMap.put(CORRELATION_VIOLATION, makeBpelFault(CORRELATION_VIOLATION));
      mFaultMap.put(UNINITIALIZED_VARIABLE, makeBpelFault(UNINITIALIZED_VARIABLE));
      mFaultMap.put(REPEATED_COMPENSATION, makeBpelFault(REPEATED_COMPENSATION));
      mFaultMap.put(INVALID_REPLY, makeBpelFault(INVALID_REPLY));
      mFaultMap.put(FOREACH_COUNTER_ERROR, makeBpelFault(FOREACH_COUNTER_ERROR));
      mFaultMap.put(FOREACH_INVALID_BRANCH_CONDITION, makeBpelFault(FOREACH_INVALID_BRANCH_CONDITION));
      mFaultMap.put(FOREACH_COMPLETION_CONDITION_FAILURE, makeBpelFault(FOREACH_COMPLETION_CONDITION_FAILURE));
      mFaultMap.put(SYSTEM_ERROR, makeSystemFault(SYSTEM_ERROR));
      mFaultMap.put(BAD_PROCESS, makeSystemFault(BAD_PROCESS));
      mFaultMap.put(MISSING_REPLY, makeBpelFault(MISSING_REPLY));
      mFaultMap.put(VALIDATION_ERROR, makeSystemFault(VALIDATION_ERROR));
      mFaultMap.put(XPATH_FUNCTION_ERROR, makeSystemFault(XPATH_FUNCTION_ERROR));
      mFaultMap.put(INVALID_TRANS_CONDITION, makeSystemFault(INVALID_TRANS_CONDITION));
      mFaultMap.put(XPATH_DATE_PARSE_ERROR, makeSystemFault(XPATH_DATE_PARSE_ERROR));
      mFaultMap.put(XPATH_DURATION_FORMAT_ERROR, makeSystemFault(XPATH_DURATION_FORMAT_ERROR));
      mFaultMap.put(RETRY_FAULT, makeSystemFault(RETRY_FAULT));
   }
 
   /**
    * Given the name of a BPEL fault, this method will return a cloned <code>AeFault</code>
    * object for it. Null will be returned if an invalid name is given. See public
    * constants for valid names. 
    * @param aFaultName the name of the fault we are interested in
    */
   public static IAeFault getFault(String aFaultName)
   {
      AeFault sFault = (AeFault)mFaultMap.get(aFaultName);
      AeFault dFault = null ;
      if ( sFault != null )
         dFault = new AeFault( sFault.getFaultName(), null );
      
      return dFault ;
   }

   /**
    * Creates a system error fault from an exception and a message.
    * 
    * @param aThrowable
    * @param aMessage
    */
   public static IAeFault getSystemErrorFault(Throwable aThrowable, String aMessage)
   {
      IAeFault fault = AeFaultFactory.getFault(AeFaultFactory.SYSTEM_ERROR);
      fault.setInfo(aMessage);
      fault.setDetailedInfo(AeUtil.getStacktrace(aThrowable));
      return fault;
   }
   
   /**
    * Creates a new system error fault from an exception.
    * 
    * @param aThrowable
    */
   public static IAeFault getSystemErrorFault(Throwable aThrowable)
   {
      return getSystemErrorFault(aThrowable, aThrowable.getLocalizedMessage());
   }
}
