// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeInMemoryProcessManager.java,v 1.29 2006/01/19 16:16:5
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeBusinessProcess;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.IAePlanManager;
import org.activebpel.rt.bpel.def.visitors.AeDefToImplVisitor;
import org.activebpel.rt.bpel.impl.list.AeProcessFilter;
import org.activebpel.rt.bpel.impl.list.AeProcessFilterAdapter;
import org.activebpel.rt.bpel.impl.list.AeProcessInstanceDetail;
import org.activebpel.rt.bpel.impl.list.AeProcessListResult;
import org.activebpel.rt.bpel.impl.queue.AeInboundReceive;
import org.activebpel.rt.bpel.impl.queue.AeReply;
import org.activebpel.rt.message.IAeMessageData;

/**
 * Implements a simple in-memory process manager.
 */
public class AeInMemoryProcessManager extends AeManagerAdapter implements IAeProcessManager
{
   /** Count of completed processes to be left in <code>mProcesses</code>. */
   private static final int COMPLETED_PROCESS_CACHE = 25;

   /** Next available process id */
   private long mNextProcessId = 1;

   /** Maps process ids to processes */
   private Hashtable mProcesses = new Hashtable();

   /** Process ids for processes temporarily left in <code>mProcesses</code> */
   private List mCompletedProcessIds = new LinkedList();

   /** The next journal ID to use for journaling methods. */
   private long mNextJournalId = 1;
   
   /**
    * No-arg ctor
    */
   public AeInMemoryProcessManager()
   {
      this(Collections.EMPTY_MAP);
   }

   /**
    * Constructs an in-memory process manager.
    *
    * @param aConfig The configuration map for this manager.
    */
   public AeInMemoryProcessManager(Map aConfig)
   {
      // Map is ignored because the in-memory version does not require config.
   }

   /**
    * Gets the next journal id.
    */
   public synchronized long getNextJournalId()
   {
      return mNextJournalId++;
   }
      
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#createBusinessProcess(org.activebpel.rt.bpel.impl.IAeProcessPlan)
    */
   public IAeBusinessProcess createBusinessProcess(IAeProcessPlan aPlan) throws AeBusinessProcessException
   {
      long pid = getNextProcessId();
      IAeBusinessProcess process = AeDefToImplVisitor.createProcess(pid, getEngine(), aPlan);
      putProcess(pid, process);
      return process;
   }

   /**
    * Returns a key compatible with a Java <code>Map</code> for the specified
    * process id.
    */
   protected Object getKey(long aProcessId)
   {
      return new Long(aProcessId);
   }

   /**
    * Returns next available process id.
    */
   protected long getNextProcessId()
   {
      synchronized (mProcesses)
      {
      	return mNextProcessId++;
   	}
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#getProcess(long)
    */
   public IAeBusinessProcess getProcess(long aProcessId)
   {
      return (IAeBusinessProcess) getProcessMap().get(getKey(aProcessId));
   }

   /**
    * Returns process map.
    */
   protected Map getProcessMap()
   {
      return mProcesses;
   }

   /**
    * Creates a process instance detail for the given process
    *
    * @param aProcess
    */
   private AeProcessInstanceDetail createProcessInstanceDetail(IAeBusinessProcess aProcess)
   {
      AeProcessInstanceDetail detail = new AeProcessInstanceDetail();
      detail.setName(aProcess.getName());
      detail.setProcessId(aProcess.getProcessId());
      detail.setState(aProcess.getProcessState());
      detail.setStateReason(aProcess.getProcessStateReason());
      detail.setStarted(aProcess.getStartDate());
      detail.setEnded(aProcess.getEndDate());
      return detail;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#getProcessInstanceDetails(long)
    */
   public AeProcessInstanceDetail getProcessInstanceDetails(long aProcessId)
   {
      IAeBusinessProcess process = getProcess(aProcessId);
      return (process == null) ? null : createProcessInstanceDetail(process);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#getProcesses(org.activebpel.rt.bpel.impl.list.AeProcessFilter)
    */
   public AeProcessListResult getProcesses(AeProcessFilter aFilter)
   {
      ArrayList results = new ArrayList();
      int totalCount = aFilter == null ? 0 : aFilter.getListStart();
      int matches = 0;

      synchronized(getProcessMap())
      {
         AeProcessFilterAdapter filter = new AeProcessFilterAdapter( aFilter );

         SortedMap map = new TreeMap(new AeReverseComparator());
         map.putAll(getProcessMap());

         for (Iterator iter=map.keySet().iterator(); iter.hasNext();)
         {
            IAeBusinessProcess process = (IAeBusinessProcess)map.get(iter.next());
            if( filter.accept(process) )
            {
               totalCount++;
               if( aFilter.isWithinRange(totalCount) )
               {
                  results.add(createProcessInstanceDetail(process));
               }
               matches++;
            }
         }
      }

      // Convert the list into an array as expected
      AeProcessInstanceDetail[] details = new AeProcessInstanceDetail[results.size()];
      results.toArray(details);
      return new AeProcessListResult(totalCount, details);
   }

   /**
    * Puts process into process map.
    *
    * @param aProcessId
    * @param aProcess
    */
   protected void putProcess(long aProcessId, IAeBusinessProcess aProcess)
   {
      getProcessMap().put(getKey(aProcessId), aProcess);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#processEnded(long)
    */
   public void processEnded(long aProcessId)
   {
      removeProcess(aProcessId);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#releaseProcess(org.activebpel.rt.bpel.IAeBusinessProcess)
    */
   public void releaseProcess(IAeBusinessProcess aProcess)
   {
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#removeProcess(long)
    */
   public void removeProcess(long aProcessId)
   {
      if (COMPLETED_PROCESS_CACHE <= 0)
      {
         // If we're not caching completed processes, then just remove the
         // process from the process map.
         getProcessMap().remove(getKey(aProcessId));
      }
      else
      {
         synchronized (mCompletedProcessIds)
         {
            // Otherwise, pare the list of cached processes to the cache size.
            while (mCompletedProcessIds.size() >= COMPLETED_PROCESS_CACHE)
               getProcessMap().remove(mCompletedProcessIds.remove(0));

            // And then add the specified process to the cache.
            mCompletedProcessIds.add(getKey(aProcessId));
         }
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#removeProcesses(org.activebpel.rt.bpel.impl.list.AeProcessFilter)
    */
   public int removeProcesses(AeProcessFilter aFilter)
   {
      synchronized (getProcessMap())
      {
         List processes = new ArrayList(getProcessMap().values());
         AeProcessFilterAdapter filter = new AeProcessFilterAdapter(aFilter);
         int removed = 0;

         for (Iterator i = processes.iterator(); i.hasNext(); )
         {
            IAeBusinessProcess process = (IAeBusinessProcess) i.next();

            if (filter.accept(process))
            {
               removeProcess(process.getProcessId());
               ++removed;
            }
         }

         return removed;
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#setPlanManager(org.activebpel.rt.bpel.IAePlanManager)
    */
   public void setPlanManager(IAePlanManager aPlanManager)
   {
      // Don't need plan manager for in-memory process manager.
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#getProcessQName(long)
    */
   public QName getProcessQName(long aProcessId)
   {
      QName processName = null;
      Object key = getKey( aProcessId );
      IAeBusinessProcess process = (IAeBusinessProcess)getProcessMap().get( key );
      if( process != null )
      {
         processName = process.getName();
      }
      return processName;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#journalEntryDone(long, long)
    */
   public void journalEntryDone(long aProcessId, long aJournalId)
   {
      // Nothing to do for in-memory process manager.
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#journalInvokeData(long, int, org.activebpel.rt.message.IAeMessageData, java.util.Map)
    */
   public long journalInvokeData(long aProcessId, int aLocationId, IAeMessageData aMessageData, Map aProcessProperties)
   {
      // Don't need to save received invoke data for in-memory process manager.
      return getNextJournalId();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#journalInvokeFault(long, int, org.activebpel.rt.bpel.IAeFault, java.util.Map)
    */
   public long journalInvokeFault(long aProcessId, int aLocationId, IAeFault aFault, Map aProcessProperties)
   {
      // Don't need to save received invoke fault for in-memory process manager.
      return getNextJournalId();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#journalInboundReceive(long, int, org.activebpel.rt.bpel.impl.queue.AeInboundReceive, boolean)
    */
   public long journalInboundReceive(long aProcessId, int aLocationId, AeInboundReceive aInboundReceive, boolean aReplyWaiting)
   {
      // Don't need to save received message for in-memory process manager.
      return getNextJournalId();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeProcessManager#journalSentReply(long, org.activebpel.rt.bpel.impl.queue.AeReply, java.util.Map)
    */
   public void journalSentReply(long aProcessId, AeReply aSentReply, Map aProcessProperties)
   {
      // Don't need to save sent reply for in-memory process manager.
   }

   /**
    * Implements a <code>Comparator</code> that reverses the natural order of
    * <code>Comparable</code> objects.
    */
   protected static class AeReverseComparator implements Comparator
   {
      /**
       * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
       */
      public int compare(Object o1, Object o2)
      {
         return -((Comparable) o1).compareTo(o2);
      }
   }
}
