// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeMessageDataDeserializer.java,v 1.9 2006/02/17 22:41:5
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessageDataFactory;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.def.AeVariableDef;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeUtil;
import org.activebpel.rt.util.AeXmlUtil;
import org.exolab.castor.xml.schema.XMLType;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Constructs a message data object from serialized message data.
 * @todo should pass type mapper even though this is just for transient fault messages
 */
public class AeMessageDataDeserializer implements IAeImplStateNames
{
   /** The XML <code>Element</code> containing the serialized message data. */
   private Element mMessageDataElement;

   /** The message type specifying the message data to produce. */
   private QName mMessageType;

   /** The resulting message data object. */
   private IAeMessageData mMessageData;

   /** Optional variable to use for part type definitions. */
   private IAeVariable mVariable;

   /** Optional type mapping to use to deserialize simple types. */
   private AeTypeMapping mTypeMapping;

   /**
    * Default constructor.
    */
   public AeMessageDataDeserializer()
   {
   }

   /**
    * Constructs a message data object from the specified serialized message
    * data and message type.
    */
   protected IAeMessageData createMessageData(Element aMessageDataElement, QName aMessageType) throws AeBusinessProcessException
   {
      List partElements = selectNodes(aMessageDataElement, "./" + STATE_PART); //$NON-NLS-1$
      IAeMessageData messageData = AeMessageDataFactory.instance().createMessageData(aMessageType);

      for (Iterator i = partElements.iterator(); i.hasNext(); )
      {
         Element partElement = (Element) i.next();
         String name = partElement.getAttribute(STATE_NAME);
         Object value;

         Element data = AeXmlUtil.getFirstSubElement(partElement);
         if (data != null)
         {
            value = createPartDocument(data);
         }
         else
         {
            value = AeXmlUtil.getText(partElement);
         }

         messageData.setData(name, value);
      }

      // If there is a variable to supply type definitions and a type mapping
      // to deserialize simple types, then fix parts that are simple types.
      if ((getVariable() != null) && (getTypeMapping() != null))
      {
         AeVariableDef variableDef = getVariable().getDefinition();
         AeTypeMapping typeMapping = getTypeMapping();

         for (Iterator i = messageData.getPartNames(); i.hasNext(); )
         {
            String name = (String) i.next();
            Object value = messageData.getData(name);

            if (value instanceof String && AeUtil.notNullOrEmpty((String)value))
            {
               XMLType type = variableDef.getPartType(name);
               if (type.getName() != null)
               {
                  QName typeName = new QName(type.getSchema().getSchemaNamespace(), type.getName());
                  value = typeMapping.deserialize(typeName, (String) value);
                  messageData.setData(name, value);
               }
            }
         }
      }

      return messageData;
   }

   /**
    * Returns a new part <code>Document</code> with the specified data.
    *
    * @param aPartData The root of the part data.
    */
   protected Document createPartDocument(Element aPartData)
   {
      Document document = AeXmlUtil.newDocument();
      Element root = (Element) document.importNode(aPartData, true);
      document.appendChild(root);
      return document;
   }

   /**
    * Returns a message data object constructed from the serialized message
    * data and message type specified with <code>setMessageDataElement</code>
    * and <code>setMessageType</code>.
    */
   public IAeMessageData getMessageData() throws AeBusinessProcessException
   {
      if (mMessageData == null)
      {
         if (mMessageDataElement == null)
         {
            throw new AeBusinessProcessException(AeMessages.getString("AeMessageDataDeserializer.ERROR_1")); //$NON-NLS-1$
         }

         mMessageData = createMessageData(mMessageDataElement, getMessageType());
      }

      return mMessageData;
   }

   /**
    * Returns optional type mapping to use to deserialize simple types.
    */
   protected AeTypeMapping getTypeMapping()
   {
      return mTypeMapping;
   }

   /**
    * Returns optional variable to use for part type definitions.
    */
   protected IAeVariable getVariable()
   {
      return mVariable;
   }

   /**
    * Returns the message type specifying the message data to produce.
    */
   protected QName getMessageType()
   {
      if (mMessageType == null)
      {
         String localPart = mMessageDataElement.getAttribute(STATE_NAME);
         String namespace = mMessageDataElement.getAttribute(STATE_NAMESPACEURI);

         mMessageType = new QName(namespace, localPart);
      }

      return mMessageType;
   }

   /**
    * Resets all output variables.
    */
   protected void reset()
   {
      mMessageData = null;
   }

   /**
    * Selects nodes by XPath.
    *
    * @param aNode The node to select from.
    * @param aPath The XPath expression.
    * @return List The list of matching nodes.
    * @throws AeBusinessProcessException
    */
   protected static List selectNodes(Node aNode, String aPath) throws AeBusinessProcessException
   {
      try
      {
         XPath xpath = new DOMXPath(aPath);
         return xpath.selectNodes(aNode);
      }
      catch (JaxenException e)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeMessageDataDeserializer.ERROR_2") + aPath, e); //$NON-NLS-1$
      }
   }

   /**
    * Sets the XML <code>Element</code> containing the serialized message data.
    *
    * @param aMessageDataElement
    */
   public void setMessageDataElement(Element aMessageDataElement)
   {
      reset();

      mMessageDataElement = aMessageDataElement;
   }

   /**
    * Sets the message type specifying the message data to produce.
    *
    * @param aMessageType
    */
   public void setMessageType(QName aMessageType)
   {
      reset();

      mMessageType = aMessageType;
   }
   
   /**
    * Sets optional type mapping to use to deserialize simple types.
    */
   public void setTypeMapping(AeTypeMapping aTypeMapping)
   {
      reset();

      mTypeMapping = aTypeMapping;
   }

   /**
    * Sets optional variable to use for part type definitions.
    */
   public void setVariable(IAeVariable aVariable)
   {
      reset();

      mVariable = aVariable;
   }
}
