// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeTypeMapping.java,v 1.12 2006/02/03 19:21:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;

import javax.xml.namespace.QName;

import org.activebpel.rt.IAeConstants;
import org.activebpel.rt.util.AeSchemaDate;
import org.activebpel.rt.util.AeSchemaDateTime;
import org.activebpel.rt.util.AeSchemaDuration;
import org.activebpel.rt.util.AeSchemaTime;
import org.exolab.castor.xml.schema.XMLType;

/**
 * Supports type mappings for standard xsd to java types.
 * 
 * Standard mappings from WSDL to Java
 *  xsd:base64Binary  byte[] 
 *  xsd:boolean  boolean 
 *  xsd:byte  byte 
 *  xsd:date   java.util.Date 
 *  xsd:dateTime  java.util.Calendar 
 *  xsd:time   java.util.Time 
 *  xsd:decimal  java.math.BigDecimal 
 *  xsd:double  double 
 *  xsd:float  float 
 *  xsd:hexBinary  byte[] 
 *  xsd:int int 
 *  xsd:integer java.math.BigInteger 
 *  xsd:long  long 
 *  xsd:QName  javax.xml.namespace.QName 
 *  xsd:short  short 
 *  xsd:string java.lang.String 
 *
 * TODO Should have an interface and a factory for this.
 * TODO Need to do binary types (byte arrays).
 */
public class AeTypeMapping
{
   /** Java to schema mappings. */
   protected HashMap mJava2SchemaMappings = new HashMap();
   /** Schema to java mappings. */
   protected HashMap mSchema2JavaMappings = new HashMap();
   /** The default mapper if we don't have a specific one. */
   protected IAeTypeMapper defaultMapper = new AeBasicMapper();
   
   public static final QName XSD_BOOLEAN = new QName(IAeConstants.W3C_XML_SCHEMA, "boolean"); //$NON-NLS-1$
   public static final QName XSD_BYTE = new QName(IAeConstants.W3C_XML_SCHEMA, "byte"); //$NON-NLS-1$
   public static final QName XSD_DATE = new QName(IAeConstants.W3C_XML_SCHEMA, "date"); //$NON-NLS-1$
   public static final QName XSD_DATETIME = new QName(IAeConstants.W3C_XML_SCHEMA, "dateTime"); //$NON-NLS-1$
   public static final QName XSD_TIME = new QName(IAeConstants.W3C_XML_SCHEMA, "time"); //$NON-NLS-1$
   public static final QName XSD_DURATION = new QName(IAeConstants.W3C_XML_SCHEMA, "duration"); //$NON-NLS-1$
   public static final QName XSD_DECIMAL = new QName(IAeConstants.W3C_XML_SCHEMA, "decimal"); //$NON-NLS-1$
   public static final QName XSD_DOUBLE = new QName(IAeConstants.W3C_XML_SCHEMA, "double"); //$NON-NLS-1$
   public static final QName XSD_FLOAT = new QName(IAeConstants.W3C_XML_SCHEMA, "float"); //$NON-NLS-1$
   public static final QName XSD_INT = new QName(IAeConstants.W3C_XML_SCHEMA, "int"); //$NON-NLS-1$
   public static final QName XSD_INTEGER = new QName(IAeConstants.W3C_XML_SCHEMA, "integer"); //$NON-NLS-1$
   public static final QName XSD_LONG = new QName(IAeConstants.W3C_XML_SCHEMA, "long"); //$NON-NLS-1$
   public static final QName XSD_QNAME = new QName(IAeConstants.W3C_XML_SCHEMA, "QName"); //$NON-NLS-1$
   public static final QName XSD_SHORT = new QName(IAeConstants.W3C_XML_SCHEMA, "short"); //$NON-NLS-1$
   public static final QName XSD_STRING = new QName(IAeConstants.W3C_XML_SCHEMA, "string"); //$NON-NLS-1$

   /**
    * Ctor for aType mapping populates the mappings.
    */
   public AeTypeMapping()
   {
      mJava2SchemaMappings.put(Boolean.class, new AeBooleanMapper());
      mSchema2JavaMappings.put(XSD_BOOLEAN, new AeBooleanMapper());

      mJava2SchemaMappings.put(Byte.class, new AeByteMapper());
      mSchema2JavaMappings.put(XSD_BYTE, new AeByteMapper());

      mJava2SchemaMappings.put(AeSchemaDate.class, new AeDateMapper());
      mSchema2JavaMappings.put(XSD_DATE, new AeDateMapper());

      mJava2SchemaMappings.put(AeSchemaDateTime.class, new AeDateTimeMapper());
      mJava2SchemaMappings.put(Date.class, new AeDateTimeMapper());
      mSchema2JavaMappings.put(XSD_DATETIME, new AeDateTimeMapper());

      mJava2SchemaMappings.put(AeSchemaTime.class, new AeTimeMapper());
      mSchema2JavaMappings.put(XSD_TIME, new AeTimeMapper());

      mJava2SchemaMappings.put(AeSchemaDuration.class, new AeDurationMapper());
      mSchema2JavaMappings.put(XSD_DURATION, new AeDurationMapper());
      
      mJava2SchemaMappings.put(BigDecimal.class, new AeBigDecimalMapper());
      mSchema2JavaMappings.put(XSD_DECIMAL, new AeBigDecimalMapper());

      mJava2SchemaMappings.put(Double.class, new AeDoubleMapper());
      mSchema2JavaMappings.put(XSD_DOUBLE, new AeDoubleMapper());

      mJava2SchemaMappings.put(Float.class, new AeFloatMapper());
      mSchema2JavaMappings.put(XSD_FLOAT, new AeFloatMapper());

      mJava2SchemaMappings.put(Integer.class, new AeIntMapper());
      mSchema2JavaMappings.put(XSD_INT, new AeIntMapper());

      mJava2SchemaMappings.put(BigInteger.class, new AeBigIntegerMapper());
      mSchema2JavaMappings.put(XSD_INTEGER, new AeBigIntegerMapper());

      mJava2SchemaMappings.put(Long.class, new AeLongMapper());
      mSchema2JavaMappings.put(XSD_LONG, new AeLongMapper());

      mJava2SchemaMappings.put(QName.class, new AeQNameMapper());
      mSchema2JavaMappings.put(XSD_QNAME, new AeQNameMapper());

      mJava2SchemaMappings.put(Short.class, new AeShortMapper());
      mSchema2JavaMappings.put(XSD_SHORT, new AeShortMapper());
   }
   
   /**
    * Serializes a schema object to a string.
    * @param aObj The object to serialize as string 
    * @return String representing object.
    */
   public String serialize(Object aObj)
   {
      if(aObj != null)
      {
         IAeTypeMapper mapper = (IAeTypeMapper)mJava2SchemaMappings.get(aObj.getClass());
         if(mapper == null)
            mapper = defaultMapper;
         return mapper.serialize(aObj);
      }
      return ""; //$NON-NLS-1$
   }
   
   /**
    * Serializes a schema object to a string using the type as a hint.
    * 
    * @param aType The type of the schema object.
    * @param aObj The object to serialize.
    * @return String representing object.
    */
   public String serialize(QName aType, Object aObj)
   {
      if (aObj != null)
      {
         IAeTypeMapper mapper = (IAeTypeMapper) mSchema2JavaMappings.get(aType);
         if (mapper == null)
            return serialize(aObj);
         return mapper.serialize(aObj);
      }
      return ""; //$NON-NLS-1$
   }
   
   /**
    * Deserializes a schema object from a string.
    * @param aType The qname of the schema type
    * @param aObj The object ti be deserialized
    * @return The Object in java form
    */
   public Object deserialize(QName aType, String aObj)
   {
      if(aType != null && aObj != null)
      {
         IAeTypeMapper mapper = (IAeTypeMapper)mSchema2JavaMappings.get(aType);
         // TODO check for base type mapping
         if (mapper == null)
            mapper = defaultMapper;
         return mapper.deserialize(aObj);
      }
      return aObj;
   }
   
   /**
    * Deserializes a schema object from a string and will check for base type
    * deserializers if specific one is not avaiable.
    * @param aType The xml type of the schema type
    * @param aObj The object ti be deserialized
    * @return The Object in java form
    */
   public Object deserialize(XMLType aType, String aObj)
   {
      if(aObj != null)
      {
         QName typeName = new QName(aType.getSchema().getSchemaNamespace(), aType.getName());
         IAeTypeMapper mapper = (IAeTypeMapper)mSchema2JavaMappings.get(typeName);
         while(mapper == null && aType.getBaseType() != null)
         {
            aType = aType.getBaseType();
            typeName = new QName(aType.getSchema().getSchemaNamespace(), aType.getName());
            mapper = (IAeTypeMapper)mSchema2JavaMappings.get(typeName);
         }
         if(mapper == null)
            mapper = defaultMapper;
         return mapper.deserialize(aObj);
      }
      return aObj;
   }
}

/**
 * Interface which all type mappers should adhere to.
 */
interface IAeTypeMapper
{
   /**
    * Serializes a schema object to a string.
    * @param aObj The object being serialized
    * @return The string representing the object
    */
   public String serialize(Object aObj);

   /**
    * Deserializes a java object from a string.
    * @param aObj The string being deserialized
    * @return The object representing the schema type
    */
   public Object deserialize(String aObj);
}

/**
 * Simple object to string handler.
 */
class AeBasicMapper implements IAeTypeMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#serialize(java.lang.Object)
    */
   public String serialize(Object aObj)
   {
      if(aObj != null)
         return aObj.toString();
      return ""; //$NON-NLS-1$
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return aObj;
   }
}

/**
 * Handles xsd boolean types.
 */
class AeBooleanMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      if ("1".equals(aObj)) //$NON-NLS-1$
      {
         return Boolean.TRUE;
      }
      else
      {
         return Boolean.valueOf(aObj);
      }
   }
}

/**
 * Handles xsd integer types.
 */
class AeBigIntegerMapper extends AeBigDecimalMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return ((BigDecimal)super.deserialize(aObj)).toBigInteger();
   }
}

/**
 * Handles xsd byte types.
 */
class AeByteMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return new Byte(aObj);
   }
}

/**
 * Handles xsd date types.
 */
class AeDateMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#serialize(java.lang.Object)
    */
   public String serialize(Object aObj)
   {
      if (aObj instanceof AeSchemaDate)
         return ((AeSchemaDate) aObj).toString();
      if (aObj instanceof AeSchemaDateTime)
         return new AeSchemaDate(((AeSchemaDateTime) aObj).toDate()).toString();
      // Note: the following two cases should never happen, but they're easy to handle just in case.
      if (aObj instanceof Calendar)
         return new AeSchemaDate((Calendar) aObj).toString();
      if (aObj instanceof Date)
         return new AeSchemaDate((Date) aObj).toString();

      return super.serialize(aObj);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new AeSchemaDate(aObj);
      }
      catch (Exception e)
      {
         // Also accept xsd:dateTime formatted strings.  This code was added for the following use case:
         //  1) javascript expression of the form "new Date(95, 2, 10)" - returns a java.util.Date object
         //  2) we serialize the Date object to a String formatted as an xsd:dateTime (because we don't
         //     necessarily know the type of the variable we are assigning to)
         //  3) we deserialize the value here but we KNOW we want an xsd:date
         try
         {
            return new AeSchemaDate(new AeSchemaDateTime(aObj).toDate());
         }
         catch (Exception e2)
         {
         }
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new AeSchemaDate(new Date());
      }
   }
}

/**
 * Handles xsd date types.
 */
class AeDateTimeMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#serialize(java.lang.Object)
    */
   public String serialize(Object aObj)
   {
      if (aObj instanceof AeSchemaDateTime)
         return ((AeSchemaDateTime) aObj).toString();
      // Note: the following two cases should never happen, but they're easy to handle just in case.
      if (aObj instanceof Calendar)
         return new AeSchemaDateTime((Calendar) aObj).toString();
      if (aObj instanceof Date)
         return new AeSchemaDateTime((Date) aObj).toString();
      return super.serialize(aObj);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new AeSchemaDateTime(aObj);
      }
      catch (Exception e)
      {
         // This is unlikely - should only happen if a badly formatted xsd:dateTime is sent.
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new AeSchemaDateTime(new Date());
      }
   }
}

/**
 * Handles xsd date types.
 */
class AeTimeMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#serialize(java.lang.Object)
    */
   public String serialize(Object aObj)
   {
      if(aObj instanceof AeSchemaTime)
         return ((AeSchemaTime) aObj).toString();
      // Note: the following two cases should never happen, but they're easy to handle just in case.
      if (aObj instanceof Calendar)
         return new AeSchemaTime((Calendar) aObj).toString();
      if (aObj instanceof Date)
         return new AeSchemaTime((Date) aObj).toString();
      return super.serialize(aObj);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new AeSchemaTime(aObj);
      }
      catch (Exception e)
      {
         // This is unlikely - should only happen if a badly formatted xsd:time is sent.
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new AeSchemaTime(new Date());
      }
   }
}

/**
 * Handles xsd duration types.
 */
class AeDurationMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#serialize(java.lang.Object)
    */
   public String serialize(Object aObj)
   {
      if (aObj instanceof AeSchemaDuration)
         return ((AeSchemaDuration) aObj).toString();
      return super.serialize(aObj);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new AeSchemaDuration(aObj);
      }
      catch (Exception e)
      {
         // This is unlikely - should only happen if a badly formatted xsd:duration is sent.
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new AeSchemaTime(new Date());
      }
   }
}

/**
 * Handles xsd double types.
 */
class AeDoubleMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new Double(aObj);
      }
      catch (Exception ex)
      {
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new Double(0);
      }
   }
}

/**
 * Handles xsd float types.
 */
class AeFloatMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new Float(aObj);
      }
      catch (Exception ex)
      {
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new Float(0);
      }
   }
}

/**
 * Handles xsd decimal types.
 */
class AeBigDecimalMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      try
      {
         return new BigDecimal(aObj);
      }
      catch (Exception ex)
      {
         // TODO (MF) I removed the log statement to avoid generating error messages when setting variables via remote debugger. Restore once the designer code changes. 
         return new BigDecimal(0);
      }
   }
}

/**
 * Handles xsd int types.
 */
class AeIntMapper extends AeDoubleMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return new Integer(((Double)super.deserialize(aObj)).intValue());
   }
}

/**
 * Handles xsd long types.
 */
class AeLongMapper extends AeDoubleMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return new Long(((Double)super.deserialize(aObj)).longValue());
   }
}

/**
 * Handles xsd QName types.
 */
class AeQNameMapper extends AeBasicMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return new QName(aObj);
   }
}

/**
 * Handles xsd QName types.
 */
class AeShortMapper extends AeDoubleMapper
{
   /**
    * @see org.activebpel.rt.bpel.impl.IAeTypeMapper#deserialize(java.lang.String)
    */
   public Object deserialize(String aObj)
   {
      return new Short(((Double)super.deserialize(aObj)).shortValue());
   }
}
