// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeVariable.java,v 1.38 2006/02/13 23:09:1
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.util.Iterator;

import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.activebpel.rt.AeException;
import org.activebpel.rt.IAeConstants;
import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.AeWSDLDefHelper;
import org.activebpel.rt.bpel.IAeBusinessProcess;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.def.AeVariableDef;
import org.activebpel.rt.message.AeEmptyMessage;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeXmlUtil;
import org.activebpel.rt.wsdl.def.AeBPELExtendedWSDLDef;
import org.activebpel.rt.xml.AeXMLParserBase;
import org.exolab.castor.xml.schema.XMLType;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/** The class implementing all variable interactions. */
public class AeVariable implements IAeVariable, Cloneable
{
   /** The scope which owns this variable. */
   private IAeBpelObject mScope;
   /** The associated variable definition. */
   private AeVariableDef mVariableDef;
   /** The associated message data if any. */
   private IAeMessageData mMessageData;
   /** The associated type data if any. */
   private Object mTypeData;
   /** The associated element if any. */
   private Element mElementData;
   /** Location path override if different from the definition object */
   private String mLocationPath;
   
   // Note on versionNumber: The versionNumber field is not really a version number
   // per se but rather a unique id assigned to each version of a variable within
   // a process. The process maintains an int to use and increment with each new
   // variable created. 

   /** The version number of the variable. Increments with each change */
   private int mVersionNumber;
   
   

   /**
    * Constructs a new variable from a definition object
    * @param aVariableDef variable definition which is basis for this object
    */
   public AeVariable(IAeBpelObject aScope, AeVariableDef aVariableDef)
   {
      setDefinition(aVariableDef);
      setScope(aScope);
      // Incrementing the version number here to avoid a problem introduced with
      // parallel forEach's. In that case, unassigned variables will have the same
      // locationId and versionNumber which will be a problem with the persistence layer
      incrementVersionNumber();
   }

   /**
    * Get the definition associated with this process
    * @see org.activebpel.rt.bpel.IAeVariable#getDefinition()
    */
   public AeVariableDef getDefinition()
   {
      return mVariableDef;
   }

   /**
    * Set the definition associated with this process
    * @param aVariableDef definition object to be set
    */
   protected void setDefinition(AeVariableDef aVariableDef)
   {
      mVariableDef = aVariableDef;
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeLocatableObject#hasLocationId()
    */
   public boolean hasLocationId()
   {
      return true;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#getLocationPath()
    */
   public String getLocationPath()
   {
      if (hasCustomLocationPath())
      {
         return mLocationPath;
      }
      return getDefinition().getLocationPath();
   }
   
   /**
    * Setter for the location path.
    * 
    * @param aLocationPath
    */
   public void setLocationPath(String aLocationPath)
   {
      mLocationPath = aLocationPath;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeLocatableObject#getLocationId()
    */
   public int getLocationId()
   {
      return getDefinition().getLocationId();
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeLocatableObject#hasCustomLocationPath()
    */
   public boolean hasCustomLocationPath()
   {
      return mLocationPath != null;
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeLocatableObject#setLocationId(int)
    */
   public void setLocationId(int aId)
   {
      // no op here
   }
   
   /**
    * Returns the name of the variable
    * @see org.activebpel.rt.bpel.IAeVariable#getName()
    */
   public String getName()
   {
      return mVariableDef.getName();
   }

   /**
    * Get the message data associated with the variable. Returns null if not a message.
    * @see org.activebpel.rt.bpel.IAeVariable#getMessageData()
    */
   public IAeMessageData getMessageData() throws AeBusinessProcessException
   {
      if (mMessageData == null && !mVariableDef.hasParts() && mVariableDef.getMessageType() != null)
      {
         mMessageData = new AeEmptyMessage(mVariableDef.getMessageType());
      }

      if (mMessageData == null)
      {
         throw new AeBpelException(AeMessages.getString("AeVariable.ERROR_0"), AeFaultFactory.getFault(AeFaultFactory.UNINITIALIZED_VARIABLE)); //$NON-NLS-1$
      }
      
      return mMessageData;
   }

   /**
    * Sets the message data associated with the variable.
    * @see org.activebpel.rt.bpel.IAeVariable#setMessageData(org.activebpel.rt.message.IAeMessageData)
    */
   public void setMessageData(IAeMessageData aData)
   {
      if (mMessageData != null)
         mMessageData.removeMessageDataListener(this);

      mMessageData = aData;
      mMessageData.addMessageDataListener(this);
      dataUpdated();
   }

   /**
    * Returns the associated type data if any.
    * @see org.activebpel.rt.bpel.IAeVariable#getTypeData()
    */
   public Object getTypeData()
   {
      return mTypeData;
   }

   /**
    * Set the type data associated with this variable.
    * @see org.activebpel.rt.bpel.IAeVariable#setTypeData(java.lang.Object)
    */
   public void setTypeData(Object aData)
   {
      dataUpdated();
      mTypeData = aData;
   }

   /**
    * Get the data if the variable is specified as element. Returns null if not element.
    * @see org.activebpel.rt.bpel.IAeVariable#getElementData()
    */
   public Element getElementData()
   {
      return mElementData;
   }

   /**
    * Set the element data associated with this variable.
    * @see org.activebpel.rt.bpel.IAeVariable#setElementData(org.w3c.dom.Element)
    */
   public void setElementData(Element aData)
   {
      dataUpdated();
      mElementData = aData;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#getMessageType()
    */
   public QName getMessageType()
   {
      return getDefinition().getMessageType();
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#getElement()
    */
   public QName getElement()
   {
      return getDefinition().getElement();
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#getType()
    */
   public QName getType()
   {
      return getDefinition().getType();
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#validate()
    */
   public void validate() throws AeBusinessProcessException
   {
      validate(false);
   }
   
   /**
    * @see org.activebpel.rt.bpel.IAeVariable#validate(boolean)
    */
   public void validate(boolean aAllowEmptyPartData) throws AeBusinessProcessException
   {
      Exception cause = null;
      boolean isValid = true;

      if ( getMessageType() == null )
      {
         //TODO need to add validation logic for Type and Element data. 
         return;
      }
      
      try
      {
         AeXMLParserBase parser = new AeXMLParserBase();
         for (Iterator iter=getMessageData().getPartNames(); iter.hasNext() && isValid;)
         {
            String partName = (String)iter.next();
            XMLType partType = getDefinition().getPartType(partName);
            Object data = getMessageData().getData(partName);

            // Data which is of simple type must be wrappered in DOM for validation
            Document doc = null;
            if (getDefinition().isPartElement(partName) || AeXmlUtil.isComplexOrAny(partType))
            {
               // Complex data can only be validated if data is a DOM
               if (data instanceof Document)
               {
                  // Must add xsi:type for type elements
                  doc = (Document)data;
                  if (! getDefinition().isPartElement(partName))
                     declarePartType((Document)data, getDefinition().getPart(partName).getTypeName());
               }
            }
            else
            {
               doc = declareSimpleType(parser, partName, data);
            }

            // If no doc to validate, data is invalid otherwise let parser do the validation
            if (doc == null)
            {
               if (! aAllowEmptyPartData)
                  isValid = false;
            }
            else
            {
               IAeProcessPlan plan = getProcess().getProcessPlan();
               AeBPELExtendedWSDLDef wsdlDef = AeWSDLDefHelper.getWSDLDefinitionForMsg(plan, getDefinition().getMessageType());
               parser.setWSDLLocator(wsdlDef.getLocator());
               parser.validateDocument(doc, wsdlDef.getSchemas());
               isValid = ! parser.hasParseWarnings();
            }
         }
      }
      catch(Exception e)
      {
         isValid = false;
         cause = e;
      }

      if (! isValid)
      {
         String msg = "'" + getName() + "'"; //$NON-NLS-1$ //$NON-NLS-2$
         if(cause != null && cause.getMessage() != null)
            msg += AeMessages.getString("AeVariable.3") + cause.getMessage(); //$NON-NLS-1$
         throw new AeBpelException( AeMessages.getString("AeVariable.ERROR_4") + msg,  //$NON-NLS-1$
                                     AeFaultFactory.VALIDATION_ERROR );    
      }
   }

   /**
    * Process the document for complex data declared as "type". We must insert
    * the xsi:type attribute in order to pass validation.
    * @param aDoc the document we are processing
    */
   private void declarePartType(Document aDoc, QName aTypeName)
   {
      AeXmlUtil.declarePartType(aDoc, aTypeName);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#declarePartTypes()
    */
   public void declarePartTypes()
   {
      try
      {
         for (Iterator iter = getMessageData().getPartNames(); iter.hasNext();)
         {
            String partName = (String)iter.next();
            XMLType partType = getDefinition().getPartType(partName);
            Object data = getMessageData().getData(partName);

            // If the part type is a non-element complex type
            if (AeXmlUtil.isComplexOrAny(partType) && !getDefinition().isPartElement(partName))
            {
               declarePartType((Document)data, getDefinition().getPart(partName).getTypeName());
            }
         }
      }
      catch (AeException e)
      {
         e.logError();
      }
   }
   
   /**
    * This method is used to create and prepare a DOM for validation. We will obtain
    * the XML type information from the part and build the node tree based upon that.
    * The data will be added to this node for validation. If any required information
    * cannot be determined null will be returned for the document.
    *
    * @param aParser The instance of the parser we are using.
    * @param aPartName The name of the part we are processing
    * @param aData The data for the part we wish to validate
    * @throws AeBusinessProcessException
    */
   private Document declareSimpleType(AeXMLParserBase aParser, String aPartName, Object aData) throws AeBusinessProcessException
   {
      Document doc = null;

      Part part = getDefinition().getPart(aPartName);
      QName typeName = part.getTypeName();
      if (typeName == null)
         typeName = part.getElementName();

      if (typeName != null)
      {
         doc = aParser.createDocument();
         Element docElement = doc.createElement(aPartName);
         doc.appendChild(docElement);

         // Set the schema and schema instance namespace
         docElement.setAttributeNS(IAeConstants.W3C_XMLNS, "xmlns:ns", typeName.getNamespaceURI()); //$NON-NLS-1$
         docElement.setAttributeNS(IAeConstants.W3C_XMLNS, "xmlns:xsi", IAeConstants.W3C_XML_SCHEMA_INSTANCE); //$NON-NLS-1$

         // Set the xsi:type attribute
         docElement.setAttributeNS(IAeConstants.W3C_XML_SCHEMA_INSTANCE, "xsi:type", "ns:" + typeName.getLocalPart()); //$NON-NLS-1$ //$NON-NLS-2$

         String value = getProcess().getEngine().getTypeMapping().serialize(aData);

         docElement.appendChild(doc.createTextNode(value));
      }

      return doc;
   }

   /**
    * Getter for the version number
    */
   public int getVersionNumber()
   {
      return mVersionNumber;
   }

   /**
    * Returns a copy of this variable so it can be persisted in our snapshot for
    * any future compensation logic.
    */
   public Object clone()
   {
      try
      {
         return super.clone();
      }
      catch (CloneNotSupportedException e)
      {
         throw new InternalError("Unexpected error during clone: " + e.getLocalizedMessage()); //$NON-NLS-1$
      }
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#clear()
    */
   public void clear()
   {
      if (hasData())
      {
         dataUpdated();
      }

      mElementData = null;
      mMessageData = null;
      mTypeData = null;
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#hasMessageData()
    */
   public boolean hasMessageData()
   {
      return mMessageData != null;
   }

   /**
    * @see org.activebpel.rt.message.IAeMessageDataListener#dataUpdated()
    */
   public void dataUpdated()
   {
      incrementVersionNumber();
   }

   /**
    * increments the variable's version number 
    */
   protected void incrementVersionNumber()
   {
      // not thread safe but this isn't an issue with the current impl since only
      // one activity will be executing at a time.
      IAeBusinessProcess process = getScope().getProcess();
      setVersionNumber(process.getNextVariableVersionNumber());
      process.setNextVariableVersionNumber(getVersionNumber() + 1);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#hasData()
    */
   public boolean hasData()
   {
      return (mElementData != null) || (mMessageData != null) || (mTypeData != null);
   }

   /**
    * @see org.activebpel.rt.bpel.IAeVariable#setVersionNumber(int)
    */
   public void setVersionNumber(int aVersionNumber)
   {
      mVersionNumber = aVersionNumber;
   }
   
   /**
    * @return The process which owns this variable.
    */
   public IAeBusinessProcess getProcess()
   {
      return getScope().getProcess();
   }

   /**
    * @return Returns the scope which owns this variable.
    */
   public IAeBpelObject getScope()
   {
      return mScope;
   }
   
   /**
    * @param aScope The scope to set.
    */
   protected void setScope(IAeBpelObject aScope)
   {
      mScope = aScope;
   }
}
