// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeVariableDeserializer.java,v 1.11 2005/12/29 20:01:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.text.MessageFormat;
import java.util.List;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeBusinessProcessEngine;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeXmlUtil;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * Restores the state of a variable from a serialized variable document.
 */
public class AeVariableDeserializer implements IAeImplStateNames
{
   /** The engine we are running inside of, supplys type mapper. */
   protected IAeBusinessProcessEngine mEngine;

   /**
    * Constructs the deserializer with an engine since engine has type mapping info.
    * @param aEngine the engine associated with derializer for mapping usages.
    */
   public AeVariableDeserializer(IAeBusinessProcessEngine aEngine)
   {
      mEngine = aEngine;
   }

   /** The serialized variable document. */
   private Document mDocument;

   /** The variable to restore. */
   private IAeVariable mVariable;

   /**
    * Returns the serialized variable document.
    */
   protected Document getDocument()
   {
      return mDocument;
   }

   /**
    * Returns element data from the serialized variable document.
    */
   protected Element getElementData() throws AeBusinessProcessException
   {
      // Return the first child of the document root.
      return (Element) selectSingleNode(getDocument(), "/*/*"); //$NON-NLS-1$
   }

   /**
    * Returns message data from the serialized variable document.
    *
    * @param aVariable the variable to get message data for
    */
   protected IAeMessageData getMessageData(IAeVariable aVariable) throws AeBusinessProcessException
   {
      AeMessageDataDeserializer deserializer = new AeMessageDataDeserializer();
      deserializer.setMessageDataElement(getDocument().getDocumentElement());
      deserializer.setMessageType(aVariable.getMessageType());
      deserializer.setTypeMapping(getTypeMapping());
      deserializer.setVariable(aVariable);

      return deserializer.getMessageData();
   }

   /**
    * Returns type data from the serialized variable document.
    */
   protected String getTypeData() throws AeBusinessProcessException
   {
      // Return the contents of the first and only text node.
      Text text = (Text) selectOptionalNode(getDocument(), "/*/text()"); //$NON-NLS-1$
      return (text == null) ? "" : text.getNodeValue(); //$NON-NLS-1$
   }

   /**
    * Returns the variable to restore.
    */
   protected IAeVariable getVariable()
   {
      return mVariable;
   }

   /**
    * Returns the variable version number from the serialized variable
    * document.
    *
    * @return boolean
    * @throws AeBusinessProcessException
    */
   protected int getVersionNumber() throws AeBusinessProcessException
   {
      Attr attr = (Attr) selectSingleNode(getDocument(), "/*/@" + STATE_VERSION); //$NON-NLS-1$
      String version = attr.getValue();

      try
      {
         return Integer.parseInt(version);
      }
      catch (NumberFormatException e)
      {
         throw new AeBusinessProcessException(MessageFormat.format(AeMessages.getString("AeVariableDeserializer.ERROR_1"), new Object[] {version}), e); //$NON-NLS-1$
      }
   }

   /**
    * Returns <code>true</code> if and only if the serialized variable document
    * has data to restore.
    *
    * @return boolean
    * @throws AeBusinessProcessException
    */
   protected boolean hasData() throws AeBusinessProcessException
   {
      Attr attr = (Attr) selectSingleNode(getDocument(), "/*/@" + STATE_HASDATA); //$NON-NLS-1$
      String hasData = attr.getValue();
      return "true".equals(hasData); //$NON-NLS-1$
   }

   /**
    * Restores the variable from the serialized variable document.
    *
    * @throws AeBusinessProcessException
    */
   public void restoreVariable() throws AeBusinessProcessException
   {
      IAeVariable variable = getVariable();

      if (!hasData())
      {
         variable.clear();
      }
      else if (variable.getType() != null)
      {
         variable.setTypeData(getTypeMapping().deserialize(variable.getType(), getTypeData()));
      }
      else if (variable.getElement() != null)
      {
         // The element data needs to be in its own document
         Document document = AeXmlUtil.newDocument();
         Element element = (Element) document.importNode(getElementData(), true);
         document.appendChild(element);
         variable.setElementData(element);
      }
      else
      {
         IAeMessageData messageData = getMessageData(variable);
         variable.setMessageData(messageData);
      }
      
      // Restore the version number *after* setting the variable data, because
      // setting the variable data updates the version number.
      restoreVersionNumber();
   }
   
   /**
    * Restore the version number from the serialized data.
    * @throws AeBusinessProcessException
    */
   protected void restoreVersionNumber() throws AeBusinessProcessException
   {
      getVariable().setVersionNumber(getVersionNumber());
   }

   /**
    * Restores the variable from the specified serialized variable document.
    *
    * @param aDocument
    * @throws AeBusinessProcessException
    */
   public void restoreVariable(Document aDocument) throws AeBusinessProcessException
   {
      setDocument(aDocument);
      restoreVariable();
   }

   /**
    * Selects a single node by XPath. Returns <code>null</code> if there is no
    * matching node.
    *
    * @param aNode The node to select from.
    * @param aPath The XPath expression.
    * @return Node The matching node.
    * @throws AeBusinessProcessException
    */
   protected static Node selectOptionalNode(Node aNode, String aPath) throws AeBusinessProcessException
   {
      List nodes;

      try
      {
         XPath xpath = new DOMXPath(aPath);
         nodes = xpath.selectNodes(aNode);
      }
      catch (JaxenException e)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeVariableDeserializer.ERROR_0") + aPath, e); //$NON-NLS-1$
      }

      // Check for no more than 1 matching node.
      if (nodes.size() > 1)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeVariableDeserializer.ERROR_9") + aPath); //$NON-NLS-1$
      }

      return (nodes.size() == 0) ? null : (Node) nodes.get(0);
   }

   /**
    * Select a single node by XPath.
    *
    * @param aNode The node to select from.
    * @param aPath The XPath expression.
    * @return Node The matching node.
    * @throws AeBusinessProcessException
    */
   protected static Node selectSingleNode(Node aNode, String aPath) throws AeBusinessProcessException
   {
      List nodes;

      try
      {
         XPath xpath = new DOMXPath(aPath);
         nodes = xpath.selectNodes(aNode);
      }
      catch (JaxenException e)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeVariableDeserializer.ERROR_0") + aPath, e); //$NON-NLS-1$
      }

      // Check for exactly 1 matching node.
      if (nodes.size() == 0)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeVariableDeserializer.ERROR_11") + aPath); //$NON-NLS-1$
      }

      if (nodes.size() > 1)
      {
         throw new AeBusinessProcessException(AeMessages.getString("AeVariableDeserializer.ERROR_9") + aPath); //$NON-NLS-1$
      }

      return (Node) nodes.get(0);
   }

   /**
    * Sets the serialized variable document to use.
    *
    * @param aDocument
    */
   protected void setDocument(Document aDocument)
   {
      mDocument = aDocument;
   }

   /**
    * Sets the variable to restore.
    *
    * @param aVariable
    */
   public void setVariable(IAeVariable aVariable)
   {
      mVariable = aVariable;
   }

   /** Helper method to get type mapping from engine. */
   protected AeTypeMapping getTypeMapping()
   {
      return mEngine.getTypeMapping();
   }
}
