// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/AeWaitableReplyReceiver.java,v 1.1 2005/08/04 20:01:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl;

import java.util.Map;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.message.IAeMessageData;

/**
 * Class used to handle the reply in a request response operation. 
 * This class's onFault and onMessage method notifies any waiting 
 * threads by calling notifyAll.
 */
public class AeWaitableReplyReceiver implements IAeReplyReceiver
{
   /** The fault returned from this response receiver */
   private IAeFault mFault;
   /** The message data returned from this response receiver */
   private IAeMessageData mMessageData;
   /** The Business Process properties. */
   private Map mProcessProperties;
   
   /**
    * Notifies any waiting threads on when this callback is invoked.
    * @param aFault fault
    * @param aProcessProperties business process properties.
    * @see org.activebpel.rt.bpel.impl.IAeReplyReceiver#onFault(org.activebpel.rt.bpel.IAeFault, java.util.Map)  
    */
   public synchronized void onFault(IAeFault aFault, Map aProcessProperties) throws AeBusinessProcessException
   {
      setFault(aFault);         
      setBusinessProcessProperties(aProcessProperties);
      notifyAll();
   }

   /**
    * Notifies any waiting threads on when this callback is invoked.
    * @param aMessageData message data.
    * @param aProcessProperties business process properties.
    * @see org.activebpel.rt.bpel.impl.IAeReplyReceiver#onMessage(org.activebpel.rt.message.IAeMessageData, java.util.Map)     
    */
   public synchronized void onMessage(IAeMessageData aMessageData, Map aProcessProperties) throws AeBusinessProcessException
   {
      setMessageData(aMessageData);         
      setBusinessProcessProperties(aProcessProperties);
      notifyAll();
   }
   
   /**
    * This object is considered waitable if it its callback methods for receiving
    * the message data, fault, or void signal haven't been invoked. This method
    * should be used to determine if the object is in a waitable state before 
    * waiting on it. Failing to call this method (or testing the object's state
    * through the aforementioned calls) risks waiting forever for a callback that
    * will never be issued.
    */
   public synchronized boolean isWaitable()
   {
      return getMessageData() == null && getFault() == null;
   }
   
   /**
    * Returns the fault or null if none was returned.
    */
   public IAeFault getFault()
   {
      return mFault;
   }
   
   /**
    * @param aFault The fault to set.
    */
   protected void setFault(IAeFault aFault)
   {
      mFault = aFault;
   }   
   
   /**
    * Returns the message data or null if none was returned.
    */
   public IAeMessageData getMessageData()
   {
      return mMessageData;
   }

   /**
    * @param aMessageData The messageData to set.
    */
   protected void setMessageData(IAeMessageData aMessageData)
   {
      mMessageData = aMessageData;
   }
   
   /**
    * Returns the process process properties or null of none were set.
    */
   public Map getBusinessProcessProperties()
   {
      return mProcessProperties;
   }
   
   /**
    * @param aProcessProperties The processProperties to set.
    */
   protected void setBusinessProcessProperties(Map aProcessProperties)
   {
      mProcessProperties = aProcessProperties;
   }
   
}
