// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/activity/AeActivityCompensateImpl.java,v 1.27 2006/01/14 22:40:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.activity;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.def.AeBaseDef;
import org.activebpel.rt.bpel.def.activity.AeActivityCompensateDef;
import org.activebpel.rt.bpel.impl.AeBpelState;
import org.activebpel.rt.bpel.impl.IAeActivityParent;
import org.activebpel.rt.bpel.impl.IAeBpelObject;
import org.activebpel.rt.bpel.impl.activity.support.AeCompInfo;
import org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeDefaultFaultHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeFaultHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeSupportActivityLocationPathSuffix;
import org.activebpel.rt.bpel.impl.activity.support.IAeCompensationCallback;
import org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor;
import org.activebpel.rt.util.AeUtil;

/**
 * The compensate activity gets called from a fault handler or from within an
 * already executing compensation handler. The role of the compensation activity
 * is to identify the scope or scopes that are getting compensated and then to
 * queue the identified scope's compensation handlers.
 */
public class AeActivityCompensateImpl extends AeActivityImpl implements IAeCompensationCallback
{
   /** The <code>List</code> of matching <code>AeCompInfo</code> objects for iteration. */
   private List mIterationScopes;

   /** The next index in the list of matching <code>AeCompInfo</code> objects. */
   private int mNextIndex = 0;
   
   /** If true, the compensate activity should run only the coordinated <code>AeCompInfo</code> objects. */
   private boolean mMatchCoordinated; 

   /** 
    * default constructor for activity which compensates all compInfo objects  
    */
   public AeActivityCompensateImpl(AeActivityCompensateDef aActivityDef, IAeActivityParent aParent)
   {
      this(aActivityDef, aParent, false);
   }

   /**
    * Constructor for activity.
    * @param aActivityDef activity definition
    * @param aParent enclosing scope or fault handler
    * @param aCoordinated if true, then only the coordinated compInfo objects are used.
    */
   public AeActivityCompensateImpl(AeActivityCompensateDef aActivityDef, 
         IAeActivityParent aParent, boolean aMatchCoordinated)
   {
      super(aActivityDef, aParent);
      mMatchCoordinated = aMatchCoordinated;
   }
   
   /**
    * @return Returns the matchCoordinated.
    */
   protected boolean isMatchCoordinated()
   {
      return mMatchCoordinated;
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeVisitable#accept(org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor)
    */
   public void accept( IAeImplVisitor aVisitor ) throws AeBusinessProcessException
   {
      aVisitor.visit(this);
   }

   /**
    * factory method that creates the implicit compensation activity for use
    * by the implicit fault handler and the implicit compensation handler
    * @param aParent
    */
   public static AeActivityCompensateImpl createImplicitCompensation(IAeActivityParent aParent)
   {
      return createImplicitCompensation(aParent, false);
   }
   
   /**
    * Factory method that creates the implicit compensation activity for use
    * by the implicit fault handler and the implicit compensation handler. 
    * Setting aMatchCoordinated to true compensates only the coordinated activities.
    *  
    * @param aParent enclosing scope.
    * @param aMatchCoordinated filters the list of CompInfo objects by matching only the coordinated activities.
    */
   public static AeActivityCompensateImpl createImplicitCompensation(IAeActivityParent aParent, boolean aMatchCoordinated)
   { 
      AeActivityCompensateImpl implicitComp = new AeActivityCompensateImpl(null, aParent, aMatchCoordinated)
      {
         public AeBaseDef getDefinition()
         {
            throw new UnsupportedOperationException(AeMessages.getString("AeActivityCompensateImpl.ERROR_0")); //$NON-NLS-1$
         }

         public String getLocationPath()
         {
            IAeBpelObject parent = getParent();
            StringBuffer buffer = new StringBuffer(parent.getLocationPath());
            if (isMatchCoordinated())
            {
               buffer.append(AeSupportActivityLocationPathSuffix.IMPLICIT_CC_COMPENSATE_ACTIVITY);
            }
            else
            {
               buffer.append(AeSupportActivityLocationPathSuffix.IMPLICIT_COMPENSATE_ACTIVITY); 
            }
            return buffer.toString();
         }

         public boolean hasLocationId()
         {
            // There is no definition object to give us a location id.
            return false;
         }

         /**
          * The implicit compensation activity can only execute the default
          * compensation logic so this method always returns true.
          */
         protected boolean isDefaultCompensation()
         {
            return true;
         }
      };
      return implicitComp;
   }

   /**
    * A <code>compensate</code> Activity can either target a scope by name or
    * it is the default compensation routine which calls compensate on all of
    * the enclosed scopes. This method with set the <code>iterator</code>
    * field with all of the instances that need compensation. We might have
    * multiple compensation calls to make even with the named scope construct
    * since that named scope could have executed in loop multiple times.
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#execute()
    */
   public void execute() throws AeBusinessProcessException
   {
      // clearing the scopes here since we could exist within a loop and get 
      // executed multiple times. Want to force the scopes to get rebuilt each time
      mIterationScopes = null;
      
      // Verify that each compInfo is enabled.
      for (Iterator i = getIterationScopes().iterator(); i.hasNext(); )
      {
         ((AeCompInfo) i.next()).checkEnabled();
      }

      // Set iterator to start at index 0.
      setNextIndex(0);

      compScope();
   }

   /**
    * executes the compensation handler for the next compensatable scope instance
    * from the Iterator. If the iterator is empty, then we're done.
    * @throws AeBusinessProcessException
    */
   private void compScope() throws AeBusinessProcessException
   {
      if (hasMoreCompInfos())
      {
         AeCompInfo compInfo = getNextCompInfo();
         
         compInfo.checkEnabled();

         compInfo.prepareForCompensation(this);

         getProcess().queueObjectToExecute(compInfo.getCompensationHandler());

      }
      else
      {
         objectCompleted();
      }
   }
   
   
   /**
    * Gets the next comp info object for compensation
    */
   private AeCompInfo getNextCompInfo() throws AeBusinessProcessException
   {
      int i = getNextIndex();
      setNextIndex(i + 1);

      return (AeCompInfo) getIterationScopes().get(i);
   }

   /**
    * returns true if there are more comp info objects that need executing
    */
   private boolean hasMoreCompInfos() throws AeBusinessProcessException
   {
      return getNextIndex() < getIterationScopes().size();
   }

   /**
    * Convenience method to pull the comp info object from the enclosed scope
    */
   private AeCompInfo getCompInfo() throws AeBusinessProcessException
   {
      return findRootScopeForCompensation().getCompInfo();
   }

   /**
    * Walks up the parent hiearchy until it comes across a catch/catchAll or compensationHandler
    * and then it returns that object's enclosing scope.
    */
   private AeActivityScopeImpl findRootScopeForCompensation() throws AeBusinessProcessException
   {
      boolean foundCatchOrCompensationHandler = false;
      for (IAeBpelObject parent = getParent(); parent != null; parent = parent.getParent())
      {
         // skip over scopes til u find a catch/catchAll or compensation handler
         if (foundCatchOrCompensationHandler && parent instanceof AeActivityScopeImpl)
         {
            return (AeActivityScopeImpl) parent;
         }
         // we found a catch/catchAll
         else if (parent instanceof AeDefaultFaultHandler || parent instanceof AeFaultHandler)
         {
            foundCatchOrCompensationHandler = true;
         }
         // we found a compensation handler
         else if (parent instanceof AeCompensationHandler)
         {
            foundCatchOrCompensationHandler = true;
         }
      }

      // should never get here unless the process is invalid, at which point we
      // should have caught it during static analysis.
      staticAnalysisFailure(AeMessages.getString("AeActivityCompensateImpl.1")); //$NON-NLS-1$
      return null;
   }

   /**
    * Convenience method that returns true if we're applying the default compensation
    * logic. This is determined by checking the definition object to see if there
    * is a scope attribute defined for the compensate activity.
    */
   protected boolean isDefaultCompensation()
   {
      return AeUtil.isNullOrEmpty(getScopeNameForCompensation());
   }

   /**
    * Gets the name of the scope that we're targeting for compensation. This assumes
    * that you've already checked to see if it's default compensation or not.
    * Invoking this method without checking if it's default compensation first
    * risks generating an exception since the implicit compensation handler
    * won't have a definition object.
    */
   protected String getScopeNameForCompensation()
   {
      AeActivityCompensateDef def = (AeActivityCompensateDef) getDefinition();

      return def.getScope();
   }

   /**
    * A fault during compensation gets propagated to the parent like any other fault.
    * We'll remove any installed compensation information here before propagating
    * the fault.
    * @see org.activebpel.rt.bpel.impl.activity.support.IAeCompensationCallback#compensationCompleteWithFault(org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler, org.activebpel.rt.bpel.IAeFault)
    */
   public void compensationCompleteWithFault(AeCompensationHandler aHandler, IAeFault aFault) throws AeBusinessProcessException
   {
      aHandler.getCompInfo().compensationComplete();
      objectCompletedWithFault(aFault);
   }
   
   /**
    * Callback method from the compensation handler to indicate that it has
    * successfully completed without a fault. We'll remove the installed compensation
    * information and then move onto the next instance that needs compensation.
    * @see org.activebpel.rt.bpel.impl.activity.support.IAeCompensationCallback#compensationComplete(org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler)
    */
   public void compensationComplete(AeCompensationHandler aHandler) throws AeBusinessProcessException
   {
      AeCompInfo ci = aHandler.getCompInfo(); 
      ci.compensationComplete();
      compScope();
   }
   
   /** 
    * Overrides method to call this activity's childTerminated.
    * @see org.activebpel.rt.bpel.impl.activity.support.IAeCompensationCallback#compensationTerminated(org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler)
    */
   public void compensationTerminated(AeCompensationHandler aHandler) throws AeBusinessProcessException
   {
      // reset comp info
      aHandler.getCompInfo().compensationComplete();      
      childTerminated(aHandler);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.activity.AeActivityImpl#acquireVariableLocks()
    */
   protected boolean acquireVariableLocks()
   {
      return true;
   }

   /**
    * Returns the list of matching scopes for iteration.
    */
   protected List getIterationScopes() throws AeBusinessProcessException
   {
      if (mIterationScopes == null)
      {
         List matchingScopes;

         if (isMatchCoordinated())
         {
            // get all effected comp info objects which are coordinated.
            matchingScopes = getCompInfo().getCoordinatedEnclosedScopes();
         }
         else if (isDefaultCompensation())
         {
            // default compensation affects all enclosed scopes
            matchingScopes = getCompInfo().getEnclosedScopes();
         }
         else
         {
            // we'll pull all of the execution instances of a named scope
            String scopeName = getScopeNameForCompensation();
            matchingScopes = getCompInfo().getEnclosedInfoByScopeName(scopeName);
         }

         mIterationScopes = new ArrayList(matchingScopes);
      }

      return mIterationScopes;
   }

   /**
    * Returns the next index in the list of matching scopes.
    */
   public int getNextIndex()
   {
      return mNextIndex;
   }

   /**
    * Sets the next index for iteration.
    */
   public void setNextIndex(int aNextIndex)
   {
      mNextIndex = aNextIndex;
   }
   
   /**
    * Returns the next child eligible for termination or null if there are no
    * child objects that need to be terminated.
    */
   protected IAeBpelObject getChildForTermination() throws AeBusinessProcessException
   {
      IAeBpelObject childToTerminate = null;
      if (mNextIndex > 0)
      {
         // we need to terminate the active compensationHandler
         AeCompensationHandler compHandler = getExecutingCompensationHandlerForTermination();
         if (compHandler != null)
         {
            if (compHandler.getState().canTransitionToDeadPath())
            {
               compHandler.setState(AeBpelState.DEAD_PATH);
            }
            else
            {
               childToTerminate = compHandler;
            }
         }
      }
      return childToTerminate;
   }
   
   /**
    * Gets the executing compensation handler to terminate.
    */
   protected AeCompensationHandler getExecutingCompensationHandlerForTermination() throws AeBusinessProcessException
   {
      AeCompensationHandler compHandler = null;
      int offset = mNextIndex - 1;
      // precautionary range check - the offset should never be out of range
      // since that would mean that we're being asked to terminate but we have
      // no compensation handlers running. 
      if (offset < getIterationScopes().size())
      {
         AeCompInfo compInfo = (AeCompInfo) getIterationScopes().get(offset);
         if (compInfo.getCompensationHandler() != null && !compInfo.getCompensationHandler().getState().isFinal())
         {
            compHandler = compInfo.getCompensationHandler();
         }
      }
      return compHandler;
   }

   /** 
    * Returns true if this IAeCompensationCallback is a coordinated. This is normally used by
    * during saving and restoring state information.
    * @return true if this is a coordinated comp info.
    */      
   public boolean isCoordinated()
   { 
      return false;
   }
   
   /**
    * Returns the coordinationId for this IAeCompensationCallback.
    * This information is normally used during state save/restore procedures.
    * @return the coordinationId if this is a coordinated comp info.
    */
   public String getCoordinationId()
   {   
      return ""; //$NON-NLS-1$
   }
}
