// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/activity/AeActivityReceiveImpl.java,v 1.36 2006/01/11 21:01:0
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.activity;

import java.util.Iterator;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.IAePartnerLink;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.def.activity.AeActivityReceiveDef;
import org.activebpel.rt.bpel.def.activity.support.AeCorrelationDef;
import org.activebpel.rt.bpel.impl.AeBpelState;
import org.activebpel.rt.bpel.impl.IAeActivityParent;
import org.activebpel.rt.bpel.impl.activity.support.AeCorrelationSet;
import org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor;
import org.activebpel.rt.message.IAeMessageData;

/**
 * Implementation of the bpel receive activity.
 */
public class AeActivityReceiveImpl extends AeActivityImpl 
   implements IAeMessageReceiverActivity
{
   /** True if we have queued our receiver */
   private boolean mQueued = false;
   
   /** Constructor for receive activity */
   public AeActivityReceiveImpl(AeActivityReceiveDef aActivityDef, IAeActivityParent aParent)
   {
      super(aActivityDef, aParent);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeVisitable#accept(org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor)
    */
   public void accept( IAeImplVisitor aVisitor ) throws AeBusinessProcessException
   {
      aVisitor.visit(this);
   }
   
   /**
    * Implements receive execution logic.  The receive queues itself waiting for 
    * message with the engine.  The engine will then call back the receive 
    * when its message is ready.
    * 
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#execute()
    */
   public void execute() throws AeBusinessProcessException
   {
      boolean okToQueue = AeMultiStartHelper.checkForMultiStartBehavior(this);
      if (okToQueue)
      {
         queueReceive();
      }
   }

   /**
    * Prepares the correlation sets for the activity and then queues the activity.
    */
   private void queueReceive() throws AeBusinessProcessException
   {
      setQueued(true);
      
      String messageExchange = findEnclosingScope().getMessageExchangePath(getDef().getMessageExchange());
      
      getProcess().queueMessageReceiver(this, getLocationId(), messageExchange);
   }
   
   /**
    * Convenience method to avoid having to cast the definition
    */
   private AeActivityReceiveDef getDef()
   {
      return (AeActivityReceiveDef) getDefinition();
   }

   /**
    * Handle the receipt of our correlated message.
    * @see org.activebpel.rt.bpel.impl.IAeMessageReceiver#onMessage(org.activebpel.rt.message.IAeMessageData)
    */
   public void onMessage(IAeMessageData aMessage) throws AeBusinessProcessException
   {
      getProcess().removeReceiverKeyForConflictingReceives(this);
      
      setQueued(false);

      // Increment the "outstanding receives" counter on the process.
      if (!getDef().isOneWay())
      {
         findEnclosingScope().incrementOutstandingReplyCount(getDef().getMessageExchange());
      }

      IAeVariable variable = getVariable(getDef().getVariable());
      variable.setMessageData(aMessage);
      if( isValidationOfServiceMessagesEnabled() )
      {
         variable.validate();
      }
      
      for(Iterator iter = getDef().getCorrelationList(); iter.hasNext(); )
      {
         AeCorrelationDef corrDef = (AeCorrelationDef)iter.next();
         AeCorrelationSet corrSet = getCorrelationSet(corrDef.getCorrelationSetName());
         corrSet.initiateOrValidate(variable, corrDef.isInitiate());
      }
      
      // we are done
      objectCompleted();
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity#onFault(org.activebpel.rt.bpel.IAeFault)
    */
   public void onFault(IAeFault aFault) throws AeBusinessProcessException
   {
      objectCompletedWithFault(aFault);
   }
   
   /**
    * One of our correlation sets has been initialized. If all of our sets are
    * initialized then we should queue ourselves.
    * @see org.activebpel.rt.bpel.impl.activity.support.IAeCorrelationListener#correlationSetInitialized(org.activebpel.rt.bpel.impl.activity.support.AeCorrelationSet)
    */
   public void correlationSetInitialized(AeCorrelationSet aSet) throws AeBusinessProcessException
   {
      aSet.removeCorrelationListener(this);
      
      if (AeMultiStartHelper.isCorrelatedDataAvailable(this))
      {
         queueReceive();
      }

   }

   /**
    * Overrides method to extend base in order to dequeue any queued receives. So if we are 
    * terminating or completing with queued receives they should be dequeued. 
    * @see org.activebpel.rt.bpel.impl.IAeExecutableQueueItem#setState(org.activebpel.rt.bpel.impl.AeBpelState)
    */
   public void setState(AeBpelState aNewState) throws AeBusinessProcessException
   {
      dequeue();
      super.setState(aNewState);
   }
   
   /**
    * Dequeues the receive if it is queued.
    */
   protected void dequeue() throws AeBusinessProcessException
   {
      if (isQueued())
      {
         getProcess().dequeueMessageReceiver(this);
         setQueued(false);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity#getCorrelationList()
    */
   public Iterator getCorrelationList()
   {
      return getDef().getCorrelationList();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity#getOperation()
    */
   public String getOperation()
   {
      return getDef().getOperation();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity#getPartnerLink()
    */
   public IAePartnerLink getPartnerLink()
   {
      return getPartnerLink(getDef().getPartnerLink());
   }

   /**
    * @see org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity#getPortType()
    */
   public QName getPortType()
   {
      return getDef().getPortType();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.activity.IAeMessageReceiverActivity#canCreateInstance()
    */
   public boolean canCreateInstance()
   {
      return getDef().isCreateInstance();
   }

   /**
    * Getter for the queued flag
    */
   public boolean isQueued()
   {
      return mQueued;
   }

   /**
    * Setter for the queued flag
    * @param aB
    */
   public void setQueued(boolean aB)
   {
      mQueued = aB;
   }
}
