// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/activity/support/AeCompensationHandler.java,v 1.17 2006/01/24 23:46:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.activity.support;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.def.AeScopeCompensatorContainerDef;
import org.activebpel.rt.bpel.impl.AeBpelState;
import org.activebpel.rt.bpel.impl.IAeBpelObject;
import org.activebpel.rt.bpel.impl.activity.AeActivityScopeImpl;
import org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor;

/**
 * Models the <code>compensateHandler</code> that is part of a scope or process.
 */
public class AeCompensationHandler extends AeBaseHandler
{
   /** Reference back to the object that invoked the compensation handler */
   private IAeCompensationCallback mCallback;

   /** reference to the current set of compensation information. If a scope
    * was executed in a loop then it'll likely be compensated for each iteration
    * through the loop. The compensate activity will call into the handler
    * to execute the compensation behavior once for each time the scope
    * reported its execution as complete. Each time through, the comp info
    * object will be set in place so the comp handler will have access to the
    * correct variable snapshot  */
   private AeCompInfo mCompInfo;

   /**
    * Requires the definition object and the scope impl.
    * @param aContainerDef
    * @param aParent
    */
   public AeCompensationHandler(AeScopeCompensatorContainerDef aContainerDef,
                                 AeActivityScopeImpl aParent)
   {
      super(aContainerDef, aParent);
   }

   /**
    * Explicit compensation handlers run only the CompInfo objects specified by
    * Compensate activities. After running all of the activities with in this handler,
    * any remaining coordinatated activites (invokes) must also be (implicitly) compensated.
    * This method returns true so that the coordinated activities are also compensated.
    * 
    * @return Returns true if the coordinated activities (invokes) 
    * should also be compensated after executing the normal list of CompInfo object.
    */
   protected boolean isEnableCoordinatedActivityCompensation()
   {
      // Explicit compensation handlers should always run the comp for the coord. activities
      // after running the normal list of comps.
      return true;
   }
      
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeVisitable#accept(org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor)
    */
   public void accept( IAeImplVisitor aVisitor ) throws AeBusinessProcessException
   {
      aVisitor.visit(this);
   }

   /**
    * Overridden here so we can pull the variable data we need from the snapshot.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getVariable(java.lang.String)
    */
   public IAeVariable getVariable(String aName)
   {
      AeCompInfo compInfo = getCompInfo();
      IAeVariable var = compInfo.getVariable(aName);
      if (var == null)
         var = super.getVariable(aName);
      return var;
   }

   /**
    * Overridden here so we can pull the corelation set we need from the snapshot.
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getCorrelationSet(java.lang.String)
    */
   public AeCorrelationSet getCorrelationSet(String aName) throws AeBusinessProcessException
   {
      AeCorrelationSet corrSet = getCompInfo().getSnapshot().getCorrelationSet(aName);
      if (corrSet == null)
         corrSet = super.getCorrelationSet(aName);
      return corrSet;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#execute()
    */
   public void execute() throws AeBusinessProcessException
   {
      getProcess().queueObjectToExecute(getActivity());
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#childComplete(org.activebpel.rt.bpel.impl.IAeBpelObject)
    */
   public void childComplete(IAeBpelObject aChild)
      throws AeBusinessProcessException
   {
      if (queueCoordinatedCompensator(aChild, null))
      {
         return;
      }      
      if (getActivity().getState() != AeBpelState.DEAD_PATH)
      {
         setState(AeBpelState.FINISHED);
         mCallback.compensationComplete(this);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#childCompleteWithFault(org.activebpel.rt.bpel.impl.IAeBpelObject, org.activebpel.rt.bpel.IAeFault)
    */
   public void childCompleteWithFault(
      IAeBpelObject aChild,
      IAeFault aFaultObject)
      throws AeBusinessProcessException
   {
      if (queueCoordinatedCompensator(aChild, aFaultObject))
      {
         return;
      }       
      IAeFault fault = getFault();
      setFaultedState( fault );
      // need to clear the fault once it is propagated to the compensate
      // if not done, then the comp handler will get saved w/ a fault
      setFault(null);
      mCallback.compensationCompleteWithFault(this, fault );
   }

   /**
    * Called by the compensate activity during compensation. This will install
    * the compInfo object and provide access to the variable snapshot data
    * @param aCompInfo
    */
   public void setCompInfo(AeCompInfo aCompInfo)
   {
      mCompInfo = aCompInfo;
   }

   /**
    * Getter for the compensation info object
    */
   public AeCompInfo getCompInfo()
   {
      return mCompInfo;
   }

   /**
    * Sets the callback reference so when we complete we'll know where to
    * resume execution.
    * @param aImpl
    */
   public void setCallback(IAeCompensationCallback aCallback)
   {
      mCallback = aCallback;
   }

   /**
    * Getter for the compensation callback reference.
    */
   public IAeCompensationCallback getCallback()
   {
      return mCallback;
   }
   
   /** 
    * Overrides method to call the base class <code>noMoreChildrenToTerminate()</code>
    * followed by invoking the <code>IAeCompensationCallback</code>'s <code>compensationTerminated</code>
    * method.
    * @see org.activebpel.rt.bpel.impl.AeAbstractBpelObject#noMoreChildrenToTerminate()
    */
   protected void noMoreChildrenToTerminate() throws AeBusinessProcessException
   {
      super.noMoreChildrenToTerminate();
      // noMoreChildrenToTerminate() calls the process objectCompletedWithFault (moving up the parent-child hierarchy).
      // In the case of the compensationHandler, we also have to call back "laterally" i.e the compensation
      // handlers's callback - which is typically a CompensateActivity (explicit or implicit).
      if (getCallback() != null && getState().isFinal())
      {
         getCallback().compensationTerminated(this);
      }      
   }   

}
