// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/activity/support/AeLink.java,v 1.18 2005/06/08 12:50:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.activity.support;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeActivity;
import org.activebpel.rt.bpel.IAeProcessInfoEvent;
import org.activebpel.rt.bpel.def.activity.support.AeLinkDef;
import org.activebpel.rt.bpel.impl.AeAbstractBpelObject;
import org.activebpel.rt.bpel.impl.IAeLink;
import org.activebpel.rt.bpel.impl.activity.AeActivityFlowImpl;
import org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor;
import org.activebpel.rt.util.AeUtil;

import java.util.Collections;
import java.util.Iterator;

/**
 * Models the <code>link</code>s that are defined within a <code>flow</code>.
 * A link has a single source activity and a single target activity. The 
 * use of a link ensures that the target activity will not execute until the
 * status of the link is known and it's transition condition is evaluated.
 * 
 * The link implementation keeps a reference to the target activity so it can
 * notify the engine when the link's status changes.
 */
public class AeLink extends AeAbstractBpelObject implements IAeLink
{
   /** the link's status, null if unknown */
   private Boolean mStatus;
   
   /** target activity for this link */
   private IAeActivity mTargetActivity;
   
   /** source activity for this link */
   private IAeActivity mSourceActivity;

   /** transition condition for the source */   
   private String mTransitionCondition;
   
   /** flag controls whether or not we notify the process upon our state change */
   private boolean mNotifyProcess = true;
   
   /**
    * Creates the link from the def object and flow.
    * @param aDef
    * @param aParent
    */
   public AeLink(AeLinkDef aDef, AeActivityFlowImpl aParent)
   {
      super(aDef, aParent);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeVisitable#accept(org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor)
    */
   public void accept( IAeImplVisitor aVisitor ) throws AeBusinessProcessException
   {
      aVisitor.visit(this);
   }
   
   /**
    * Gets the name of the link from the def.
    * @return String
    */
   public String getName()
   {
      AeLinkDef def = (AeLinkDef) getDefinition();
      return def.getName();
   }
   
   /**
    * Returns true if the status of this link is known.
    */
   public boolean isStatusKnown()
   {
      return mStatus != null;
   }
   
   /**
    * Returns the status. If the status of this link is unknown then it'll 
    * generate an exception since the spec dictates that you don't evaluate join
    * conditions until the status of all inbound links is known.
    * @return true if transition condition met, false if not, exception if not known
    */
   public boolean getStatus()
   {
      if ( ! isStatusKnown() )
      {
         throw new IllegalStateException("call isStatusKnown() before checking status"); //$NON-NLS-1$
      }
      
      return mStatus.booleanValue();
   }
   
   /**
    * Sets status <i>without</i> notifying the process. This is used when
    * restoring the state of a process from its serialized representation.
    *
    * @param aNewStatus the new status to set
    */
   public void setRawStatus(boolean aNewStatus)
   {
      mStatus = new Boolean(aNewStatus);
      mNotifyProcess = false;
   }
   
   /**
    * Setter for the status. The links status gets set after its source activity
    * completes.
    * @param aBool
    */
   public void setStatus(boolean aBool)
   {
      // status is immutable
      if(mStatus == null)
      {
         mStatus = new Boolean(aBool);
         if (mNotifyProcess)
            getProcess().linkStatusChanged(this);
      }
   }
   
   /**
    * Clears the status of the link.  This is used by flow to reset links 
    * before execute.  
    * TODO that we may need to trigger an event so listeners can reflact this unknown state.
    */
   public void clearStatus()
   {
      mStatus = null;
      mNotifyProcess = true;
   }
   
   /**
    * Links are evaluated when their source activity completes.   
    */
   public void evaluate() throws AeBusinessProcessException
   {
      mNotifyProcess = false;
      if (AeUtil.isNullOrEmpty(getTransitionCondition()))
      {
         setStatus(true);
      }
      else
      {
         boolean result = executeBooleanExpression(getTransitionCondition());
         getProcess().getEngine().fireEvaluationEvent(
            getProcess().getProcessId(), getTransitionCondition(), this, IAeProcessInfoEvent.INFO_LINK_XTN, getLocationPath(), Boolean.toString(result));
         setStatus( result );
      }
   }
   
   /**
    * Notifies the process of the link's state change. 
    */
   public void notifyProcess()
   {
      getProcess().linkStatusChanged(this);
   }

   /**
    * Getter for the target activity.
    */
   public IAeActivity getTargetActivity()
   {
      return mTargetActivity;
   }

   /**
    * Setter for the target activity.
    * @param aActivity
    */
   public void setTargetActivity(IAeActivity aActivity)
   {
      mTargetActivity = aActivity;
   }

   /**
    * Getter for the transition condition
    */
   public String getTransitionCondition()
   {
      return mTransitionCondition;
   }

   /**
    * Setter for the transition condition
    * @param aString
    */
   public void setTransitionCondition(String aString)
   {
      mTransitionCondition = aString;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeBpelObject#getChildrenForStateChange()
    */
   public Iterator getChildrenForStateChange()
   {
      return Collections.EMPTY_SET.iterator();
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeLink#getSourceActivity()
    */
   public IAeActivity getSourceActivity()
   {
      return mSourceActivity;
   }

   /**
    * Setter for the source activity.
    */
   public void setSourceActivity(IAeActivity aActivity)
   {
      mSourceActivity = aActivity;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.IAeExecutableBpelObject#execute()
    */
   public void execute() throws AeBusinessProcessException
   {
      // no-op
   }

}
