// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/activity/support/AeScopeSnapshot.java,v 1.6 2005/12/29 20:01:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.activity.support;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.impl.AeVariable;

/**
 * A snapshot of the process variables for use in compensation. Compensation handlers
 * expect to see and use the exact same variable contents that were in existence
 * at the time of the completion of the scope. In order to facilitate this,
 * we create a snapshot of the variable data and record all of the variable names
 * and their version numbers so we'll be able to pull the correct version of
 * the variable when asked.
 */
public class AeScopeSnapshot
{
   /** Maps the variable name to its version number */
   private final Map mVariableMap;

   /** Maps the correlation set to its version number */
   private final Map mCorrelationSetMap;

   /**
    * Default constructor.
    */
   public AeScopeSnapshot()
   {
      this(Collections.EMPTY_MAP, Collections.EMPTY_MAP);
   }

   /**
    * Constructor.
    *
    * @param aVariableMap
    * @param aCorrelationSetMap
    */
   public AeScopeSnapshot(Map aVariableMap, Map aCorrelationSetMap)
   {
      mVariableMap = new HashMap(aVariableMap);
      mCorrelationSetMap = new HashMap(aCorrelationSetMap);
   }

   /**
    * Adds the map of <code>AeVariable</code> by reference. The current map is cleared before
    * adding the new collection.
    * @param aVariableMap 
    */   
   public void addVariables(Map aVariableMap)
   {
      // Note: this method is used by the AeBusinessProcess when it needs to record the current
      // variable information for Process instance compensation.
      mVariableMap.clear();
      mVariableMap.putAll(aVariableMap);
   }
   
   /**
    * Adds the map of <code>CorrelationSet</code> by reference. The current correlation set map
    * is first cleared before this set is added.
    * @param aVariableMap 
    */   
   public void addCorrelationSet(Map aCorrelationSetMap)
   {
      // Note: this method is used by the AeBusinessProcess when it needs to record the current
      // correlation set information for Process instance compensation.
      mCorrelationSetMap.clear();
      mCorrelationSetMap.putAll(aCorrelationSetMap);      
   }   
   
   /**
    * Walks the map of <code>AeVariable</code> and records the variable name and
    * version number for each of the variables.
    * @param aVariableIter 
    */
   public void addVariables(Iterator aVariableIter)
   {
      while(aVariableIter.hasNext())
      {
         AeVariable var = (AeVariable) aVariableIter.next();
         // used to clone the variable here but changed to do the clone
         // of the variable within the scope if it ever executes again
         mVariableMap.put(var.getName(), var);
      }
   }

   /**
    * Gets the variable version number from the map
    * @param aName
    */
   public IAeVariable getVariable(String aName)
   {
      return (IAeVariable) mVariableMap.get(aName);
   }

   /**
    * Returns the <code>Set</code> of this snapshot's variables.
    */
   public Set getVariables()
   {
      return new HashSet(mVariableMap.values());
   }

   /**
    * Walks the map of <code>AeCorrelationSet</code> and records the CorrelationSet name and
    * value for each of the CorrelationSets.
    * @param aCorrelationSetMap if null, we'll do nothing.
    */
   public void addCorrelationSets(Iterator aCorrelationSetIter)
   {
      while (aCorrelationSetIter.hasNext())
      {
         AeCorrelationSet corrSet = (AeCorrelationSet) aCorrelationSetIter.next();
         // used to clone the corrSet here but changed to do the clone
         // of the corrSet within the scope if it ever executes again
         mCorrelationSetMap.put(corrSet.getName(), corrSet);
      }
   }

   /**
    * Gets the CorrelationSet version number from the map
    * @param aName
    */
   public AeCorrelationSet getCorrelationSet(String aName)
   {
      return (AeCorrelationSet) mCorrelationSetMap.get(aName);
   }

   /**
    * Returns the <code>Set</code> of this snapshot's correlation sets.
    */
   public Set getCorrelationSets()
   {
      return new HashSet(mCorrelationSetMap.values());
   }
}
