// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/storage/AeFaultDeserializer.java,v 1.3 2006/01/24 23:46:2
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.storage;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.IAeFault;
import org.activebpel.rt.bpel.def.IAeBPELConstants;
import org.activebpel.rt.bpel.impl.AeBusinessProcess;
import org.activebpel.rt.bpel.impl.AeFaultFactory;
import org.activebpel.rt.bpel.impl.AeMessageDataDeserializer;
import org.activebpel.rt.bpel.impl.IAeBpelObject;
import org.activebpel.rt.bpel.impl.IAeImplStateNames;
import org.activebpel.rt.bpel.impl.activity.support.AeFault;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeUtil;
import org.activebpel.rt.util.AeXmlUtil;
import org.w3c.dom.Element;

/**
 * Deserializes a fault from its serialization.
 */
public class AeFaultDeserializer implements IAeImplStateNames
{
   /** The fault once deserialized. */
   private IAeFault mFault;

   /** The fault's serialization. */
   private Element mFaultElement;

   /** The fault's process: used to locate the fault's source object. */
   private AeBusinessProcess mProcess;

   /**
    * Creates an instance of {@link org.activebpel.rt.bpel.IAeFault} from its
    * serialization.
    * 
    * @param aFaultElement
    * @throws AeBusinessProcessException
    */
   protected IAeFault createFault(Element aFaultElement) throws AeBusinessProcessException
   {
      String localPart = aFaultElement.getAttribute(STATE_NAME);
      String namespace = aFaultElement.getAttribute(STATE_NAMESPACEURI);
      QName faultName = new QName(namespace, localPart);
      
      IAeFault result;

      // If there's message data, then create the fault object from scratch
      // with the message data.
      String hasMessageData = aFaultElement.getAttribute(STATE_HASMESSAGEDATA);
      if ("true".equals(hasMessageData)) //$NON-NLS-1$
      {
         AeMessageDataDeserializer deserializer = new AeMessageDataDeserializer();
         deserializer.setMessageDataElement(AeXmlUtil.getFirstSubElement(aFaultElement));

         IAeMessageData messageData = deserializer.getMessageData();
         result = new AeFault(faultName, messageData);
      }
      // Otherwise, try getting the fault from the fault factory.
      else
      {
         result = AeFaultFactory.getFault(localPart);
         if ((result == null) || !namespace.equals(result.getFaultName().getNamespaceURI()))
         {
            // The fault factory doesn't have this fault or has a fault from a
            // different namespace.
            result = new AeFault(faultName, null);
         }
      }
      
      result.setSuspendable( readBooleanAttribute(aFaultElement, STATE_SUSPENDABLE, true) );

      boolean faultIsForcedTermination = faultName.getNamespaceURI().equals(IAeBPELConstants.BPWS_NAMESPACE_URI) && faultName.getLocalPart().equals(AeFaultFactory.FORCED_TERMINATION);
      result.setRethrowable( readBooleanAttribute(aFaultElement, STATE_RETHROWABLE, !faultIsForcedTermination) );

      String sourceLocationPath = aFaultElement.getAttribute(STATE_SOURCE);
      if (!AeUtil.isNullOrEmpty(sourceLocationPath))
      {
         if (getProcess() == null)
         {
            throw new IllegalStateException(AeMessages.getString("AeFaultDeserializer.ERROR_0")); //$NON-NLS-1$
         }

         IAeBpelObject source = getProcess().findBpelObject(sourceLocationPath);
         result.setSource(source);
      }

      return result;
   }

   /**
    * Reads the boolean attribute from the element, defaulting to the value provided if not 
    * available from the element.
    * @param aFaultElement
    * @param aAttributeName
    * @param aDefaultValue
    */
   protected boolean readBooleanAttribute(Element aFaultElement, String aAttributeName, boolean aDefaultValue)
   {
      String flag = aFaultElement.getAttribute( aAttributeName );

      // if this fault was serialized before this feature was available,
      // then use the default value
      if( AeUtil.isNullOrEmpty(flag) )
      {
         return aDefaultValue;
      }
      else
      {
         return AeUtil.toBoolean( flag );
      }
   }

   /**
    * Returns the fault deserialized from the serialization that was set with
    * {@link #setFaultElement}.
    */
   public IAeFault getFault() throws AeBusinessProcessException
   {
      if (mFault == null)
      {
         if (getFaultElement() == null)
         {
            throw new IllegalStateException(AeMessages.getString("AeFaultDeserializer.ERROR_1")); //$NON-NLS-1$
         }

         mFault = createFault(getFaultElement());
      }

      return mFault;
   }

   /**
    * Returns the fault serialization to use.
    */
   protected Element getFaultElement()
   {
      return mFaultElement;
   }

   /**
    * Returns the process to use to locate fault source objects.
    */
   protected AeBusinessProcess getProcess()
   {
      return mProcess;
   }

   /**
    * Resets all output variables.
    */
   protected void reset()
   {
      mFault = null;
   }

   /**
    * Sets the fault serialization to use.
    *
    * @param aFaultElement
    */
   public void setFaultElement(Element aFaultElement)
   {
      reset();

      mFaultElement = aFaultElement;
   }

   /**
    * Sets the process to use to locate fault source objects.
    * 
    * @param aProcess
    */
   public void setProcess(AeBusinessProcess aProcess)
   {
      reset();

      mProcess = aProcess;
   }
}
