// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/visitors/AeProcessSnapshotVisitor.java,v 1.4 2005/12/02 21:22:0
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.visitors;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.IAeVariable;
import org.activebpel.rt.bpel.def.AeCorrelationSetDef;
import org.activebpel.rt.bpel.def.AeProcessDef;
import org.activebpel.rt.bpel.def.AeScopeDef;
import org.activebpel.rt.bpel.def.AeVariableDef;
import org.activebpel.rt.bpel.def.activity.AeActivityScopeDef;
import org.activebpel.rt.bpel.impl.AeBpelState;
import org.activebpel.rt.bpel.impl.AeBusinessProcess;
import org.activebpel.rt.bpel.impl.activity.AeActivityInvokeImpl;
import org.activebpel.rt.bpel.impl.activity.AeActivityScopeImpl;
import org.activebpel.rt.bpel.impl.activity.support.AeCompInfo;
import org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeCorrelationSet;
import org.activebpel.rt.bpel.impl.activity.support.AeScopeSnapshot;

/**
 * Visits a tree of BPEL implementation objects to collect all versions of all
 * variables and correlation sets that are still live for normal or
 * compensation processing.
 */
public class AeProcessSnapshotVisitor extends AeImplTraversingVisitor
{
   /** The <code>Set</code> of <code>AeCorrelationSet</code> instances. */
   private final Set mCorrelationSets = new HashSet();

   /** The <code>Set</code> of <code>IAeVariable</code> instances. */
   private final Set mVariables = new HashSet();

   /** The <code>Set</code> of pending </code>AeActivityInvokeImpl</code> instances. */
   private final Set mPendingInvokes = new HashSet();

   /**
    * Adds all live variable and correlation set instances reachable from the
    * specified compensation info object.
    *
    * @param aCompInfo
    */
   protected void addCompInfo(AeCompInfo aCompInfo)
   {
      AeScopeSnapshot snapshot = aCompInfo.getSnapshot();
      if (snapshot != null)
      {
         Set correlationSets = snapshot.getCorrelationSets();
         Set variables = snapshot.getVariables();
   
         addCorrelationSets(correlationSets);
         addVariables(variables);
      }

      // Add variables and correlation sets from enclosed scopes' compensation
      // infos.
      for (Iterator i = aCompInfo.getEnclosedScopes().iterator(); i.hasNext(); )
      {
         AeCompInfo enclosedScope = (AeCompInfo) i.next();
         addCompInfo(enclosedScope);
      }
   }

   /**
    * Adds the specified correlation set instance to the set that will be
    * returned by <code>getCorrelationSets</code>.
    *
    * @param aCorrelationSet
    */
   protected void addCorrelationSet(AeCorrelationSet aCorrelationSet)
   {
      getCorrelationSets().add(aCorrelationSet);
   }

   /**
    * Adds the specified set of correlation set instances to the set that will
    * be returned by <code>getCorrelationSets</code>.
    *
    * @param aCorrelationSets
    */
   protected void addCorrelationSets(Set aCorrelationSets)
   {
      getCorrelationSets().addAll(aCorrelationSets);
   }

   /**
    * Adds the specified invoke instance to the set of pending invokes that
    * will be returned by <code>getPendingInvokes</code>.
    *
    * @param aInvoke
    */
   protected void addPendingInvoke(AeActivityInvokeImpl aInvoke)
   {
      getPendingInvokes().add(aInvoke);
   }

   /**
    * Adds all live variable and correlation set instances reachable from the
    * specified activity scope implementation object.
    *
    * @param aImpl
    * @param aScopeDef
    */
   protected void addScope(AeActivityScopeImpl aImpl, AeScopeDef aScopeDef) throws AeBusinessProcessException
   {
      // Add the scope's variables.
      for (Iterator i = aScopeDef.getVariables(); i.hasNext(); )
      {
         AeVariableDef variableDef = (AeVariableDef) i.next();
         IAeVariable variable = aImpl.getVariable(variableDef.getName());

         addVariable(variable);
      }

      // Add the scope's correlation sets.
      for (Iterator i = aScopeDef.getCorrelationSets(); i.hasNext(); )
      {
         AeCorrelationSetDef correlationSetDef = (AeCorrelationSetDef) i.next();
         AeCorrelationSet correlationSet = aImpl.getCorrelationSet(correlationSetDef.getName());

         addCorrelationSet(correlationSet);
      }

      if (aImpl.hasCompInfo())
      {
         // Add variables and correlation sets from the scope's compensation
         // info.
         addCompInfo(aImpl.getCompInfo());
      }
   }

   /**
    * Adds the specified variable instance to the set that will be returned by
    * <code>getVariables</code>.
    *
    * @param aVariable
    */
   protected void addVariable(IAeVariable aVariable)
   {
      getVariables().add(aVariable);
   }

   /**
    * Adds the specified set of variable instances to the set that will be
    * returned by <code>getVariables</code>.
    *
    * @param aVariables
    */
   protected void addVariables(Set aVariables)
   {
      getVariables().addAll(aVariables);
   }

   /**
    * Returns the <code>Set</code> of <code>AeCorrelationSet</code> instances.
    */
   public Set getCorrelationSets()
   {
      return mCorrelationSets;
   }

   /**
    * Returns the <code>Set</code> of pending </code>AeActivityInvokeImpl</code> instances.
    */
   public Set getPendingInvokes()
   {
      return mPendingInvokes;
   }

   /**
    * Returns the <code>Set</code> of <code>IAeVariable</code> instances.
    */
   public Set getVariables()
   {
      return mVariables;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.AeActivityInvokeImpl)
    */
   public void visit(AeActivityInvokeImpl aImpl) throws AeBusinessProcessException
   {
      traverse(aImpl);   

      // If this invoke is in the executing state and it is not queued for retry, then it's a pending invoke.
      if (aImpl.getState() == AeBpelState.EXECUTING && (! aImpl.isQueued()) )
      {
         addPendingInvoke(aImpl);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.AeActivityScopeImpl)
    */
   public void visit(AeActivityScopeImpl aImpl) throws AeBusinessProcessException
   {
      traverse(aImpl);

      AeActivityScopeDef activityScopeDef = (AeActivityScopeDef) aImpl.getDefinition();
      AeScopeDef scopeDef = activityScopeDef.getScope();
      addScope(aImpl, scopeDef);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.AeBusinessProcess)
    */
   public void visit(AeBusinessProcess aImpl) throws AeBusinessProcessException
   {
      traverse(aImpl);

      AeProcessDef processDef = (AeProcessDef) aImpl.getDefinition();
      addScope(aImpl, processDef);
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler)
    */
   public void visit(AeCompensationHandler aImpl) throws AeBusinessProcessException
   {
      traverse(aImpl);

      AeCompInfo compInfo = aImpl.getCompInfo();
      if (compInfo != null)
      {
         addCompInfo(compInfo);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.support.AeCorrelationSet)
    */
   public void visit(AeCorrelationSet aImpl) throws AeBusinessProcessException
   {
      traverse(aImpl);

      addCorrelationSet(aImpl);
   }
}
