// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/impl/visitors/AeRestoreCompInfoVisitor.java,v 1.3 2005/12/29 20:01:3
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.impl.visitors;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.impl.AeAbstractBpelObject;
import org.activebpel.rt.bpel.impl.AeBusinessProcess;
import org.activebpel.rt.bpel.impl.AeVariable;
import org.activebpel.rt.bpel.impl.activity.AeActivityForEachParallelImpl;
import org.activebpel.rt.bpel.impl.activity.AeActivityScopeImpl;
import org.activebpel.rt.bpel.impl.activity.support.AeCompInfo;
import org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeCoordinatorCompInfo;
import org.activebpel.rt.bpel.impl.activity.support.AeCoordinatorCompensationHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeCorrelationSet;
import org.activebpel.rt.bpel.impl.activity.support.AeImplicitCompensationHandler;
import org.activebpel.rt.bpel.impl.activity.support.AeScopeSnapshot;
import org.activebpel.rt.bpel.impl.storage.AeRestoreImplState;
import org.w3c.dom.Element;

/**
 * Visits a tree of BPEL implementation objects to restore {@link AeCompInfo}
 * objects from an instance of <code>AeRestoreImplState</code>.
 */
public class AeRestoreCompInfoVisitor extends AeBaseRestoreVisitor
{
   /**
    * Constructor.
    *
    * @param aImplState The state object from which to restore state.
    */
   public AeRestoreCompInfoVisitor(AeRestoreImplState aImplState)
   {
      super(aImplState);
   }

   /**
    * Creates the compensation info object for the specified implementation
    * object, if the compensation info exists.
    *
    * @param aImpl
    * @return AeCompInfo
    * @throws AeBusinessProcessException
    */
   protected AeCompInfo createCompInfo(AeAbstractBpelObject aImpl) throws AeBusinessProcessException
   {
      Element implElement = getImplState().getElement(aImpl);
      String xpath = "./" + STATE_COMPINFO; //$NON-NLS-1$
      Element compInfoElement = (Element) selectOptionalNode(implElement, xpath, "Error restoring compensation info"); //$NON-NLS-1$

      return (compInfoElement == null) ? null : createCompInfo(compInfoElement);
   }
   
   /**
    * Creates an instance of <code>AeCompInfo</code> from its serialization.
    *
    * @param aElement
    * @return AeCompInfo
    * @throws AeBusinessProcessException
    */
   protected AeCompInfo createCompInfo(Element aElement) throws AeBusinessProcessException
   {
      int id = getAttributeInt(aElement, STATE_ID);
      AeCompInfo compInfo = (AeCompInfo) getObject(id);

      if (compInfo == null)
      {
         // Haven't seen this object id yet. Create the compensation info
         // object.
         int scopeLocationId = getAttributeInt(aElement, STATE_SCOPE);
         AeActivityScopeImpl scope = (AeActivityScopeImpl) getImplState().getProcess().findBpelObject(scopeLocationId);
         
         if (scope == null)
         {
         	// TODO (MF) should defer the restore of the scope until the compInfo executes. This requires reworking the restore of the scope snapshot data below.
         
            // The scope could be null if it references a previous iteration of
            // a forEach that is no longer accessible. Attempt to restore this
            // scope and have it parented by its forEach.
            String scopeLocation = getAttribute(aElement, STATE_SCOPE_LOCATION);
            scope = restoreScopeInstance(scopeLocation, true);
         }
         

         boolean hasCoordination = getAttributeBoolean(aElement, STATE_HASCOORDINATIONS);
         String coordinationId = getAttribute(aElement, STATE_COORDINATION_ID);
         if (hasCoordination)
         {
            compInfo = new AeCoordinatorCompInfo(scope, coordinationId);
         }
         else
         {
            compInfo = new AeCompInfo(scope);
         }

         putIdObject(id, compInfo);

         boolean enabled = getAttributeBoolean(aElement, STATE_ENABLED);
         compInfo.setEnabled(enabled);

         if (enabled)
         {
            // This element contains the compensation info object's inner data.
            // Reconstruct the variables, correlation sets, and enclosed
            // compensation info objects.
            Map variables = createCompInfoVariables(aElement, scope);
            Map correlationSets = createCompInfoCorrelationSets(aElement, scope);
            AeScopeSnapshot snapshot = new AeScopeSnapshot(variables, correlationSets);
            compInfo.setSnapshot(snapshot);

            String xpath = "./" + STATE_COMPINFO; //$NON-NLS-1$
            List enclosedScopeElements = selectNodes(aElement, xpath, "Error restoring enclosed compensation info"); //$NON-NLS-1$
            LinkedList enclosedScopes = new LinkedList();

            for (Iterator i = enclosedScopeElements.iterator(); i.hasNext(); )
            {
               Element element = (Element) i.next();
               AeCompInfo enclosedScope = createCompInfo(element);
               enclosedScopes.addLast(enclosedScope);
            }

            compInfo.setEnclosedScopes(enclosedScopes);
         }
      }

      return compInfo;
   }

   /**
    * Creates a <code>Map</code> of correlation sets from their serialization
    * in a compensation info serialization.
    *
    * @param aCompInfoElement
    * @param aScope
    * @return Map
    * @throws AeBusinessProcessException
    */
   protected Map createCompInfoCorrelationSets(Element aCompInfoElement, AeActivityScopeImpl aScope) throws AeBusinessProcessException
   {
      String xpath = "./" + STATE_CORRSET; //$NON-NLS-1$
      List elements = selectNodes(aCompInfoElement, xpath, "Error restoring compensation info correlation sets"); //$NON-NLS-1$
      Map map = new HashMap();

      for (Iterator i = elements.iterator(); i.hasNext(); )
      {
         Element element = (Element) i.next();
         String name = getAttribute(element, STATE_NAME);

         int versionNumber = getAttributeInt(element, STATE_VERSION);
         AeCorrelationSet correlationSet = (AeCorrelationSet) aScope.getCorrelationSet(name);

         if (versionNumber != correlationSet.getVersionNumber())
         {
            // Different version from the correlation set in the scope. Clone a
            // new correlation set from the scope correlation set.
            correlationSet = (AeCorrelationSet) correlationSet.clone();
            restoreCorrelationSet(element, correlationSet);
         }

         map.put(name, correlationSet);
      }

      return map;
   }

   /**
    * Creates a <code>Map</code> of variables from their serialization in a
    * compensation info serialization.
    *
    * @param aCompInfoElement
    * @param aScope
    * @return Map
    * @throws AeBusinessProcessException
    */
   protected Map createCompInfoVariables(Element aCompInfoElement, AeActivityScopeImpl aScope) throws AeBusinessProcessException
   {
      String xpath = "./" + STATE_VAR; //$NON-NLS-1$
      List elements = selectNodes(aCompInfoElement, xpath, "Error restoring compensation info variables"); //$NON-NLS-1$
      Map map = new HashMap();

      for (Iterator i = elements.iterator(); i.hasNext(); )
      {
         Element element = (Element) i.next();
         String name = getAttribute(element, STATE_NAME);

         int versionNumber = getAttributeInt(element, STATE_VERSION);
         AeVariable variable = (AeVariable) aScope.getVariable(name);

         if (versionNumber != variable.getVersionNumber())
         {
            // Different version from the variable in the scope. Clone a new
            // variable from the scope variable.
            variable = (AeVariable) variable.clone();
            restoreVariable(element, variable);
         }

         map.put(name, variable);
      }

      return map;
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.AeImplTraversingVisitor#traverse(org.activebpel.rt.bpel.impl.AeAbstractBpelObject)
    */
   protected void traverse(AeAbstractBpelObject aImpl) throws AeBusinessProcessException
   {
      Element element = getImplState().getElement(aImpl);

      boolean skipChildren = getAttributeBoolean(element, STATE_SKIPCHILDREN);
      if (!skipChildren)
      {
         super.traverse(aImpl);
      }
   }

   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.AeActivityScopeImpl)
    */
   public void visit(AeActivityScopeImpl aImpl) throws AeBusinessProcessException
   {
      super.visit(aImpl);

      AeCompInfo info = createCompInfo(aImpl);
      aImpl.setCompInfo(info);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.AeBusinessProcess)
    */
   public void visit(AeBusinessProcess aImpl) throws AeBusinessProcessException
   {
      visit((AeActivityScopeImpl) aImpl);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.support.AeCompensationHandler)
    */
   public void visit(AeCompensationHandler aImpl) throws AeBusinessProcessException
   {
      super.visit(aImpl);

      AeCompInfo info = createCompInfo(aImpl);
      aImpl.setCompInfo(info);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.AeActivityForEachParallelImpl)
    */
   public void visit(AeActivityForEachParallelImpl aImpl)
         throws AeBusinessProcessException
   {
      super.visit(aImpl);
      
      // now visit the compensatable scopes
      for (Iterator iter = aImpl.getCompensatableChildren().iterator(); iter.hasNext();)
      {
         AeActivityScopeImpl scope = (AeActivityScopeImpl) iter.next();
         visit(scope);
      }
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.support.AeImplicitCompensationHandler)
    */
   public void visit(AeImplicitCompensationHandler aImpl) throws AeBusinessProcessException
   {
      visit((AeCompensationHandler) aImpl);
   }
   
   /**
    * @see org.activebpel.rt.bpel.impl.visitors.IAeImplVisitor#visit(org.activebpel.rt.bpel.impl.activity.support.AeCoordinatorCompensationHandler)
    */
   public void visit(AeCoordinatorCompensationHandler aImpl) throws AeBusinessProcessException
   {
      visit((AeCompensationHandler) aImpl);
   }      
}
