// $Header: /Development/AEDevelopment/projects/org.activebpel.rt.bpel/src/org/activebpel/rt/bpel/xpath/AeXPathHelper.java,v 1.9 2006/01/19 17:24:4
/*
 * Copyright (c) 2004-2006 Active Endpoints, Inc.
 *
 * This program is licensed under the terms of the GNU General Public License
 * Version 2 (the "License") as published by the Free Software Foundation, and 
 * the ActiveBPEL Licensing Policies (the "Policies").  A copy of the License 
 * and the Policies were distributed with this program.  
 *
 * The License is available at:
 * http: *www.gnu.org/copyleft/gpl.html
 *
 * The Policies are available at:
 * http: *www.activebpel.org/licensing/index.html
 *
 * Unless required by applicable law or agreed to in writing, this program is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
 * OF ANY KIND, either express or implied.  See the License and the Policies
 * for specific language governing the use of this program.
 */
package org.activebpel.rt.bpel.xpath;

import java.text.MessageFormat;
import java.util.List;

import javax.xml.namespace.QName;

import org.activebpel.rt.bpel.AeBusinessProcessException;
import org.activebpel.rt.bpel.AeMessages;
import org.activebpel.rt.bpel.function.AeFunctionCallException;
import org.activebpel.rt.bpel.impl.AeBpelException;
import org.activebpel.rt.bpel.impl.AeCorrelationViolationException;
import org.activebpel.rt.bpel.impl.AeFaultFactory;
import org.activebpel.rt.bpel.impl.AeNamespaceResolver;
import org.activebpel.rt.bpel.impl.AeTypeMapping;
import org.activebpel.rt.bpel.impl.expr.xpath.AeXPathNamespaceContext;
import org.activebpel.rt.message.IAeMessageData;
import org.activebpel.rt.util.AeUtil;
import org.activebpel.rt.wsdl.def.IAePropertyAlias;
import org.jaxen.FunctionCallException;
import org.jaxen.FunctionContext;
import org.jaxen.NamespaceContext;
import org.jaxen.UnresolvableException;
import org.jaxen.XPath;
import org.jaxen.XPathFunctionContext;
import org.jaxen.dom.DOMXPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * A helper class for executing XPath expressions. 
 */
public class AeXPathHelper
{
   /** private ctor to prevent instantiation */
   private AeXPathHelper()  {}

   /**
    * If the argument is a java.util.List then it must have a single entry, any
    * more than that results in a selection failure exception.
    * @param aData
    * @param aAllowEmptySelection If true a empty string is returned for no selection, else an exception is thrown.
    * @throws AeBpelException
    */
   public static Object unwrapXPathValue(Object aData, boolean aAllowEmptySelection) throws AeBpelException
   {
      Object returnValue = aData;
      if (aData instanceof List)
      {
         List list = (List)aData;
         if (list.size() == 1)
            returnValue = list.get(0);
         else if(list.size() == 0 && aAllowEmptySelection)
            returnValue = null;
         else
            throw new AeBpelException(AeMessages.getString("AeXPathHelper.ERROR_0"), AeFaultFactory.SELECTION_FAILURE); //$NON-NLS-1$
      }
      
      if (returnValue instanceof Double)
      {
         // if the double can safely be converted to a Long then do it.
         // this handles the case where we are attempting to set an int
         // value on a complex type and introducing a floating point number
         // where we don't want one. 
         Double d = (Double) returnValue;
         if (Double.compare(d.doubleValue(), Math.ceil(d.doubleValue()) ) == 0 )
         {
            returnValue = new Long(d.longValue());
         }
      }
      
      return returnValue;
   }

   /**
    * Evaluates the property, potentially executing an XPath expression on the message
    * data to extract the property value from the specified part. If the part identified
    * by the alias is a simple type then we simply return that value instead of 
    * executing an xpath.
    * @param aPropAlias
    * @param aMsgData
    */
   public static Object evaluateProperty(IAePropertyAlias aPropAlias, IAeMessageData aMsgData) throws AeBusinessProcessException
   {
      Object propData = null;
      Object varData = aMsgData.getData(aPropAlias.getPart());
      if (varData instanceof Document)
      {
         Element docElem = ((Document) varData).getDocumentElement();
         if (AeUtil.notNullOrEmpty(aPropAlias.getQuery()))
         {
            NamespaceContext namespaceContext = new AeXPathNamespaceContext(new AeNamespaceResolver(aPropAlias), docElem);
            FunctionContext functionContext = XPathFunctionContext.getInstance(); // TODO should we allow custom functions to be used here?
            Object result = executeXPathExpression(aPropAlias.getQuery(), varData, functionContext, namespaceContext);
            propData = unwrapXPathValue(result, false); // TODO should we pass config here?
         }
         else
         {
            propData = ((Document)varData).getDocumentElement();
         }
      }
      else
      {
         propData = varData;
      }
      return propData;
   }

   /**
    * Convenience method that evaluates the property and then converts it to a simple type.
    * @param aPropAlias
    * @param aMsgData
    * @throws AeBusinessProcessException
    */
   public static Object evaluatePropertyToSimpleType(
      IAePropertyAlias aPropAlias,
      IAeMessageData aMsgData,
      AeTypeMapping aTypeMapping,
      QName aType)
      throws AeBusinessProcessException
   {
      Object value = evaluateProperty(aPropAlias, aMsgData);
      return convertPropertyResultToSimpleType(value, aTypeMapping, aType);
   }

   /**
    * Executes the xpath expression using the contextual information passed in AND the default Jaxen
    * XPathFunctionContext.  This should only be used if the XPath expression being executed does
    * not need access to the custom function contexts.
    * 
    * @param aExpr
    * @param aContext
    * @param aNamespaceResolver
    */
   public static Object executeXPathExpression(String aExpr, Object aContext,
         NamespaceContext aNamespaceResolver) throws AeBusinessProcessException
   {
      return executeXPathExpression(aExpr, aContext, XPathFunctionContext.getInstance(), aNamespaceResolver);
   }

   /**
    * Executes the xpath expression using the contextual information passed in. 
    * @param aExpr
    * @param aContext
    * @param aFunctionContext
    * @param aNamespaceResolver
    * @throws AeBusinessProcessException
    */
   public static Object executeXPathExpression(String aExpr, Object aContext, 
                                               FunctionContext aFunctionContext, 
                                               NamespaceContext aNamespaceResolver)
      throws AeBusinessProcessException
   {
      Object result = null;
      
      try
      {
         XPath path = new DOMXPath(aExpr);
         path.setFunctionContext(aFunctionContext);
         path.setNamespaceContext(aNamespaceResolver);
         result = path.evaluate(aContext);
      }
      catch (FunctionCallException fe)
      {
         rethrowFunctionCallException(aExpr, fe);
      }
      catch (UnresolvableException ue)
      {
         String msg = AeMessages.format("AeXPathHelper.ERROR_EVALUATING_EXPRESSION_UNRESOLVEABLE", new Object[] { aExpr, ue.getLocalizedMessage() }); //$NON-NLS-1$
         throw new AeBpelException(msg, AeFaultFactory.XPATH_FUNCTION_ERROR, ue);
      }
      catch (Throwable ex)
      {
         throw new AeBpelException( MessageFormat.format(AeMessages.getString("AeXPathHelper.ERROR_1"), //$NON-NLS-1$
                                                         new Object[] {aExpr, ex.getMessage()}), 
                                    AeFaultFactory.XPATH_FUNCTION_ERROR, ex );    
      }
      
      return result;
   }
   
   /**
    * Determines the root cause of a function call exception and rethrows it in the form of a 
    * AeBpelException.
    * 
    * @param aExpression
    * @param aFunctionCallException
    * @throws AeBpelException
    */
   private static void rethrowFunctionCallException(String aExpression, FunctionCallException aFunctionCallException) throws AeBpelException
   {
      if (aFunctionCallException.getCause() instanceof AeBusinessProcessException)
         throw new AeBpelException(MessageFormat.format(AeMessages.getString("AeXPathHelper.ERROR_1"), //$NON-NLS-1$
               new Object[] { aExpression, aFunctionCallException.getCause().getMessage() }),
               AeFaultFactory.XPATH_FUNCTION_ERROR, aFunctionCallException);
      else if (aFunctionCallException.getCause() instanceof AeFunctionCallException)
      {
         Throwable rootException = aFunctionCallException.getCause().getCause();
         throw new AeBpelException(MessageFormat.format(AeMessages.getString("AeXPathHelper.ERROR_1"), //$NON-NLS-1$
               new Object[] { aExpression, rootException.getMessage() }),
               AeFaultFactory.XPATH_FUNCTION_ERROR, aFunctionCallException);
         
      }
      else
         throw new AeBpelException( MessageFormat.format(AeMessages.getString("AeXPathHelper.ERROR_1"), //$NON-NLS-1$
               new Object[] {aExpression, aFunctionCallException.getMessage()}), 
               AeFaultFactory.XPATH_FUNCTION_ERROR, aFunctionCallException );
   }

   /**
    * Converts the argument to a simple type. If the argument is already a simple 
    * type (java.lang.String or java.lang.Number) then it is returned unchanged. 
    * If not then we're expecting a dom Element in which case we'll extract the 
    * text from the element and return that as our value.  
    * @param aObject - if you pass null, you'll get an exception
    */
   private static Object convertPropertyResultToSimpleType(Object aObject, AeTypeMapping aTypeMapping, QName aType)
      throws AeBusinessProcessException
   {
      // a null signals a problem
      if (aObject == null)
      {
         throw new AeCorrelationViolationException(AeCorrelationViolationException.NULL_VALUE);
      }
      Object simpleType = null;
      if (aObject instanceof Element)
      {
         Element element = (Element)aObject;
         element.normalize();
         String xmlString = element.getFirstChild()!=null? element.getFirstChild().getNodeValue() : null;
         simpleType = aTypeMapping.deserialize(aType, xmlString);
      }
      else if (aObject instanceof Node)
      {
         simpleType = ((Node)aObject).getNodeValue();
      }
      else
      {
         simpleType = aObject;
      }

      return simpleType.toString();
   }
}
